/* Generated by Cython 0.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {}
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__asynq__async_task
#define __PYX_HAVE_API__asynq__async_task
#include "string.h"
#include "stdio.h"
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "asynq/async_task.py",
  "asynq/async_task.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "events.pxd",
  "helpers.pxd",
  "asynq/_debug.pxd",
  "asynq/futures.pxd",
  "asynq/batching.pxd",
  "asynq/scheduler.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5qcore_6events_EventHook;
struct __pyx_obj_5qcore_6events_SinkingEventHook;
struct __pyx_obj_5qcore_6events_EventInterceptor;
struct __pyx_obj_5qcore_7helpers_MarkerObject;
struct __pyx_obj_5qcore_7helpers_EmptyContext;
struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper;
struct __pyx_obj_5qcore_7helpers_ScopedValue;
struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext;
struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext;
struct __pyx_obj_5asynq_6_debug_DebugOptions;
struct __pyx_obj_5asynq_7futures_FutureBase;
struct __pyx_obj_5asynq_7futures_Future;
struct __pyx_obj_5asynq_7futures_ConstFuture;
struct __pyx_obj_5asynq_7futures_ErrorFuture;
struct __pyx_obj_5asynq_8batching_BatchBase;
struct __pyx_obj_5asynq_8batching_BatchItemBase;
struct __pyx_obj_5asynq_8batching_DebugBatchItem;
struct __pyx_obj_5asynq_8batching_DebugBatch;
struct __pyx_obj_5asynq_9scheduler_TaskScheduler;
struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator;
struct __pyx_obj_5asynq_10async_task_AsyncTask;
struct __pyx_opt_args_5qcore_7helpers_safe_str;
struct __pyx_opt_args_5qcore_7helpers_safe_repr;

/* "qcore/helpers.pxd":75
 * 
 * cpdef object ellipsis(object source, int max_length)
 * cpdef object safe_str(object source, int max_length=?)             # <<<<<<<<<<<<<<
 * cpdef object safe_repr(object source, int max_length=?)
 * 
 */
struct __pyx_opt_args_5qcore_7helpers_safe_str {
  int __pyx_n;
  int max_length;
};

/* "qcore/helpers.pxd":76
 * cpdef object ellipsis(object source, int max_length)
 * cpdef object safe_str(object source, int max_length=?)
 * cpdef object safe_repr(object source, int max_length=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5qcore_7helpers_safe_repr {
  int __pyx_n;
  int max_length;
};
struct __pyx_opt_args_5qcore_10inspection_wraps;

/* "qcore/inspection.pxd":38
 * cpdef bint is_classmethod(object fn) except -1
 * 
 * cpdef wraps(object wrapped, object assigned=?, object updated=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5qcore_10inspection_wraps {
  int __pyx_n;
  PyObject *assigned;
  PyObject *updated;
};
struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL;

/* "_debug.pxd":44
 *     cdef public object ENABLE_COMPLEX_ASSERTIONS
 * 
 *     cpdef DUMP_ALL(self, value=?)             # <<<<<<<<<<<<<<
 * 
 * cdef public DebugOptions options
 */
struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL {
  int __pyx_n;
  PyObject *value;
};
struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump;

/* "futures.pxd":46
 *     cdef inline raise_if_error(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump {
  int __pyx_n;
  int indent;
};
struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel;
struct __pyx_opt_args_5asynq_8batching_9BatchBase_dump;
struct __pyx_opt_args_5asynq_8batching_sync;

/* "batching.pxd":35
 * 
 *     cpdef flush(self)
 *     cpdef cancel(self, object error=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _compute(self)
 */
struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel {
  int __pyx_n;
  PyObject *error;
};

/* "batching.pxd":43
 *     cpdef _try_switch_active_batch(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class BatchItemBase(futures.FutureBase):
 */
struct __pyx_opt_args_5asynq_8batching_9BatchBase_dump {
  int __pyx_n;
  int indent;
};

/* "batching.pxd":66
 * cdef object _debug_batch_state
 * 
 * cpdef sync(tag=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5asynq_8batching_sync {
  int __pyx_n;
  PyObject *tag;
};
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump;

/* "scheduler.pxd":45
 *     cpdef reset(self)
 * 
 *     cpdef async_task.AsyncTask start(self, async_task.AsyncTask task,             # <<<<<<<<<<<<<<
 *             bint run_immediately=?)
 *     cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start {
  int __pyx_n;
  int run_immediately;
};

/* "scheduler.pxd":48
 *             bint run_immediately=?)
 *     cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
 *     cpdef int flush_batches(self, bint must_await=?) except -1             # <<<<<<<<<<<<<<
 * 
 *     cpdef add_dependencies(self, async_task.AsyncTask task, object yield_result)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches {
  int __pyx_n;
  int must_await;
};

/* "scheduler.pxd":76
 *     @cython.locals(dump_continue_task=bint, task=async_task.AsyncTask)
 *     cdef async_task.AsyncTask _continue_with_task(self)
 *     cdef batching.BatchBase _continue_with_batch(self, bint assert_no_batch=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef batching.BatchBase _select_batch_to_flush(self)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch {
  int __pyx_n;
  int assert_no_batch;
};

/* "scheduler.pxd":84
 *     cpdef _accept_changes(self, TaskScheduler clone)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 *     @cython.locals(current_time=float)
 *     cpdef try_time_based_dump(self, object last_task=?)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump {
  int __pyx_n;
  int indent;
};

/* "scheduler.pxd":86
 *     cpdef dump(self, int indent=?)
 *     @cython.locals(current_time=float)
 *     cpdef try_time_based_dump(self, object last_task=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class TaskSchedulerActivator(object):
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump {
  int __pyx_n;
  PyObject *last_task;
};
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start;
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump;

/* "asynq/async_task.pxd":53
 * 
 *     @cython.locals(s=scheduler.TaskScheduler)
 *     cpdef AsyncTask start(self, bint run_immediately=?)             # <<<<<<<<<<<<<<
 *     @cython.locals(s=scheduler.TaskScheduler)
 *     cpdef AsyncTask after(self, futures.FutureBase future)
 */
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start {
  int __pyx_n;
  int run_immediately;
};

/* "asynq/async_task.pxd":84
 *     cpdef list traceback(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 *     # Contexts support
 */
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump {
  int __pyx_n;
  int indent;
};

/* "qcore/events.pxd":15
 * # limitations under the License.
 * 
 * cdef class EventHook(object):             # <<<<<<<<<<<<<<
 *     cdef list handlers
 * 
 */
struct __pyx_obj_5qcore_6events_EventHook {
  PyObject_HEAD
  PyObject *handlers;
};


/* "qcore/events.pxd":19
 * 
 * 
 * cdef class SinkingEventHook(EventHook):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5qcore_6events_SinkingEventHook {
  struct __pyx_obj_5qcore_6events_EventHook __pyx_base;
};


/* "qcore/events.pxd":25
 * 
 * 
 * cdef class EventInterceptor(object):             # <<<<<<<<<<<<<<
 *     cdef object source
 *     cdef dict events
 */
struct __pyx_obj_5qcore_6events_EventInterceptor {
  PyObject_HEAD
  PyObject *source;
  PyObject *events;
};


/* "qcore/helpers.pxd":24
 * 
 * 
 * cdef class MarkerObject(object):             # <<<<<<<<<<<<<<
 *     cpdef unicode name
 * 
 */
struct __pyx_obj_5qcore_7helpers_MarkerObject {
  PyObject_HEAD
  PyObject *name;
};


/* "qcore/helpers.pxd":33
 * 
 * 
 * cdef class EmptyContext(object):             # <<<<<<<<<<<<<<
 *     cpdef __enter__(self)
 *     cpdef __exit__(self, exc_type, exc_val, exc_tb)
 */
struct __pyx_obj_5qcore_7helpers_EmptyContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext *__pyx_vtab;
};


/* "qcore/helpers.pxd":40
 * 
 * 
 * cdef class CythonCachedHashWrapper(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 *     cdef int _hash
 */
struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper *__pyx_vtab;
  PyObject *_value;
  int _hash;
};


/* "qcore/helpers.pxd":52
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration
 * 
 * cdef class ScopedValue(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 * 
 */
struct __pyx_obj_5qcore_7helpers_ScopedValue {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue *__pyx_vtab;
  PyObject *_value;
};


/* "qcore/helpers.pxd":50
 * 
 * 
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class ScopedValue(object):
 */
struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext {
  PyObject_HEAD
  struct __pyx_obj_5qcore_7helpers_ScopedValue *_target;
  PyObject *_value;
  PyObject *_old_value;
};


/* "qcore/helpers.pxd":65
 * 
 * 
 * cdef class _PropertyOverrideContext(object):             # <<<<<<<<<<<<<<
 *     cdef object _target
 *     cdef object _property_name
 */
struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext {
  PyObject_HEAD
  PyObject *_target;
  PyObject *_property_name;
  PyObject *_value;
  PyObject *_old_value;
};


/* "_debug.pxd":18
 * 
 * 
 * cdef class DebugOptions(object):             # <<<<<<<<<<<<<<
 *     cdef public bint DUMP_PRE_ERROR_STATE
 *     cdef public bint DUMP_EXCEPTIONS
 */
struct __pyx_obj_5asynq_6_debug_DebugOptions {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions *__pyx_vtab;
  int DUMP_PRE_ERROR_STATE;
  int DUMP_EXCEPTIONS;
  int DUMP_AWAIT_RECURSION;
  int DUMP_SCHEDULE_TASK;
  int DUMP_CONTINUE_TASK;
  int DUMP_SCHEDULE_BATCH;
  int DUMP_FLUSH_BATCH;
  int DUMP_DEPENDENCIES;
  int DUMP_COMPUTED;
  int DUMP_NEW_TASKS;
  int DUMP_YIELD_RESULTS;
  int DUMP_QUEUED_RESULTS;
  int DUMP_CONTEXTS;
  int DUMP_SCHEDULER_CHANGE;
  int DUMP_SYNC;
  int DUMP_PRIMER;
  int DUMP_STACK;
  int DUMP_SCHEDULER_STATE;
  float SCHEDULER_STATE_DUMP_INTERVAL;
  int DEBUG_STR_REPR_MAX_LENGTH;
  PyObject *STACK_DUMP_LIMIT;
  PyObject *ENABLE_COMPLEX_ASSERTIONS;
};


/* "futures.pxd":26
 * 
 * 
 * cdef class FutureBase(object):             # <<<<<<<<<<<<<<
 *     cdef public object _value
 *     cdef public object _error
 */
struct __pyx_obj_5asynq_7futures_FutureBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase *__pyx_vtab;
  PyObject *_value;
  PyObject *_error;
  int _in_repr;
  struct __pyx_obj_5qcore_6events_EventHook *on_computed;
};


/* "futures.pxd":49
 * 
 * 
 * cdef class Future(FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public object _value_provider
 *     cpdef _compute(self)
 */
struct __pyx_obj_5asynq_7futures_Future {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  PyObject *_value_provider;
};


/* "futures.pxd":54
 * 
 * 
 * cdef class ConstFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5asynq_7futures_ConstFuture {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
};


/* "futures.pxd":57
 *     pass
 * 
 * cdef class ErrorFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_5asynq_7futures_ErrorFuture {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
};


/* "batching.pxd":26
 * 
 * 
 * cdef class BatchBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public list items
 * 
 */
struct __pyx_obj_5asynq_8batching_BatchBase {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  PyObject *items;
};


/* "batching.pxd":45
 *     cpdef dump(self, int indent=?)
 * 
 * cdef class BatchItemBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public BatchBase batch
 *     cdef public long index
 */
struct __pyx_obj_5asynq_8batching_BatchItemBase {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  struct __pyx_obj_5asynq_8batching_BatchBase *batch;
  long index;
};


/* "batching.pxd":53
 * 
 * 
 * cdef class DebugBatchItem(BatchItemBase):             # <<<<<<<<<<<<<<
 *     cdef public object _result
 * 
 */
struct __pyx_obj_5asynq_8batching_DebugBatchItem {
  struct __pyx_obj_5asynq_8batching_BatchItemBase __pyx_base;
  PyObject *_result;
};


/* "batching.pxd":56
 *     cdef public object _result
 * 
 * cdef class DebugBatch(BatchBase):             # <<<<<<<<<<<<<<
 *     cdef public str name
 *     cdef public long index
 */
struct __pyx_obj_5asynq_8batching_DebugBatch {
  struct __pyx_obj_5asynq_8batching_BatchBase __pyx_base;
  PyObject *name;
  long index;
};


/* "scheduler.pxd":29
 * 
 * 
 * cdef class TaskScheduler(object):             # <<<<<<<<<<<<<<
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time
 */
struct __pyx_obj_5asynq_9scheduler_TaskScheduler {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *__pyx_vtab;
  PyObject *_empty_set;
  float _last_dump_time;
  PyObject *_new_tasks;
  PyObject *_tasks;
  PyObject *_batches;
  int await_recursion_depth;
  int max_await_recursion_depth;
  struct __pyx_obj_5qcore_6events_EventHook *on_before_batch_flush;
  struct __pyx_obj_5qcore_6events_EventHook *on_after_batch_flush;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *active_task;
  PyObject *name;
};


/* "scheduler.pxd":88
 *     cpdef try_time_based_dump(self, object last_task=?)
 * 
 * cdef class TaskSchedulerActivator(object):             # <<<<<<<<<<<<<<
 *     cdef TaskScheduler _scheduler
 *     cdef TaskScheduler _old_scheduler
 */
struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator {
  PyObject_HEAD
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *_scheduler;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *_old_scheduler;
  int _must_await_on_exit;
  int _must_flush_batches_on_exit;
};


/* "asynq/async_task.pxd":31
 * 
 * 
 * cdef class AsyncTask(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index
 */
struct __pyx_obj_5asynq_10async_task_AsyncTask {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  int is_scheduled;
  long iteration_index;
  int group_cancel;
  PyObject *fn;
  PyObject *args;
  PyObject *kwargs;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *caller;
  int depth;
  int max_depth;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *scheduler;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *creator;
  PyObject *_generator;
  PyObject *_last_value;
  PyObject *_frame_info;
  PyObject *_dependencies;
  PyObject *_contexts;
};



/* "qcore/helpers.pxd":33
 * 
 * 
 * cdef class EmptyContext(object):             # <<<<<<<<<<<<<<
 *     cpdef __enter__(self)
 *     cpdef __exit__(self, exc_type, exc_val, exc_tb)
 */

struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext {
  PyObject *(*__pyx___enter__)(struct __pyx_obj_5qcore_7helpers_EmptyContext *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___exit__)(struct __pyx_obj_5qcore_7helpers_EmptyContext *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext *__pyx_vtabptr_5qcore_7helpers_EmptyContext;


/* "qcore/helpers.pxd":40
 * 
 * 
 * cdef class CythonCachedHashWrapper(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 *     cdef int _hash
 */

struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper {
  PyObject *(*value)(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper *, int __pyx_skip_dispatch);
  PyObject *(*hash)(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper *__pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper;


/* "qcore/helpers.pxd":52
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration
 * 
 * cdef class ScopedValue(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 * 
 */

struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue {
  PyObject *(*get)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, int __pyx_skip_dispatch);
  PyObject *(*set)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*override)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue *__pyx_vtabptr_5qcore_7helpers_ScopedValue;


/* "_debug.pxd":18
 * 
 * 
 * cdef class DebugOptions(object):             # <<<<<<<<<<<<<<
 *     cdef public bint DUMP_PRE_ERROR_STATE
 *     cdef public bint DUMP_EXCEPTIONS
 */

struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions {
  PyObject *(*DUMP_ALL)(struct __pyx_obj_5asynq_6_debug_DebugOptions *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions *__pyx_vtabptr_5asynq_6_debug_DebugOptions;


/* "futures.pxd":26
 * 
 * 
 * cdef class FutureBase(object):             # <<<<<<<<<<<<<<
 *     cdef public object _value
 *     cdef public object _error
 */

struct __pyx_vtabstruct_5asynq_7futures_FutureBase {
  PyObject *(*value)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*set_value)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*reset_unsafe)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*error)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*set_error)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, int __pyx_skip_dispatch);
  int (*is_computed)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*_compute)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*_computed)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*make_dependency)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*raise_if_error)(struct __pyx_obj_5asynq_7futures_FutureBase *);
  PyObject *(*dump)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_7futures_FutureBase *__pyx_vtabptr_5asynq_7futures_FutureBase;


/* "futures.pxd":49
 * 
 * 
 * cdef class Future(FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public object _value_provider
 *     cpdef _compute(self)
 */

struct __pyx_vtabstruct_5asynq_7futures_Future {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_Future *__pyx_vtabptr_5asynq_7futures_Future;


/* "futures.pxd":54
 * 
 * 
 * cdef class ConstFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5asynq_7futures_ConstFuture {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_ConstFuture *__pyx_vtabptr_5asynq_7futures_ConstFuture;


/* "futures.pxd":57
 *     pass
 * 
 * cdef class ErrorFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture *__pyx_vtabptr_5asynq_7futures_ErrorFuture;


/* "batching.pxd":26
 * 
 * 
 * cdef class BatchBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public list items
 * 
 */

struct __pyx_vtabstruct_5asynq_8batching_BatchBase {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
  int (*is_flushed)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*is_cancelled)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*is_empty)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*get_priority)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*flush)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*cancel)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel *__pyx_optional_args);
  PyObject *(*_flush)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*_cancel)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*_try_switch_active_batch)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5asynq_8batching_BatchBase *__pyx_vtabptr_5asynq_8batching_BatchBase;


/* "batching.pxd":45
 *     cpdef dump(self, int indent=?)
 * 
 * cdef class BatchItemBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public BatchBase batch
 *     cdef public long index
 */

struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase *__pyx_vtabptr_5asynq_8batching_BatchItemBase;


/* "batching.pxd":53
 * 
 * 
 * cdef class DebugBatchItem(BatchItemBase):             # <<<<<<<<<<<<<<
 *     cdef public object _result
 * 
 */

struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem {
  struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem *__pyx_vtabptr_5asynq_8batching_DebugBatchItem;


/* "batching.pxd":56
 *     cdef public object _result
 * 
 * cdef class DebugBatch(BatchBase):             # <<<<<<<<<<<<<<
 *     cdef public str name
 *     cdef public long index
 */

struct __pyx_vtabstruct_5asynq_8batching_DebugBatch {
  struct __pyx_vtabstruct_5asynq_8batching_BatchBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_DebugBatch *__pyx_vtabptr_5asynq_8batching_DebugBatch;


/* "scheduler.pxd":29
 * 
 * 
 * cdef class TaskScheduler(object):             # <<<<<<<<<<<<<<
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time
 */

struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler {
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *(*activate)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, int, int __pyx_skip_dispatch);
  PyObject *(*reset)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*start)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start *__pyx_optional_args);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*continue_with)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_7futures_FutureBase *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  int (*flush_batches)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches *__pyx_optional_args);
  PyObject *(*add_dependencies)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_dependency)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*schedule_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*_await)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch);
  int (*_execute)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_try_execute_on_new_scheduler)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch);
  int (*_schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  int (*_schedule_without_checks)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  int (*_try_schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int);
  int (*_schedule_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *);
  int (*_flush_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *);
  PyObject *(*_enqueue_new_tasks)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*_continue_with_task)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);
  struct __pyx_obj_5asynq_8batching_BatchBase *(*_continue_with_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch *__pyx_optional_args);
  struct __pyx_obj_5asynq_8batching_BatchBase *(*_select_batch_to_flush)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*_clone)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*_accept_changes)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*dump)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump *__pyx_optional_args);
  PyObject *(*try_time_based_dump)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *__pyx_vtabptr_5asynq_9scheduler_TaskScheduler;


/* "asynq/async_task.py":47
 * 
 * 
 * class AsyncTask(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     """Asynchronous task implementation.
 *     Uses passed generator (normally produced by generator method)
 */

struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
  int (*is_blocked)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  int (*can_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*start)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start *__pyx_optional_args);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*after)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*cancel_dependencies)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*_continue_on_generator)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, PyObject *);
  PyObject *(*_accept_yield_result)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_accept_error)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_queue_exit)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_queue_throw_error)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_remove_dependency_cython)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_before_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*_after_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*traceback)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*_enter_context)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_leave_context)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_pause_contexts)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*_resume_contexts)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
};
static struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *__pyx_vtabptr_5asynq_10async_task_AsyncTask;
static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__continue_on_generator(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__accept_yield_result(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__accept_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* PyThreadStateGet.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SaveResetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyIntBinop.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* ListCompAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_dealloc);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_5asynq_10async_task_9AsyncTask_can_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5asynq_10async_task_9AsyncTask_is_blocked(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_10async_task_9AsyncTask_start(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_10async_task_9AsyncTask_after(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_cancel_dependencies(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__compute(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__computed(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__continue_on_generator(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_error); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__accept_yield_result(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_result); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__accept_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__queue_exit(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_result, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__queue_throw_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__remove_dependency_cython(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_dependency); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_make_dependency(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_task, PyObject *__pyx_v_scheduler, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__before_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__after_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_traceback(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_dump(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__enter_context(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__leave_context(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_context); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__pause_contexts(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__resume_contexts(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'qcore.events' */
static PyTypeObject *__pyx_ptype_5qcore_6events_EventHook = 0;
static PyTypeObject *__pyx_ptype_5qcore_6events_SinkingEventHook = 0;
static PyTypeObject *__pyx_ptype_5qcore_6events_EventInterceptor = 0;
static struct __pyx_obj_5qcore_6events_SinkingEventHook **__pyx_vp_5qcore_6events_sinking_event_hook = 0;
#define __pyx_v_5qcore_6events_sinking_event_hook (*__pyx_vp_5qcore_6events_sinking_event_hook)

/* Module declarations from 'qcore.helpers' */
static PyTypeObject *__pyx_ptype_5qcore_7helpers_MarkerObject = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_EmptyContext = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_ScopedValue = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers__PropertyOverrideContext = 0;
static PyObject **__pyx_vp_5qcore_7helpers_empty_list = 0;
#define __pyx_v_5qcore_7helpers_empty_list (*__pyx_vp_5qcore_7helpers_empty_list)
static PyObject **__pyx_vp_5qcore_7helpers_empty_tuple = 0;
#define __pyx_v_5qcore_7helpers_empty_tuple (*__pyx_vp_5qcore_7helpers_empty_tuple)
static PyObject **__pyx_vp_5qcore_7helpers_empty_dict = 0;
#define __pyx_v_5qcore_7helpers_empty_dict (*__pyx_vp_5qcore_7helpers_empty_dict)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_none = 0;
#define __pyx_v_5qcore_7helpers_none (*__pyx_vp_5qcore_7helpers_none)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_miss = 0;
#define __pyx_v_5qcore_7helpers_miss (*__pyx_vp_5qcore_7helpers_miss)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_same = 0;
#define __pyx_v_5qcore_7helpers_same (*__pyx_vp_5qcore_7helpers_same)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_unspecified = 0;
#define __pyx_v_5qcore_7helpers_unspecified (*__pyx_vp_5qcore_7helpers_unspecified)
static struct __pyx_obj_5qcore_7helpers_EmptyContext **__pyx_vp_5qcore_7helpers_empty_context = 0;
#define __pyx_v_5qcore_7helpers_empty_context (*__pyx_vp_5qcore_7helpers_empty_context)
static PyObject **__pyx_vp_5qcore_7helpers_CachedHashWrapper = 0;
#define __pyx_v_5qcore_7helpers_CachedHashWrapper (*__pyx_vp_5qcore_7helpers_CachedHashWrapper)
static PyObject **__pyx_vp_5qcore_7helpers_override = 0;
#define __pyx_v_5qcore_7helpers_override (*__pyx_vp_5qcore_7helpers_override)
static PyObject *(*__pyx_f_5qcore_7helpers_safe_str)(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5qcore_7helpers_safe_str *__pyx_optional_args); /*proto*/

/* Module declarations from 'qcore.inspection' */
static PyObject *(*__pyx_f_5qcore_10inspection_get_function_call_str)(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static int (*__pyx_f_5qcore_10inspection_is_cython_or_generator)(PyObject *, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'asynq._debug' */
static PyTypeObject *__pyx_ptype_5asynq_6_debug_DebugOptions = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_6_debug_options = 0;
#define __pyx_v_5asynq_6_debug_options (*__pyx_vp_5asynq_6_debug_options)

/* Module declarations from 'asynq.futures' */
static PyTypeObject *__pyx_ptype_5asynq_7futures_FutureBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_Future = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_ConstFuture = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_ErrorFuture = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_7futures__debug_options = 0;
#define __pyx_v_5asynq_7futures__debug_options (*__pyx_vp_5asynq_7futures__debug_options)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5asynq_7futures__none = 0;
#define __pyx_v_5asynq_7futures__none (*__pyx_vp_5asynq_7futures__none)

/* Module declarations from 'asynq.batching' */
static PyTypeObject *__pyx_ptype_5asynq_8batching_BatchBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_BatchItemBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_DebugBatchItem = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_DebugBatch = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_8batching__debug_options = 0;
#define __pyx_v_5asynq_8batching__debug_options (*__pyx_vp_5asynq_8batching__debug_options)
static PyObject **__pyx_vp_5asynq_8batching__debug_batch_state = 0;
#define __pyx_v_5asynq_8batching__debug_batch_state (*__pyx_vp_5asynq_8batching__debug_batch_state)

/* Module declarations from 'asynq.scheduler' */
static PyTypeObject *__pyx_ptype_5asynq_9scheduler_TaskScheduler = 0;
static PyTypeObject *__pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_9scheduler__debug_options = 0;
#define __pyx_v_5asynq_9scheduler__debug_options (*__pyx_vp_5asynq_9scheduler__debug_options)
static PyObject **__pyx_vp_5asynq_9scheduler__futures_none = 0;
#define __pyx_v_5asynq_9scheduler__futures_none (*__pyx_vp_5asynq_9scheduler__futures_none)
static PyObject **__pyx_vp_5asynq_9scheduler__state = 0;
#define __pyx_v_5asynq_9scheduler__state (*__pyx_vp_5asynq_9scheduler__state)
static struct __pyx_obj_5asynq_9scheduler_TaskScheduler *(*__pyx_f_5asynq_9scheduler_get_scheduler)(int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *(*__pyx_f_5asynq_9scheduler_get_active_task)(int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'asynq.async_task' */
static PyTypeObject *__pyx_ptype_5asynq_10async_task_AsyncTask = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions *__pyx_v_5asynq_10async_task__debug_options = 0;
static PyObject *__pyx_v_5asynq_10async_task__futures_none = 0;
static PyObject *__pyx_v_5asynq_10async_task__none_future = 0;
static PyObject *__pyx_v_5asynq_10async_task__empty_tuple = 0;
static PyObject *__pyx_v_5asynq_10async_task__empty_dictionary = 0;
static PyObject *__pyx_f_5asynq_10async_task_unwrap(PyObject *); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task__cancel_futures(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "asynq.async_task"
int __pyx_module_is_main_asynq__async_task = 0;

/* Implementation of 'asynq.async_task' */
static PyObject *__pyx_builtin_GeneratorExit;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_BaseException;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_Exception;
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_[] = "\n";
static const char __pyx_k__2[] = "= ";
static const char __pyx_k__3[] = "...";
static const char __pyx_k__4[] = "*";
static const char __pyx_k__5[] = "";
static const char __pyx_k_fn[] = "fn";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_new[] = "new";
static const char __pyx_k_six[] = "six";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dump[] = "dump";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_repr[] = "repr";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_task[] = "_task";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "_type_";
static const char __pyx_k_after[] = "after";
static const char __pyx_k_await[] = "await";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_depth[] = "depth";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_pause[] = "__pause__";
static const char __pyx_k_s_s_s[] = "%s (%s, %s)";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indent[] = "indent";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_lineno[] = "lineno";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_resume[] = "__resume__";
static const char __pyx_k_task_2[] = "task";
static const char __pyx_k_async_s[] = "@async %s";
static const char __pyx_k_compute[] = "_compute";
static const char __pyx_k_debug_2[] = "_debug";
static const char __pyx_k_error_2[] = "error = ";
static const char __pyx_k_futures[] = "futures";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_tb_next[] = "tb_next";
static const char __pyx_k_waiting[] = "waiting";
static const char __pyx_k_codeline[] = "codeline";
static const char __pyx_k_computed[] = "_computed";
static const char __pyx_k_continue[] = "_continue";
static const char __pyx_k_exc_info[] = "exc_info";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_funcname[] = "funcname";
static const char __pyx_k_function[] = "function";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_schedule[] = "schedule";
static const char __pyx_k_tb_frame[] = "tb_frame";
static const char __pyx_k_Exception[] = "Exception";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_generator[] = "generator";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_max_depth[] = "max_depth";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_scheduled[] = "scheduled";
static const char __pyx_k_scheduler[] = "scheduler";
static const char __pyx_k_set_error[] = "set_error";
static const char __pyx_k_subscribe[] = "subscribe";
static const char __pyx_k_traceback[] = "_traceback";
static const char __pyx_k_computed_2[] = "computed, ";
static const char __pyx_k_dump_error[] = "dump_error";
static const char __pyx_k_is_blocked[] = "is_blocked";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_queue_exit[] = "_queue_exit";
static const char __pyx_k_blocked_x_i[] = "blocked x%i";
static const char __pyx_k_core_errors[] = "core_errors";
static const char __pyx_k_empty_tuple[] = "_empty_tuple";
static const char __pyx_k_is_computed[] = "is_computed";
static const char __pyx_k_none_future[] = "_none_future";
static const char __pyx_k_scheduled_2[] = ", scheduled";
static const char __pyx_k_traceback_2[] = "traceback";
static const char __pyx_k_Dependencies[] = "Dependencies:";
static const char __pyx_k_can_continue[] = "can_continue";
static const char __pyx_k_code_context[] = "code_context";
static const char __pyx_k_core_helpers[] = "core_helpers";
static const char __pyx_k_dependencies[] = "_dependencies";
static const char __pyx_k_futures_none[] = "_futures_none";
static const char __pyx_k_getframeinfo[] = "getframeinfo";
static const char __pyx_k_group_cancel[] = "group_cancel";
static const char __pyx_k_qcore_errors[] = "qcore.errors";
static const char __pyx_k_BaseException[] = "BaseException";
static const char __pyx_k_GeneratorExit[] = "GeneratorExit";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_debug_options[] = "_debug_options";
static const char __pyx_k_none_future_2[] = "none_future";
static const char __pyx_k_qcore_helpers[] = "qcore.helpers";
static const char __pyx_k_get_frame_info[] = "get_frame_info";
static const char __pyx_k_passed_yield_i[] = "passed yield #%i";
static const char __pyx_k_traceback_hide[] = "__traceback_hide__";
static const char __pyx_k_traceback_line[] = "_traceback_line";
static const char __pyx_k_AsyncTaskResult[] = "AsyncTaskResult";
static const char __pyx_k_MAX_DUMP_INDENT[] = "MAX_DUMP_INDENT";
static const char __pyx_k_No_dependencies[] = "No dependencies.";
static const char __pyx_k_async_context_s[] = "@async: +context: %s";
static const char __pyx_k_async_yield_s_s[] = "@async: yield: %s -> %s";
static const char __pyx_k_core_inspection[] = "core_inspection";
static const char __pyx_k_make_dependency[] = "make_dependency";
static const char __pyx_k_run_immediately[] = "run_immediately";
static const char __pyx_k_asynq_async_task[] = "asynq.async_task";
static const char __pyx_k_before_1st_yield[] = "before 1st yield";
static const char __pyx_k_empty_dictionary[] = "_empty_dictionary";
static const char __pyx_k_qcore_inspection[] = "qcore.inspection";
static const char __pyx_k_async_context_s_2[] = "@async: -context: %s";
static const char __pyx_k_queue_throw_error[] = "_queue_throw_error";
static const char __pyx_k_remove_dependency[] = "_remove_dependency";
static const char __pyx_k_cancel_dependencies[] = "cancel_dependencies";
static const char __pyx_k_prepare_for_reraise[] = "prepare_for_reraise";
static const char __pyx_k_AsyncTaskResult___init[] = "AsyncTaskResult.__init__";
static const char __pyx_k_async_queuing_exit_s_s[] = "@async: queuing exit: %s <- %s";
static const char __pyx_k_AsyncTaskCancelledError[] = "AsyncTaskCancelledError";
static const char __pyx_k_FutureIsAlreadyComputed[] = "FutureIsAlreadyComputed";
static const char __pyx_k_async_dependency_s_got_s[] = "@async: -dependency: %s got %s";
static const char __pyx_k_async_new_task_s_created_by_s[] = "@async: new task: %s, created by %s";
static const char __pyx_k_Cannot_unwrap_an_object_of_type[] = "Cannot unwrap an object of type '%s': only futures and None are allowed.";
static const char __pyx_k_async_cancelled_dependencies_of[] = "@async: <- cancelled dependencies of %s";
static const char __pyx_k_async_cancelling_dependencies_o[] = "@async: -> cancelling dependencies of %s:";
static const char __pyx_k_async_queuing_throw_error_s_x_s[] = "@async: queuing throw error: %s <-x- %s";
static const char __pyx_k_home_jelle_github_asynq_asynq_a[] = "/home/jelle/github/asynq/asynq/async_task.py";
static const char __pyx_k_File_file_s_line_lineno_s_in_fun[] = "File \"%(file)s\", line %(lineno)s, in %(funcname)s\n    %(codeline)s";
static const char __pyx_k_Task_stack_depth_exceeded_specif[] = "Task stack depth exceeded specified maximum (%i)";
static const char __pyx_k_almost_finished_generator_is_clo[] = "almost finished (generator is closed)";
static const char __pyx_k_generator_is_expected_but_s_is_p[] = "generator is expected, but %s is provided";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_n_s_AsyncTaskCancelledError;
static PyObject *__pyx_n_s_AsyncTaskResult;
static PyObject *__pyx_n_s_AsyncTaskResult___init;
static PyObject *__pyx_n_s_BaseException;
static PyObject *__pyx_kp_s_Cannot_unwrap_an_object_of_type;
static PyObject *__pyx_kp_s_Dependencies;
static PyObject *__pyx_n_s_Exception;
static PyObject *__pyx_kp_s_File_file_s_line_lineno_s_in_fun;
static PyObject *__pyx_n_s_FutureIsAlreadyComputed;
static PyObject *__pyx_n_s_GeneratorExit;
static PyObject *__pyx_n_s_MAX_DUMP_INDENT;
static PyObject *__pyx_kp_s_No_dependencies;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_kp_s_Task_stack_depth_exceeded_specif;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_n_s__4;
static PyObject *__pyx_n_s__5;
static PyObject *__pyx_n_s_add;
static PyObject *__pyx_n_s_after;
static PyObject *__pyx_kp_s_almost_finished_generator_is_clo;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_kp_s_async_cancelled_dependencies_of;
static PyObject *__pyx_kp_s_async_cancelling_dependencies_o;
static PyObject *__pyx_kp_s_async_context_s;
static PyObject *__pyx_kp_s_async_context_s_2;
static PyObject *__pyx_kp_s_async_dependency_s_got_s;
static PyObject *__pyx_kp_s_async_new_task_s_created_by_s;
static PyObject *__pyx_kp_s_async_queuing_exit_s_s;
static PyObject *__pyx_kp_s_async_queuing_throw_error_s_x_s;
static PyObject *__pyx_kp_s_async_s;
static PyObject *__pyx_kp_s_async_yield_s_s;
static PyObject *__pyx_n_s_asynq_async_task;
static PyObject *__pyx_n_s_await;
static PyObject *__pyx_kp_s_before_1st_yield;
static PyObject *__pyx_kp_s_blocked_x_i;
static PyObject *__pyx_n_s_can_continue;
static PyObject *__pyx_n_s_cancel_dependencies;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_code_context;
static PyObject *__pyx_n_s_codeline;
static PyObject *__pyx_n_s_compute;
static PyObject *__pyx_n_s_computed;
static PyObject *__pyx_kp_s_computed_2;
static PyObject *__pyx_n_s_continue;
static PyObject *__pyx_n_s_core_errors;
static PyObject *__pyx_n_s_core_helpers;
static PyObject *__pyx_n_s_core_inspection;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_debug_2;
static PyObject *__pyx_n_s_debug_options;
static PyObject *__pyx_n_s_dependencies;
static PyObject *__pyx_n_s_depth;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dump;
static PyObject *__pyx_n_s_dump_error;
static PyObject *__pyx_n_s_empty_dictionary;
static PyObject *__pyx_n_s_empty_tuple;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_kp_s_error_2;
static PyObject *__pyx_n_s_exc_info;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_filename;
static PyObject *__pyx_n_s_fn;
static PyObject *__pyx_n_s_funcname;
static PyObject *__pyx_n_s_function;
static PyObject *__pyx_n_s_futures;
static PyObject *__pyx_n_s_futures_none;
static PyObject *__pyx_n_s_generator;
static PyObject *__pyx_kp_s_generator_is_expected_but_s_is_p;
static PyObject *__pyx_n_s_get_frame_info;
static PyObject *__pyx_n_s_getframeinfo;
static PyObject *__pyx_n_s_group_cancel;
static PyObject *__pyx_kp_s_home_jelle_github_asynq_asynq_a;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_indent;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_inspect;
static PyObject *__pyx_n_s_is_blocked;
static PyObject *__pyx_n_s_is_computed;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_lineno;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_dependency;
static PyObject *__pyx_n_s_max_depth;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_none_future;
static PyObject *__pyx_n_s_none_future_2;
static PyObject *__pyx_kp_s_passed_yield_i;
static PyObject *__pyx_n_s_pause;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_prepare_for_reraise;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qcore_errors;
static PyObject *__pyx_n_s_qcore_helpers;
static PyObject *__pyx_n_s_qcore_inspection;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_queue_exit;
static PyObject *__pyx_n_s_queue_throw_error;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_remove_dependency;
static PyObject *__pyx_n_s_repr;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_resume;
static PyObject *__pyx_n_s_run_immediately;
static PyObject *__pyx_kp_s_s_s_s;
static PyObject *__pyx_n_s_schedule;
static PyObject *__pyx_n_s_scheduled;
static PyObject *__pyx_kp_s_scheduled_2;
static PyObject *__pyx_n_s_scheduler;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set_error;
static PyObject *__pyx_n_s_six;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_subscribe;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_task;
static PyObject *__pyx_n_s_task_2;
static PyObject *__pyx_n_s_tb_frame;
static PyObject *__pyx_n_s_tb_next;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_traceback;
static PyObject *__pyx_n_s_traceback_2;
static PyObject *__pyx_n_s_traceback_hide;
static PyObject *__pyx_n_s_traceback_line;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_waiting;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_pf_5asynq_10async_task_15AsyncTaskResult___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_result); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask___init__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_generator, PyObject *__pyx_v_fn, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, PyObject *__pyx_v_group_cancel, PyObject *__pyx_v_max_depth); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_2can_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_4is_blocked(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_6start(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_v_run_immediately); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_8after(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_10cancel_dependencies(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_12_compute(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_14_computed(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_16_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_18_queue_exit(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_20_queue_throw_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_22_remove_dependency(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_dependency); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_24make_dependency(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_task, PyObject *__pyx_v_scheduler); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_26traceback(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_28_traceback_line(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_30__str__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_32dump(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_v_indent); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_12is_scheduled___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_12is_scheduled_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_15iteration_index___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_15iteration_index_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_12group_cancel___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_12group_cancel_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_2fn___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_2fn_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_2fn_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_4args___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_4args_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_4args_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_6caller___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_6caller_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_6caller_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_5depth___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_5depth_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_9max_depth___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_9max_depth_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_7creator___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_7creator_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_7creator_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_10_generator___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_10async_task__cancel_futures(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value, PyObject *__pyx_v_error); /* proto */
static PyObject *__pyx_tp_new_5asynq_10async_task_AsyncTask(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_40;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_codeobj__7;

/* "asynq/async_task.py":42
 * 
 * class AsyncTaskResult(GeneratorExit):
 *     def __init__(self, result):             # <<<<<<<<<<<<<<
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 *         self.result = result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_15AsyncTaskResult_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5asynq_10async_task_15AsyncTaskResult_1__init__ = {"__init__", (PyCFunction)__pyx_pw_5asynq_10async_task_15AsyncTaskResult_1__init__, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5asynq_10async_task_15AsyncTaskResult_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 42, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 42, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_result = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 42, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTaskResult.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_10async_task_15AsyncTaskResult___init__(__pyx_self, __pyx_v_self, __pyx_v_result);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_15AsyncTaskResult___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "asynq/async_task.py":43
 * class AsyncTaskResult(GeneratorExit):
 *     def __init__(self, result):
 *         GeneratorExit.__init__(self, "AsyncTaskResult")             # <<<<<<<<<<<<<<
 *         self.result = result
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_builtin_GeneratorExit, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_5 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3) {
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_v_self);
  __Pyx_INCREF(__pyx_n_s_AsyncTaskResult);
  __Pyx_GIVEREF(__pyx_n_s_AsyncTaskResult);
  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_n_s_AsyncTaskResult);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":44
 *     def __init__(self, result):
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 *         self.result = result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_result, __pyx_v_result) < 0) __PYX_ERR(0, 44, __pyx_L1_error)

  /* "asynq/async_task.py":42
 * 
 * class AsyncTaskResult(GeneratorExit):
 *     def __init__(self, result):             # <<<<<<<<<<<<<<
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 *         self.result = result
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.async_task.AsyncTaskResult.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":54
 *     """
 * 
 *     def __init__(self, generator, fn, args, kwargs, group_cancel=True, max_depth=0):             # <<<<<<<<<<<<<<
 *         super(AsyncTask, self).__init__()
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:
 */

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_generator = 0;
  PyObject *__pyx_v_fn = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_v_group_cancel = 0;
  PyObject *__pyx_v_max_depth = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_generator,&__pyx_n_s_fn,&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_group_cancel,&__pyx_n_s_max_depth,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_True);
    values[5] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_generator)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_fn)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 1); __PYX_ERR(0, 54, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 2); __PYX_ERR(0, 54, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 3); __PYX_ERR(0, 54, __pyx_L3_error)
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_group_cancel);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_max_depth);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 54, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_generator = values[0];
    __pyx_v_fn = values[1];
    __pyx_v_args = values[2];
    __pyx_v_kwargs = values[3];
    __pyx_v_group_cancel = values[4];
    __pyx_v_max_depth = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 54, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask___init__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), __pyx_v_generator, __pyx_v_fn, __pyx_v_args, __pyx_v_kwargs, __pyx_v_group_cancel, __pyx_v_max_depth);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask___init__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_generator, PyObject *__pyx_v_fn, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, PyObject *__pyx_v_group_cancel, PyObject *__pyx_v_max_depth) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "asynq/async_task.py":55
 * 
 *     def __init__(self, generator, fn, args, kwargs, group_cancel=True, max_depth=0):
 *         super(AsyncTask, self).__init__()             # <<<<<<<<<<<<<<
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:
 *             assert core_inspection.is_cython_or_generator(generator), \
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_5asynq_10async_task_AsyncTask));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_5asynq_10async_task_AsyncTask));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_5asynq_10async_task_AsyncTask));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":56
 *     def __init__(self, generator, fn, args, kwargs, group_cancel=True, max_depth=0):
 *         super(AsyncTask, self).__init__()
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:             # <<<<<<<<<<<<<<
 *             assert core_inspection.is_cython_or_generator(generator), \
 *                 'generator is expected, but %s is provided' % generator
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_5asynq_10async_task__debug_options->ENABLE_COMPLEX_ASSERTIONS); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 56, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "asynq/async_task.py":57
 *         super(AsyncTask, self).__init__()
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:
 *             assert core_inspection.is_cython_or_generator(generator), \             # <<<<<<<<<<<<<<
 *                 'generator is expected, but %s is provided' % generator
 *         self.fn = fn
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_4 = __pyx_f_5qcore_10inspection_is_cython_or_generator(__pyx_v_generator, 0); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 57, __pyx_L1_error)
      if (unlikely(!(__pyx_t_4 != 0))) {

        /* "asynq/async_task.py":58
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:
 *             assert core_inspection.is_cython_or_generator(generator), \
 *                 'generator is expected, but %s is provided' % generator             # <<<<<<<<<<<<<<
 *         self.fn = fn
 *         self.args = args
 */
        __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_generator_is_expected_but_s_is_p, __pyx_v_generator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        PyErr_SetObject(PyExc_AssertionError, __pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __PYX_ERR(0, 57, __pyx_L1_error)
      }
    }
    #endif

    /* "asynq/async_task.py":56
 *     def __init__(self, generator, fn, args, kwargs, group_cancel=True, max_depth=0):
 *         super(AsyncTask, self).__init__()
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:             # <<<<<<<<<<<<<<
 *             assert core_inspection.is_cython_or_generator(generator), \
 *                 'generator is expected, but %s is provided' % generator
 */
  }

  /* "asynq/async_task.py":59
 *             assert core_inspection.is_cython_or_generator(generator), \
 *                 'generator is expected, but %s is provided' % generator
 *         self.fn = fn             # <<<<<<<<<<<<<<
 *         self.args = args
 *         self.kwargs = kwargs
 */
  __Pyx_INCREF(__pyx_v_fn);
  __Pyx_GIVEREF(__pyx_v_fn);
  __Pyx_GOTREF(__pyx_v_self->fn);
  __Pyx_DECREF(__pyx_v_self->fn);
  __pyx_v_self->fn = __pyx_v_fn;

  /* "asynq/async_task.py":60
 *                 'generator is expected, but %s is provided' % generator
 *         self.fn = fn
 *         self.args = args             # <<<<<<<<<<<<<<
 *         self.kwargs = kwargs
 *         self.is_scheduled = False
 */
  __Pyx_INCREF(__pyx_v_args);
  __Pyx_GIVEREF(__pyx_v_args);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = __pyx_v_args;

  /* "asynq/async_task.py":61
 *         self.fn = fn
 *         self.args = args
 *         self.kwargs = kwargs             # <<<<<<<<<<<<<<
 *         self.is_scheduled = False
 *         self.iteration_index = 0
 */
  __Pyx_INCREF(__pyx_v_kwargs);
  __Pyx_GIVEREF(__pyx_v_kwargs);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = __pyx_v_kwargs;

  /* "asynq/async_task.py":62
 *         self.args = args
 *         self.kwargs = kwargs
 *         self.is_scheduled = False             # <<<<<<<<<<<<<<
 *         self.iteration_index = 0
 *         self.group_cancel = group_cancel
 */
  __pyx_v_self->is_scheduled = 0;

  /* "asynq/async_task.py":63
 *         self.kwargs = kwargs
 *         self.is_scheduled = False
 *         self.iteration_index = 0             # <<<<<<<<<<<<<<
 *         self.group_cancel = group_cancel
 *         self.caller = None
 */
  __pyx_v_self->iteration_index = 0;

  /* "asynq/async_task.py":64
 *         self.is_scheduled = False
 *         self.iteration_index = 0
 *         self.group_cancel = group_cancel             # <<<<<<<<<<<<<<
 *         self.caller = None
 *         self.depth = 0
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_group_cancel); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L1_error)
  __pyx_v_self->group_cancel = __pyx_t_4;

  /* "asynq/async_task.py":65
 *         self.iteration_index = 0
 *         self.group_cancel = group_cancel
 *         self.caller = None             # <<<<<<<<<<<<<<
 *         self.depth = 0
 *         self.max_depth = max_depth
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->caller);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->caller));
  __pyx_v_self->caller = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);

  /* "asynq/async_task.py":66
 *         self.group_cancel = group_cancel
 *         self.caller = None
 *         self.depth = 0             # <<<<<<<<<<<<<<
 *         self.max_depth = max_depth
 *         self.scheduler = None
 */
  __pyx_v_self->depth = 0;

  /* "asynq/async_task.py":67
 *         self.caller = None
 *         self.depth = 0
 *         self.max_depth = max_depth             # <<<<<<<<<<<<<<
 *         self.scheduler = None
 *         self.creator = scheduler.get_active_task()
 */
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_max_depth); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_v_self->max_depth = __pyx_t_5;

  /* "asynq/async_task.py":68
 *         self.depth = 0
 *         self.max_depth = max_depth
 *         self.scheduler = None             # <<<<<<<<<<<<<<
 *         self.creator = scheduler.get_active_task()
 *         self._generator = generator
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->scheduler);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scheduler));
  __pyx_v_self->scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None);

  /* "asynq/async_task.py":69
 *         self.max_depth = max_depth
 *         self.scheduler = None
 *         self.creator = scheduler.get_active_task()             # <<<<<<<<<<<<<<
 *         self._generator = generator
 *         self._frame_info = None
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_active_task(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->creator);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->creator));
  __pyx_v_self->creator = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":70
 *         self.scheduler = None
 *         self.creator = scheduler.get_active_task()
 *         self._generator = generator             # <<<<<<<<<<<<<<
 *         self._frame_info = None
 *         self._last_value = None
 */
  __Pyx_INCREF(__pyx_v_generator);
  __Pyx_GIVEREF(__pyx_v_generator);
  __Pyx_GOTREF(__pyx_v_self->_generator);
  __Pyx_DECREF(__pyx_v_self->_generator);
  __pyx_v_self->_generator = __pyx_v_generator;

  /* "asynq/async_task.py":71
 *         self.creator = scheduler.get_active_task()
 *         self._generator = generator
 *         self._frame_info = None             # <<<<<<<<<<<<<<
 *         self._last_value = None
 *         self._dependencies = set()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_frame_info);
  __Pyx_DECREF(__pyx_v_self->_frame_info);
  __pyx_v_self->_frame_info = Py_None;

  /* "asynq/async_task.py":72
 *         self._generator = generator
 *         self._frame_info = None
 *         self._last_value = None             # <<<<<<<<<<<<<<
 *         self._dependencies = set()
 *         if self.creator is None:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_value);
  __Pyx_DECREF(__pyx_v_self->_last_value);
  __pyx_v_self->_last_value = Py_None;

  /* "asynq/async_task.py":73
 *         self._frame_info = None
 *         self._last_value = None
 *         self._dependencies = set()             # <<<<<<<<<<<<<<
 *         if self.creator is None:
 *             self._contexts = []
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_dependencies);
  __Pyx_DECREF(__pyx_v_self->_dependencies);
  __pyx_v_self->_dependencies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":74
 *         self._last_value = None
 *         self._dependencies = set()
 *         if self.creator is None:             # <<<<<<<<<<<<<<
 *             self._contexts = []
 *         else:
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_self->creator) == Py_None);
  __pyx_t_6 = (__pyx_t_4 != 0);
  if (__pyx_t_6) {

    /* "asynq/async_task.py":75
 *         self._dependencies = set()
 *         if self.creator is None:
 *             self._contexts = []             # <<<<<<<<<<<<<<
 *         else:
 *             self._contexts = list(self.creator._contexts)
 */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_contexts);
    __Pyx_DECREF(__pyx_v_self->_contexts);
    __pyx_v_self->_contexts = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynq/async_task.py":74
 *         self._last_value = None
 *         self._dependencies = set()
 *         if self.creator is None:             # <<<<<<<<<<<<<<
 *             self._contexts = []
 *         else:
 */
    goto __pyx_L4;
  }

  /* "asynq/async_task.py":77
 *             self._contexts = []
 *         else:
 *             self._contexts = list(self.creator._contexts)             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_NEW_TASKS:
 *             debug.write('@async: new task: %s, created by %s' %
 */
  /*else*/ {
    __pyx_t_1 = PySequence_List(__pyx_v_self->creator->_contexts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_contexts);
    __Pyx_DECREF(__pyx_v_self->_contexts);
    __pyx_v_self->_contexts = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "asynq/async_task.py":78
 *         else:
 *             self._contexts = list(self.creator._contexts)
 *         if _debug_options.DUMP_NEW_TASKS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: new task: %s, created by %s' %
 *                 (debug.str(self), debug.str(self.creator)))
 */
  __pyx_t_6 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_NEW_TASKS != 0);
  if (__pyx_t_6) {

    /* "asynq/async_task.py":79
 *             self._contexts = list(self.creator._contexts)
 *         if _debug_options.DUMP_NEW_TASKS:
 *             debug.write('@async: new task: %s, created by %s' %             # <<<<<<<<<<<<<<
 *                 (debug.str(self), debug.str(self.creator)))
 * 
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/async_task.py":80
 *         if _debug_options.DUMP_NEW_TASKS:
 *             debug.write('@async: new task: %s, created by %s' %
 *                 (debug.str(self), debug.str(self.creator)))             # <<<<<<<<<<<<<<
 * 
 *     def can_continue(self):
 */
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_str); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_8, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_9, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_str); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (!__pyx_t_9) {
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_self->creator)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
    } else {
      __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self->creator));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self->creator));
      PyTuple_SET_ITEM(__pyx_t_10, 0+1, ((PyObject *)__pyx_v_self->creator));
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
    __pyx_t_2 = 0;
    __pyx_t_8 = 0;

    /* "asynq/async_task.py":79
 *             self._contexts = list(self.creator._contexts)
 *         if _debug_options.DUMP_NEW_TASKS:
 *             debug.write('@async: new task: %s, created by %s' %             # <<<<<<<<<<<<<<
 *                 (debug.str(self), debug.str(self.creator)))
 * 
 */
    __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_async_new_task_s_created_by_s, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_8);
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/async_task.py":78
 *         else:
 *             self._contexts = list(self.creator._contexts)
 *         if _debug_options.DUMP_NEW_TASKS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: new task: %s, created by %s' %
 *                 (debug.str(self), debug.str(self.creator)))
 */
  }

  /* "asynq/async_task.py":54
 *     """
 * 
 *     def __init__(self, generator, fn, args, kwargs, group_cancel=True, max_depth=0):             # <<<<<<<<<<<<<<
 *         super(AsyncTask, self).__init__()
 *         if _debug_options.ENABLE_COMPLEX_ASSERTIONS:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":82
 *                 (debug.str(self), debug.str(self.creator)))
 * 
 *     def can_continue(self):             # <<<<<<<<<<<<<<
 *         """Indicates whether this async task has more steps to execute.
 * 
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_3can_continue(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_5asynq_10async_task_9AsyncTask_can_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("can_continue", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_can_continue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_3can_continue)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":88
 * 
 *         """
 *         return self._generator is not None             # <<<<<<<<<<<<<<
 * 
 *     def is_blocked(self):
 */
  __pyx_t_5 = (__pyx_v_self->_generator != Py_None);
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "asynq/async_task.py":82
 *                 (debug.str(self), debug.str(self.creator)))
 * 
 *     def can_continue(self):             # <<<<<<<<<<<<<<
 *         """Indicates whether this async task has more steps to execute.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.can_continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_3can_continue(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5asynq_10async_task_9AsyncTask_2can_continue[] = "Indicates whether this async task has more steps to execute.\n\n        Task can't continue, if its underlying generator is disposed.\n\n        ";
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_3can_continue(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("can_continue (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_2can_continue(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_2can_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("can_continue", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask_can_continue(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.can_continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":90
 *         return self._generator is not None
 * 
 *     def is_blocked(self):             # <<<<<<<<<<<<<<
 *         """Indicates whether this async task is currently blocked.
 *         Blocked tasks are tasks waiting for completion of other
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_5is_blocked(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_5asynq_10async_task_9AsyncTask_is_blocked(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("is_blocked", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_blocked); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_5is_blocked)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":96
 * 
 *         """
 *         return True if self._dependencies else False             # <<<<<<<<<<<<<<
 * 
 *     def start(self, run_immediately=False):
 */
  __pyx_t_6 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
  if (__pyx_t_6) {
    __pyx_t_5 = 1;
  } else {
    __pyx_t_5 = 0;
  }
  __pyx_r = __pyx_t_5;
  goto __pyx_L0;

  /* "asynq/async_task.py":90
 *         return self._generator is not None
 * 
 *     def is_blocked(self):             # <<<<<<<<<<<<<<
 *         """Indicates whether this async task is currently blocked.
 *         Blocked tasks are tasks waiting for completion of other
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.is_blocked", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_5is_blocked(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5asynq_10async_task_9AsyncTask_4is_blocked[] = "Indicates whether this async task is currently blocked.\n        Blocked tasks are tasks waiting for completion of other\n        tasks or batches.\n\n        ";
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_5is_blocked(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_blocked (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_4is_blocked(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_4is_blocked(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("is_blocked", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask_is_blocked(__pyx_v_self, 1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.is_blocked", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":98
 *         return True if self._dependencies else False
 * 
 *     def start(self, run_immediately=False):             # <<<<<<<<<<<<<<
 *         """
 *         Starts the task on the current task scheduler.
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_7start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_10async_task_9AsyncTask_start(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start *__pyx_optional_args) {
  int __pyx_v_run_immediately = ((int)0);
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_s = 0;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start __pyx_t_7;
  __Pyx_RefNannySetupContext("start", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_run_immediately = __pyx_optional_args->run_immediately;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_7start)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_run_immediately); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 98, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":108
 * 
 *         """
 *         s = scheduler.get_scheduler()             # <<<<<<<<<<<<<<
 *         return s.start(self, run_immediately)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":109
 *         """
 *         s = scheduler.get_scheduler()
 *         return s.start(self, run_immediately)             # <<<<<<<<<<<<<<
 * 
 *     def after(self, future):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.run_immediately = __pyx_v_run_immediately;
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_s->__pyx_vtab)->start(__pyx_v_s, __pyx_v_self, 0, &__pyx_t_7)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/async_task.py":98
 *         return True if self._dependencies else False
 * 
 *     def start(self, run_immediately=False):             # <<<<<<<<<<<<<<
 *         """
 *         Starts the task on the current task scheduler.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_7start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_10async_task_9AsyncTask_6start[] = "\n        Starts the task on the current task scheduler.\n        Newly created tasks aren't started by default.\n\n        :param run_immediately: indicates whether first iteration\n            over the generator must be performed immediately\n        :return: self\n\n        ";
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_7start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_run_immediately;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_run_immediately,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_run_immediately);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "start") < 0)) __PYX_ERR(0, 98, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_run_immediately = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_run_immediately == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L3_error)
    } else {
      __pyx_v_run_immediately = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 98, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6start(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), __pyx_v_run_immediately);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_6start(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_v_run_immediately) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start __pyx_t_2;
  __Pyx_RefNannySetupContext("start", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.run_immediately = __pyx_v_run_immediately;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_5asynq_10async_task_AsyncTask->start(__pyx_v_self, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":111
 *         return s.start(self, run_immediately)
 * 
 *     def after(self, future):             # <<<<<<<<<<<<<<
 *         """Starts the task after specified future gets computed.
 * 
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9after(PyObject *__pyx_v_self, PyObject *__pyx_v_future); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_10async_task_9AsyncTask_after(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future, int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_s = 0;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("after", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_after); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_9after)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_future)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_future));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_future));
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_future));
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 111, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 111, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":118
 * 
 *         """
 *         s = scheduler.get_scheduler()             # <<<<<<<<<<<<<<
 *         return s.continue_with(future, self)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":119
 *         """
 *         s = scheduler.get_scheduler()
 *         return s.continue_with(future, self)             # <<<<<<<<<<<<<<
 * 
 *     def cancel_dependencies(self, error):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_s->__pyx_vtab)->continue_with(__pyx_v_s, __pyx_v_future, __pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/async_task.py":111
 *         return s.start(self, run_immediately)
 * 
 *     def after(self, future):             # <<<<<<<<<<<<<<
 *         """Starts the task after specified future gets computed.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.after", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9after(PyObject *__pyx_v_self, PyObject *__pyx_v_future); /*proto*/
static char __pyx_doc_5asynq_10async_task_9AsyncTask_8after[] = "Starts the task after specified future gets computed.\n\n        :param future: future that must be computed\n        :return: self\n\n        ";
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9after(PyObject *__pyx_v_self, PyObject *__pyx_v_future) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("after (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_future), __pyx_ptype_5asynq_7futures_FutureBase, 1, "future", 0))) __PYX_ERR(0, 111, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_8after(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_future));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_8after(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("after", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_10async_task_9AsyncTask_after(__pyx_v_self, __pyx_v_future, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.after", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":121
 *         return s.continue_with(future, self)
 * 
 *     def cancel_dependencies(self, error):             # <<<<<<<<<<<<<<
 *         if not self._dependencies:
 *             return
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11cancel_dependencies(PyObject *__pyx_v_self, PyObject *__pyx_v_error); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_cancel_dependencies(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_dependencies = 0;
  PyObject *__pyx_v_cancellation_error = NULL;
  PyObject *__pyx_v_dependency = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  __Pyx_RefNannySetupContext("cancel_dependencies", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_cancel_dependencies); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_11cancel_dependencies)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_error);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":122
 * 
 *     def cancel_dependencies(self, error):
 *         if not self._dependencies:             # <<<<<<<<<<<<<<
 *             return
 *         dependencies = list(self._dependencies)
 */
  __pyx_t_6 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (__pyx_t_7) {

    /* "asynq/async_task.py":123
 *     def cancel_dependencies(self, error):
 *         if not self._dependencies:
 *             return             # <<<<<<<<<<<<<<
 *         dependencies = list(self._dependencies)
 *         cancellation_error = error if isinstance(error, AsyncTaskCancelledError) else AsyncTaskCancelledError(error)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/async_task.py":122
 * 
 *     def cancel_dependencies(self, error):
 *         if not self._dependencies:             # <<<<<<<<<<<<<<
 *             return
 *         dependencies = list(self._dependencies)
 */
  }

  /* "asynq/async_task.py":124
 *         if not self._dependencies:
 *             return
 *         dependencies = list(self._dependencies)             # <<<<<<<<<<<<<<
 *         cancellation_error = error if isinstance(error, AsyncTaskCancelledError) else AsyncTaskCancelledError(error)
 *         if _debug_options.DUMP_DEPENDENCIES:
 */
  __pyx_t_1 = PySequence_List(__pyx_v_self->_dependencies); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_dependencies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":125
 *             return
 *         dependencies = list(self._dependencies)
 *         cancellation_error = error if isinstance(error, AsyncTaskCancelledError) else AsyncTaskCancelledError(error)             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskCancelledError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PyObject_IsInstance(__pyx_v_error, __pyx_t_2); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if ((__pyx_t_7 != 0)) {
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_1 = __pyx_v_error;
  } else {
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskCancelledError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_error);
      __Pyx_GIVEREF(__pyx_v_error);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_error);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_v_cancellation_error = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":126
 *         dependencies = list(self._dependencies)
 *         cancellation_error = error if isinstance(error, AsyncTaskCancelledError) else AsyncTaskCancelledError(error)
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))
 *         for dependency in dependencies:
 */
  __pyx_t_7 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_DEPENDENCIES != 0);
  if (__pyx_t_7) {

    /* "asynq/async_task.py":127
 *         cancellation_error = error if isinstance(error, AsyncTaskCancelledError) else AsyncTaskCancelledError(error)
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))             # <<<<<<<<<<<<<<
 *         for dependency in dependencies:
 *             _cancel_futures(dependency, cancellation_error)
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_async_cancelling_dependencies_o, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_2) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/async_task.py":126
 *         dependencies = list(self._dependencies)
 *         cancellation_error = error if isinstance(error, AsyncTaskCancelledError) else AsyncTaskCancelledError(error)
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))
 *         for dependency in dependencies:
 */
  }

  /* "asynq/async_task.py":128
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))
 *         for dependency in dependencies:             # <<<<<<<<<<<<<<
 *             _cancel_futures(dependency, cancellation_error)
 *         if _debug_options.DUMP_DEPENDENCIES:
 */
  __pyx_t_1 = __pyx_v_dependencies; __Pyx_INCREF(__pyx_t_1); __pyx_t_9 = 0;
  for (;;) {
    if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 128, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_dependency, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynq/async_task.py":129
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))
 *         for dependency in dependencies:
 *             _cancel_futures(dependency, cancellation_error)             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: <- cancelled dependencies of %s' % debug.str(self))
 */
    __pyx_t_3 = __pyx_f_5asynq_10async_task__cancel_futures(__pyx_v_dependency, __pyx_v_cancellation_error, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynq/async_task.py":128
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -> cancelling dependencies of %s:' % debug.str(self))
 *         for dependency in dependencies:             # <<<<<<<<<<<<<<
 *             _cancel_futures(dependency, cancellation_error)
 *         if _debug_options.DUMP_DEPENDENCIES:
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":130
 *         for dependency in dependencies:
 *             _cancel_futures(dependency, cancellation_error)
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: <- cancelled dependencies of %s' % debug.str(self))
 * 
 */
  __pyx_t_7 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_DEPENDENCIES != 0);
  if (__pyx_t_7) {

    /* "asynq/async_task.py":131
 *             _cancel_futures(dependency, cancellation_error)
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: <- cancelled dependencies of %s' % debug.str(self))             # <<<<<<<<<<<<<<
 * 
 *     def _compute(self):
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_async_cancelled_dependencies_of, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/async_task.py":130
 *         for dependency in dependencies:
 *             _cancel_futures(dependency, cancellation_error)
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: <- cancelled dependencies of %s' % debug.str(self))
 * 
 */
  }

  /* "asynq/async_task.py":121
 *         return s.continue_with(future, self)
 * 
 *     def cancel_dependencies(self, error):             # <<<<<<<<<<<<<<
 *         if not self._dependencies:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.cancel_dependencies", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dependencies);
  __Pyx_XDECREF(__pyx_v_cancellation_error);
  __Pyx_XDECREF(__pyx_v_dependency);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11cancel_dependencies(PyObject *__pyx_v_self, PyObject *__pyx_v_error); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11cancel_dependencies(PyObject *__pyx_v_self, PyObject *__pyx_v_error) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_dependencies (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_10cancel_dependencies(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_error));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_10cancel_dependencies(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cancel_dependencies", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask_cancel_dependencies(__pyx_v_self, __pyx_v_error, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.cancel_dependencies", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":133
 *             debug.write('@async: <- cancelled dependencies of %s' % debug.str(self))
 * 
 *     def _compute(self):             # <<<<<<<<<<<<<<
 *         # Forwards the call to task scheduler
 *         scheduler.get_scheduler().await([self])
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_13_compute(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__compute(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_compute", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_compute); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_13_compute)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":135
 *     def _compute(self):
 *         # Forwards the call to task scheduler
 *         scheduler.get_scheduler().await([self])             # <<<<<<<<<<<<<<
 *         # No need to assign a value/error here, since
 *         # _continue method (called by TaskScheduler) does this.
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_await); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyList_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self));
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":133
 *             debug.write('@async: <- cancelled dependencies of %s' % debug.str(self))
 * 
 *     def _compute(self):             # <<<<<<<<<<<<<<
 *         # Forwards the call to task scheduler
 *         scheduler.get_scheduler().await([self])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._compute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_13_compute(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_13_compute(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_compute (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_12_compute(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_12_compute(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_compute", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask__compute(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._compute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":139
 *         # _continue method (called by TaskScheduler) does this.
 * 
 *     def _computed(self):             # <<<<<<<<<<<<<<
 *         try:
 *             if self._generator is not None:
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_15_computed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__computed(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  __Pyx_RefNannySetupContext("_computed", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_computed); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_15_computed)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":140
 * 
 *     def _computed(self):
 *         try:             # <<<<<<<<<<<<<<
 *             if self._generator is not None:
 *                 self._generator.close()
 */
  /*try:*/ {

    /* "asynq/async_task.py":141
 *     def _computed(self):
 *         try:
 *             if self._generator is not None:             # <<<<<<<<<<<<<<
 *                 self._generator.close()
 *                 self._generator = None
 */
    __pyx_t_5 = (__pyx_v_self->_generator != Py_None);
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "asynq/async_task.py":142
 *         try:
 *             if self._generator is not None:
 *                 self._generator.close()             # <<<<<<<<<<<<<<
 *                 self._generator = None
 *         finally:
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_generator, __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 142, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L4_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 142, __pyx_L4_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/async_task.py":143
 *             if self._generator is not None:
 *                 self._generator.close()
 *                 self._generator = None             # <<<<<<<<<<<<<<
 *         finally:
 *             # super() doesn't work in Cython-ed version here
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->_generator);
      __Pyx_DECREF(__pyx_v_self->_generator);
      __pyx_v_self->_generator = Py_None;

      /* "asynq/async_task.py":141
 *     def _computed(self):
 *         try:
 *             if self._generator is not None:             # <<<<<<<<<<<<<<
 *                 self._generator.close()
 *                 self._generator = None
 */
    }
  }

  /* "asynq/async_task.py":146
 *         finally:
 *             # super() doesn't work in Cython-ed version here
 *             futures.FutureBase._computed(self)             # <<<<<<<<<<<<<<
 *             error = self.error()
 *             if error is not None and self.group_cancel:
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = __pyx_vtabptr_5asynq_7futures_FutureBase->_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/async_task.py":147
 *             # super() doesn't work in Cython-ed version here
 *             futures.FutureBase._computed(self)
 *             error = self.error()             # <<<<<<<<<<<<<<
 *             if error is not None and self.group_cancel:
 *                 self.cancel_dependencies(error)
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.error(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_v_error = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "asynq/async_task.py":148
 *             futures.FutureBase._computed(self)
 *             error = self.error()
 *             if error is not None and self.group_cancel:             # <<<<<<<<<<<<<<
 *                 self.cancel_dependencies(error)
 * 
 */
      __pyx_t_5 = (__pyx_v_error != Py_None);
      __pyx_t_7 = (__pyx_t_5 != 0);
      if (__pyx_t_7) {
      } else {
        __pyx_t_6 = __pyx_t_7;
        goto __pyx_L8_bool_binop_done;
      }
      __pyx_t_7 = (__pyx_v_self->group_cancel != 0);
      __pyx_t_6 = __pyx_t_7;
      __pyx_L8_bool_binop_done:;
      if (__pyx_t_6) {

        /* "asynq/async_task.py":149
 *             error = self.error()
 *             if error is not None and self.group_cancel:
 *                 self.cancel_dependencies(error)             # <<<<<<<<<<<<<<
 * 
 *     def _continue(self):
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->cancel_dependencies(__pyx_v_self, __pyx_v_error, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/async_task.py":148
 *             futures.FutureBase._computed(self)
 *             error = self.error()
 *             if error is not None and self.group_cancel:             # <<<<<<<<<<<<<<
 *                 self.cancel_dependencies(error)
 * 
 */
      }
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __pyx_t_8 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
      {

        /* "asynq/async_task.py":146
 *         finally:
 *             # super() doesn't work in Cython-ed version here
 *             futures.FutureBase._computed(self)             # <<<<<<<<<<<<<<
 *             error = self.error()
 *             if error is not None and self.group_cancel:
 */
        __pyx_t_1 = __pyx_vtabptr_5asynq_7futures_FutureBase->_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/async_task.py":147
 *             # super() doesn't work in Cython-ed version here
 *             futures.FutureBase._computed(self)
 *             error = self.error()             # <<<<<<<<<<<<<<
 *             if error is not None and self.group_cancel:
 *                 self.cancel_dependencies(error)
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.error(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_v_error = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "asynq/async_task.py":148
 *             futures.FutureBase._computed(self)
 *             error = self.error()
 *             if error is not None and self.group_cancel:             # <<<<<<<<<<<<<<
 *                 self.cancel_dependencies(error)
 * 
 */
        __pyx_t_7 = (__pyx_v_error != Py_None);
        __pyx_t_5 = (__pyx_t_7 != 0);
        if (__pyx_t_5) {
        } else {
          __pyx_t_6 = __pyx_t_5;
          goto __pyx_L13_bool_binop_done;
        }
        __pyx_t_5 = (__pyx_v_self->group_cancel != 0);
        __pyx_t_6 = __pyx_t_5;
        __pyx_L13_bool_binop_done:;
        if (__pyx_t_6) {

          /* "asynq/async_task.py":149
 *             error = self.error()
 *             if error is not None and self.group_cancel:
 *                 self.cancel_dependencies(error)             # <<<<<<<<<<<<<<
 * 
 *     def _continue(self):
 */
          __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->cancel_dependencies(__pyx_v_self, __pyx_v_error, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/async_task.py":148
 *             futures.FutureBase._computed(self)
 *             error = self.error()
 *             if error is not None and self.group_cancel:             # <<<<<<<<<<<<<<
 *                 self.cancel_dependencies(error)
 * 
 */
        }
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      }
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
      goto __pyx_L1_error;
      __pyx_L11_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      }
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "asynq/async_task.py":139
 *         # _continue method (called by TaskScheduler) does this.
 * 
 *     def _computed(self):             # <<<<<<<<<<<<<<
 *         try:
 *             if self._generator is not None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._computed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_15_computed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_15_computed(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_computed (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_14_computed(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_14_computed(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_computed", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask__computed(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._computed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":151
 *                 self.cancel_dependencies(error)
 * 
 *     def _continue(self):             # <<<<<<<<<<<<<<
 *         self._before_continue()
 *         # Obvious optimization: we don't need to return from
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_17_continue(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_e = NULL;
  PyTypeObject *__pyx_v_error_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("_continue", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_continue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_17_continue)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":152
 * 
 *     def _continue(self):
 *         self._before_continue()             # <<<<<<<<<<<<<<
 *         # Obvious optimization: we don't need to return from
 *         # this method, if we still can go further after yield.
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_before_continue(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":157
 *         # So we return only if there are dependencies or
 *         # current task is computed.
 *         while True:             # <<<<<<<<<<<<<<
 *             try:
 *                 value = unwrap(self._last_value)
 */
  while (1) {

    /* "asynq/async_task.py":158
 *         # current task is computed.
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = unwrap(self._last_value)
 *                 error = None
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      /*try:*/ {

        /* "asynq/async_task.py":159
 *         while True:
 *             try:
 *                 value = unwrap(self._last_value)             # <<<<<<<<<<<<<<
 *                 error = None
 *             except BaseException as e:
 */
        __pyx_t_1 = __pyx_v_self->_last_value;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_2 = __pyx_f_5asynq_10async_task_unwrap(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 159, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "asynq/async_task.py":160
 *             try:
 *                 value = unwrap(self._last_value)
 *                 error = None             # <<<<<<<<<<<<<<
 *             except BaseException as e:
 *                 value = None
 */
        __Pyx_INCREF(Py_None);
        __Pyx_XDECREF_SET(__pyx_v_error, Py_None);

        /* "asynq/async_task.py":158
 *         # current task is computed.
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = unwrap(self._last_value)
 *                 error = None
 */
      }
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynq/async_task.py":161
 *                 value = unwrap(self._last_value)
 *                 error = None
 *             except BaseException as e:             # <<<<<<<<<<<<<<
 *                 value = None
 *                 error = e
 */
      __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_8) {
        __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_3) < 0) __PYX_ERR(0, 161, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_1);

        /* "asynq/async_task.py":162
 *                 error = None
 *             except BaseException as e:
 *                 value = None             # <<<<<<<<<<<<<<
 *                 error = e
 *             try:
 */
        __Pyx_INCREF(Py_None);
        __Pyx_XDECREF_SET(__pyx_v_value, Py_None);

        /* "asynq/async_task.py":163
 *             except BaseException as e:
 *                 value = None
 *                 error = e             # <<<<<<<<<<<<<<
 *             try:
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 */
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_XDECREF_SET(__pyx_v_error, __pyx_v_e);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "asynq/async_task.py":158
 *         # current task is computed.
 *         while True:
 *             try:             # <<<<<<<<<<<<<<
 *                 value = unwrap(self._last_value)
 *                 error = None
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
      __pyx_L12_try_end:;
    }

    /* "asynq/async_task.py":164
 *                 value = None
 *                 error = e
 *             try:             # <<<<<<<<<<<<<<
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 *             except StopIteration as error:  # Most frequent, so it's the first one
 */
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_6, &__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_5);
        /*try:*/ {

          /* "asynq/async_task.py":165
 *                 error = e
 *             try:
 *                 self._accept_yield_result(self._continue_on_generator(value, error))             # <<<<<<<<<<<<<<
 *             except StopIteration as error:  # Most frequent, so it's the first one
 *                 if hasattr(error, 'value'):
 */
          __pyx_t_3 = __pyx_f_5asynq_10async_task_9AsyncTask__continue_on_generator(__pyx_v_self, __pyx_v_value, __pyx_v_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask__accept_yield_result(__pyx_v_self, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L20_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/async_task.py":164
 *                 value = None
 *                 error = e
 *             try:             # <<<<<<<<<<<<<<
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 *             except StopIteration as error:  # Most frequent, so it's the first one
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        goto __pyx_L27_try_end;
        __pyx_L20_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/async_task.py":166
 *             try:
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 *             except StopIteration as error:  # Most frequent, so it's the first one             # <<<<<<<<<<<<<<
 *                 if hasattr(error, 'value'):
 *                     # We're on a Python version that supports adding a value to StopIteration
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration);
        if (__pyx_t_8) {
          __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 166, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_DECREF_SET(__pyx_v_error, __pyx_t_3);

          /* "asynq/async_task.py":167
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 *             except StopIteration as error:  # Most frequent, so it's the first one
 *                 if hasattr(error, 'value'):             # <<<<<<<<<<<<<<
 *                     # We're on a Python version that supports adding a value to StopIteration
 *                     self._queue_exit(error.value)
 */
          __pyx_t_9 = PyObject_HasAttr(__pyx_v_error, __pyx_n_s_value); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 167, __pyx_L22_except_error)
          __pyx_t_10 = (__pyx_t_9 != 0);
          if (__pyx_t_10) {

            /* "asynq/async_task.py":169
 *                 if hasattr(error, 'value'):
 *                     # We're on a Python version that supports adding a value to StopIteration
 *                     self._queue_exit(error.value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     # This means there was no asynq.result() call, so the value of
 */
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_error, __pyx_n_s_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 169, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_11 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_queue_exit(__pyx_v_self, __pyx_t_4, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 169, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "asynq/async_task.py":167
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 *             except StopIteration as error:  # Most frequent, so it's the first one
 *                 if hasattr(error, 'value'):             # <<<<<<<<<<<<<<
 *                     # We're on a Python version that supports adding a value to StopIteration
 *                     self._queue_exit(error.value)
 */
            goto __pyx_L30;
          }

          /* "asynq/async_task.py":173
 *                     # This means there was no asynq.result() call, so the value of
 *                     # this task should be None
 *                     self._queue_exit(None)             # <<<<<<<<<<<<<<
 *             except GeneratorExit as error:
 *                 error_type = type(error)
 */
          /*else*/ {
            __pyx_t_11 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_queue_exit(__pyx_v_self, Py_None, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 173, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __pyx_L30:;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L21_exception_handled;
        }

        /* "asynq/async_task.py":174
 *                     # this task should be None
 *                     self._queue_exit(None)
 *             except GeneratorExit as error:             # <<<<<<<<<<<<<<
 *                 error_type = type(error)
 *                 if error_type is AsyncTaskResult:
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_GeneratorExit);
        if (__pyx_t_8) {
          __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 174, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_DECREF_SET(__pyx_v_error, __pyx_t_3);

          /* "asynq/async_task.py":175
 *                     self._queue_exit(None)
 *             except GeneratorExit as error:
 *                 error_type = type(error)             # <<<<<<<<<<<<<<
 *                 if error_type is AsyncTaskResult:
 *                     self._queue_exit(error.result)
 */
          __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_error)));
          __Pyx_XDECREF_SET(__pyx_v_error_type, ((PyTypeObject*)((PyObject *)Py_TYPE(__pyx_v_error))));

          /* "asynq/async_task.py":176
 *             except GeneratorExit as error:
 *                 error_type = type(error)
 *                 if error_type is AsyncTaskResult:             # <<<<<<<<<<<<<<
 *                     self._queue_exit(error.result)
 *                 elif error_type is AsyncTaskCancelledError:
 */
          __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskResult); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 176, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = (__pyx_v_error_type == ((PyTypeObject*)__pyx_t_11));
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_9 = (__pyx_t_10 != 0);
          if (__pyx_t_9) {

            /* "asynq/async_task.py":177
 *                 error_type = type(error)
 *                 if error_type is AsyncTaskResult:
 *                     self._queue_exit(error.result)             # <<<<<<<<<<<<<<
 *                 elif error_type is AsyncTaskCancelledError:
 *                     self._accept_error(error)
 */
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_error, __pyx_n_s_result); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 177, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_4 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_queue_exit(__pyx_v_self, __pyx_t_11, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 177, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "asynq/async_task.py":176
 *             except GeneratorExit as error:
 *                 error_type = type(error)
 *                 if error_type is AsyncTaskResult:             # <<<<<<<<<<<<<<
 *                     self._queue_exit(error.result)
 *                 elif error_type is AsyncTaskCancelledError:
 */
            goto __pyx_L33;
          }

          /* "asynq/async_task.py":178
 *                 if error_type is AsyncTaskResult:
 *                     self._queue_exit(error.result)
 *                 elif error_type is AsyncTaskCancelledError:             # <<<<<<<<<<<<<<
 *                     self._accept_error(error)
 *                 else:
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskCancelledError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 178, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_9 = (__pyx_v_error_type == ((PyTypeObject*)__pyx_t_4));
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_10 = (__pyx_t_9 != 0);
          if (__pyx_t_10) {

            /* "asynq/async_task.py":179
 *                     self._queue_exit(error.result)
 *                 elif error_type is AsyncTaskCancelledError:
 *                     self._accept_error(error)             # <<<<<<<<<<<<<<
 *                 else:
 *                     self._queue_exit(None)
 */
            __pyx_t_4 = __pyx_f_5asynq_10async_task_9AsyncTask__accept_error(__pyx_v_self, __pyx_v_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 179, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "asynq/async_task.py":178
 *                 if error_type is AsyncTaskResult:
 *                     self._queue_exit(error.result)
 *                 elif error_type is AsyncTaskCancelledError:             # <<<<<<<<<<<<<<
 *                     self._accept_error(error)
 *                 else:
 */
            goto __pyx_L33;
          }

          /* "asynq/async_task.py":181
 *                     self._accept_error(error)
 *                 else:
 *                     self._queue_exit(None)             # <<<<<<<<<<<<<<
 *             except BaseException as error:
 *                 self._accept_error(error)
 */
          /*else*/ {
            __pyx_t_4 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_queue_exit(__pyx_v_self, Py_None, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 181, __pyx_L22_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __pyx_L33:;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L21_exception_handled;
        }

        /* "asynq/async_task.py":182
 *                 else:
 *                     self._queue_exit(None)
 *             except BaseException as error:             # <<<<<<<<<<<<<<
 *                 self._accept_error(error)
 *             finally:
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_8) {
          __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_2) < 0) __PYX_ERR(0, 182, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_DECREF_SET(__pyx_v_error, __pyx_t_3);

          /* "asynq/async_task.py":183
 *                     self._queue_exit(None)
 *             except BaseException as error:
 *                 self._accept_error(error)             # <<<<<<<<<<<<<<
 *             finally:
 *                 if self._dependencies or self.is_computed():
 */
          __pyx_t_4 = __pyx_f_5asynq_10async_task_9AsyncTask__accept_error(__pyx_v_self, __pyx_v_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 183, __pyx_L22_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L21_exception_handled;
        }
        goto __pyx_L22_except_error;
        __pyx_L22_except_error:;

        /* "asynq/async_task.py":164
 *                 value = None
 *                 error = e
 *             try:             # <<<<<<<<<<<<<<
 *                 self._accept_yield_result(self._continue_on_generator(value, error))
 *             except StopIteration as error:  # Most frequent, so it's the first one
 */
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
        goto __pyx_L18_error;
        __pyx_L21_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_6, __pyx_t_5);
        __pyx_L27_try_end:;
      }
    }

    /* "asynq/async_task.py":185
 *                 self._accept_error(error)
 *             finally:
 *                 if self._dependencies or self.is_computed():             # <<<<<<<<<<<<<<
 *                     self._after_continue()
 *                     return
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_t_9 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
        if (!__pyx_t_9) {
        } else {
          __pyx_t_10 = __pyx_t_9;
          goto __pyx_L37_bool_binop_done;
        }
        __pyx_t_9 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __pyx_t_12 = (__pyx_t_9 != 0);
        __pyx_t_10 = __pyx_t_12;
        __pyx_L37_bool_binop_done:;
        if (__pyx_t_10) {

          /* "asynq/async_task.py":186
 *             finally:
 *                 if self._dependencies or self.is_computed():
 *                     self._after_continue()             # <<<<<<<<<<<<<<
 *                     return
 * 
 */
          __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_after_continue(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "asynq/async_task.py":187
 *                 if self._dependencies or self.is_computed():
 *                     self._after_continue()
 *                     return             # <<<<<<<<<<<<<<
 * 
 *     def _continue_on_generator(self, value, error):
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_r = Py_None; __Pyx_INCREF(Py_None);
          goto __pyx_L0;

          /* "asynq/async_task.py":185
 *                 self._accept_error(error)
 *             finally:
 *                 if self._dependencies or self.is_computed():             # <<<<<<<<<<<<<<
 *                     self._after_continue()
 *                     return
 */
        }
        goto __pyx_L19;
      }
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __pyx_L18_error:;
        __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_17);
        __pyx_t_8 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
        {
          __pyx_t_12 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
          if (!__pyx_t_12) {
          } else {
            __pyx_t_10 = __pyx_t_12;
            goto __pyx_L44_bool_binop_done;
          }
          __pyx_t_12 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 185, __pyx_L42_error)
          __pyx_t_9 = (__pyx_t_12 != 0);
          __pyx_t_10 = __pyx_t_9;
          __pyx_L44_bool_binop_done:;
          if (__pyx_t_10) {

            /* "asynq/async_task.py":186
 *             finally:
 *                 if self._dependencies or self.is_computed():
 *                     self._after_continue()             # <<<<<<<<<<<<<<
 *                     return
 * 
 */
            __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_after_continue(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L42_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "asynq/async_task.py":187
 *                 if self._dependencies or self.is_computed():
 *                     self._after_continue()
 *                     return             # <<<<<<<<<<<<<<
 * 
 *     def _continue_on_generator(self, value, error):
 */
            __Pyx_XDECREF(__pyx_r);
            __pyx_r = Py_None; __Pyx_INCREF(Py_None);
            goto __pyx_L41_return;

            /* "asynq/async_task.py":185
 *                 self._accept_error(error)
 *             finally:
 *                 if self._dependencies or self.is_computed():             # <<<<<<<<<<<<<<
 *                     self._after_continue()
 *                     return
 */
          }
        }
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        }
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_ErrRestore(__pyx_t_5, __pyx_t_6, __pyx_t_7);
        __pyx_t_5 = 0; __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
        __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
        goto __pyx_L1_error;
        __pyx_L41_return:;
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
        goto __pyx_L0;
        __pyx_L42_error:;
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ExceptionReset(__pyx_t_15, __pyx_t_16, __pyx_t_17);
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0;
        goto __pyx_L1_error;
      }
      __pyx_L19:;
    }
  }

  /* "asynq/async_task.py":151
 *                 self.cancel_dependencies(error)
 * 
 *     def _continue(self):             # <<<<<<<<<<<<<<
 *         self._before_continue()
 *         # Obvious optimization: we don't need to return from
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XDECREF(__pyx_v_error_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_17_continue(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_17_continue(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_continue (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_16_continue(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_16_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_continue", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask__continue(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":189
 *                     return
 * 
 *     def _continue_on_generator(self, value, error):             # <<<<<<<<<<<<<<
 *         try:
 *             if error is None:
 */

static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__continue_on_generator(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value, PyObject *__pyx_v_error) {
  PyObject *__pyx_v_tb = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  __Pyx_RefNannySetupContext("_continue_on_generator", 0);

  /* "asynq/async_task.py":190
 * 
 *     def _continue_on_generator(self, value, error):
 *         try:             # <<<<<<<<<<<<<<
 *             if error is None:
 *                 if self._generator is None:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "asynq/async_task.py":191
 *     def _continue_on_generator(self, value, error):
 *         try:
 *             if error is None:             # <<<<<<<<<<<<<<
 *                 if self._generator is None:
 *                     raise StopIteration()
 */
      __pyx_t_4 = (__pyx_v_error == Py_None);
      __pyx_t_5 = (__pyx_t_4 != 0);
      if (__pyx_t_5) {

        /* "asynq/async_task.py":192
 *         try:
 *             if error is None:
 *                 if self._generator is None:             # <<<<<<<<<<<<<<
 *                     raise StopIteration()
 *             elif self._generator is None:
 */
        __pyx_t_5 = (__pyx_v_self->_generator == Py_None);
        __pyx_t_4 = (__pyx_t_5 != 0);
        if (__pyx_t_4) {

          /* "asynq/async_task.py":193
 *             if error is None:
 *                 if self._generator is None:
 *                     raise StopIteration()             # <<<<<<<<<<<<<<
 *             elif self._generator is None:
 *                 raise error
 */
          __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_builtin_StopIteration); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 193, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_Raise(__pyx_t_6, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __PYX_ERR(0, 193, __pyx_L3_error)

          /* "asynq/async_task.py":192
 *         try:
 *             if error is None:
 *                 if self._generator is None:             # <<<<<<<<<<<<<<
 *                     raise StopIteration()
 *             elif self._generator is None:
 */
        }

        /* "asynq/async_task.py":191
 *     def _continue_on_generator(self, value, error):
 *         try:
 *             if error is None:             # <<<<<<<<<<<<<<
 *                 if self._generator is None:
 *                     raise StopIteration()
 */
        goto __pyx_L11;
      }

      /* "asynq/async_task.py":194
 *                 if self._generator is None:
 *                     raise StopIteration()
 *             elif self._generator is None:             # <<<<<<<<<<<<<<
 *                 raise error
 * 
 */
      __pyx_t_4 = (__pyx_v_self->_generator == Py_None);
      __pyx_t_5 = (__pyx_t_4 != 0);
      if (__pyx_t_5) {

        /* "asynq/async_task.py":195
 *                     raise StopIteration()
 *             elif self._generator is None:
 *                 raise error             # <<<<<<<<<<<<<<
 * 
 *             self.iteration_index += 1
 */
        __Pyx_Raise(__pyx_v_error, 0, 0, 0);
        __PYX_ERR(0, 195, __pyx_L3_error)

        /* "asynq/async_task.py":194
 *                 if self._generator is None:
 *                     raise StopIteration()
 *             elif self._generator is None:             # <<<<<<<<<<<<<<
 *                 raise error
 * 
 */
      }
      __pyx_L11:;

      /* "asynq/async_task.py":197
 *                 raise error
 * 
 *             self.iteration_index += 1             # <<<<<<<<<<<<<<
 *             if error is None:
 *                 return self._generator.send(value)
 */
      __pyx_v_self->iteration_index = (__pyx_v_self->iteration_index + 1);

      /* "asynq/async_task.py":198
 * 
 *             self.iteration_index += 1
 *             if error is None:             # <<<<<<<<<<<<<<
 *                 return self._generator.send(value)
 *             else:
 */
      __pyx_t_5 = (__pyx_v_error == Py_None);
      __pyx_t_4 = (__pyx_t_5 != 0);
      if (__pyx_t_4) {

        /* "asynq/async_task.py":199
 *             self.iteration_index += 1
 *             if error is None:
 *                 return self._generator.send(value)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._last_value = None  # No need to keep it further
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_generator, __pyx_n_s_send); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 199, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        if (!__pyx_t_8) {
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 199, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
        } else {
          __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 199, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_v_value);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 199, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_r = __pyx_t_6;
        __pyx_t_6 = 0;
        goto __pyx_L7_try_return;

        /* "asynq/async_task.py":198
 * 
 *             self.iteration_index += 1
 *             if error is None:             # <<<<<<<<<<<<<<
 *                 return self._generator.send(value)
 *             else:
 */
      }

      /* "asynq/async_task.py":201
 *                 return self._generator.send(value)
 *             else:
 *                 self._last_value = None  # No need to keep it further             # <<<<<<<<<<<<<<
 *                 self._frame_info = debug.get_frame_info(self._generator)
 *                 if hasattr(error, '_task'):
 */
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __Pyx_GIVEREF(Py_None);
        __Pyx_GOTREF(__pyx_v_self->_last_value);
        __Pyx_DECREF(__pyx_v_self->_last_value);
        __pyx_v_self->_last_value = Py_None;

        /* "asynq/async_task.py":202
 *             else:
 *                 self._last_value = None  # No need to keep it further
 *                 self._frame_info = debug.get_frame_info(self._generator)             # <<<<<<<<<<<<<<
 *                 if hasattr(error, '_task'):
 *                     return self._generator.throw(error._type_, error, error._traceback)
 */
        __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 202, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_get_frame_info); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 202, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        if (!__pyx_t_7) {
          __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_self->_generator); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
        } else {
          __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 202, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
          __Pyx_INCREF(__pyx_v_self->_generator);
          __Pyx_GIVEREF(__pyx_v_self->_generator);
          PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_v_self->_generator);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GIVEREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_v_self->_frame_info);
        __Pyx_DECREF(__pyx_v_self->_frame_info);
        __pyx_v_self->_frame_info = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "asynq/async_task.py":203
 *                 self._last_value = None  # No need to keep it further
 *                 self._frame_info = debug.get_frame_info(self._generator)
 *                 if hasattr(error, '_task'):             # <<<<<<<<<<<<<<
 *                     return self._generator.throw(error._type_, error, error._traceback)
 *                 else:
 */
        __pyx_t_4 = PyObject_HasAttr(__pyx_v_error, __pyx_n_s_task); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 203, __pyx_L3_error)
        __pyx_t_5 = (__pyx_t_4 != 0);
        if (__pyx_t_5) {

          /* "asynq/async_task.py":204
 *                 self._frame_info = debug.get_frame_info(self._generator)
 *                 if hasattr(error, '_task'):
 *                     return self._generator.throw(error._type_, error, error._traceback)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return self._generator.throw(type(error), error)
 */
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_generator, __pyx_n_s_throw); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 204, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_error, __pyx_n_s_type); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 204, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_error, __pyx_n_s_traceback); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 204, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_10 = NULL;
          __pyx_t_11 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
              __pyx_t_11 = 1;
            }
          }
          __pyx_t_12 = PyTuple_New(3+__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 204, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_10) {
            __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10); __pyx_t_10 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_11, __pyx_t_8);
          __Pyx_INCREF(__pyx_v_error);
          __Pyx_GIVEREF(__pyx_v_error);
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_11, __pyx_v_error);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_12, 2+__pyx_t_11, __pyx_t_7);
          __pyx_t_8 = 0;
          __pyx_t_7 = 0;
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_12, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 204, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_r = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L7_try_return;

          /* "asynq/async_task.py":203
 *                 self._last_value = None  # No need to keep it further
 *                 self._frame_info = debug.get_frame_info(self._generator)
 *                 if hasattr(error, '_task'):             # <<<<<<<<<<<<<<
 *                     return self._generator.throw(error._type_, error, error._traceback)
 *                 else:
 */
        }

        /* "asynq/async_task.py":206
 *                     return self._generator.throw(error._type_, error, error._traceback)
 *                 else:
 *                     return self._generator.throw(type(error), error)             # <<<<<<<<<<<<<<
 *         except (StopIteration, GeneratorExit):
 *             # Returning leads to a StopIteration exception, which is
 */
        /*else*/ {
          __Pyx_XDECREF(__pyx_r);
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_generator, __pyx_n_s_throw); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 206, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_12 = NULL;
          __pyx_t_11 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_9))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_9);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_9, function);
              __pyx_t_11 = 1;
            }
          }
          __pyx_t_7 = PyTuple_New(2+__pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 206, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_error)));
          __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_error)));
          PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_11, ((PyObject *)Py_TYPE(__pyx_v_error)));
          __Pyx_INCREF(__pyx_v_error);
          __Pyx_GIVEREF(__pyx_v_error);
          PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_11, __pyx_v_error);
          __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 206, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_r = __pyx_t_6;
          __pyx_t_6 = 0;
          goto __pyx_L7_try_return;
        }
      }

      /* "asynq/async_task.py":190
 * 
 *     def _continue_on_generator(self, value, error):
 *         try:             # <<<<<<<<<<<<<<
 *             if error is None:
 *                 if self._generator is None:
 */
    }
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "asynq/async_task.py":207
 *                 else:
 *                     return self._generator.throw(type(error), error)
 *         except (StopIteration, GeneratorExit):             # <<<<<<<<<<<<<<
 *             # Returning leads to a StopIteration exception, which is
 *             # handled here. In this case we shouldn't need to extract frame
 */
    __pyx_t_13 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_StopIteration) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_GeneratorExit);
    if (__pyx_t_13) {
      __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue_on_generator", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_6, &__pyx_t_9, &__pyx_t_7) < 0) __PYX_ERR(0, 207, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_7);

      /* "asynq/async_task.py":211
 *             # handled here. In this case we shouldn't need to extract frame
 *             # info, because that is slow.
 *             self._generator = None             # <<<<<<<<<<<<<<
 *             raise
 *         except:
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->_generator);
      __Pyx_DECREF(__pyx_v_self->_generator);
      __pyx_v_self->_generator = Py_None;

      /* "asynq/async_task.py":212
 *             # info, because that is slow.
 *             self._generator = None
 *             raise             # <<<<<<<<<<<<<<
 *         except:
 *             # Any exception means generator is closed at this point,
 */
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestoreWithState(__pyx_t_6, __pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0; __pyx_t_9 = 0; __pyx_t_7 = 0; 
      __PYX_ERR(0, 212, __pyx_L5_except_error)
    }

    /* "asynq/async_task.py":213
 *             self._generator = None
 *             raise
 *         except:             # <<<<<<<<<<<<<<
 *             # Any exception means generator is closed at this point,
 *             # since it fall through all exception handlers there;
 */
    /*except:*/ {
      __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue_on_generator", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_9, &__pyx_t_6) < 0) __PYX_ERR(0, 213, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_6);

      /* "asynq/async_task.py":221
 *             # show where in the async task the failure happened. However, if the error was thrown
 *             # into the generator, we'll already have set the frame info.
 *             if self._frame_info is None:             # <<<<<<<<<<<<<<
 *                 tb = sys.exc_info()[2]
 * 
 */
      __pyx_t_5 = (__pyx_v_self->_frame_info == Py_None);
      __pyx_t_4 = (__pyx_t_5 != 0);
      if (__pyx_t_4) {

        /* "asynq/async_task.py":222
 *             # into the generator, we'll already have set the frame info.
 *             if self._frame_info is None:
 *                 tb = sys.exc_info()[2]             # <<<<<<<<<<<<<<
 * 
 *                 while tb.tb_next is not None:
 */
        __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_sys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 222, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_exc_info); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 222, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        if (__pyx_t_8) {
          __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_8); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 222, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else {
          __pyx_t_12 = __Pyx_PyObject_CallNoArg(__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 222, __pyx_L5_except_error)
        }
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_12, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 222, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_v_tb = __pyx_t_10;
        __pyx_t_10 = 0;

        /* "asynq/async_task.py":224
 *                 tb = sys.exc_info()[2]
 * 
 *                 while tb.tb_next is not None:             # <<<<<<<<<<<<<<
 *                     tb = tb.tb_next
 *                 self._frame_info = inspect.getframeinfo(tb.tb_frame)
 */
        while (1) {
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_next); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 224, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_4 = (__pyx_t_10 != Py_None);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_5 = (__pyx_t_4 != 0);
          if (!__pyx_t_5) break;

          /* "asynq/async_task.py":225
 * 
 *                 while tb.tb_next is not None:
 *                     tb = tb.tb_next             # <<<<<<<<<<<<<<
 *                 self._frame_info = inspect.getframeinfo(tb.tb_frame)
 *             self._generator = None
 */
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_next); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 225, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF_SET(__pyx_v_tb, __pyx_t_10);
          __pyx_t_10 = 0;
        }

        /* "asynq/async_task.py":226
 *                 while tb.tb_next is not None:
 *                     tb = tb.tb_next
 *                 self._frame_info = inspect.getframeinfo(tb.tb_frame)             # <<<<<<<<<<<<<<
 *             self._generator = None
 *             raise
 */
        __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_inspect); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 226, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_getframeinfo); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 226, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_tb, __pyx_n_s_tb_frame); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 226, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_14 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_14)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_14);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        if (!__pyx_t_14) {
          __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 226, __pyx_L5_except_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_10);
        } else {
          __pyx_t_15 = PyTuple_New(1+1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 226, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_14); __pyx_t_14 = NULL;
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_15, 0+1, __pyx_t_12);
          __pyx_t_12 = 0;
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_15, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 226, __pyx_L5_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GIVEREF(__pyx_t_10);
        __Pyx_GOTREF(__pyx_v_self->_frame_info);
        __Pyx_DECREF(__pyx_v_self->_frame_info);
        __pyx_v_self->_frame_info = __pyx_t_10;
        __pyx_t_10 = 0;

        /* "asynq/async_task.py":221
 *             # show where in the async task the failure happened. However, if the error was thrown
 *             # into the generator, we'll already have set the frame info.
 *             if self._frame_info is None:             # <<<<<<<<<<<<<<
 *                 tb = sys.exc_info()[2]
 * 
 */
      }

      /* "asynq/async_task.py":227
 *                     tb = tb.tb_next
 *                 self._frame_info = inspect.getframeinfo(tb.tb_frame)
 *             self._generator = None             # <<<<<<<<<<<<<<
 *             raise
 * 
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_self->_generator);
      __Pyx_DECREF(__pyx_v_self->_generator);
      __pyx_v_self->_generator = Py_None;

      /* "asynq/async_task.py":228
 *                 self._frame_info = inspect.getframeinfo(tb.tb_frame)
 *             self._generator = None
 *             raise             # <<<<<<<<<<<<<<
 * 
 *     def _accept_yield_result(self, result):
 */
      __Pyx_GIVEREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_9, __pyx_t_6);
      __pyx_t_7 = 0; __pyx_t_9 = 0; __pyx_t_6 = 0; 
      __PYX_ERR(0, 228, __pyx_L5_except_error)
    }
    __pyx_L5_except_error:;

    /* "asynq/async_task.py":190
 * 
 *     def _continue_on_generator(self, value, error):
 *         try:             # <<<<<<<<<<<<<<
 *             if error is None:
 *                 if self._generator is None:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "asynq/async_task.py":189
 *                     return
 * 
 *     def _continue_on_generator(self, value, error):             # <<<<<<<<<<<<<<
 *         try:
 *             if error is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._continue_on_generator", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tb);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":230
 *             raise
 * 
 *     def _accept_yield_result(self, result):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_YIELD_RESULTS:
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
 */

static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__accept_yield_result(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("_accept_yield_result", 0);

  /* "asynq/async_task.py":231
 * 
 *     def _accept_yield_result(self, result):
 *         if _debug_options.DUMP_YIELD_RESULTS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
 *         self._last_value = result
 */
  __pyx_t_1 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_YIELD_RESULTS != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":232
 *     def _accept_yield_result(self, result):
 *         if _debug_options.DUMP_YIELD_RESULTS:
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))             # <<<<<<<<<<<<<<
 *         self._last_value = result
 *         self.scheduler.add_dependencies(self, result)
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_repr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_result); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_v_result);
      __Pyx_GIVEREF(__pyx_v_result);
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_v_result);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_yield_s_s, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/async_task.py":231
 * 
 *     def _accept_yield_result(self, result):
 *         if _debug_options.DUMP_YIELD_RESULTS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
 *         self._last_value = result
 */
  }

  /* "asynq/async_task.py":233
 *         if _debug_options.DUMP_YIELD_RESULTS:
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
 *         self._last_value = result             # <<<<<<<<<<<<<<
 *         self.scheduler.add_dependencies(self, result)
 * 
 */
  __Pyx_INCREF(__pyx_v_result);
  __Pyx_GIVEREF(__pyx_v_result);
  __Pyx_GOTREF(__pyx_v_self->_last_value);
  __Pyx_DECREF(__pyx_v_self->_last_value);
  __pyx_v_self->_last_value = __pyx_v_result;

  /* "asynq/async_task.py":234
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
 *         self._last_value = result
 *         self.scheduler.add_dependencies(self, result)             # <<<<<<<<<<<<<<
 * 
 *     def _accept_error(self, error):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->scheduler->__pyx_vtab)->add_dependencies(__pyx_v_self->scheduler, __pyx_v_self, __pyx_v_result, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynq/async_task.py":230
 *             raise
 * 
 *     def _accept_yield_result(self, result):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_YIELD_RESULTS:
 *             debug.write('@async: yield: %s -> %s' % (debug.str(self), debug.repr(result)))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._accept_yield_result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":236
 *         self.scheduler.add_dependencies(self, result)
 * 
 *     def _accept_error(self, error):             # <<<<<<<<<<<<<<
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             # We can't change the value in this case, so we can just return here.
 */

static CYTHON_INLINE PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__accept_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_accept_error", 0);

  /* "asynq/async_task.py":237
 * 
 *     def _accept_error(self, error):
 *         if self._value is not _futures_none:  # is_computed(), but faster             # <<<<<<<<<<<<<<
 *             # We can't change the value in this case, so we can just return here.
 *             pass
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base._value != __pyx_v_5asynq_10async_task__futures_none);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {
    goto __pyx_L3;
  }

  /* "asynq/async_task.py":243
 *             # when there is no _task it means that this is the bottommost level of the async
 *             # task. We must attach the traceback as soon as possible
 *             if not hasattr(error, '_task'):             # <<<<<<<<<<<<<<
 *                 error._task = self
 *                 core_errors.prepare_for_reraise(error)
 */
  /*else*/ {
    __pyx_t_2 = PyObject_HasAttr(__pyx_v_error, __pyx_n_s_task); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 243, __pyx_L1_error)
    __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
    if (__pyx_t_1) {

      /* "asynq/async_task.py":244
 *             # task. We must attach the traceback as soon as possible
 *             if not hasattr(error, '_task'):
 *                 error._task = self             # <<<<<<<<<<<<<<
 *                 core_errors.prepare_for_reraise(error)
 *             else:
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_error, __pyx_n_s_task, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 244, __pyx_L1_error)

      /* "asynq/async_task.py":245
 *             if not hasattr(error, '_task'):
 *                 error._task = self
 *                 core_errors.prepare_for_reraise(error)             # <<<<<<<<<<<<<<
 *             else:
 *                 # when we already have the _task on the error, it means that
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_core_errors); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_prepare_for_reraise); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 245, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_error);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 245, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/async_task.py":243
 *             # when there is no _task it means that this is the bottommost level of the async
 *             # task. We must attach the traceback as soon as possible
 *             if not hasattr(error, '_task'):             # <<<<<<<<<<<<<<
 *                 error._task = self
 *                 core_errors.prepare_for_reraise(error)
 */
      goto __pyx_L4;
    }

    /* "asynq/async_task.py":257
 *                 # we can glue all the different tasks' tracebacks and make it look like
 *                 # the error came from there.
 *                 error._traceback = sys.exc_info()[2]             # <<<<<<<<<<<<<<
 * 
 *             if _debug_options.DUMP_EXCEPTIONS:
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_sys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_exc_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      if (__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 257, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_error, __pyx_n_s_traceback, __pyx_t_6) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_L4:;

    /* "asynq/async_task.py":259
 *                 error._traceback = sys.exc_info()[2]
 * 
 *             if _debug_options.DUMP_EXCEPTIONS:             # <<<<<<<<<<<<<<
 *                 debug.dump_error(error)
 *             # Must queue, since there can be captured dependencies to resolve
 */
    __pyx_t_1 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_EXCEPTIONS != 0);
    if (__pyx_t_1) {

      /* "asynq/async_task.py":260
 * 
 *             if _debug_options.DUMP_EXCEPTIONS:
 *                 debug.dump_error(error)             # <<<<<<<<<<<<<<
 *             # Must queue, since there can be captured dependencies to resolve
 *             self._queue_throw_error(error)
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_dump_error); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 260, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
      } else {
        __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_error);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 260, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "asynq/async_task.py":259
 *                 error._traceback = sys.exc_info()[2]
 * 
 *             if _debug_options.DUMP_EXCEPTIONS:             # <<<<<<<<<<<<<<
 *                 debug.dump_error(error)
 *             # Must queue, since there can be captured dependencies to resolve
 */
    }

    /* "asynq/async_task.py":262
 *                 debug.dump_error(error)
 *             # Must queue, since there can be captured dependencies to resolve
 *             self._queue_throw_error(error)             # <<<<<<<<<<<<<<
 * 
 *     def _queue_exit(self, result):
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_queue_throw_error(__pyx_v_self, __pyx_v_error, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_L3:;

  /* "asynq/async_task.py":236
 *         self.scheduler.add_dependencies(self, result)
 * 
 *     def _accept_error(self, error):             # <<<<<<<<<<<<<<
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             # We can't change the value in this case, so we can just return here.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._accept_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":264
 *             self._queue_throw_error(error)
 * 
 *     def _queue_exit(self, result):             # <<<<<<<<<<<<<<
 *         # Result must be unwrapped first,
 *         # and since there can be dependencies, the error result (exception)
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_19_queue_exit(PyObject *__pyx_v_self, PyObject *__pyx_v_result); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__queue_exit(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("_queue_exit", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_queue_exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_19_queue_exit)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_result); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_result);
        __Pyx_GIVEREF(__pyx_v_result);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_result);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":272
 *         # an error during dependency computation, we must re-throw it
 *         # from the generator.
 *         if self._value is not _futures_none:  # is_computed(), but faster             # <<<<<<<<<<<<<<
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None:
 */
  __pyx_t_6 = (__pyx_v_self->__pyx_base._value != __pyx_v_5asynq_10async_task__futures_none);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "asynq/async_task.py":273
 *         # from the generator.
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)             # <<<<<<<<<<<<<<
 *         if self._generator is not None:
 *             self._generator.close()
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_futures); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FutureIsAlreadyComputed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_2) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 273, __pyx_L1_error)

    /* "asynq/async_task.py":272
 *         # an error during dependency computation, we must re-throw it
 *         # from the generator.
 *         if self._value is not _futures_none:  # is_computed(), but faster             # <<<<<<<<<<<<<<
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None:
 */
  }

  /* "asynq/async_task.py":274
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None:             # <<<<<<<<<<<<<<
 *             self._generator.close()
 *             self._generator = None
 */
  __pyx_t_7 = (__pyx_v_self->_generator != Py_None);
  __pyx_t_6 = (__pyx_t_7 != 0);
  if (__pyx_t_6) {

    /* "asynq/async_task.py":275
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None:
 *             self._generator.close()             # <<<<<<<<<<<<<<
 *             self._generator = None
 *         if self._dependencies:  # is_blocked(), but faster
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_generator, __pyx_n_s_close); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/async_task.py":276
 *         if self._generator is not None:
 *             self._generator.close()
 *             self._generator = None             # <<<<<<<<<<<<<<
 *         if self._dependencies:  # is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_generator);
    __Pyx_DECREF(__pyx_v_self->_generator);
    __pyx_v_self->_generator = Py_None;

    /* "asynq/async_task.py":274
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None:             # <<<<<<<<<<<<<<
 *             self._generator.close()
 *             self._generator = None
 */
  }

  /* "asynq/async_task.py":277
 *             self._generator.close()
 *             self._generator = None
 *         if self._dependencies:  # is_blocked(), but faster             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))
 */
  __pyx_t_6 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
  if (__pyx_t_6) {

    /* "asynq/async_task.py":278
 *             self._generator = None
 *         if self._dependencies:  # is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))
 *             self._last_value = result
 */
    __pyx_t_6 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_QUEUED_RESULTS != 0);
    if (__pyx_t_6) {

      /* "asynq/async_task.py":279
 *         if self._dependencies:  # is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))             # <<<<<<<<<<<<<<
 *             self._last_value = result
 *         else:
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_repr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_2); __pyx_t_2 = NULL;
        __Pyx_INCREF(__pyx_v_result);
        __Pyx_GIVEREF(__pyx_v_result);
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_v_result);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (!__pyx_t_8) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_queuing_exit_s_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/async_task.py":278
 *             self._generator = None
 *         if self._dependencies:  # is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))
 *             self._last_value = result
 */
    }

    /* "asynq/async_task.py":280
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))
 *             self._last_value = result             # <<<<<<<<<<<<<<
 *         else:
 *             self.set_value(result)
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_last_value);
    __Pyx_DECREF(__pyx_v_self->_last_value);
    __pyx_v_self->_last_value = __pyx_v_result;

    /* "asynq/async_task.py":277
 *             self._generator.close()
 *             self._generator = None
 *         if self._dependencies:  # is_blocked(), but faster             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing exit: %s <- %s' % (debug.repr(result), debug.str(self)))
 */
    goto __pyx_L5;
  }

  /* "asynq/async_task.py":282
 *             self._last_value = result
 *         else:
 *             self.set_value(result)             # <<<<<<<<<<<<<<
 * 
 *     def _queue_throw_error(self, error):
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_value(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), __pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L5:;

  /* "asynq/async_task.py":264
 *             self._queue_throw_error(error)
 * 
 *     def _queue_exit(self, result):             # <<<<<<<<<<<<<<
 *         # Result must be unwrapped first,
 *         # and since there can be dependencies, the error result (exception)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._queue_exit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_19_queue_exit(PyObject *__pyx_v_self, PyObject *__pyx_v_result); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_19_queue_exit(PyObject *__pyx_v_self, PyObject *__pyx_v_result) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_queue_exit (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_18_queue_exit(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_result));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_18_queue_exit(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_queue_exit", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask__queue_exit(__pyx_v_self, __pyx_v_result, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._queue_exit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":284
 *             self.set_value(result)
 * 
 *     def _queue_throw_error(self, error):             # <<<<<<<<<<<<<<
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_21_queue_throw_error(PyObject *__pyx_v_self, PyObject *__pyx_v_error); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__queue_throw_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("_queue_throw_error", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_queue_throw_error); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_21_queue_throw_error)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_error); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_error);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":285
 * 
 *     def _queue_throw_error(self, error):
 *         if self._value is not _futures_none:  # is_computed(), but faster             # <<<<<<<<<<<<<<
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster
 */
  __pyx_t_6 = (__pyx_v_self->__pyx_base._value != __pyx_v_5asynq_10async_task__futures_none);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "asynq/async_task.py":286
 *     def _queue_throw_error(self, error):
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)             # <<<<<<<<<<<<<<
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_futures); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_FutureIsAlreadyComputed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_2) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 286, __pyx_L1_error)

    /* "asynq/async_task.py":285
 * 
 *     def _queue_throw_error(self, error):
 *         if self._value is not _futures_none:  # is_computed(), but faster             # <<<<<<<<<<<<<<
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster
 */
  }

  /* "asynq/async_task.py":287
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 */
  __pyx_t_6 = (__pyx_v_self->_generator != Py_None);
  __pyx_t_8 = (__pyx_t_6 != 0);
  if (!__pyx_t_8) {
  } else {
    __pyx_t_7 = __pyx_t_8;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_8 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
  __pyx_t_7 = __pyx_t_8;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_7) {

    /* "asynq/async_task.py":288
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap
 */
    __pyx_t_7 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_QUEUED_RESULTS != 0);
    if (__pyx_t_7) {

      /* "asynq/async_task.py":289
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))             # <<<<<<<<<<<<<<
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap
 *             if self.group_cancel:
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_repr); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_2); __pyx_t_2 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_v_error);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (!__pyx_t_9) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_10, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_queuing_throw_error_s_x_s, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_2) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/async_task.py":288
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster
 *             if _debug_options.DUMP_QUEUED_RESULTS:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap
 */
    }

    /* "asynq/async_task.py":290
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap             # <<<<<<<<<<<<<<
 *             if self.group_cancel:
 *                 self.cancel_dependencies(error)
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_error);
    __Pyx_GIVEREF(__pyx_v_error);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_error);
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5asynq_7futures_ErrorFuture), __pyx_t_1, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->_last_value);
    __Pyx_DECREF(__pyx_v_self->_last_value);
    __pyx_v_self->_last_value = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "asynq/async_task.py":291
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap
 *             if self.group_cancel:             # <<<<<<<<<<<<<<
 *                 self.cancel_dependencies(error)
 *         else:
 */
    __pyx_t_7 = (__pyx_v_self->group_cancel != 0);
    if (__pyx_t_7) {

      /* "asynq/async_task.py":292
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap
 *             if self.group_cancel:
 *                 self.cancel_dependencies(error)             # <<<<<<<<<<<<<<
 *         else:
 *             self.set_error(error)
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->cancel_dependencies(__pyx_v_self, __pyx_v_error, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "asynq/async_task.py":291
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 *             self._last_value = futures.ErrorFuture(error)  # To get it re-thrown on unwrap
 *             if self.group_cancel:             # <<<<<<<<<<<<<<
 *                 self.cancel_dependencies(error)
 *         else:
 */
    }

    /* "asynq/async_task.py":287
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)
 *         if self._generator is not None or self._dependencies:  # can_continue() or is_blocked(), but faster             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_QUEUED_RESULTS:
 *                 debug.write('@async: queuing throw error: %s <-x- %s' % (debug.repr(error), debug.str(self)))
 */
    goto __pyx_L4;
  }

  /* "asynq/async_task.py":294
 *                 self.cancel_dependencies(error)
 *         else:
 *             self.set_error(error)             # <<<<<<<<<<<<<<
 * 
 *     def _remove_dependency(self, dependency):
 */
  /*else*/ {
    __pyx_t_5 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_error(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), __pyx_v_error, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_L4:;

  /* "asynq/async_task.py":284
 *             self.set_value(result)
 * 
 *     def _queue_throw_error(self, error):             # <<<<<<<<<<<<<<
 *         if self._value is not _futures_none:  # is_computed(), but faster
 *             raise futures.FutureIsAlreadyComputed(self)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._queue_throw_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_21_queue_throw_error(PyObject *__pyx_v_self, PyObject *__pyx_v_error); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_21_queue_throw_error(PyObject *__pyx_v_self, PyObject *__pyx_v_error) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_queue_throw_error (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_20_queue_throw_error(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_error));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_20_queue_throw_error(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_error) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_queue_throw_error", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask__queue_throw_error(__pyx_v_self, __pyx_v_error, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._queue_throw_error", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":296
 *             self.set_error(error)
 * 
 *     def _remove_dependency(self, dependency):             # <<<<<<<<<<<<<<
 *         self._remove_dependency_cython(dependency)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_23_remove_dependency(PyObject *__pyx_v_self, PyObject *__pyx_v_dependency); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_23_remove_dependency(PyObject *__pyx_v_self, PyObject *__pyx_v_dependency) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_remove_dependency (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_22_remove_dependency(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_dependency));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_22_remove_dependency(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_dependency) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_remove_dependency", 0);

  /* "asynq/async_task.py":297
 * 
 *     def _remove_dependency(self, dependency):
 *         self._remove_dependency_cython(dependency)             # <<<<<<<<<<<<<<
 * 
 *     def _remove_dependency_cython(self, dependency):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_remove_dependency_cython(__pyx_v_self, __pyx_v_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":296
 *             self.set_error(error)
 * 
 *     def _remove_dependency(self, dependency):             # <<<<<<<<<<<<<<
 *         self._remove_dependency_cython(dependency)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._remove_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":299
 *         self._remove_dependency_cython(dependency)
 * 
 *     def _remove_dependency_cython(self, dependency):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__remove_dependency_cython(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_dependency) {
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("_remove_dependency_cython", 0);

  /* "asynq/async_task.py":300
 * 
 *     def _remove_dependency_cython(self, dependency):
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 *         self._dependencies.remove(dependency)
 */
  __pyx_t_1 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_DEPENDENCIES != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":301
 *     def _remove_dependency_cython(self, dependency):
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))             # <<<<<<<<<<<<<<
 *         self._dependencies.remove(dependency)
 *         if self._value is _futures_none:  # not is_computed(), but faster
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, ((PyObject *)__pyx_v_self));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_dependency); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_v_dependency);
      __Pyx_GIVEREF(__pyx_v_dependency);
      PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_v_dependency);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_dependency_s_got_s, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/async_task.py":300
 * 
 *     def _remove_dependency_cython(self, dependency):
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 *         self._dependencies.remove(dependency)
 */
  }

  /* "asynq/async_task.py":302
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 *         self._dependencies.remove(dependency)             # <<<<<<<<<<<<<<
 *         if self._value is _futures_none:  # not is_computed(), but faster
 *             try:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_dependencies, __pyx_n_s_remove); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_dependency);
    __Pyx_GIVEREF(__pyx_v_dependency);
    PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_dependency);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynq/async_task.py":303
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 *         self._dependencies.remove(dependency)
 *         if self._value is _futures_none:  # not is_computed(), but faster             # <<<<<<<<<<<<<<
 *             try:
 *                 error = dependency.error()
 */
  __pyx_t_1 = (__pyx_v_self->__pyx_base._value == __pyx_v_5asynq_10async_task__futures_none);
  __pyx_t_9 = (__pyx_t_1 != 0);
  if (__pyx_t_9) {

    /* "asynq/async_task.py":304
 *         self._dependencies.remove(dependency)
 *         if self._value is _futures_none:  # not is_computed(), but faster
 *             try:             # <<<<<<<<<<<<<<
 *                 error = dependency.error()
 *                 if not (error is None or isinstance(error, AsyncTaskCancelledError)):
 */
    /*try:*/ {

      /* "asynq/async_task.py":305
 *         if self._value is _futures_none:  # not is_computed(), but faster
 *             try:
 *                 error = dependency.error()             # <<<<<<<<<<<<<<
 *                 if not (error is None or isinstance(error, AsyncTaskCancelledError)):
 *                     self._queue_throw_error(error)
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_dependency, __pyx_n_s_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 305, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_6) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 305, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 305, __pyx_L6_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_error = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "asynq/async_task.py":306
 *             try:
 *                 error = dependency.error()
 *                 if not (error is None or isinstance(error, AsyncTaskCancelledError)):             # <<<<<<<<<<<<<<
 *                     self._queue_throw_error(error)
 *             finally:
 */
      __pyx_t_1 = (__pyx_v_error == Py_None);
      __pyx_t_10 = (__pyx_t_1 != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_9 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskCancelledError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = PyObject_IsInstance(__pyx_v_error, __pyx_t_2); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 306, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_1 = (__pyx_t_10 != 0);
      __pyx_t_9 = __pyx_t_1;
      __pyx_L9_bool_binop_done:;
      __pyx_t_1 = ((!__pyx_t_9) != 0);
      if (__pyx_t_1) {

        /* "asynq/async_task.py":307
 *                 error = dependency.error()
 *                 if not (error is None or isinstance(error, AsyncTaskCancelledError)):
 *                     self._queue_throw_error(error)             # <<<<<<<<<<<<<<
 *             finally:
 *                 if not self._dependencies:  # not is_blocked(), but faster
 */
        __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_queue_throw_error(__pyx_v_self, __pyx_v_error, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynq/async_task.py":306
 *             try:
 *                 error = dependency.error()
 *                 if not (error is None or isinstance(error, AsyncTaskCancelledError)):             # <<<<<<<<<<<<<<
 *                     self._queue_throw_error(error)
 *             finally:
 */
      }
    }

    /* "asynq/async_task.py":309
 *                     self._queue_throw_error(error)
 *             finally:
 *                 if not self._dependencies:  # not is_blocked(), but faster             # <<<<<<<<<<<<<<
 *                     self.scheduler._schedule_without_checks(self)
 * 
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_t_1 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
        __pyx_t_9 = ((!__pyx_t_1) != 0);
        if (__pyx_t_9) {

          /* "asynq/async_task.py":310
 *             finally:
 *                 if not self._dependencies:  # not is_blocked(), but faster
 *                     self.scheduler._schedule_without_checks(self)             # <<<<<<<<<<<<<<
 * 
 *     def make_dependency(self, task, scheduler):
 */
          __pyx_t_9 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->scheduler->__pyx_vtab)->_schedule_without_checks(__pyx_v_self->scheduler, __pyx_v_self); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 310, __pyx_L1_error)

          /* "asynq/async_task.py":309
 *                     self._queue_throw_error(error)
 *             finally:
 *                 if not self._dependencies:  # not is_blocked(), but faster             # <<<<<<<<<<<<<<
 *                     self.scheduler._schedule_without_checks(self)
 * 
 */
        }
        goto __pyx_L7;
      }
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __pyx_L6_error:;
        __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_17);
        __Pyx_XGOTREF(__pyx_t_18);
        __Pyx_XGOTREF(__pyx_t_19);
        __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
        {
          __pyx_t_9 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
          __pyx_t_1 = ((!__pyx_t_9) != 0);
          if (__pyx_t_1) {

            /* "asynq/async_task.py":310
 *             finally:
 *                 if not self._dependencies:  # not is_blocked(), but faster
 *                     self.scheduler._schedule_without_checks(self)             # <<<<<<<<<<<<<<
 * 
 *     def make_dependency(self, task, scheduler):
 */
            __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->scheduler->__pyx_vtab)->_schedule_without_checks(__pyx_v_self->scheduler, __pyx_v_self); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 310, __pyx_L13_error)

            /* "asynq/async_task.py":309
 *                     self._queue_throw_error(error)
 *             finally:
 *                 if not self._dependencies:  # not is_blocked(), but faster             # <<<<<<<<<<<<<<
 *                     self.scheduler._schedule_without_checks(self)
 * 
 */
          }
        }
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
        }
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
        __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
        goto __pyx_L1_error;
        __pyx_L13_error:;
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
        }
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
        goto __pyx_L1_error;
      }
      __pyx_L7:;
    }

    /* "asynq/async_task.py":303
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 *         self._dependencies.remove(dependency)
 *         if self._value is _futures_none:  # not is_computed(), but faster             # <<<<<<<<<<<<<<
 *             try:
 *                 error = dependency.error()
 */
  }

  /* "asynq/async_task.py":299
 *         self._remove_dependency_cython(dependency)
 * 
 *     def _remove_dependency_cython(self, dependency):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: -dependency: %s got %s' % (debug.str(self), debug.str(dependency)))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._remove_dependency_cython", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":312
 *                     self.scheduler._schedule_without_checks(self)
 * 
 *     def make_dependency(self, task, scheduler):             # <<<<<<<<<<<<<<
 *         """Mark self as a dependency on task.
 * 
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_25make_dependency(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_make_dependency(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_task, PyObject *__pyx_v_scheduler, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("make_dependency", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_make_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_25make_dependency)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_v_task);
      __Pyx_GIVEREF(__pyx_v_task);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_task);
      __Pyx_INCREF(__pyx_v_scheduler);
      __Pyx_GIVEREF(__pyx_v_scheduler);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_scheduler);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":318
 * 
 *         """
 *         self.depth = task.depth + 1             # <<<<<<<<<<<<<<
 *         if self.max_depth == 0:
 *             self.max_depth = task.max_depth
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->depth = __pyx_t_7;

  /* "asynq/async_task.py":319
 *         """
 *         self.depth = task.depth + 1
 *         if self.max_depth == 0:             # <<<<<<<<<<<<<<
 *             self.max_depth = task.max_depth
 *         if self.max_depth != 0:
 */
  __pyx_t_8 = ((__pyx_v_self->max_depth == 0) != 0);
  if (__pyx_t_8) {

    /* "asynq/async_task.py":320
 *         self.depth = task.depth + 1
 *         if self.max_depth == 0:
 *             self.max_depth = task.max_depth             # <<<<<<<<<<<<<<
 *         if self.max_depth != 0:
 *             if self.depth > self.max_depth:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_max_depth); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_self->max_depth = __pyx_t_7;

    /* "asynq/async_task.py":319
 *         """
 *         self.depth = task.depth + 1
 *         if self.max_depth == 0:             # <<<<<<<<<<<<<<
 *             self.max_depth = task.max_depth
 *         if self.max_depth != 0:
 */
  }

  /* "asynq/async_task.py":321
 *         if self.max_depth == 0:
 *             self.max_depth = task.max_depth
 *         if self.max_depth != 0:             # <<<<<<<<<<<<<<
 *             if self.depth > self.max_depth:
 *                 debug.dump(scheduler)
 */
  __pyx_t_8 = ((__pyx_v_self->max_depth != 0) != 0);
  if (__pyx_t_8) {

    /* "asynq/async_task.py":322
 *             self.max_depth = task.max_depth
 *         if self.max_depth != 0:
 *             if self.depth > self.max_depth:             # <<<<<<<<<<<<<<
 *                 debug.dump(scheduler)
 *                 assert False, \
 */
    __pyx_t_8 = ((__pyx_v_self->depth > __pyx_v_self->max_depth) != 0);
    if (__pyx_t_8) {

      /* "asynq/async_task.py":323
 *         if self.max_depth != 0:
 *             if self.depth > self.max_depth:
 *                 debug.dump(scheduler)             # <<<<<<<<<<<<<<
 *                 assert False, \
 *                     "Task stack depth exceeded specified maximum (%i)" % self.max_depth
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_dump); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_1) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_scheduler); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
        __Pyx_INCREF(__pyx_v_scheduler);
        __Pyx_GIVEREF(__pyx_v_scheduler);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_scheduler);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynq/async_task.py":324
 *             if self.depth > self.max_depth:
 *                 debug.dump(scheduler)
 *                 assert False, \             # <<<<<<<<<<<<<<
 *                     "Task stack depth exceeded specified maximum (%i)" % self.max_depth
 *         self.caller = task
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        if (unlikely(!0)) {

          /* "asynq/async_task.py":325
 *                 debug.dump(scheduler)
 *                 assert False, \
 *                     "Task stack depth exceeded specified maximum (%i)" % self.max_depth             # <<<<<<<<<<<<<<
 *         self.caller = task
 *         task._dependencies.add(self)
 */
          __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 325, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Task_stack_depth_exceeded_specif, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          PyErr_SetObject(PyExc_AssertionError, __pyx_t_3);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 324, __pyx_L1_error)
        }
      }
      #endif

      /* "asynq/async_task.py":322
 *             self.max_depth = task.max_depth
 *         if self.max_depth != 0:
 *             if self.depth > self.max_depth:             # <<<<<<<<<<<<<<
 *                 debug.dump(scheduler)
 *                 assert False, \
 */
    }

    /* "asynq/async_task.py":321
 *         if self.max_depth == 0:
 *             self.max_depth = task.max_depth
 *         if self.max_depth != 0:             # <<<<<<<<<<<<<<
 *             if self.depth > self.max_depth:
 *                 debug.dump(scheduler)
 */
  }

  /* "asynq/async_task.py":326
 *                 assert False, \
 *                     "Task stack depth exceeded specified maximum (%i)" % self.max_depth
 *         self.caller = task             # <<<<<<<<<<<<<<
 *         task._dependencies.add(self)
 *         self.on_computed.subscribe(task._remove_dependency)
 */
  if (!(likely(((__pyx_v_task) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_task, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 326, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_task;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->caller);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->caller));
  __pyx_v_self->caller = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynq/async_task.py":327
 *                     "Task stack depth exceeded specified maximum (%i)" % self.max_depth
 *         self.caller = task
 *         task._dependencies.add(self)             # <<<<<<<<<<<<<<
 *         self.on_computed.subscribe(task._remove_dependency)
 *         scheduler.schedule(self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_dependencies); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_add); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2); __pyx_t_2 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_1, 0+1, ((PyObject *)__pyx_v_self));
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 327, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynq/async_task.py":328
 *         self.caller = task
 *         task._dependencies.add(self)
 *         self.on_computed.subscribe(task._remove_dependency)             # <<<<<<<<<<<<<<
 *         scheduler.schedule(self)
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.on_computed), __pyx_n_s_subscribe); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_remove_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2); __pyx_t_2 = NULL;
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynq/async_task.py":329
 *         task._dependencies.add(self)
 *         self.on_computed.subscribe(task._remove_dependency)
 *         scheduler.schedule(self)             # <<<<<<<<<<<<<<
 * 
 *     def _before_continue(self):
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_scheduler, __pyx_n_s_schedule); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4); __pyx_t_4 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_1, 0+1, ((PyObject *)__pyx_v_self));
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_1, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynq/async_task.py":312
 *                     self.scheduler._schedule_without_checks(self)
 * 
 *     def make_dependency(self, task, scheduler):             # <<<<<<<<<<<<<<
 *         """Mark self as a dependency on task.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.make_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_25make_dependency(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_10async_task_9AsyncTask_24make_dependency[] = "Mark self as a dependency on task.\n\n        i.e. self needs to be computed before task can be continued further.\n\n        ";
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_25make_dependency(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_task = 0;
  PyObject *__pyx_v_scheduler = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("make_dependency (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_task_2,&__pyx_n_s_scheduler,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_task_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_scheduler)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("make_dependency", 1, 2, 2, 1); __PYX_ERR(0, 312, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "make_dependency") < 0)) __PYX_ERR(0, 312, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_task = values[0];
    __pyx_v_scheduler = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("make_dependency", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 312, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.make_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_24make_dependency(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), __pyx_v_task, __pyx_v_scheduler);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_24make_dependency(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_task, PyObject *__pyx_v_scheduler) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("make_dependency", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask_make_dependency(__pyx_v_self, __pyx_v_task, __pyx_v_scheduler, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.make_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":331
 *         scheduler.schedule(self)
 * 
 *     def _before_continue(self):             # <<<<<<<<<<<<<<
 *         self._resume_contexts()
 * 
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__before_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_before_continue", 0);

  /* "asynq/async_task.py":332
 * 
 *     def _before_continue(self):
 *         self._resume_contexts()             # <<<<<<<<<<<<<<
 * 
 *     def _after_continue(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_resume_contexts(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":331
 *         scheduler.schedule(self)
 * 
 *     def _before_continue(self):             # <<<<<<<<<<<<<<
 *         self._resume_contexts()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._before_continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":334
 *         self._resume_contexts()
 * 
 *     def _after_continue(self):             # <<<<<<<<<<<<<<
 *         self._pause_contexts()
 * 
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__after_continue(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_after_continue", 0);

  /* "asynq/async_task.py":335
 * 
 *     def _after_continue(self):
 *         self._pause_contexts()             # <<<<<<<<<<<<<<
 * 
 *     def traceback(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->_pause_contexts(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/async_task.py":334
 *         self._resume_contexts()
 * 
 *     def _after_continue(self):             # <<<<<<<<<<<<<<
 *         self._pause_contexts()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._after_continue", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":337
 *         self._pause_contexts()
 * 
 *     def traceback(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self_str = self._traceback_line()
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_27traceback(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_traceback(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_self_str = NULL;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("traceback", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_traceback_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_27traceback)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 337, __pyx_L1_error)
      __pyx_r = ((PyObject*)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":338
 * 
 *     def traceback(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self_str = self._traceback_line()
 *         except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "asynq/async_task.py":339
 *     def traceback(self):
 *         try:
 *             self_str = self._traceback_line()             # <<<<<<<<<<<<<<
 *         except Exception:
 *             # If _traceback_line failed for whatever reason (e.g. there is no correct frame_info),
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_traceback_line); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 339, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_3) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L3_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_self_str = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "asynq/async_task.py":338
 * 
 *     def traceback(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self_str = self._traceback_line()
 *         except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/async_task.py":340
 *         try:
 *             self_str = self._traceback_line()
 *         except Exception:             # <<<<<<<<<<<<<<
 *             # If _traceback_line failed for whatever reason (e.g. there is no correct frame_info),
 *             # fall back to __str__ so that we can still provide useful information for debugging
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_Exception);
    if (__pyx_t_8) {
      __Pyx_AddTraceback("asynq.async_task.AsyncTask.traceback", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 340, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "asynq/async_task.py":343
 *             # If _traceback_line failed for whatever reason (e.g. there is no correct frame_info),
 *             # fall back to __str__ so that we can still provide useful information for debugging
 *             self_str = core_helpers.safe_str(self)             # <<<<<<<<<<<<<<
 *         if self.caller is None:
 *             return [self_str]
 */
      __pyx_t_4 = __pyx_f_5qcore_7helpers_safe_str(((PyObject *)__pyx_v_self), 0, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 343, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_self_str, __pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "asynq/async_task.py":338
 * 
 *     def traceback(self):
 *         try:             # <<<<<<<<<<<<<<
 *             self_str = self._traceback_line()
 *         except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_6, __pyx_t_7);
    __pyx_L10_try_end:;
  }

  /* "asynq/async_task.py":344
 *             # fall back to __str__ so that we can still provide useful information for debugging
 *             self_str = core_helpers.safe_str(self)
 *         if self.caller is None:             # <<<<<<<<<<<<<<
 *             return [self_str]
 *         result = self.caller.traceback()
 */
  __pyx_t_9 = (((PyObject *)__pyx_v_self->caller) == Py_None);
  __pyx_t_10 = (__pyx_t_9 != 0);
  if (__pyx_t_10) {

    /* "asynq/async_task.py":345
 *             self_str = core_helpers.safe_str(self)
 *         if self.caller is None:
 *             return [self_str]             # <<<<<<<<<<<<<<
 *         result = self.caller.traceback()
 *         result.append(self_str)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_self_str);
    __Pyx_GIVEREF(__pyx_v_self_str);
    PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_self_str);
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "asynq/async_task.py":344
 *             # fall back to __str__ so that we can still provide useful information for debugging
 *             self_str = core_helpers.safe_str(self)
 *         if self.caller is None:             # <<<<<<<<<<<<<<
 *             return [self_str]
 *         result = self.caller.traceback()
 */
  }

  /* "asynq/async_task.py":346
 *         if self.caller is None:
 *             return [self_str]
 *         result = self.caller.traceback()             # <<<<<<<<<<<<<<
 *         result.append(self_str)
 *         return result
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->caller->__pyx_base.__pyx_vtab)->traceback(__pyx_v_self->caller, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "asynq/async_task.py":347
 *             return [self_str]
 *         result = self.caller.traceback()
 *         result.append(self_str)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  if (unlikely(__pyx_v_result == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(0, 347, __pyx_L1_error)
  }
  __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_self_str); if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 347, __pyx_L1_error)

  /* "asynq/async_task.py":348
 *         result = self.caller.traceback()
 *         result.append(self_str)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def _traceback_line(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "asynq/async_task.py":337
 *         self._pause_contexts()
 * 
 *     def traceback(self):             # <<<<<<<<<<<<<<
 *         try:
 *             self_str = self._traceback_line()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.traceback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_self_str);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_27traceback(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_27traceback(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("traceback (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_26traceback(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_26traceback(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("traceback", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task_9AsyncTask_traceback(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.traceback", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":350
 *         return result
 * 
 *     def _traceback_line(self):             # <<<<<<<<<<<<<<
 *         frame_info = self._frame_info
 *         if frame_info is None and self._generator is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_29_traceback_line(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_29_traceback_line(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_traceback_line (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_28_traceback_line(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_28_traceback_line(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_v_frame_info = NULL;
  PyObject *__pyx_v_template = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_traceback_line", 0);

  /* "asynq/async_task.py":351
 * 
 *     def _traceback_line(self):
 *         frame_info = self._frame_info             # <<<<<<<<<<<<<<
 *         if frame_info is None and self._generator is not None:
 *             frame_info = debug.get_frame_info(self._generator)
 */
  __pyx_t_1 = __pyx_v_self->_frame_info;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_frame_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":352
 *     def _traceback_line(self):
 *         frame_info = self._frame_info
 *         if frame_info is None and self._generator is not None:             # <<<<<<<<<<<<<<
 *             frame_info = debug.get_frame_info(self._generator)
 *         if frame_info is not None:
 */
  __pyx_t_3 = (__pyx_v_frame_info == Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (__pyx_v_self->_generator != Py_None);
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "asynq/async_task.py":353
 *         frame_info = self._frame_info
 *         if frame_info is None and self._generator is not None:
 *             frame_info = debug.get_frame_info(self._generator)             # <<<<<<<<<<<<<<
 *         if frame_info is not None:
 *             template = '''File "%(file)s", line %(lineno)s, in %(funcname)s
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_get_frame_info); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_self->_generator); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_self->_generator);
      __Pyx_GIVEREF(__pyx_v_self->_generator);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_self->_generator);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_frame_info, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynq/async_task.py":352
 *     def _traceback_line(self):
 *         frame_info = self._frame_info
 *         if frame_info is None and self._generator is not None:             # <<<<<<<<<<<<<<
 *             frame_info = debug.get_frame_info(self._generator)
 *         if frame_info is not None:
 */
  }

  /* "asynq/async_task.py":354
 *         if frame_info is None and self._generator is not None:
 *             frame_info = debug.get_frame_info(self._generator)
 *         if frame_info is not None:             # <<<<<<<<<<<<<<
 *             template = '''File "%(file)s", line %(lineno)s, in %(funcname)s
 *     %(codeline)s'''
 */
  __pyx_t_2 = (__pyx_v_frame_info != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "asynq/async_task.py":355
 *             frame_info = debug.get_frame_info(self._generator)
 *         if frame_info is not None:
 *             template = '''File "%(file)s", line %(lineno)s, in %(funcname)s             # <<<<<<<<<<<<<<
 *     %(codeline)s'''
 *             return template % {
 */
    __Pyx_INCREF(__pyx_kp_s_File_file_s_line_lineno_s_in_fun);
    __pyx_v_template = __pyx_kp_s_File_file_s_line_lineno_s_in_fun;

    /* "asynq/async_task.py":357
 *             template = '''File "%(file)s", line %(lineno)s, in %(funcname)s
 *     %(codeline)s'''
 *             return template % {             # <<<<<<<<<<<<<<
 *                 'file': frame_info.filename,
 *                 'lineno': frame_info.lineno,
 */
    __Pyx_XDECREF(__pyx_r);

    /* "asynq/async_task.py":358
 *     %(codeline)s'''
 *             return template % {
 *                 'file': frame_info.filename,             # <<<<<<<<<<<<<<
 *                 'lineno': frame_info.lineno,
 *                 'funcname': frame_info.function,
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame_info, __pyx_n_s_filename); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_file, __pyx_t_6) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "asynq/async_task.py":359
 *             return template % {
 *                 'file': frame_info.filename,
 *                 'lineno': frame_info.lineno,             # <<<<<<<<<<<<<<
 *                 'funcname': frame_info.function,
 *                 'codeline': '\n'.join(frame_info.code_context).strip()
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame_info, __pyx_n_s_lineno); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_lineno, __pyx_t_6) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "asynq/async_task.py":360
 *                 'file': frame_info.filename,
 *                 'lineno': frame_info.lineno,
 *                 'funcname': frame_info.function,             # <<<<<<<<<<<<<<
 *                 'codeline': '\n'.join(frame_info.code_context).strip()
 *             }
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame_info, __pyx_n_s_function); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_funcname, __pyx_t_6) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "asynq/async_task.py":361
 *                 'lineno': frame_info.lineno,
 *                 'funcname': frame_info.function,
 *                 'codeline': '\n'.join(frame_info.code_context).strip()             # <<<<<<<<<<<<<<
 *             }
 *         else:
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_frame_info, __pyx_n_s_code_context); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = __Pyx_PyString_Join(__pyx_kp_s_, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_strip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 361, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_6 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 361, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_codeline, __pyx_t_6) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "asynq/async_task.py":357
 *             template = '''File "%(file)s", line %(lineno)s, in %(funcname)s
 *     %(codeline)s'''
 *             return template % {             # <<<<<<<<<<<<<<
 *                 'file': frame_info.filename,
 *                 'lineno': frame_info.lineno,
 */
    __pyx_t_6 = PyNumber_Remainder(__pyx_v_template, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "asynq/async_task.py":354
 *         if frame_info is None and self._generator is not None:
 *             frame_info = debug.get_frame_info(self._generator)
 *         if frame_info is not None:             # <<<<<<<<<<<<<<
 *             template = '''File "%(file)s", line %(lineno)s, in %(funcname)s
 *     %(codeline)s'''
 */
  }

  /* "asynq/async_task.py":364
 *             }
 *         else:
 *             return str(self)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_self));
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "asynq/async_task.py":350
 *         return result
 * 
 *     def _traceback_line(self):             # <<<<<<<<<<<<<<
 *         frame_info = self._frame_info
 *         if frame_info is None and self._generator is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._traceback_line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_frame_info);
  __Pyx_XDECREF(__pyx_v_template);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":366
 *             return str(self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         fn_str = core_inspection.get_function_call_str(self.fn, self.args, self.kwargs)
 *         name = '@async %s' % fn_str
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_31__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_31__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_30__str__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_30__str__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_v_fn_str = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_step = NULL;
  PyObject *__pyx_v_status = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "asynq/async_task.py":367
 * 
 *     def __str__(self):
 *         fn_str = core_inspection.get_function_call_str(self.fn, self.args, self.kwargs)             # <<<<<<<<<<<<<<
 *         name = '@async %s' % fn_str
 * 
 */
  __pyx_t_1 = __pyx_v_self->fn;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyTuple_CheckExact(__pyx_v_self->args))||((__pyx_v_self->args) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_self->args)->tp_name), 0))) __PYX_ERR(0, 367, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_self->args;
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(PyDict_CheckExact(__pyx_v_self->kwargs))||((__pyx_v_self->kwargs) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_self->kwargs)->tp_name), 0))) __PYX_ERR(0, 367, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_self->kwargs;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = __pyx_f_5qcore_10inspection_get_function_call_str(__pyx_t_1, ((PyObject*)__pyx_t_2), ((PyObject*)__pyx_t_3), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_fn_str = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "asynq/async_task.py":368
 *     def __str__(self):
 *         fn_str = core_inspection.get_function_call_str(self.fn, self.args, self.kwargs)
 *         name = '@async %s' % fn_str             # <<<<<<<<<<<<<<
 * 
 *         # we subtract one because by the time the stacktrace is printed
 */
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_s, __pyx_v_fn_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_name = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "asynq/async_task.py":372
 *         # we subtract one because by the time the stacktrace is printed
 *         # the iteration_index has already been incremented
 *         if self.iteration_index - 1 == 0:             # <<<<<<<<<<<<<<
 *             step = 'before 1st yield'
 *         else:
 */
  __pyx_t_5 = (((__pyx_v_self->iteration_index - 1) == 0) != 0);
  if (__pyx_t_5) {

    /* "asynq/async_task.py":373
 *         # the iteration_index has already been incremented
 *         if self.iteration_index - 1 == 0:
 *             step = 'before 1st yield'             # <<<<<<<<<<<<<<
 *         else:
 *             step = 'passed yield #%i' % (self.iteration_index - 1)
 */
    __Pyx_INCREF(__pyx_kp_s_before_1st_yield);
    __pyx_v_step = __pyx_kp_s_before_1st_yield;

    /* "asynq/async_task.py":372
 *         # we subtract one because by the time the stacktrace is printed
 *         # the iteration_index has already been incremented
 *         if self.iteration_index - 1 == 0:             # <<<<<<<<<<<<<<
 *             step = 'before 1st yield'
 *         else:
 */
    goto __pyx_L3;
  }

  /* "asynq/async_task.py":375
 *             step = 'before 1st yield'
 *         else:
 *             step = 'passed yield #%i' % (self.iteration_index - 1)             # <<<<<<<<<<<<<<
 * 
 *         if self.is_computed():
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_self->iteration_index - 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_passed_yield_i, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_step = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "asynq/async_task.py":377
 *             step = 'passed yield #%i' % (self.iteration_index - 1)
 * 
 *         if self.is_computed():             # <<<<<<<<<<<<<<
 *             status = 'computed, '
 *             if self.error() is None:
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "asynq/async_task.py":378
 * 
 *         if self.is_computed():
 *             status = 'computed, '             # <<<<<<<<<<<<<<
 *             if self.error() is None:
 *                 status += '= ' + repr(self.value())
 */
    __Pyx_INCREF(__pyx_kp_s_computed_2);
    __pyx_v_status = __pyx_kp_s_computed_2;

    /* "asynq/async_task.py":379
 *         if self.is_computed():
 *             status = 'computed, '
 *             if self.error() is None:             # <<<<<<<<<<<<<<
 *                 status += '= ' + repr(self.value())
 *             else:
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.error(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = (__pyx_t_3 == Py_None);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "asynq/async_task.py":380
 *             status = 'computed, '
 *             if self.error() is None:
 *                 status += '= ' + repr(self.value())             # <<<<<<<<<<<<<<
 *             else:
 *                 status += 'error = ' + repr(self.error())
 */
      __pyx_t_3 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.value(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_Repr(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_Add(__pyx_kp_s__2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_status, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF_SET(__pyx_v_status, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "asynq/async_task.py":379
 *         if self.is_computed():
 *             status = 'computed, '
 *             if self.error() is None:             # <<<<<<<<<<<<<<
 *                 status += '= ' + repr(self.value())
 *             else:
 */
      goto __pyx_L5;
    }

    /* "asynq/async_task.py":382
 *                 status += '= ' + repr(self.value())
 *             else:
 *                 status += 'error = ' + repr(self.error())             # <<<<<<<<<<<<<<
 *         else:
 *             if self.is_blocked():
 */
    /*else*/ {
      __pyx_t_4 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.error(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_Repr(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Add(__pyx_kp_s_error_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_status, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_status, __pyx_t_3);
      __pyx_t_3 = 0;
    }
    __pyx_L5:;

    /* "asynq/async_task.py":377
 *             step = 'passed yield #%i' % (self.iteration_index - 1)
 * 
 *         if self.is_computed():             # <<<<<<<<<<<<<<
 *             status = 'computed, '
 *             if self.error() is None:
 */
    goto __pyx_L4;
  }

  /* "asynq/async_task.py":384
 *                 status += 'error = ' + repr(self.error())
 *         else:
 *             if self.is_blocked():             # <<<<<<<<<<<<<<
 *                 status = 'blocked x%i' % len(self._dependencies)
 *             elif self.can_continue():
 */
  /*else*/ {
    __pyx_t_5 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->is_blocked(__pyx_v_self, 0); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 384, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "asynq/async_task.py":385
 *         else:
 *             if self.is_blocked():
 *                 status = 'blocked x%i' % len(self._dependencies)             # <<<<<<<<<<<<<<
 *             elif self.can_continue():
 *                 if self.is_scheduled:
 */
      __pyx_t_3 = __pyx_v_self->_dependencies;
      __Pyx_INCREF(__pyx_t_3);
      if (unlikely(__pyx_t_3 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 385, __pyx_L1_error)
      }
      __pyx_t_7 = PySet_GET_SIZE(__pyx_t_3); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_blocked_x_i, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_status = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "asynq/async_task.py":384
 *                 status += 'error = ' + repr(self.error())
 *         else:
 *             if self.is_blocked():             # <<<<<<<<<<<<<<
 *                 status = 'blocked x%i' % len(self._dependencies)
 *             elif self.can_continue():
 */
      goto __pyx_L6;
    }

    /* "asynq/async_task.py":386
 *             if self.is_blocked():
 *                 status = 'blocked x%i' % len(self._dependencies)
 *             elif self.can_continue():             # <<<<<<<<<<<<<<
 *                 if self.is_scheduled:
 *                     status = 'scheduled'
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_self->__pyx_base.__pyx_vtab)->can_continue(__pyx_v_self, 0); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 386, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "asynq/async_task.py":387
 *                 status = 'blocked x%i' % len(self._dependencies)
 *             elif self.can_continue():
 *                 if self.is_scheduled:             # <<<<<<<<<<<<<<
 *                     status = 'scheduled'
 *                 elif self.scheduler:
 */
      __pyx_t_5 = (__pyx_v_self->is_scheduled != 0);
      if (__pyx_t_5) {

        /* "asynq/async_task.py":388
 *             elif self.can_continue():
 *                 if self.is_scheduled:
 *                     status = 'scheduled'             # <<<<<<<<<<<<<<
 *                 elif self.scheduler:
 *                     status = 'waiting'
 */
        __Pyx_INCREF(__pyx_n_s_scheduled);
        __pyx_v_status = __pyx_n_s_scheduled;

        /* "asynq/async_task.py":387
 *                 status = 'blocked x%i' % len(self._dependencies)
 *             elif self.can_continue():
 *                 if self.is_scheduled:             # <<<<<<<<<<<<<<
 *                     status = 'scheduled'
 *                 elif self.scheduler:
 */
        goto __pyx_L7;
      }

      /* "asynq/async_task.py":389
 *                 if self.is_scheduled:
 *                     status = 'scheduled'
 *                 elif self.scheduler:             # <<<<<<<<<<<<<<
 *                     status = 'waiting'
 *                 else:
 */
      __pyx_t_5 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->scheduler)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 389, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "asynq/async_task.py":390
 *                     status = 'scheduled'
 *                 elif self.scheduler:
 *                     status = 'waiting'             # <<<<<<<<<<<<<<
 *                 else:
 *                     status = 'new'
 */
        __Pyx_INCREF(__pyx_n_s_waiting);
        __pyx_v_status = __pyx_n_s_waiting;

        /* "asynq/async_task.py":389
 *                 if self.is_scheduled:
 *                     status = 'scheduled'
 *                 elif self.scheduler:             # <<<<<<<<<<<<<<
 *                     status = 'waiting'
 *                 else:
 */
        goto __pyx_L7;
      }

      /* "asynq/async_task.py":392
 *                     status = 'waiting'
 *                 else:
 *                     status = 'new'             # <<<<<<<<<<<<<<
 *             else:
 *                 status = 'almost finished (generator is closed)'
 */
      /*else*/ {
        __Pyx_INCREF(__pyx_n_s_new);
        __pyx_v_status = __pyx_n_s_new;
      }
      __pyx_L7:;

      /* "asynq/async_task.py":386
 *             if self.is_blocked():
 *                 status = 'blocked x%i' % len(self._dependencies)
 *             elif self.can_continue():             # <<<<<<<<<<<<<<
 *                 if self.is_scheduled:
 *                     status = 'scheduled'
 */
      goto __pyx_L6;
    }

    /* "asynq/async_task.py":394
 *                     status = 'new'
 *             else:
 *                 status = 'almost finished (generator is closed)'             # <<<<<<<<<<<<<<
 *                 if self.is_scheduled:
 *                     status += ', scheduled'
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_kp_s_almost_finished_generator_is_clo);
      __pyx_v_status = __pyx_kp_s_almost_finished_generator_is_clo;

      /* "asynq/async_task.py":395
 *             else:
 *                 status = 'almost finished (generator is closed)'
 *                 if self.is_scheduled:             # <<<<<<<<<<<<<<
 *                     status += ', scheduled'
 *         return '%s (%s, %s)' % (name, status, step)
 */
      __pyx_t_5 = (__pyx_v_self->is_scheduled != 0);
      if (__pyx_t_5) {

        /* "asynq/async_task.py":396
 *                 status = 'almost finished (generator is closed)'
 *                 if self.is_scheduled:
 *                     status += ', scheduled'             # <<<<<<<<<<<<<<
 *         return '%s (%s, %s)' % (name, status, step)
 * 
 */
        __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_status, __pyx_kp_s_scheduled_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF_SET(__pyx_v_status, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "asynq/async_task.py":395
 *             else:
 *                 status = 'almost finished (generator is closed)'
 *                 if self.is_scheduled:             # <<<<<<<<<<<<<<
 *                     status += ', scheduled'
 *         return '%s (%s, %s)' % (name, status, step)
 */
      }
    }
    __pyx_L6:;
  }
  __pyx_L4:;

  /* "asynq/async_task.py":397
 *                 if self.is_scheduled:
 *                     status += ', scheduled'
 *         return '%s (%s, %s)' % (name, status, step)             # <<<<<<<<<<<<<<
 * 
 *     def dump(self, indent=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_name);
  __Pyx_INCREF(__pyx_v_status);
  __Pyx_GIVEREF(__pyx_v_status);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_status);
  __Pyx_INCREF(__pyx_v_step);
  __Pyx_GIVEREF(__pyx_v_step);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_step);
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_s_s, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "asynq/async_task.py":366
 *             return str(self)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         fn_str = core_inspection.get_function_call_str(self.fn, self.args, self.kwargs)
 *         name = '@async %s' % fn_str
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fn_str);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_step);
  __Pyx_XDECREF(__pyx_v_status);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":399
 *         return '%s (%s, %s)' % (name, status, step)
 * 
 *     def dump(self, indent=0):             # <<<<<<<<<<<<<<
 *         if indent > MAX_DUMP_INDENT:
 *             debug.write('...', indent + 1)
 */

static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_33dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask_dump(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump *__pyx_optional_args) {
  int __pyx_v_indent = ((int)0);
  PyObject *__pyx_v_dependency = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  __Pyx_RefNannySetupContext("dump", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_indent = __pyx_optional_args->indent;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_33dump)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_indent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/async_task.py":400
 * 
 *     def dump(self, indent=0):
 *         if indent > MAX_DUMP_INDENT:             # <<<<<<<<<<<<<<
 *             debug.write('...', indent + 1)
 *             return
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_indent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_DUMP_INDENT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_7) {

    /* "asynq/async_task.py":401
 *     def dump(self, indent=0):
 *         if indent > MAX_DUMP_INDENT:
 *             debug.write('...', indent + 1)             # <<<<<<<<<<<<<<
 *             return
 *         debug.write(debug.str(self), indent)
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_3 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s__3);
    __Pyx_GIVEREF(__pyx_kp_s__3);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_8, __pyx_kp_s__3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_8, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "asynq/async_task.py":402
 *         if indent > MAX_DUMP_INDENT:
 *             debug.write('...', indent + 1)
 *             return             # <<<<<<<<<<<<<<
 *         debug.write(debug.str(self), indent)
 *         if self._dependencies:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/async_task.py":400
 * 
 *     def dump(self, indent=0):
 *         if indent > MAX_DUMP_INDENT:             # <<<<<<<<<<<<<<
 *             debug.write('...', indent + 1)
 *             return
 */
  }

  /* "asynq/async_task.py":403
 *             debug.write('...', indent + 1)
 *             return
 *         debug.write(debug.str(self), indent)             # <<<<<<<<<<<<<<
 *         if self._dependencies:
 *             debug.write('Dependencies:', indent + 1)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_self));
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_indent); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = NULL;
  __pyx_t_8 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_8 = 1;
    }
  }
  __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_5) {
    __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_t_6);
  __pyx_t_1 = 0;
  __pyx_t_6 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "asynq/async_task.py":404
 *             return
 *         debug.write(debug.str(self), indent)
 *         if self._dependencies:             # <<<<<<<<<<<<<<
 *             debug.write('Dependencies:', indent + 1)
 *             for dependency in self._dependencies:
 */
  __pyx_t_7 = (__pyx_v_self->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_dependencies) != 0);
  if (__pyx_t_7) {

    /* "asynq/async_task.py":405
 *         debug.write(debug.str(self), indent)
 *         if self._dependencies:
 *             debug.write('Dependencies:', indent + 1)             # <<<<<<<<<<<<<<
 *             for dependency in self._dependencies:
 *                 dependency.dump(indent + 2)  # Recursive
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_1 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Dependencies);
    __Pyx_GIVEREF(__pyx_kp_s_Dependencies);
    PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_8, __pyx_kp_s_Dependencies);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_8, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "asynq/async_task.py":406
 *         if self._dependencies:
 *             debug.write('Dependencies:', indent + 1)
 *             for dependency in self._dependencies:             # <<<<<<<<<<<<<<
 *                 dependency.dump(indent + 2)  # Recursive
 *                 # debug.write(debug.str(dependency), indent + 2)
 */
    __pyx_t_4 = PyObject_GetIter(__pyx_v_self->_dependencies); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 406, __pyx_L1_error)
    for (;;) {
      {
        __pyx_t_2 = __pyx_t_9(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 406, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_dependency, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynq/async_task.py":407
 *             debug.write('Dependencies:', indent + 1)
 *             for dependency in self._dependencies:
 *                 dependency.dump(indent + 2)  # Recursive             # <<<<<<<<<<<<<<
 *                 # debug.write(debug.str(dependency), indent + 2)
 *         else:
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_dependency, __pyx_n_s_dump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyInt_From_long((__pyx_v_indent + 2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 407, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynq/async_task.py":406
 *         if self._dependencies:
 *             debug.write('Dependencies:', indent + 1)
 *             for dependency in self._dependencies:             # <<<<<<<<<<<<<<
 *                 dependency.dump(indent + 2)  # Recursive
 *                 # debug.write(debug.str(dependency), indent + 2)
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "asynq/async_task.py":404
 *             return
 *         debug.write(debug.str(self), indent)
 *         if self._dependencies:             # <<<<<<<<<<<<<<
 *             debug.write('Dependencies:', indent + 1)
 *             for dependency in self._dependencies:
 */
    goto __pyx_L4;
  }

  /* "asynq/async_task.py":410
 *                 # debug.write(debug.str(dependency), indent + 2)
 *         else:
 *             debug.write('No dependencies.', indent + 1)             # <<<<<<<<<<<<<<
 * 
 *     # Contexts support
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_3 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_No_dependencies);
    __Pyx_GIVEREF(__pyx_kp_s_No_dependencies);
    PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_8, __pyx_kp_s_No_dependencies);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_8, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L4:;

  /* "asynq/async_task.py":399
 *         return '%s (%s, %s)' % (name, status, step)
 * 
 *     def dump(self, indent=0):             # <<<<<<<<<<<<<<
 *         if indent > MAX_DUMP_INDENT:
 *             debug.write('...', indent + 1)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dependency);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_33dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_33dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_indent;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dump (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_indent,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_indent);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dump") < 0)) __PYX_ERR(0, 399, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_indent = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_indent == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
    } else {
      __pyx_v_indent = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dump", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 399, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_32dump(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), __pyx_v_indent);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_32dump(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, int __pyx_v_indent) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump __pyx_t_2;
  __Pyx_RefNannySetupContext("dump", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.indent = __pyx_v_indent;
  __pyx_t_1 = __pyx_vtabptr_5asynq_10async_task_AsyncTask->__pyx_base.dump(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_self), 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":414
 *     # Contexts support
 * 
 *     def _enter_context(self, context):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: +context: %s' % debug.str(context))
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__enter_context(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("_enter_context", 0);

  /* "asynq/async_task.py":415
 * 
 *     def _enter_context(self, context):
 *         if _debug_options.DUMP_CONTEXTS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: +context: %s' % debug.str(context))
 *         self._contexts.append(context)
 */
  __pyx_t_1 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_CONTEXTS != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":416
 *     def _enter_context(self, context):
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: +context: %s' % debug.str(context))             # <<<<<<<<<<<<<<
 *         self._contexts.append(context)
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_context);
      __Pyx_GIVEREF(__pyx_v_context);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_context);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_context_s, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 416, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/async_task.py":415
 * 
 *     def _enter_context(self, context):
 *         if _debug_options.DUMP_CONTEXTS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: +context: %s' % debug.str(context))
 *         self._contexts.append(context)
 */
  }

  /* "asynq/async_task.py":417
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: +context: %s' % debug.str(context))
 *         self._contexts.append(context)             # <<<<<<<<<<<<<<
 * 
 *     def _leave_context(self, context):
 */
  if (unlikely(__pyx_v_self->_contexts == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
    __PYX_ERR(0, 417, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_self->_contexts, __pyx_v_context); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 417, __pyx_L1_error)

  /* "asynq/async_task.py":414
 *     # Contexts support
 * 
 *     def _enter_context(self, context):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: +context: %s' % debug.str(context))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._enter_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":419
 *         self._contexts.append(context)
 * 
 *     def _leave_context(self, context):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: -context: %s' % debug.str(context))
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__leave_context(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("_leave_context", 0);

  /* "asynq/async_task.py":420
 * 
 *     def _leave_context(self, context):
 *         if _debug_options.DUMP_CONTEXTS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: -context: %s' % debug.str(context))
 *         self._contexts.remove(context)
 */
  __pyx_t_1 = (__pyx_v_5asynq_10async_task__debug_options->DUMP_CONTEXTS != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":421
 *     def _leave_context(self, context):
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: -context: %s' % debug.str(context))             # <<<<<<<<<<<<<<
 *         self._contexts.remove(context)
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_context); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_context);
      __Pyx_GIVEREF(__pyx_v_context);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_context);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_context_s_2, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 421, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/async_task.py":420
 * 
 *     def _leave_context(self, context):
 *         if _debug_options.DUMP_CONTEXTS:             # <<<<<<<<<<<<<<
 *             debug.write('@async: -context: %s' % debug.str(context))
 *         self._contexts.remove(context)
 */
  }

  /* "asynq/async_task.py":422
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: -context: %s' % debug.str(context))
 *         self._contexts.remove(context)             # <<<<<<<<<<<<<<
 * 
 *     def _pause_contexts(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_contexts, __pyx_n_s_remove); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (!__pyx_t_7) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_context); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
    __Pyx_INCREF(__pyx_v_context);
    __Pyx_GIVEREF(__pyx_v_context);
    PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_context);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 422, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "asynq/async_task.py":419
 *         self._contexts.append(context)
 * 
 *     def _leave_context(self, context):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_CONTEXTS:
 *             debug.write('@async: -context: %s' % debug.str(context))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._leave_context", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":424
 *         self._contexts.remove(context)
 * 
 *     def _pause_contexts(self):             # <<<<<<<<<<<<<<
 *         contexts = self._contexts
 *         i = len(contexts) - 1
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__pause_contexts(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_contexts = NULL;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("_pause_contexts", 0);

  /* "asynq/async_task.py":425
 * 
 *     def _pause_contexts(self):
 *         contexts = self._contexts             # <<<<<<<<<<<<<<
 *         i = len(contexts) - 1
 *         # execute each __pause__() in a try/except and if 1 or more of them
 */
  __pyx_t_1 = __pyx_v_self->_contexts;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_contexts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":426
 *     def _pause_contexts(self):
 *         contexts = self._contexts
 *         i = len(contexts) - 1             # <<<<<<<<<<<<<<
 *         # execute each __pause__() in a try/except and if 1 or more of them
 *         # raise an exception, then save the last exception raised so that it
 */
  if (unlikely(__pyx_v_contexts == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 426, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_contexts); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 426, __pyx_L1_error)
  __pyx_v_i = (__pyx_t_2 - 1);

  /* "asynq/async_task.py":431
 *         # can be re-raised later. We re-raise the last exception to make the
 *         # behavior consistent with __exit__.
 *         error = None             # <<<<<<<<<<<<<<
 *         while i >= 0:
 *             try:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_error = Py_None;

  /* "asynq/async_task.py":432
 *         # behavior consistent with __exit__.
 *         error = None
 *         while i >= 0:             # <<<<<<<<<<<<<<
 *             try:
 *                 contexts[i].__pause__()
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_i >= 0) != 0);
    if (!__pyx_t_3) break;

    /* "asynq/async_task.py":433
 *         error = None
 *         while i >= 0:
 *             try:             # <<<<<<<<<<<<<<
 *                 contexts[i].__pause__()
 *             except BaseException as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "asynq/async_task.py":434
 *         while i >= 0:
 *             try:
 *                 contexts[i].__pause__()             # <<<<<<<<<<<<<<
 *             except BaseException as e:
 *                 error = e
 */
        if (unlikely(__pyx_v_contexts == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 434, __pyx_L5_error)
        }
        __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_contexts, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 434, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_pause); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 434, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L5_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/async_task.py":433
 *         error = None
 *         while i >= 0:
 *             try:             # <<<<<<<<<<<<<<
 *                 contexts[i].__pause__()
 *             except BaseException as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/async_task.py":435
 *             try:
 *                 contexts[i].__pause__()
 *             except BaseException as e:             # <<<<<<<<<<<<<<
 *                 error = e
 *                 core_errors.prepare_for_reraise(error)
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("asynq.async_task.AsyncTask._pause_contexts", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(0, 435, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_8);

        /* "asynq/async_task.py":436
 *                 contexts[i].__pause__()
 *             except BaseException as e:
 *                 error = e             # <<<<<<<<<<<<<<
 *                 core_errors.prepare_for_reraise(error)
 *             i -= 1
 */
        __Pyx_INCREF(__pyx_v_e);
        __Pyx_DECREF_SET(__pyx_v_error, __pyx_v_e);

        /* "asynq/async_task.py":437
 *             except BaseException as e:
 *                 error = e
 *                 core_errors.prepare_for_reraise(error)             # <<<<<<<<<<<<<<
 *             i -= 1
 *         if error is not None:
 */
        __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_core_errors); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 437, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_prepare_for_reraise); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 437, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        if (!__pyx_t_11) {
          __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_error); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 437, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_10);
        } else {
          __pyx_t_13 = PyTuple_New(1+1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 437, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
          __Pyx_INCREF(__pyx_v_error);
          __Pyx_GIVEREF(__pyx_v_error);
          PyTuple_SET_ITEM(__pyx_t_13, 0+1, __pyx_v_error);
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_13, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 437, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "asynq/async_task.py":433
 *         error = None
 *         while i >= 0:
 *             try:             # <<<<<<<<<<<<<<
 *                 contexts[i].__pause__()
 *             except BaseException as e:
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_L12_try_end:;
    }

    /* "asynq/async_task.py":438
 *                 error = e
 *                 core_errors.prepare_for_reraise(error)
 *             i -= 1             # <<<<<<<<<<<<<<
 *         if error is not None:
 *             self._accept_error(error)
 */
    __pyx_v_i = (__pyx_v_i - 1);
  }

  /* "asynq/async_task.py":439
 *                 core_errors.prepare_for_reraise(error)
 *             i -= 1
 *         if error is not None:             # <<<<<<<<<<<<<<
 *             self._accept_error(error)
 * 
 */
  __pyx_t_3 = (__pyx_v_error != Py_None);
  __pyx_t_14 = (__pyx_t_3 != 0);
  if (__pyx_t_14) {

    /* "asynq/async_task.py":440
 *             i -= 1
 *         if error is not None:
 *             self._accept_error(error)             # <<<<<<<<<<<<<<
 * 
 *     def _resume_contexts(self):
 */
    __pyx_t_7 = __pyx_f_5asynq_10async_task_9AsyncTask__accept_error(__pyx_v_self, __pyx_v_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "asynq/async_task.py":439
 *                 core_errors.prepare_for_reraise(error)
 *             i -= 1
 *         if error is not None:             # <<<<<<<<<<<<<<
 *             self._accept_error(error)
 * 
 */
  }

  /* "asynq/async_task.py":424
 *         self._contexts.remove(context)
 * 
 *     def _pause_contexts(self):             # <<<<<<<<<<<<<<
 *         contexts = self._contexts
 *         i = len(contexts) - 1
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._pause_contexts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_contexts);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":442
 *             self._accept_error(error)
 * 
 *     def _resume_contexts(self):             # <<<<<<<<<<<<<<
 *         i = 0
 *         contexts = self._contexts
 */

static PyObject *__pyx_f_5asynq_10async_task_9AsyncTask__resume_contexts(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_v_i;
  PyObject *__pyx_v_contexts = NULL;
  Py_ssize_t __pyx_v_l;
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("_resume_contexts", 0);

  /* "asynq/async_task.py":443
 * 
 *     def _resume_contexts(self):
 *         i = 0             # <<<<<<<<<<<<<<
 *         contexts = self._contexts
 *         l = len(contexts)
 */
  __pyx_v_i = 0;

  /* "asynq/async_task.py":444
 *     def _resume_contexts(self):
 *         i = 0
 *         contexts = self._contexts             # <<<<<<<<<<<<<<
 *         l = len(contexts)
 *         # same try/except deal as with _pause_contexts, but in this case
 */
  __pyx_t_1 = __pyx_v_self->_contexts;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_contexts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/async_task.py":445
 *         i = 0
 *         contexts = self._contexts
 *         l = len(contexts)             # <<<<<<<<<<<<<<
 *         # same try/except deal as with _pause_contexts, but in this case
 *         # we re-raise the first exception raised.
 */
  if (unlikely(__pyx_v_contexts == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 445, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_v_contexts); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_v_l = __pyx_t_2;

  /* "asynq/async_task.py":448
 *         # same try/except deal as with _pause_contexts, but in this case
 *         # we re-raise the first exception raised.
 *         error = None             # <<<<<<<<<<<<<<
 *         while i < l:
 *             try:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_error = Py_None;

  /* "asynq/async_task.py":449
 *         # we re-raise the first exception raised.
 *         error = None
 *         while i < l:             # <<<<<<<<<<<<<<
 *             try:
 *                 contexts[i].__resume__()
 */
  while (1) {
    __pyx_t_3 = ((__pyx_v_i < __pyx_v_l) != 0);
    if (!__pyx_t_3) break;

    /* "asynq/async_task.py":450
 *         error = None
 *         while i < l:
 *             try:             # <<<<<<<<<<<<<<
 *                 contexts[i].__resume__()
 *             except BaseException as e:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "asynq/async_task.py":451
 *         while i < l:
 *             try:
 *                 contexts[i].__resume__()             # <<<<<<<<<<<<<<
 *             except BaseException as e:
 *                 if error is None:
 */
        if (unlikely(__pyx_v_contexts == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 451, __pyx_L5_error)
        }
        __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_contexts, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 451, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_resume); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 451, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_7)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_7);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        if (__pyx_t_7) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L5_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L5_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/async_task.py":450
 *         error = None
 *         while i < l:
 *             try:             # <<<<<<<<<<<<<<
 *                 contexts[i].__resume__()
 *             except BaseException as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L12_try_end;
      __pyx_L5_error:;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/async_task.py":452
 *             try:
 *                 contexts[i].__resume__()
 *             except BaseException as e:             # <<<<<<<<<<<<<<
 *                 if error is None:
 *                     error = e
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("asynq.async_task.AsyncTask._resume_contexts", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(0, 452, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_8);

        /* "asynq/async_task.py":453
 *                 contexts[i].__resume__()
 *             except BaseException as e:
 *                 if error is None:             # <<<<<<<<<<<<<<
 *                     error = e
 *                     core_errors.prepare_for_reraise(error)
 */
        __pyx_t_3 = (__pyx_v_error == Py_None);
        __pyx_t_10 = (__pyx_t_3 != 0);
        if (__pyx_t_10) {

          /* "asynq/async_task.py":454
 *             except BaseException as e:
 *                 if error is None:
 *                     error = e             # <<<<<<<<<<<<<<
 *                     core_errors.prepare_for_reraise(error)
 *             i += 1
 */
          __Pyx_INCREF(__pyx_v_e);
          __Pyx_DECREF_SET(__pyx_v_error, __pyx_v_e);

          /* "asynq/async_task.py":455
 *                 if error is None:
 *                     error = e
 *                     core_errors.prepare_for_reraise(error)             # <<<<<<<<<<<<<<
 *             i += 1
 *         if error is not None:
 */
          __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_core_errors); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 455, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_prepare_for_reraise); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 455, __pyx_L7_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
            }
          }
          if (!__pyx_t_12) {
            __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_error); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 455, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_11);
          } else {
            __pyx_t_14 = PyTuple_New(1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 455, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_12); __pyx_t_12 = NULL;
            __Pyx_INCREF(__pyx_v_error);
            __Pyx_GIVEREF(__pyx_v_error);
            PyTuple_SET_ITEM(__pyx_t_14, 0+1, __pyx_v_error);
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_14, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 455, __pyx_L7_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "asynq/async_task.py":453
 *                 contexts[i].__resume__()
 *             except BaseException as e:
 *                 if error is None:             # <<<<<<<<<<<<<<
 *                     error = e
 *                     core_errors.prepare_for_reraise(error)
 */
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "asynq/async_task.py":450
 *         error = None
 *         while i < l:
 *             try:             # <<<<<<<<<<<<<<
 *                 contexts[i].__resume__()
 *             except BaseException as e:
 */
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_PyThreadState_assign
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_4, __pyx_t_5, __pyx_t_6);
      __pyx_L12_try_end:;
    }

    /* "asynq/async_task.py":456
 *                     error = e
 *                     core_errors.prepare_for_reraise(error)
 *             i += 1             # <<<<<<<<<<<<<<
 *         if error is not None:
 *             self._accept_error(error)
 */
    __pyx_v_i = (__pyx_v_i + 1);
  }

  /* "asynq/async_task.py":457
 *                     core_errors.prepare_for_reraise(error)
 *             i += 1
 *         if error is not None:             # <<<<<<<<<<<<<<
 *             self._accept_error(error)
 * 
 */
  __pyx_t_10 = (__pyx_v_error != Py_None);
  __pyx_t_3 = (__pyx_t_10 != 0);
  if (__pyx_t_3) {

    /* "asynq/async_task.py":458
 *             i += 1
 *         if error is not None:
 *             self._accept_error(error)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_7 = __pyx_f_5asynq_10async_task_9AsyncTask__accept_error(__pyx_v_self, __pyx_v_error); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "asynq/async_task.py":457
 *                     core_errors.prepare_for_reraise(error)
 *             i += 1
 *         if error is not None:             # <<<<<<<<<<<<<<
 *             self._accept_error(error)
 * 
 */
  }

  /* "asynq/async_task.py":442
 *             self._accept_error(error)
 * 
 *     def _resume_contexts(self):             # <<<<<<<<<<<<<<
 *         i = 0
 *         contexts = self._contexts
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._resume_contexts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_contexts);
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":32
 * 
 * cdef class AsyncTask(futures.FutureBase):
 *     cdef public bint is_scheduled             # <<<<<<<<<<<<<<
 *     cdef public long iteration_index
 *     cdef public bint group_cancel
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_12is_scheduled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_12is_scheduled_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_12is_scheduled___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_12is_scheduled___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_scheduled); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.is_scheduled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_12is_scheduled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_12is_scheduled_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_12is_scheduled_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_12is_scheduled_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 32, __pyx_L1_error)
  __pyx_v_self->is_scheduled = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.is_scheduled.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":33
 * cdef class AsyncTask(futures.FutureBase):
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index             # <<<<<<<<<<<<<<
 *     cdef public bint group_cancel
 *     cdef public object fn
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_15iteration_index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_15iteration_index_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_15iteration_index___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_15iteration_index___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_self->iteration_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.iteration_index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_15iteration_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_15iteration_index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_15iteration_index_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_15iteration_index_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_long(__pyx_v_value); if (unlikely((__pyx_t_1 == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 33, __pyx_L1_error)
  __pyx_v_self->iteration_index = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.iteration_index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":34
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index
 *     cdef public bint group_cancel             # <<<<<<<<<<<<<<
 *     cdef public object fn
 *     cdef public object args
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_12group_cancel_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_12group_cancel_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_12group_cancel___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_12group_cancel___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->group_cancel); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.group_cancel.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_12group_cancel_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_12group_cancel_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_12group_cancel_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_12group_cancel_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 34, __pyx_L1_error)
  __pyx_v_self->group_cancel = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.group_cancel.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":35
 *     cdef public long iteration_index
 *     cdef public bint group_cancel
 *     cdef public object fn             # <<<<<<<<<<<<<<
 *     cdef public object args
 *     cdef public object kwargs
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_2fn_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_2fn_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_2fn___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_2fn___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->fn);
  __pyx_r = __pyx_v_self->fn;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_2fn_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_2fn_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->fn);
  __Pyx_DECREF(__pyx_v_self->fn);
  __pyx_v_self->fn = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_2fn_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_2fn_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->fn);
  __Pyx_DECREF(__pyx_v_self->fn);
  __pyx_v_self->fn = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":36
 *     cdef public bint group_cancel
 *     cdef public object fn
 *     cdef public object args             # <<<<<<<<<<<<<<
 *     cdef public object kwargs
 *     cdef public AsyncTask caller
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_4args_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_4args_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_4args___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_4args___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->args);
  __pyx_r = __pyx_v_self->args;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_4args_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_4args_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_4args_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_4args_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_4args_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_4args_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_4args_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_4args_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->args);
  __Pyx_DECREF(__pyx_v_self->args);
  __pyx_v_self->args = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":37
 *     cdef public object fn
 *     cdef public object args
 *     cdef public object kwargs             # <<<<<<<<<<<<<<
 *     cdef public AsyncTask caller
 *     cdef public int depth
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->kwargs);
  __pyx_r = __pyx_v_self->kwargs;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_6kwargs_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->kwargs);
  __Pyx_DECREF(__pyx_v_self->kwargs);
  __pyx_v_self->kwargs = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":38
 *     cdef public object args
 *     cdef public object kwargs
 *     cdef public AsyncTask caller             # <<<<<<<<<<<<<<
 *     cdef public int depth
 *     cdef public int max_depth
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_6caller_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_6caller_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6caller___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_6caller___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->caller));
  __pyx_r = ((PyObject *)__pyx_v_self->caller);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6caller_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_6caller_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(1, 38, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->caller);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->caller));
  __pyx_v_self->caller = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.caller.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_6caller_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_6caller_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->caller);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->caller));
  __pyx_v_self->caller = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":39
 *     cdef public object kwargs
 *     cdef public AsyncTask caller
 *     cdef public int depth             # <<<<<<<<<<<<<<
 *     cdef public int max_depth
 *     cdef public scheduler.TaskScheduler scheduler
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_5depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_5depth_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_5depth___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_5depth___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_5depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_5depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_5depth_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_5depth_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 39, __pyx_L1_error)
  __pyx_v_self->depth = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":40
 *     cdef public AsyncTask caller
 *     cdef public int depth
 *     cdef public int max_depth             # <<<<<<<<<<<<<<
 *     cdef public scheduler.TaskScheduler scheduler
 *     cdef public AsyncTask creator
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9max_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9max_depth_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9max_depth___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_9max_depth___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->max_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.max_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9max_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9max_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9max_depth_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_9max_depth_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_v_self->max_depth = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.max_depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":41
 *     cdef public int depth
 *     cdef public int max_depth
 *     cdef public scheduler.TaskScheduler scheduler             # <<<<<<<<<<<<<<
 *     cdef public AsyncTask creator
 *     cdef public object _generator
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->scheduler));
  __pyx_r = ((PyObject *)__pyx_v_self->scheduler);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5asynq_9scheduler_TaskScheduler))))) __PYX_ERR(1, 41, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->scheduler);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scheduler));
  __pyx_v_self->scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.scheduler.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_9scheduler_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->scheduler);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->scheduler));
  __pyx_v_self->scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":42
 *     cdef public int max_depth
 *     cdef public scheduler.TaskScheduler scheduler
 *     cdef public AsyncTask creator             # <<<<<<<<<<<<<<
 *     cdef public object _generator
 *     cdef public object _last_value
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_7creator_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_7creator_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_7creator___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_7creator___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->creator));
  __pyx_r = ((PyObject *)__pyx_v_self->creator);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_7creator_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_7creator_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(1, 42, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->creator);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->creator));
  __pyx_v_self->creator = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask.creator.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_7creator_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_7creator_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->creator);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->creator));
  __pyx_v_self->creator = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":43
 *     cdef public scheduler.TaskScheduler scheduler
 *     cdef public AsyncTask creator
 *     cdef public object _generator             # <<<<<<<<<<<<<<
 *     cdef public object _last_value
 *     cdef public object _frame_info
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_10_generator___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_generator);
  __pyx_r = __pyx_v_self->_generator;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_generator);
  __Pyx_DECREF(__pyx_v_self->_generator);
  __pyx_v_self->_generator = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_10_generator_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_generator);
  __Pyx_DECREF(__pyx_v_self->_generator);
  __pyx_v_self->_generator = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":44
 *     cdef public AsyncTask creator
 *     cdef public object _generator
 *     cdef public object _last_value             # <<<<<<<<<<<<<<
 *     cdef public object _frame_info
 *     cdef public set _dependencies
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_last_value);
  __pyx_r = __pyx_v_self->_last_value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_last_value);
  __Pyx_DECREF(__pyx_v_self->_last_value);
  __pyx_v_self->_last_value = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_last_value_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_last_value);
  __Pyx_DECREF(__pyx_v_self->_last_value);
  __pyx_v_self->_last_value = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":45
 *     cdef public object _generator
 *     cdef public object _last_value
 *     cdef public object _frame_info             # <<<<<<<<<<<<<<
 *     cdef public set _dependencies
 *     cdef public list _contexts
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_frame_info);
  __pyx_r = __pyx_v_self->_frame_info;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_frame_info);
  __Pyx_DECREF(__pyx_v_self->_frame_info);
  __pyx_v_self->_frame_info = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_11_frame_info_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_frame_info);
  __Pyx_DECREF(__pyx_v_self->_frame_info);
  __pyx_v_self->_frame_info = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":46
 *     cdef public object _last_value
 *     cdef public object _frame_info
 *     cdef public set _dependencies             # <<<<<<<<<<<<<<
 *     cdef public list _contexts
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_dependencies);
  __pyx_r = __pyx_v_self->_dependencies;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 46, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_dependencies);
  __Pyx_DECREF(__pyx_v_self->_dependencies);
  __pyx_v_self->_dependencies = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._dependencies.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_13_dependencies_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_dependencies);
  __Pyx_DECREF(__pyx_v_self->_dependencies);
  __pyx_v_self->_dependencies = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.pxd":47
 *     cdef public object _frame_info
 *     cdef public set _dependencies
 *     cdef public list _contexts             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_blocked(self) except -1
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts___get__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts___get__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_contexts);
  __pyx_r = __pyx_v_self->_contexts;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts_2__set__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts_2__set__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 47, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_contexts);
  __Pyx_DECREF(__pyx_v_self->_contexts);
  __pyx_v_self->_contexts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task.AsyncTask._contexts.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts_4__del__(((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_10async_task_9AsyncTask_9_contexts_4__del__(struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_contexts);
  __Pyx_DECREF(__pyx_v_self->_contexts);
  __pyx_v_self->_contexts = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":461
 * 
 * 
 * def unwrap(value):             # <<<<<<<<<<<<<<
 *     """
 *     'Unwraps' the provided arguments by converting:
 */

static PyObject *__pyx_f_5asynq_10async_task_unwrap(PyObject *__pyx_v_value) {
  PyObject *__pyx_v_dct = 0;
  struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future = 0;
  int __pyx_v_length;
  PyObject *__pyx_v_lst = 0;
  PyObject *__pyx_v_tpl = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  __Pyx_RefNannySetupContext("unwrap", 0);

  /* "asynq/async_task.py":474
 * 
 *     """
 *     if value is None:  # Special case             # <<<<<<<<<<<<<<
 *         return None
 *     elif isinstance(value, futures.FutureBase):
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/async_task.py":475
 *     """
 *     if value is None:  # Special case
 *         return None             # <<<<<<<<<<<<<<
 *     elif isinstance(value, futures.FutureBase):
 *         future = value
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "asynq/async_task.py":474
 * 
 *     """
 *     if value is None:  # Special case             # <<<<<<<<<<<<<<
 *         return None
 *     elif isinstance(value, futures.FutureBase):
 */
  }

  /* "asynq/async_task.py":476
 *     if value is None:  # Special case
 *         return None
 *     elif isinstance(value, futures.FutureBase):             # <<<<<<<<<<<<<<
 *         future = value
 *         return future.value()
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_7futures_FutureBase); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":477
 *         return None
 *     elif isinstance(value, futures.FutureBase):
 *         future = value             # <<<<<<<<<<<<<<
 *         return future.value()
 *     elif type(value) is tuple:
 */
    if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5asynq_7futures_FutureBase))))) __PYX_ERR(0, 477, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_value;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_future = ((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynq/async_task.py":478
 *     elif isinstance(value, futures.FutureBase):
 *         future = value
 *         return future.value()             # <<<<<<<<<<<<<<
 *     elif type(value) is tuple:
 *         tpl = value
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_5asynq_7futures_FutureBase *)__pyx_v_future->__pyx_vtab)->value(__pyx_v_future, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "asynq/async_task.py":476
 *     if value is None:  # Special case
 *         return None
 *     elif isinstance(value, futures.FutureBase):             # <<<<<<<<<<<<<<
 *         future = value
 *         return future.value()
 */
  }

  /* "asynq/async_task.py":479
 *         future = value
 *         return future.value()
 *     elif type(value) is tuple:             # <<<<<<<<<<<<<<
 *         tpl = value
 *         # Simple perf. optimization
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyTuple_Type)));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/async_task.py":480
 *         return future.value()
 *     elif type(value) is tuple:
 *         tpl = value             # <<<<<<<<<<<<<<
 *         # Simple perf. optimization
 *         length = len(tpl)
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 480, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_value;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_tpl = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynq/async_task.py":482
 *         tpl = value
 *         # Simple perf. optimization
 *         length = len(tpl)             # <<<<<<<<<<<<<<
 *         if length <= 1:
 *             if length == 0:
 */
    if (unlikely(__pyx_v_tpl == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 482, __pyx_L1_error)
    }
    __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v_tpl); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __pyx_v_length = __pyx_t_4;

    /* "asynq/async_task.py":483
 *         # Simple perf. optimization
 *         length = len(tpl)
 *         if length <= 1:             # <<<<<<<<<<<<<<
 *             if length == 0:
 *                 return ()
 */
    __pyx_t_2 = ((__pyx_v_length <= 1) != 0);
    if (__pyx_t_2) {

      /* "asynq/async_task.py":484
 *         length = len(tpl)
 *         if length <= 1:
 *             if length == 0:             # <<<<<<<<<<<<<<
 *                 return ()
 *             return (unwrap(tpl[0]),)
 */
      __pyx_t_2 = ((__pyx_v_length == 0) != 0);
      if (__pyx_t_2) {

        /* "asynq/async_task.py":485
 *         if length <= 1:
 *             if length == 0:
 *                 return ()             # <<<<<<<<<<<<<<
 *             return (unwrap(tpl[0]),)
 *         elif length == 2:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_empty_tuple);
        __pyx_r = __pyx_empty_tuple;
        goto __pyx_L0;

        /* "asynq/async_task.py":484
 *         length = len(tpl)
 *         if length <= 1:
 *             if length == 0:             # <<<<<<<<<<<<<<
 *                 return ()
 *             return (unwrap(tpl[0]),)
 */
      }

      /* "asynq/async_task.py":486
 *             if length == 0:
 *                 return ()
 *             return (unwrap(tpl[0]),)             # <<<<<<<<<<<<<<
 *         elif length == 2:
 *             return (unwrap(tpl[0]), unwrap(tpl[1]))
 */
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(__pyx_v_tpl == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 486, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_tpl, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __pyx_f_5asynq_10async_task_unwrap(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "asynq/async_task.py":483
 *         # Simple perf. optimization
 *         length = len(tpl)
 *         if length <= 1:             # <<<<<<<<<<<<<<
 *             if length == 0:
 *                 return ()
 */
    }

    /* "asynq/async_task.py":487
 *                 return ()
 *             return (unwrap(tpl[0]),)
 *         elif length == 2:             # <<<<<<<<<<<<<<
 *             return (unwrap(tpl[0]), unwrap(tpl[1]))
 *         else:
 */
    __pyx_t_2 = ((__pyx_v_length == 2) != 0);
    if (__pyx_t_2) {

      /* "asynq/async_task.py":488
 *             return (unwrap(tpl[0]),)
 *         elif length == 2:
 *             return (unwrap(tpl[0]), unwrap(tpl[1]))             # <<<<<<<<<<<<<<
 *         else:
 *             result = []
 */
      __Pyx_XDECREF(__pyx_r);
      if (unlikely(__pyx_v_tpl == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 488, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_tpl, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __pyx_f_5asynq_10async_task_unwrap(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(__pyx_v_tpl == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 488, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_tpl, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __pyx_f_5asynq_10async_task_unwrap(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "asynq/async_task.py":487
 *                 return ()
 *             return (unwrap(tpl[0]),)
 *         elif length == 2:             # <<<<<<<<<<<<<<
 *             return (unwrap(tpl[0]), unwrap(tpl[1]))
 *         else:
 */
    }

    /* "asynq/async_task.py":490
 *             return (unwrap(tpl[0]), unwrap(tpl[1]))
 *         else:
 *             result = []             # <<<<<<<<<<<<<<
 *             for item in tpl:
 *                 result.append(unwrap(item))
 */
    /*else*/ {
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_result = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "asynq/async_task.py":491
 *         else:
 *             result = []
 *             for item in tpl:             # <<<<<<<<<<<<<<
 *                 result.append(unwrap(item))
 *             return tuple(result)
 */
      if (unlikely(__pyx_v_tpl == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
        __PYX_ERR(0, 491, __pyx_L1_error)
      }
      __pyx_t_3 = __pyx_v_tpl; __Pyx_INCREF(__pyx_t_3); __pyx_t_4 = 0;
      for (;;) {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 491, __pyx_L1_error)
        #else
        __pyx_t_6 = PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "asynq/async_task.py":492
 *             result = []
 *             for item in tpl:
 *                 result.append(unwrap(item))             # <<<<<<<<<<<<<<
 *             return tuple(result)
 *     elif type(value) is list:
 */
        __pyx_t_6 = __pyx_f_5asynq_10async_task_unwrap(__pyx_v_item); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 492, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_6); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 492, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "asynq/async_task.py":491
 *         else:
 *             result = []
 *             for item in tpl:             # <<<<<<<<<<<<<<
 *                 result.append(unwrap(item))
 *             return tuple(result)
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/async_task.py":493
 *             for item in tpl:
 *                 result.append(unwrap(item))
 *             return tuple(result)             # <<<<<<<<<<<<<<
 *     elif type(value) is list:
 *         lst = value
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_r = __pyx_t_3;
      __pyx_t_3 = 0;
      goto __pyx_L0;
    }

    /* "asynq/async_task.py":479
 *         future = value
 *         return future.value()
 *     elif type(value) is tuple:             # <<<<<<<<<<<<<<
 *         tpl = value
 *         # Simple perf. optimization
 */
  }

  /* "asynq/async_task.py":494
 *                 result.append(unwrap(item))
 *             return tuple(result)
 *     elif type(value) is list:             # <<<<<<<<<<<<<<
 *         lst = value
 *         return [unwrap(item) for item in lst]
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyList_Type)));
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":495
 *             return tuple(result)
 *     elif type(value) is list:
 *         lst = value             # <<<<<<<<<<<<<<
 *         return [unwrap(item) for item in lst]
 *     elif type(value) is dict:
 */
    if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 495, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_value;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_lst = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynq/async_task.py":496
 *     elif type(value) is list:
 *         lst = value
 *         return [unwrap(item) for item in lst]             # <<<<<<<<<<<<<<
 *     elif type(value) is dict:
 *         dct = value
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(__pyx_v_lst == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 496, __pyx_L1_error)
    }
    __pyx_t_6 = __pyx_v_lst; __Pyx_INCREF(__pyx_t_6); __pyx_t_4 = 0;
    for (;;) {
      if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_6)) break;
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 496, __pyx_L1_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __pyx_f_5asynq_10async_task_unwrap(__pyx_v_item); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "asynq/async_task.py":494
 *                 result.append(unwrap(item))
 *             return tuple(result)
 *     elif type(value) is list:             # <<<<<<<<<<<<<<
 *         lst = value
 *         return [unwrap(item) for item in lst]
 */
  }

  /* "asynq/async_task.py":497
 *         lst = value
 *         return [unwrap(item) for item in lst]
 *     elif type(value) is dict:             # <<<<<<<<<<<<<<
 *         dct = value
 *         return {key: unwrap(value) for key, value in six.iteritems(dct)}
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyDict_Type)));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/async_task.py":498
 *         return [unwrap(item) for item in lst]
 *     elif type(value) is dict:
 *         dct = value             # <<<<<<<<<<<<<<
 *         return {key: unwrap(value) for key, value in six.iteritems(dct)}
 *     else:
 */
    if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(0, 498, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_value;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_dct = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "asynq/async_task.py":499
 *     elif type(value) is dict:
 *         dct = value
 *         return {key: unwrap(value) for key, value in six.iteritems(dct)}             # <<<<<<<<<<<<<<
 *     else:
 *         raise TypeError(
 */
    __Pyx_XDECREF(__pyx_r);
    { /* enter inner scope */
      PyObject *__pyx_7genexpr__pyx_v_key = NULL;
      PyObject *__pyx_7genexpr__pyx_v_value = NULL;
      __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_six); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_iteritems); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 499, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_v_dct); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_6);
      } else {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 499, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_dct);
        __Pyx_GIVEREF(__pyx_v_dct);
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_v_dct);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_8 = __pyx_t_6; __Pyx_INCREF(__pyx_t_8); __pyx_t_4 = 0;
        __pyx_t_10 = NULL;
      } else {
        __pyx_t_4 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 499, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_10 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 499, __pyx_L12_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_10)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 499, __pyx_L12_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 499, __pyx_L12_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_8, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L12_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_10(__pyx_t_8);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 499, __pyx_L12_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          #if CYTHON_COMPILING_IN_CPYTHON
          Py_ssize_t size = Py_SIZE(sequence);
          #else
          Py_ssize_t size = PySequence_Size(sequence);
          #endif
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 499, __pyx_L12_error)
          }
          #if CYTHON_COMPILING_IN_CPYTHON
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_9 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_9 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_5);
          #else
          __pyx_t_9 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 499, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_11 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 499, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
          index = 0; __pyx_t_9 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_9)) goto __pyx_L15_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_9);
          index = 1; __pyx_t_5 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_5)) goto __pyx_L15_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_5);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 2) < 0) __PYX_ERR(0, 499, __pyx_L12_error)
          __pyx_t_12 = NULL;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L16_unpacking_done;
          __pyx_L15_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_12 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 499, __pyx_L12_error)
          __pyx_L16_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_key, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_value, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_6 = __pyx_f_5asynq_10async_task_unwrap(__pyx_7genexpr__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 499, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_6);
        if (unlikely(PyDict_SetItem(__pyx_t_3, (PyObject*)__pyx_7genexpr__pyx_v_key, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 499, __pyx_L12_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_key);
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_value);
      goto __pyx_L17_exit_scope;
      __pyx_L12_error:;
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_key);
      __Pyx_XDECREF(__pyx_7genexpr__pyx_v_value);
      goto __pyx_L1_error;
      __pyx_L17_exit_scope:;
    } /* exit inner scope */
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "asynq/async_task.py":497
 *         lst = value
 *         return [unwrap(item) for item in lst]
 *     elif type(value) is dict:             # <<<<<<<<<<<<<<
 *         dct = value
 *         return {key: unwrap(value) for key, value in six.iteritems(dct)}
 */
  }

  /* "asynq/async_task.py":501
 *         return {key: unwrap(value) for key, value in six.iteritems(dct)}
 *     else:
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "Cannot unwrap an object of type '%s': only futures and None are allowed." %
 *             type(value)
 */
  /*else*/ {

    /* "asynq/async_task.py":502
 *     else:
 *         raise TypeError(
 *             "Cannot unwrap an object of type '%s': only futures and None are allowed." %             # <<<<<<<<<<<<<<
 *             type(value)
 *         )
 */
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_unwrap_an_object_of_type, ((PyObject *)Py_TYPE(__pyx_v_value))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 502, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "asynq/async_task.py":501
 *         return {key: unwrap(value) for key, value in six.iteritems(dct)}
 *     else:
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "Cannot unwrap an object of type '%s': only futures and None are allowed." %
 *             type(value)
 */
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 501, __pyx_L1_error)
  }

  /* "asynq/async_task.py":461
 * 
 * 
 * def unwrap(value):             # <<<<<<<<<<<<<<
 *     """
 *     'Unwraps' the provided arguments by converting:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynq.async_task.unwrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dct);
  __Pyx_XDECREF((PyObject *)__pyx_v_future);
  __Pyx_XDECREF(__pyx_v_lst);
  __Pyx_XDECREF(__pyx_v_tpl);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/async_task.py":515
 * 
 * 
 * def _cancel_futures(value, error):             # <<<<<<<<<<<<<<
 *     """Used by ``AsyncTask._continue`` to cancel evaluation of tasks
 *     and futures due to failure of one of them.
 */

static PyObject *__pyx_pw_5asynq_10async_task_1_cancel_futures(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_10async_task__cancel_futures(PyObject *__pyx_v_value, PyObject *__pyx_v_error, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("_cancel_futures", 0);

  /* "asynq/async_task.py":520
 * 
 *     """
 *     if value is None:             # <<<<<<<<<<<<<<
 *         return
 *     if isinstance(value, AsyncTask):
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/async_task.py":521
 *     """
 *     if value is None:
 *         return             # <<<<<<<<<<<<<<
 *     if isinstance(value, AsyncTask):
 *         if not value.is_computed():
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/async_task.py":520
 * 
 *     """
 *     if value is None:             # <<<<<<<<<<<<<<
 *         return
 *     if isinstance(value, AsyncTask):
 */
  }

  /* "asynq/async_task.py":522
 *     if value is None:
 *         return
 *     if isinstance(value, AsyncTask):             # <<<<<<<<<<<<<<
 *         if not value.is_computed():
 *             value._queue_throw_error(error)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_10async_task_AsyncTask); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":523
 *         return
 *     if isinstance(value, AsyncTask):
 *         if not value.is_computed():             # <<<<<<<<<<<<<<
 *             value._queue_throw_error(error)
 *     elif isinstance(value, futures.FutureBase):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_is_computed); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = ((!__pyx_t_1) != 0);
    if (__pyx_t_2) {

      /* "asynq/async_task.py":524
 *     if isinstance(value, AsyncTask):
 *         if not value.is_computed():
 *             value._queue_throw_error(error)             # <<<<<<<<<<<<<<
 *     elif isinstance(value, futures.FutureBase):
 *         if not value.is_computed():
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_queue_throw_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 524, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_error);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/async_task.py":523
 *         return
 *     if isinstance(value, AsyncTask):
 *         if not value.is_computed():             # <<<<<<<<<<<<<<
 *             value._queue_throw_error(error)
 *     elif isinstance(value, futures.FutureBase):
 */
    }

    /* "asynq/async_task.py":522
 *     if value is None:
 *         return
 *     if isinstance(value, AsyncTask):             # <<<<<<<<<<<<<<
 *         if not value.is_computed():
 *             value._queue_throw_error(error)
 */
    goto __pyx_L4;
  }

  /* "asynq/async_task.py":525
 *         if not value.is_computed():
 *             value._queue_throw_error(error)
 *     elif isinstance(value, futures.FutureBase):             # <<<<<<<<<<<<<<
 *         if not value.is_computed():
 *             value.set_error(error)
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_7futures_FutureBase); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/async_task.py":526
 *             value._queue_throw_error(error)
 *     elif isinstance(value, futures.FutureBase):
 *         if not value.is_computed():             # <<<<<<<<<<<<<<
 *             value.set_error(error)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_is_computed); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = ((!__pyx_t_1) != 0);
    if (__pyx_t_2) {

      /* "asynq/async_task.py":527
 *     elif isinstance(value, futures.FutureBase):
 *         if not value.is_computed():
 *             value.set_error(error)             # <<<<<<<<<<<<<<
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_set_error); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 527, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_error); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 527, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_error);
        __Pyx_GIVEREF(__pyx_v_error);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_error);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 527, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/async_task.py":526
 *             value._queue_throw_error(error)
 *     elif isinstance(value, futures.FutureBase):
 *         if not value.is_computed():             # <<<<<<<<<<<<<<
 *             value.set_error(error)
 */
    }

    /* "asynq/async_task.py":525
 *         if not value.is_computed():
 *             value._queue_throw_error(error)
 *     elif isinstance(value, futures.FutureBase):             # <<<<<<<<<<<<<<
 *         if not value.is_computed():
 *             value.set_error(error)
 */
  }
  __pyx_L4:;

  /* "asynq/async_task.py":515
 * 
 * 
 * def _cancel_futures(value, error):             # <<<<<<<<<<<<<<
 *     """Used by ``AsyncTask._continue`` to cancel evaluation of tasks
 *     and futures due to failure of one of them.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.async_task._cancel_futures", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_10async_task_1_cancel_futures(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_10async_task__cancel_futures[] = "Used by ``AsyncTask._continue`` to cancel evaluation of tasks\n    and futures due to failure of one of them.\n\n    ";
static PyObject *__pyx_pw_5asynq_10async_task_1_cancel_futures(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_error = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_cancel_futures (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_error,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_error)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_cancel_futures", 1, 2, 2, 1); __PYX_ERR(0, 515, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_cancel_futures") < 0)) __PYX_ERR(0, 515, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_value = values[0];
    __pyx_v_error = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_cancel_futures", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 515, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.async_task._cancel_futures", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_10async_task__cancel_futures(__pyx_self, __pyx_v_value, __pyx_v_error);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_10async_task__cancel_futures(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value, PyObject *__pyx_v_error) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_cancel_futures", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_10async_task__cancel_futures(__pyx_v_value, __pyx_v_error, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.async_task._cancel_futures", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask __pyx_vtable_5asynq_10async_task_AsyncTask;

static PyObject *__pyx_tp_new_5asynq_10async_task_AsyncTask(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *p;
  PyObject *o = __pyx_ptype_5asynq_7futures_FutureBase->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5asynq_7futures_FutureBase*)__pyx_vtabptr_5asynq_10async_task_AsyncTask;
  p->fn = Py_None; Py_INCREF(Py_None);
  p->args = Py_None; Py_INCREF(Py_None);
  p->kwargs = Py_None; Py_INCREF(Py_None);
  p->caller = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None); Py_INCREF(Py_None);
  p->scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None); Py_INCREF(Py_None);
  p->creator = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None); Py_INCREF(Py_None);
  p->_generator = Py_None; Py_INCREF(Py_None);
  p->_last_value = Py_None; Py_INCREF(Py_None);
  p->_frame_info = Py_None; Py_INCREF(Py_None);
  p->_dependencies = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_contexts = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5asynq_10async_task_AsyncTask(PyObject *o) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *p = (struct __pyx_obj_5asynq_10async_task_AsyncTask *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->fn);
  Py_CLEAR(p->args);
  Py_CLEAR(p->kwargs);
  Py_CLEAR(p->caller);
  Py_CLEAR(p->scheduler);
  Py_CLEAR(p->creator);
  Py_CLEAR(p->_generator);
  Py_CLEAR(p->_last_value);
  Py_CLEAR(p->_frame_info);
  Py_CLEAR(p->_dependencies);
  Py_CLEAR(p->_contexts);
  PyObject_GC_Track(o);
  if (likely(__pyx_ptype_5asynq_7futures_FutureBase)) __pyx_ptype_5asynq_7futures_FutureBase->tp_dealloc(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_5asynq_10async_task_AsyncTask);
}

static int __pyx_tp_traverse_5asynq_10async_task_AsyncTask(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *p = (struct __pyx_obj_5asynq_10async_task_AsyncTask *)o;
  e = ((likely(__pyx_ptype_5asynq_7futures_FutureBase)) ? ((__pyx_ptype_5asynq_7futures_FutureBase->tp_traverse) ? __pyx_ptype_5asynq_7futures_FutureBase->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_5asynq_10async_task_AsyncTask)); if (e) return e;
  if (p->fn) {
    e = (*v)(p->fn, a); if (e) return e;
  }
  if (p->args) {
    e = (*v)(p->args, a); if (e) return e;
  }
  if (p->kwargs) {
    e = (*v)(p->kwargs, a); if (e) return e;
  }
  if (p->caller) {
    e = (*v)(((PyObject*)p->caller), a); if (e) return e;
  }
  if (p->scheduler) {
    e = (*v)(((PyObject*)p->scheduler), a); if (e) return e;
  }
  if (p->creator) {
    e = (*v)(((PyObject*)p->creator), a); if (e) return e;
  }
  if (p->_generator) {
    e = (*v)(p->_generator, a); if (e) return e;
  }
  if (p->_last_value) {
    e = (*v)(p->_last_value, a); if (e) return e;
  }
  if (p->_frame_info) {
    e = (*v)(p->_frame_info, a); if (e) return e;
  }
  if (p->_dependencies) {
    e = (*v)(p->_dependencies, a); if (e) return e;
  }
  if (p->_contexts) {
    e = (*v)(p->_contexts, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5asynq_10async_task_AsyncTask(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *p = (struct __pyx_obj_5asynq_10async_task_AsyncTask *)o;
  if (likely(__pyx_ptype_5asynq_7futures_FutureBase)) { if (__pyx_ptype_5asynq_7futures_FutureBase->tp_clear) __pyx_ptype_5asynq_7futures_FutureBase->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_5asynq_10async_task_AsyncTask);
  tmp = ((PyObject*)p->fn);
  p->fn = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->args);
  p->args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->kwargs);
  p->kwargs = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->caller);
  p->caller = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->scheduler);
  p->scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->creator);
  p->creator = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_generator);
  p->_generator = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_value);
  p->_last_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_frame_info);
  p->_frame_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_dependencies);
  p->_dependencies = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_contexts);
  p->_contexts = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_is_scheduled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_12is_scheduled_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_is_scheduled(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_12is_scheduled_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_iteration_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_15iteration_index_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_iteration_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_15iteration_index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_group_cancel(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_12group_cancel_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_group_cancel(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_12group_cancel_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_fn(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_fn(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_2fn_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_args(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_4args_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_args(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_4args_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_4args_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_kwargs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_kwargs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_6kwargs_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_caller(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_caller(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_6caller_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_5depth_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_5depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_max_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_9max_depth_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_max_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_9max_depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_scheduler(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_scheduler(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_9scheduler_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask_creator(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask_creator(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_7creator_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask__generator(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask__generator(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_10_generator_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask__last_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask__last_value(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_11_last_value_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask__frame_info(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask__frame_info(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_11_frame_info_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask__dependencies(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask__dependencies(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_13_dependencies_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_10async_task_9AsyncTask__contexts(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_1__get__(o);
}

static int __pyx_setprop_5asynq_10async_task_9AsyncTask__contexts(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_10async_task_9AsyncTask_9_contexts_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_5asynq_10async_task_AsyncTask[] = {
  {"can_continue", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_3can_continue, METH_NOARGS, __pyx_doc_5asynq_10async_task_9AsyncTask_2can_continue},
  {"is_blocked", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_5is_blocked, METH_NOARGS, __pyx_doc_5asynq_10async_task_9AsyncTask_4is_blocked},
  {"start", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_7start, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_10async_task_9AsyncTask_6start},
  {"after", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_9after, METH_O, __pyx_doc_5asynq_10async_task_9AsyncTask_8after},
  {"cancel_dependencies", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_11cancel_dependencies, METH_O, 0},
  {"_compute", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_13_compute, METH_NOARGS, 0},
  {"_computed", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_15_computed, METH_NOARGS, 0},
  {"_continue", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_17_continue, METH_NOARGS, 0},
  {"_queue_exit", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_19_queue_exit, METH_O, 0},
  {"_queue_throw_error", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_21_queue_throw_error, METH_O, 0},
  {"_remove_dependency", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_23_remove_dependency, METH_O, 0},
  {"make_dependency", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_25make_dependency, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_10async_task_9AsyncTask_24make_dependency},
  {"traceback", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_27traceback, METH_NOARGS, 0},
  {"_traceback_line", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_29_traceback_line, METH_NOARGS, 0},
  {"dump", (PyCFunction)__pyx_pw_5asynq_10async_task_9AsyncTask_33dump, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5asynq_10async_task_AsyncTask[] = {
  {(char *)"is_scheduled", __pyx_getprop_5asynq_10async_task_9AsyncTask_is_scheduled, __pyx_setprop_5asynq_10async_task_9AsyncTask_is_scheduled, (char *)0, 0},
  {(char *)"iteration_index", __pyx_getprop_5asynq_10async_task_9AsyncTask_iteration_index, __pyx_setprop_5asynq_10async_task_9AsyncTask_iteration_index, (char *)0, 0},
  {(char *)"group_cancel", __pyx_getprop_5asynq_10async_task_9AsyncTask_group_cancel, __pyx_setprop_5asynq_10async_task_9AsyncTask_group_cancel, (char *)0, 0},
  {(char *)"fn", __pyx_getprop_5asynq_10async_task_9AsyncTask_fn, __pyx_setprop_5asynq_10async_task_9AsyncTask_fn, (char *)0, 0},
  {(char *)"args", __pyx_getprop_5asynq_10async_task_9AsyncTask_args, __pyx_setprop_5asynq_10async_task_9AsyncTask_args, (char *)0, 0},
  {(char *)"kwargs", __pyx_getprop_5asynq_10async_task_9AsyncTask_kwargs, __pyx_setprop_5asynq_10async_task_9AsyncTask_kwargs, (char *)0, 0},
  {(char *)"caller", __pyx_getprop_5asynq_10async_task_9AsyncTask_caller, __pyx_setprop_5asynq_10async_task_9AsyncTask_caller, (char *)0, 0},
  {(char *)"depth", __pyx_getprop_5asynq_10async_task_9AsyncTask_depth, __pyx_setprop_5asynq_10async_task_9AsyncTask_depth, (char *)0, 0},
  {(char *)"max_depth", __pyx_getprop_5asynq_10async_task_9AsyncTask_max_depth, __pyx_setprop_5asynq_10async_task_9AsyncTask_max_depth, (char *)0, 0},
  {(char *)"scheduler", __pyx_getprop_5asynq_10async_task_9AsyncTask_scheduler, __pyx_setprop_5asynq_10async_task_9AsyncTask_scheduler, (char *)0, 0},
  {(char *)"creator", __pyx_getprop_5asynq_10async_task_9AsyncTask_creator, __pyx_setprop_5asynq_10async_task_9AsyncTask_creator, (char *)0, 0},
  {(char *)"_generator", __pyx_getprop_5asynq_10async_task_9AsyncTask__generator, __pyx_setprop_5asynq_10async_task_9AsyncTask__generator, (char *)0, 0},
  {(char *)"_last_value", __pyx_getprop_5asynq_10async_task_9AsyncTask__last_value, __pyx_setprop_5asynq_10async_task_9AsyncTask__last_value, (char *)0, 0},
  {(char *)"_frame_info", __pyx_getprop_5asynq_10async_task_9AsyncTask__frame_info, __pyx_setprop_5asynq_10async_task_9AsyncTask__frame_info, (char *)0, 0},
  {(char *)"_dependencies", __pyx_getprop_5asynq_10async_task_9AsyncTask__dependencies, __pyx_setprop_5asynq_10async_task_9AsyncTask__dependencies, (char *)0, 0},
  {(char *)"_contexts", __pyx_getprop_5asynq_10async_task_9AsyncTask__contexts, __pyx_setprop_5asynq_10async_task_9AsyncTask__contexts, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5asynq_10async_task_AsyncTask = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynq.async_task.AsyncTask", /*tp_name*/
  sizeof(struct __pyx_obj_5asynq_10async_task_AsyncTask), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5asynq_10async_task_AsyncTask, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_5asynq_10async_task_9AsyncTask_31__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Asynchronous task implementation.\n    Uses passed generator (normally produced by generator method)\n    to implement asynchronous \"await\" behavior via 'yield' keyword.\n\n    ", /*tp_doc*/
  __pyx_tp_traverse_5asynq_10async_task_AsyncTask, /*tp_traverse*/
  __pyx_tp_clear_5asynq_10async_task_AsyncTask, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5asynq_10async_task_AsyncTask, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5asynq_10async_task_AsyncTask, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5asynq_10async_task_9AsyncTask_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5asynq_10async_task_AsyncTask, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"_cancel_futures", (PyCFunction)__pyx_pw_5asynq_10async_task_1_cancel_futures, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_10async_task__cancel_futures},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "async_task",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_n_s_AsyncTaskCancelledError, __pyx_k_AsyncTaskCancelledError, sizeof(__pyx_k_AsyncTaskCancelledError), 0, 0, 1, 1},
  {&__pyx_n_s_AsyncTaskResult, __pyx_k_AsyncTaskResult, sizeof(__pyx_k_AsyncTaskResult), 0, 0, 1, 1},
  {&__pyx_n_s_AsyncTaskResult___init, __pyx_k_AsyncTaskResult___init, sizeof(__pyx_k_AsyncTaskResult___init), 0, 0, 1, 1},
  {&__pyx_n_s_BaseException, __pyx_k_BaseException, sizeof(__pyx_k_BaseException), 0, 0, 1, 1},
  {&__pyx_kp_s_Cannot_unwrap_an_object_of_type, __pyx_k_Cannot_unwrap_an_object_of_type, sizeof(__pyx_k_Cannot_unwrap_an_object_of_type), 0, 0, 1, 0},
  {&__pyx_kp_s_Dependencies, __pyx_k_Dependencies, sizeof(__pyx_k_Dependencies), 0, 0, 1, 0},
  {&__pyx_n_s_Exception, __pyx_k_Exception, sizeof(__pyx_k_Exception), 0, 0, 1, 1},
  {&__pyx_kp_s_File_file_s_line_lineno_s_in_fun, __pyx_k_File_file_s_line_lineno_s_in_fun, sizeof(__pyx_k_File_file_s_line_lineno_s_in_fun), 0, 0, 1, 0},
  {&__pyx_n_s_FutureIsAlreadyComputed, __pyx_k_FutureIsAlreadyComputed, sizeof(__pyx_k_FutureIsAlreadyComputed), 0, 0, 1, 1},
  {&__pyx_n_s_GeneratorExit, __pyx_k_GeneratorExit, sizeof(__pyx_k_GeneratorExit), 0, 0, 1, 1},
  {&__pyx_n_s_MAX_DUMP_INDENT, __pyx_k_MAX_DUMP_INDENT, sizeof(__pyx_k_MAX_DUMP_INDENT), 0, 0, 1, 1},
  {&__pyx_kp_s_No_dependencies, __pyx_k_No_dependencies, sizeof(__pyx_k_No_dependencies), 0, 0, 1, 0},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_kp_s_Task_stack_depth_exceeded_specif, __pyx_k_Task_stack_depth_exceeded_specif, sizeof(__pyx_k_Task_stack_depth_exceeded_specif), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_n_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 1},
  {&__pyx_n_s__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 0, 1, 1},
  {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
  {&__pyx_n_s_after, __pyx_k_after, sizeof(__pyx_k_after), 0, 0, 1, 1},
  {&__pyx_kp_s_almost_finished_generator_is_clo, __pyx_k_almost_finished_generator_is_clo, sizeof(__pyx_k_almost_finished_generator_is_clo), 0, 0, 1, 0},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_kp_s_async_cancelled_dependencies_of, __pyx_k_async_cancelled_dependencies_of, sizeof(__pyx_k_async_cancelled_dependencies_of), 0, 0, 1, 0},
  {&__pyx_kp_s_async_cancelling_dependencies_o, __pyx_k_async_cancelling_dependencies_o, sizeof(__pyx_k_async_cancelling_dependencies_o), 0, 0, 1, 0},
  {&__pyx_kp_s_async_context_s, __pyx_k_async_context_s, sizeof(__pyx_k_async_context_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_context_s_2, __pyx_k_async_context_s_2, sizeof(__pyx_k_async_context_s_2), 0, 0, 1, 0},
  {&__pyx_kp_s_async_dependency_s_got_s, __pyx_k_async_dependency_s_got_s, sizeof(__pyx_k_async_dependency_s_got_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_new_task_s_created_by_s, __pyx_k_async_new_task_s_created_by_s, sizeof(__pyx_k_async_new_task_s_created_by_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_queuing_exit_s_s, __pyx_k_async_queuing_exit_s_s, sizeof(__pyx_k_async_queuing_exit_s_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_queuing_throw_error_s_x_s, __pyx_k_async_queuing_throw_error_s_x_s, sizeof(__pyx_k_async_queuing_throw_error_s_x_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_s, __pyx_k_async_s, sizeof(__pyx_k_async_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_yield_s_s, __pyx_k_async_yield_s_s, sizeof(__pyx_k_async_yield_s_s), 0, 0, 1, 0},
  {&__pyx_n_s_asynq_async_task, __pyx_k_asynq_async_task, sizeof(__pyx_k_asynq_async_task), 0, 0, 1, 1},
  {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
  {&__pyx_kp_s_before_1st_yield, __pyx_k_before_1st_yield, sizeof(__pyx_k_before_1st_yield), 0, 0, 1, 0},
  {&__pyx_kp_s_blocked_x_i, __pyx_k_blocked_x_i, sizeof(__pyx_k_blocked_x_i), 0, 0, 1, 0},
  {&__pyx_n_s_can_continue, __pyx_k_can_continue, sizeof(__pyx_k_can_continue), 0, 0, 1, 1},
  {&__pyx_n_s_cancel_dependencies, __pyx_k_cancel_dependencies, sizeof(__pyx_k_cancel_dependencies), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_code_context, __pyx_k_code_context, sizeof(__pyx_k_code_context), 0, 0, 1, 1},
  {&__pyx_n_s_codeline, __pyx_k_codeline, sizeof(__pyx_k_codeline), 0, 0, 1, 1},
  {&__pyx_n_s_compute, __pyx_k_compute, sizeof(__pyx_k_compute), 0, 0, 1, 1},
  {&__pyx_n_s_computed, __pyx_k_computed, sizeof(__pyx_k_computed), 0, 0, 1, 1},
  {&__pyx_kp_s_computed_2, __pyx_k_computed_2, sizeof(__pyx_k_computed_2), 0, 0, 1, 0},
  {&__pyx_n_s_continue, __pyx_k_continue, sizeof(__pyx_k_continue), 0, 0, 1, 1},
  {&__pyx_n_s_core_errors, __pyx_k_core_errors, sizeof(__pyx_k_core_errors), 0, 0, 1, 1},
  {&__pyx_n_s_core_helpers, __pyx_k_core_helpers, sizeof(__pyx_k_core_helpers), 0, 0, 1, 1},
  {&__pyx_n_s_core_inspection, __pyx_k_core_inspection, sizeof(__pyx_k_core_inspection), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_debug_2, __pyx_k_debug_2, sizeof(__pyx_k_debug_2), 0, 0, 1, 1},
  {&__pyx_n_s_debug_options, __pyx_k_debug_options, sizeof(__pyx_k_debug_options), 0, 0, 1, 1},
  {&__pyx_n_s_dependencies, __pyx_k_dependencies, sizeof(__pyx_k_dependencies), 0, 0, 1, 1},
  {&__pyx_n_s_depth, __pyx_k_depth, sizeof(__pyx_k_depth), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dump, __pyx_k_dump, sizeof(__pyx_k_dump), 0, 0, 1, 1},
  {&__pyx_n_s_dump_error, __pyx_k_dump_error, sizeof(__pyx_k_dump_error), 0, 0, 1, 1},
  {&__pyx_n_s_empty_dictionary, __pyx_k_empty_dictionary, sizeof(__pyx_k_empty_dictionary), 0, 0, 1, 1},
  {&__pyx_n_s_empty_tuple, __pyx_k_empty_tuple, sizeof(__pyx_k_empty_tuple), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_kp_s_error_2, __pyx_k_error_2, sizeof(__pyx_k_error_2), 0, 0, 1, 0},
  {&__pyx_n_s_exc_info, __pyx_k_exc_info, sizeof(__pyx_k_exc_info), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_fn, __pyx_k_fn, sizeof(__pyx_k_fn), 0, 0, 1, 1},
  {&__pyx_n_s_funcname, __pyx_k_funcname, sizeof(__pyx_k_funcname), 0, 0, 1, 1},
  {&__pyx_n_s_function, __pyx_k_function, sizeof(__pyx_k_function), 0, 0, 1, 1},
  {&__pyx_n_s_futures, __pyx_k_futures, sizeof(__pyx_k_futures), 0, 0, 1, 1},
  {&__pyx_n_s_futures_none, __pyx_k_futures_none, sizeof(__pyx_k_futures_none), 0, 0, 1, 1},
  {&__pyx_n_s_generator, __pyx_k_generator, sizeof(__pyx_k_generator), 0, 0, 1, 1},
  {&__pyx_kp_s_generator_is_expected_but_s_is_p, __pyx_k_generator_is_expected_but_s_is_p, sizeof(__pyx_k_generator_is_expected_but_s_is_p), 0, 0, 1, 0},
  {&__pyx_n_s_get_frame_info, __pyx_k_get_frame_info, sizeof(__pyx_k_get_frame_info), 0, 0, 1, 1},
  {&__pyx_n_s_getframeinfo, __pyx_k_getframeinfo, sizeof(__pyx_k_getframeinfo), 0, 0, 1, 1},
  {&__pyx_n_s_group_cancel, __pyx_k_group_cancel, sizeof(__pyx_k_group_cancel), 0, 0, 1, 1},
  {&__pyx_kp_s_home_jelle_github_asynq_asynq_a, __pyx_k_home_jelle_github_asynq_asynq_a, sizeof(__pyx_k_home_jelle_github_asynq_asynq_a), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_indent, __pyx_k_indent, sizeof(__pyx_k_indent), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
  {&__pyx_n_s_is_blocked, __pyx_k_is_blocked, sizeof(__pyx_k_is_blocked), 0, 0, 1, 1},
  {&__pyx_n_s_is_computed, __pyx_k_is_computed, sizeof(__pyx_k_is_computed), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_lineno, __pyx_k_lineno, sizeof(__pyx_k_lineno), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_dependency, __pyx_k_make_dependency, sizeof(__pyx_k_make_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_max_depth, __pyx_k_max_depth, sizeof(__pyx_k_max_depth), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_none_future, __pyx_k_none_future, sizeof(__pyx_k_none_future), 0, 0, 1, 1},
  {&__pyx_n_s_none_future_2, __pyx_k_none_future_2, sizeof(__pyx_k_none_future_2), 0, 0, 1, 1},
  {&__pyx_kp_s_passed_yield_i, __pyx_k_passed_yield_i, sizeof(__pyx_k_passed_yield_i), 0, 0, 1, 0},
  {&__pyx_n_s_pause, __pyx_k_pause, sizeof(__pyx_k_pause), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_prepare_for_reraise, __pyx_k_prepare_for_reraise, sizeof(__pyx_k_prepare_for_reraise), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qcore_errors, __pyx_k_qcore_errors, sizeof(__pyx_k_qcore_errors), 0, 0, 1, 1},
  {&__pyx_n_s_qcore_helpers, __pyx_k_qcore_helpers, sizeof(__pyx_k_qcore_helpers), 0, 0, 1, 1},
  {&__pyx_n_s_qcore_inspection, __pyx_k_qcore_inspection, sizeof(__pyx_k_qcore_inspection), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_queue_exit, __pyx_k_queue_exit, sizeof(__pyx_k_queue_exit), 0, 0, 1, 1},
  {&__pyx_n_s_queue_throw_error, __pyx_k_queue_throw_error, sizeof(__pyx_k_queue_throw_error), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_remove_dependency, __pyx_k_remove_dependency, sizeof(__pyx_k_remove_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_repr, __pyx_k_repr, sizeof(__pyx_k_repr), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_resume, __pyx_k_resume, sizeof(__pyx_k_resume), 0, 0, 1, 1},
  {&__pyx_n_s_run_immediately, __pyx_k_run_immediately, sizeof(__pyx_k_run_immediately), 0, 0, 1, 1},
  {&__pyx_kp_s_s_s_s, __pyx_k_s_s_s, sizeof(__pyx_k_s_s_s), 0, 0, 1, 0},
  {&__pyx_n_s_schedule, __pyx_k_schedule, sizeof(__pyx_k_schedule), 0, 0, 1, 1},
  {&__pyx_n_s_scheduled, __pyx_k_scheduled, sizeof(__pyx_k_scheduled), 0, 0, 1, 1},
  {&__pyx_kp_s_scheduled_2, __pyx_k_scheduled_2, sizeof(__pyx_k_scheduled_2), 0, 0, 1, 0},
  {&__pyx_n_s_scheduler, __pyx_k_scheduler, sizeof(__pyx_k_scheduler), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set_error, __pyx_k_set_error, sizeof(__pyx_k_set_error), 0, 0, 1, 1},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_subscribe, __pyx_k_subscribe, sizeof(__pyx_k_subscribe), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_task, __pyx_k_task, sizeof(__pyx_k_task), 0, 0, 1, 1},
  {&__pyx_n_s_task_2, __pyx_k_task_2, sizeof(__pyx_k_task_2), 0, 0, 1, 1},
  {&__pyx_n_s_tb_frame, __pyx_k_tb_frame, sizeof(__pyx_k_tb_frame), 0, 0, 1, 1},
  {&__pyx_n_s_tb_next, __pyx_k_tb_next, sizeof(__pyx_k_tb_next), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_traceback, __pyx_k_traceback, sizeof(__pyx_k_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_traceback_2, __pyx_k_traceback_2, sizeof(__pyx_k_traceback_2), 0, 0, 1, 1},
  {&__pyx_n_s_traceback_hide, __pyx_k_traceback_hide, sizeof(__pyx_k_traceback_hide), 0, 0, 1, 1},
  {&__pyx_n_s_traceback_line, __pyx_k_traceback_line, sizeof(__pyx_k_traceback_line), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_waiting, __pyx_k_waiting, sizeof(__pyx_k_waiting), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_GeneratorExit = __Pyx_GetBuiltinName(__pyx_n_s_GeneratorExit); if (!__pyx_builtin_GeneratorExit) __PYX_ERR(0, 37, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 55, __pyx_L1_error)
  __pyx_builtin_BaseException = __Pyx_GetBuiltinName(__pyx_n_s_BaseException); if (!__pyx_builtin_BaseException) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_builtin_Exception = __Pyx_GetBuiltinName(__pyx_n_s_Exception); if (!__pyx_builtin_Exception) __PYX_ERR(0, 340, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 501, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "asynq/async_task.py":42
 * 
 * class AsyncTaskResult(GeneratorExit):
 *     def __init__(self, result):             # <<<<<<<<<<<<<<
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 *         self.result = result
 */
  __pyx_tuple__6 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_result); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jelle_github_asynq_asynq_a, __pyx_n_s_init, 42, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_40 = PyInt_FromLong(40); if (unlikely(!__pyx_int_40)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initasync_task(void); /*proto*/
PyMODINIT_FUNC initasync_task(void)
#else
PyMODINIT_FUNC PyInit_async_task(void); /*proto*/
PyMODINIT_FUNC PyInit_async_task(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_async_task(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("async_task", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_asynq__async_task) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "asynq.async_task")) {
      if (unlikely(PyDict_SetItemString(modules, "asynq.async_task", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_5asynq_10async_task__debug_options = ((struct __pyx_obj_5asynq_6_debug_DebugOptions *)Py_None); Py_INCREF(Py_None);
  __pyx_v_5asynq_10async_task__futures_none = Py_None; Py_INCREF(Py_None);
  __pyx_v_5asynq_10async_task__none_future = Py_None; Py_INCREF(Py_None);
  __pyx_v_5asynq_10async_task__empty_tuple = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_5asynq_10async_task__empty_dictionary = ((PyObject*)Py_None); Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_debug_options, (void *)&__pyx_v_5asynq_10async_task__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_futures_none, (void *)&__pyx_v_5asynq_10async_task__futures_none, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_none_future, (void *)&__pyx_v_5asynq_10async_task__none_future, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_empty_tuple, (void *)&__pyx_v_5asynq_10async_task__empty_tuple, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_empty_dictionary, (void *)&__pyx_v_5asynq_10async_task__empty_dictionary, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("unwrap", (void (*)(void))__pyx_f_5asynq_10async_task_unwrap, "PyObject *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_cancel_futures", (void (*)(void))__pyx_f_5asynq_10async_task__cancel_futures, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  __pyx_ptype_5asynq_7futures_FutureBase = __Pyx_ImportType("asynq.futures", "FutureBase", sizeof(struct __pyx_obj_5asynq_7futures_FutureBase), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_FutureBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_FutureBase = (struct __pyx_vtabstruct_5asynq_7futures_FutureBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_FutureBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_FutureBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_5asynq_10async_task_AsyncTask = &__pyx_vtable_5asynq_10async_task_AsyncTask;
  __pyx_vtable_5asynq_10async_task_AsyncTask.__pyx_base = *__pyx_vtabptr_5asynq_7futures_FutureBase;
  __pyx_vtable_5asynq_10async_task_AsyncTask.__pyx_base._compute = (PyObject *(*)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask__compute;
  __pyx_vtable_5asynq_10async_task_AsyncTask.__pyx_base._computed = (PyObject *(*)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask__computed;
  __pyx_vtable_5asynq_10async_task_AsyncTask.__pyx_base.make_dependency = (PyObject *(*)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask_make_dependency;
  __pyx_vtable_5asynq_10async_task_AsyncTask.__pyx_base.dump = (PyObject *(*)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump *__pyx_optional_args))__pyx_f_5asynq_10async_task_9AsyncTask_dump;
  __pyx_vtable_5asynq_10async_task_AsyncTask.is_blocked = (int (*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask_is_blocked;
  __pyx_vtable_5asynq_10async_task_AsyncTask.can_continue = (int (*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask_can_continue;
  __pyx_vtable_5asynq_10async_task_AsyncTask.start = (struct __pyx_obj_5asynq_10async_task_AsyncTask *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start *__pyx_optional_args))__pyx_f_5asynq_10async_task_9AsyncTask_start;
  __pyx_vtable_5asynq_10async_task_AsyncTask.after = (struct __pyx_obj_5asynq_10async_task_AsyncTask *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask_after;
  __pyx_vtable_5asynq_10async_task_AsyncTask.cancel_dependencies = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask_cancel_dependencies;
  __pyx_vtable_5asynq_10async_task_AsyncTask._continue = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask__continue;
  __pyx_vtable_5asynq_10async_task_AsyncTask._continue_on_generator = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, PyObject *))__pyx_f_5asynq_10async_task_9AsyncTask__continue_on_generator;
  __pyx_vtable_5asynq_10async_task_AsyncTask._accept_yield_result = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *))__pyx_f_5asynq_10async_task_9AsyncTask__accept_yield_result;
  __pyx_vtable_5asynq_10async_task_AsyncTask._accept_error = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *))__pyx_f_5asynq_10async_task_9AsyncTask__accept_error;
  __pyx_vtable_5asynq_10async_task_AsyncTask._queue_exit = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask__queue_exit;
  __pyx_vtable_5asynq_10async_task_AsyncTask._queue_throw_error = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask__queue_throw_error;
  __pyx_vtable_5asynq_10async_task_AsyncTask._remove_dependency_cython = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *))__pyx_f_5asynq_10async_task_9AsyncTask__remove_dependency_cython;
  __pyx_vtable_5asynq_10async_task_AsyncTask._before_continue = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *))__pyx_f_5asynq_10async_task_9AsyncTask__before_continue;
  __pyx_vtable_5asynq_10async_task_AsyncTask._after_continue = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *))__pyx_f_5asynq_10async_task_9AsyncTask__after_continue;
  __pyx_vtable_5asynq_10async_task_AsyncTask.traceback = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch))__pyx_f_5asynq_10async_task_9AsyncTask_traceback;
  __pyx_vtable_5asynq_10async_task_AsyncTask._enter_context = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *))__pyx_f_5asynq_10async_task_9AsyncTask__enter_context;
  __pyx_vtable_5asynq_10async_task_AsyncTask._leave_context = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *))__pyx_f_5asynq_10async_task_9AsyncTask__leave_context;
  __pyx_vtable_5asynq_10async_task_AsyncTask._pause_contexts = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *))__pyx_f_5asynq_10async_task_9AsyncTask__pause_contexts;
  __pyx_vtable_5asynq_10async_task_AsyncTask._resume_contexts = (PyObject *(*)(struct __pyx_obj_5asynq_10async_task_AsyncTask *))__pyx_f_5asynq_10async_task_9AsyncTask__resume_contexts;
  __pyx_type_5asynq_10async_task_AsyncTask.tp_base = __pyx_ptype_5asynq_7futures_FutureBase;
  if (PyType_Ready(&__pyx_type_5asynq_10async_task_AsyncTask) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_type_5asynq_10async_task_AsyncTask.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_5asynq_10async_task_AsyncTask.tp_dict, __pyx_vtabptr_5asynq_10async_task_AsyncTask) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "AsyncTask", (PyObject *)&__pyx_type_5asynq_10async_task_AsyncTask) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_ptype_5asynq_10async_task_AsyncTask = &__pyx_type_5asynq_10async_task_AsyncTask;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(3, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(4, 15, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_EventHook = __Pyx_ImportType("qcore.events", "EventHook", sizeof(struct __pyx_obj_5qcore_6events_EventHook), 1); if (unlikely(!__pyx_ptype_5qcore_6events_EventHook)) __PYX_ERR(5, 15, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_SinkingEventHook = __Pyx_ImportType("qcore.events", "SinkingEventHook", sizeof(struct __pyx_obj_5qcore_6events_SinkingEventHook), 1); if (unlikely(!__pyx_ptype_5qcore_6events_SinkingEventHook)) __PYX_ERR(5, 19, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_EventInterceptor = __Pyx_ImportType("qcore.events", "EventInterceptor", sizeof(struct __pyx_obj_5qcore_6events_EventInterceptor), 1); if (unlikely(!__pyx_ptype_5qcore_6events_EventInterceptor)) __PYX_ERR(5, 25, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_MarkerObject = __Pyx_ImportType("qcore.helpers", "MarkerObject", sizeof(struct __pyx_obj_5qcore_7helpers_MarkerObject), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_MarkerObject)) __PYX_ERR(6, 24, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_EmptyContext = __Pyx_ImportType("qcore.helpers", "EmptyContext", sizeof(struct __pyx_obj_5qcore_7helpers_EmptyContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_EmptyContext)) __PYX_ERR(6, 33, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_EmptyContext = (struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_EmptyContext->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_EmptyContext)) __PYX_ERR(6, 33, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper = __Pyx_ImportType("qcore.helpers", "CythonCachedHashWrapper", sizeof(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper)) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper = (struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper)) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext = __Pyx_ImportType("qcore.helpers", "_ScopedValueOverrideContext", sizeof(struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext)) __PYX_ERR(6, 50, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_ScopedValue = __Pyx_ImportType("qcore.helpers", "ScopedValue", sizeof(struct __pyx_obj_5qcore_7helpers_ScopedValue), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_ScopedValue)) __PYX_ERR(6, 52, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_ScopedValue = (struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_ScopedValue->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_ScopedValue)) __PYX_ERR(6, 52, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers__PropertyOverrideContext = __Pyx_ImportType("qcore.helpers", "_PropertyOverrideContext", sizeof(struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers__PropertyOverrideContext)) __PYX_ERR(6, 65, __pyx_L1_error)
  __pyx_ptype_5asynq_6_debug_DebugOptions = __Pyx_ImportType("asynq._debug", "DebugOptions", sizeof(struct __pyx_obj_5asynq_6_debug_DebugOptions), 1); if (unlikely(!__pyx_ptype_5asynq_6_debug_DebugOptions)) __PYX_ERR(7, 18, __pyx_L1_error)
  __pyx_vtabptr_5asynq_6_debug_DebugOptions = (struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions*)__Pyx_GetVtable(__pyx_ptype_5asynq_6_debug_DebugOptions->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_6_debug_DebugOptions)) __PYX_ERR(7, 18, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_Future = __Pyx_ImportType("asynq.futures", "Future", sizeof(struct __pyx_obj_5asynq_7futures_Future), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_Future)) __PYX_ERR(8, 49, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_Future = (struct __pyx_vtabstruct_5asynq_7futures_Future*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_Future->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_Future)) __PYX_ERR(8, 49, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_ConstFuture = __Pyx_ImportType("asynq.futures", "ConstFuture", sizeof(struct __pyx_obj_5asynq_7futures_ConstFuture), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_ConstFuture)) __PYX_ERR(8, 54, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_ConstFuture = (struct __pyx_vtabstruct_5asynq_7futures_ConstFuture*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_ConstFuture->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_ConstFuture)) __PYX_ERR(8, 54, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_ErrorFuture = __Pyx_ImportType("asynq.futures", "ErrorFuture", sizeof(struct __pyx_obj_5asynq_7futures_ErrorFuture), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_ErrorFuture)) __PYX_ERR(8, 57, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_ErrorFuture = (struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_ErrorFuture->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_ErrorFuture)) __PYX_ERR(8, 57, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_BatchBase = __Pyx_ImportType("asynq.batching", "BatchBase", sizeof(struct __pyx_obj_5asynq_8batching_BatchBase), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_BatchBase)) __PYX_ERR(9, 26, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_BatchBase = (struct __pyx_vtabstruct_5asynq_8batching_BatchBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_BatchBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_BatchBase)) __PYX_ERR(9, 26, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_BatchItemBase = __Pyx_ImportType("asynq.batching", "BatchItemBase", sizeof(struct __pyx_obj_5asynq_8batching_BatchItemBase), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_BatchItemBase)) __PYX_ERR(9, 45, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_BatchItemBase = (struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_BatchItemBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_BatchItemBase)) __PYX_ERR(9, 45, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_DebugBatchItem = __Pyx_ImportType("asynq.batching", "DebugBatchItem", sizeof(struct __pyx_obj_5asynq_8batching_DebugBatchItem), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_DebugBatchItem)) __PYX_ERR(9, 53, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_DebugBatchItem = (struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_DebugBatchItem->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_DebugBatchItem)) __PYX_ERR(9, 53, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_DebugBatch = __Pyx_ImportType("asynq.batching", "DebugBatch", sizeof(struct __pyx_obj_5asynq_8batching_DebugBatch), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_DebugBatch)) __PYX_ERR(9, 56, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_DebugBatch = (struct __pyx_vtabstruct_5asynq_8batching_DebugBatch*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_DebugBatch->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_DebugBatch)) __PYX_ERR(9, 56, __pyx_L1_error)
  __pyx_ptype_5asynq_9scheduler_TaskScheduler = __Pyx_ImportType("asynq.scheduler", "TaskScheduler", sizeof(struct __pyx_obj_5asynq_9scheduler_TaskScheduler), 1); if (unlikely(!__pyx_ptype_5asynq_9scheduler_TaskScheduler)) __PYX_ERR(10, 29, __pyx_L1_error)
  __pyx_vtabptr_5asynq_9scheduler_TaskScheduler = (struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler*)__Pyx_GetVtable(__pyx_ptype_5asynq_9scheduler_TaskScheduler->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_9scheduler_TaskScheduler)) __PYX_ERR(10, 29, __pyx_L1_error)
  __pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator = __Pyx_ImportType("asynq.scheduler", "TaskSchedulerActivator", sizeof(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator), 1); if (unlikely(!__pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator)) __PYX_ERR(10, 88, __pyx_L1_error)
  /*--- Variable import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("qcore.events"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "sinking_event_hook", (void **)&__pyx_vp_5qcore_6events_sinking_event_hook, "struct __pyx_obj_5qcore_6events_SinkingEventHook *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("qcore.helpers"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_list", (void **)&__pyx_vp_5qcore_7helpers_empty_list, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_tuple", (void **)&__pyx_vp_5qcore_7helpers_empty_tuple, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_dict", (void **)&__pyx_vp_5qcore_7helpers_empty_dict, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "none", (void **)&__pyx_vp_5qcore_7helpers_none, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "miss", (void **)&__pyx_vp_5qcore_7helpers_miss, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "same", (void **)&__pyx_vp_5qcore_7helpers_same, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "unspecified", (void **)&__pyx_vp_5qcore_7helpers_unspecified, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_context", (void **)&__pyx_vp_5qcore_7helpers_empty_context, "struct __pyx_obj_5qcore_7helpers_EmptyContext *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "CachedHashWrapper", (void **)&__pyx_vp_5qcore_7helpers_CachedHashWrapper, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "override", (void **)&__pyx_vp_5qcore_7helpers_override, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_ImportModule("asynq._debug"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_3, "options", (void **)&__pyx_vp_5asynq_6_debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_ImportModule("asynq.futures"); if (!__pyx_t_4) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "_debug_options", (void **)&__pyx_vp_5asynq_7futures__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "_none", (void **)&__pyx_vp_5asynq_7futures__none, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_ImportModule("asynq.batching"); if (!__pyx_t_5) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_debug_options", (void **)&__pyx_vp_5asynq_8batching__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_debug_batch_state", (void **)&__pyx_vp_5asynq_8batching__debug_batch_state, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_ImportModule("asynq.scheduler"); if (!__pyx_t_6) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_debug_options", (void **)&__pyx_vp_5asynq_9scheduler__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_futures_none", (void **)&__pyx_vp_5asynq_9scheduler__futures_none, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_state", (void **)&__pyx_vp_5asynq_9scheduler__state, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  /*--- Function import code ---*/
  __pyx_t_7 = __Pyx_ImportModule("qcore.helpers"); if (!__pyx_t_7) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_7, "safe_str", (void (**)(void))&__pyx_f_5qcore_7helpers_safe_str, "PyObject *(PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5qcore_7helpers_safe_str *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_8 = __Pyx_ImportModule("qcore.inspection"); if (!__pyx_t_8) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_8, "get_function_call_str", (void (**)(void))&__pyx_f_5qcore_10inspection_get_function_call_str, "PyObject *(PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_8, "is_cython_or_generator", (void (**)(void))&__pyx_f_5qcore_10inspection_is_cython_or_generator, "int (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = __Pyx_ImportModule("asynq.scheduler"); if (!__pyx_t_9) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_9, "get_scheduler", (void (**)(void))&__pyx_f_5asynq_9scheduler_get_scheduler, "struct __pyx_obj_5asynq_9scheduler_TaskScheduler *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_9, "get_active_task", (void (**)(void))&__pyx_f_5asynq_9scheduler_get_active_task, "struct __pyx_obj_5asynq_10async_task_AsyncTask *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "asynq/async_task.py":15
 * # limitations under the License.
 * 
 * import inspect             # <<<<<<<<<<<<<<
 * import six
 * import sys
 */
  __pyx_t_10 = __Pyx_patch_inspect(__Pyx_Import(__pyx_n_s_inspect, 0, -1)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_inspect, __pyx_t_10) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":16
 * 
 * import inspect
 * import six             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_10 = __Pyx_Import(__pyx_n_s_six, 0, -1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_six, __pyx_t_10) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":17
 * import inspect
 * import six
 * import sys             # <<<<<<<<<<<<<<
 * 
 * import qcore.helpers as core_helpers
 */
  __pyx_t_10 = __Pyx_Import(__pyx_n_s_sys, 0, -1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_10) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":19
 * import sys
 * 
 * import qcore.helpers as core_helpers             # <<<<<<<<<<<<<<
 * import qcore.inspection as core_inspection
 * import qcore.errors as core_errors
 */
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s__4);
  __Pyx_GIVEREF(__pyx_n_s__4);
  PyList_SET_ITEM(__pyx_t_10, 0, __pyx_n_s__4);
  __pyx_t_11 = __Pyx_Import(__pyx_n_s_qcore_helpers, __pyx_t_10, -1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_core_helpers, __pyx_t_11) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "asynq/async_task.py":20
 * 
 * import qcore.helpers as core_helpers
 * import qcore.inspection as core_inspection             # <<<<<<<<<<<<<<
 * import qcore.errors as core_errors
 * 
 */
  __pyx_t_11 = PyList_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s__4);
  __Pyx_GIVEREF(__pyx_n_s__4);
  PyList_SET_ITEM(__pyx_t_11, 0, __pyx_n_s__4);
  __pyx_t_10 = __Pyx_Import(__pyx_n_s_qcore_inspection, __pyx_t_11, -1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_core_inspection, __pyx_t_10) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":21
 * import qcore.helpers as core_helpers
 * import qcore.inspection as core_inspection
 * import qcore.errors as core_errors             # <<<<<<<<<<<<<<
 * 
 * from . import debug
 */
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s__4);
  __Pyx_GIVEREF(__pyx_n_s__4);
  PyList_SET_ITEM(__pyx_t_10, 0, __pyx_n_s__4);
  __pyx_t_11 = __Pyx_Import(__pyx_n_s_qcore_errors, __pyx_t_10, -1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_core_errors, __pyx_t_11) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "asynq/async_task.py":23
 * import qcore.errors as core_errors
 * 
 * from . import debug             # <<<<<<<<<<<<<<
 * from . import futures
 * from . import scheduler
 */
  __pyx_t_11 = PyList_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_debug);
  __Pyx_GIVEREF(__pyx_n_s_debug);
  PyList_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_debug);
  __pyx_t_10 = __Pyx_Import(__pyx_n_s__5, __pyx_t_11, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_ImportFrom(__pyx_t_10, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_debug, __pyx_t_11) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":24
 * 
 * from . import debug
 * from . import futures             # <<<<<<<<<<<<<<
 * from . import scheduler
 * from . import _debug
 */
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_futures);
  __Pyx_GIVEREF(__pyx_n_s_futures);
  PyList_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_futures);
  __pyx_t_11 = __Pyx_Import(__pyx_n_s__5, __pyx_t_10, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_ImportFrom(__pyx_t_11, __pyx_n_s_futures); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_futures, __pyx_t_10) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "asynq/async_task.py":25
 * from . import debug
 * from . import futures
 * from . import scheduler             # <<<<<<<<<<<<<<
 * from . import _debug
 * 
 */
  __pyx_t_11 = PyList_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_scheduler);
  __Pyx_GIVEREF(__pyx_n_s_scheduler);
  PyList_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_scheduler);
  __pyx_t_10 = __Pyx_Import(__pyx_n_s__5, __pyx_t_11, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_ImportFrom(__pyx_t_10, __pyx_n_s_scheduler); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_scheduler, __pyx_t_11) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":26
 * from . import futures
 * from . import scheduler
 * from . import _debug             # <<<<<<<<<<<<<<
 * 
 * __traceback_hide__ = True
 */
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_debug_2);
  __Pyx_GIVEREF(__pyx_n_s_debug_2);
  PyList_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_debug_2);
  __pyx_t_11 = __Pyx_Import(__pyx_n_s__5, __pyx_t_10, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_ImportFrom(__pyx_t_11, __pyx_n_s_debug_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_debug_2, __pyx_t_10) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "asynq/async_task.py":28
 * from . import _debug
 * 
 * __traceback_hide__ = True             # <<<<<<<<<<<<<<
 * 
 * _debug_options = _debug.options
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_traceback_hide, Py_True) < 0) __PYX_ERR(0, 28, __pyx_L1_error)

  /* "asynq/async_task.py":30
 * __traceback_hide__ = True
 * 
 * _debug_options = _debug.options             # <<<<<<<<<<<<<<
 * _futures_none = futures._none
 * _none_future = futures.none_future
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_5asynq_6_debug_options));
  __Pyx_XGOTREF(((PyObject *)__pyx_v_5asynq_10async_task__debug_options));
  __Pyx_DECREF_SET(__pyx_v_5asynq_10async_task__debug_options, __pyx_v_5asynq_6_debug_options);
  __Pyx_GIVEREF(((PyObject *)__pyx_v_5asynq_6_debug_options));

  /* "asynq/async_task.py":31
 * 
 * _debug_options = _debug.options
 * _futures_none = futures._none             # <<<<<<<<<<<<<<
 * _none_future = futures.none_future
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_5asynq_7futures__none));
  __Pyx_XGOTREF(__pyx_v_5asynq_10async_task__futures_none);
  __Pyx_DECREF_SET(__pyx_v_5asynq_10async_task__futures_none, ((PyObject *)__pyx_v_5asynq_7futures__none));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_5asynq_7futures__none));

  /* "asynq/async_task.py":32
 * _debug_options = _debug.options
 * _futures_none = futures._none
 * _none_future = futures.none_future             # <<<<<<<<<<<<<<
 * 
 * MAX_DUMP_INDENT = 40
 */
  __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_futures); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_none_future_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_XGOTREF(__pyx_v_5asynq_10async_task__none_future);
  __Pyx_DECREF_SET(__pyx_v_5asynq_10async_task__none_future, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  __pyx_t_10 = 0;

  /* "asynq/async_task.py":34
 * _none_future = futures.none_future
 * 
 * MAX_DUMP_INDENT = 40             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAX_DUMP_INDENT, __pyx_int_40) < 0) __PYX_ERR(0, 34, __pyx_L1_error)

  /* "asynq/async_task.py":37
 * 
 * 
 * class AsyncTaskCancelledError(GeneratorExit):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_builtin_GeneratorExit);
  __Pyx_GIVEREF(__pyx_builtin_GeneratorExit);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_builtin_GeneratorExit);
  __pyx_t_11 = __Pyx_CalculateMetaclass(NULL, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_Py3MetaclassPrepare(__pyx_t_11, __pyx_t_10, __pyx_n_s_AsyncTaskCancelledError, __pyx_n_s_AsyncTaskCancelledError, (PyObject *) NULL, __pyx_n_s_asynq_async_task, (PyObject *) NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_11, __pyx_n_s_AsyncTaskCancelledError, __pyx_t_10, __pyx_t_12, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AsyncTaskCancelledError, __pyx_t_13) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":41
 * 
 * 
 * class AsyncTaskResult(GeneratorExit):             # <<<<<<<<<<<<<<
 *     def __init__(self, result):
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 */
  __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_builtin_GeneratorExit);
  __Pyx_GIVEREF(__pyx_builtin_GeneratorExit);
  PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_builtin_GeneratorExit);
  __pyx_t_11 = __Pyx_CalculateMetaclass(NULL, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_Py3MetaclassPrepare(__pyx_t_11, __pyx_t_10, __pyx_n_s_AsyncTaskResult, __pyx_n_s_AsyncTaskResult, (PyObject *) NULL, __pyx_n_s_asynq_async_task, (PyObject *) NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "asynq/async_task.py":42
 * 
 * class AsyncTaskResult(GeneratorExit):
 *     def __init__(self, result):             # <<<<<<<<<<<<<<
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 *         self.result = result
 */
  __pyx_t_13 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5asynq_10async_task_15AsyncTaskResult_1__init__, 0, __pyx_n_s_AsyncTaskResult___init, NULL, __pyx_n_s_asynq_async_task, __pyx_d, ((PyObject *)__pyx_codeobj__7)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyObject_SetItem(__pyx_t_12, __pyx_n_s_init, __pyx_t_13) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "asynq/async_task.py":41
 * 
 * 
 * class AsyncTaskResult(GeneratorExit):             # <<<<<<<<<<<<<<
 *     def __init__(self, result):
 *         GeneratorExit.__init__(self, "AsyncTaskResult")
 */
  __pyx_t_13 = __Pyx_Py3ClassCreate(__pyx_t_11, __pyx_n_s_AsyncTaskResult, __pyx_t_10, __pyx_t_12, NULL, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AsyncTaskResult, __pyx_t_13) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":509
 * # Private part
 * 
 * _empty_tuple = tuple()             # <<<<<<<<<<<<<<
 * _empty_dictionary = dict()
 * globals()['_empty_tuple'] = _empty_tuple
 */
  __pyx_t_10 = __Pyx_PyObject_Call(((PyObject *)(&PyTuple_Type)), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_XGOTREF(__pyx_v_5asynq_10async_task__empty_tuple);
  __Pyx_DECREF_SET(__pyx_v_5asynq_10async_task__empty_tuple, ((PyObject*)__pyx_t_10));
  __Pyx_GIVEREF(__pyx_t_10);
  __pyx_t_10 = 0;

  /* "asynq/async_task.py":510
 * 
 * _empty_tuple = tuple()
 * _empty_dictionary = dict()             # <<<<<<<<<<<<<<
 * globals()['_empty_tuple'] = _empty_tuple
 * globals()['_empty_dictionary'] = _empty_dictionary
 */
  __pyx_t_10 = PyDict_New(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_XGOTREF(__pyx_v_5asynq_10async_task__empty_dictionary);
  __Pyx_DECREF_SET(__pyx_v_5asynq_10async_task__empty_dictionary, ((PyObject*)__pyx_t_10));
  __Pyx_GIVEREF(__pyx_t_10);
  __pyx_t_10 = 0;

  /* "asynq/async_task.py":511
 * _empty_tuple = tuple()
 * _empty_dictionary = dict()
 * globals()['_empty_tuple'] = _empty_tuple             # <<<<<<<<<<<<<<
 * globals()['_empty_dictionary'] = _empty_dictionary
 * 
 */
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (unlikely(PyObject_SetItem(__pyx_t_10, __pyx_n_s_empty_tuple, __pyx_v_5asynq_10async_task__empty_tuple) < 0)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":512
 * _empty_dictionary = dict()
 * globals()['_empty_tuple'] = _empty_tuple
 * globals()['_empty_dictionary'] = _empty_dictionary             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_10 = __Pyx_Globals(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (unlikely(PyObject_SetItem(__pyx_t_10, __pyx_n_s_empty_dictionary, __pyx_v_5asynq_10async_task__empty_dictionary) < 0)) __PYX_ERR(0, 512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/async_task.py":1
 * # Copyright 2016 Quora, Inc.             # <<<<<<<<<<<<<<
 * #
 * # Licensed under the Apache License, Version 2.0 (the "License");
 */
  __pyx_t_10 = PyDict_New(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_10) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init asynq.async_task", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init asynq.async_task");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetModuleGlobalName */
    static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* ExtTypeTest */
      static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* ArgTypeTest */
      static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* PyErrFetchRestore */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetException */
      #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* SaveResetException */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetItemInt */
          static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntBinop */
          #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* StringJoin */
          #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* RaiseTooManyValuesToUnpack */
          static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
          static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
          static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
          static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* CallNextTpDealloc */
          static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_dealloc != current_tp_dealloc)
        type = type->tp_base;
    while (type && type->tp_dealloc == current_tp_dealloc)
        type = type->tp_base;
    if (type)
        type->tp_dealloc(obj);
}

/* CallNextTpTraverse */
          static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_traverse != current_tp_traverse)
        type = type->tp_base;
    while (type && type->tp_traverse == current_tp_traverse)
        type = type->tp_base;
    if (type && type->tp_traverse)
        return type->tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
          static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    while (type && type->tp_clear != current_tp_clear)
        type = type->tp_base;
    while (type && type->tp_clear == current_tp_clear)
        type = type->tp_base;
    if (type && type->tp_clear)
        type->tp_clear(obj);
}

/* GetVTable */
          static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PatchModuleWithCoroutine */
          static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchInspect */
          static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if (0) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
          static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
          static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* FetchCommonType */
          static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
          static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* GetAttr */
              static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_COMPILING_IN_CPYTHON
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
              static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* CodeObjectCache */
              static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
              #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
              #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
              static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
              static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
              static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrExport */
              static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
              static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
              #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
              #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* VoidPtrImport */
              #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
              #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
              static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
