/* Generated by Cython 0.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {}
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__asynq__utils
#define __PYX_HAVE_API__asynq__utils
#include "string.h"
#include "stdio.h"
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "asynq/utils.py",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "events.pxd",
  "helpers.pxd",
  "asynq/_debug.pxd",
  "asynq/futures.pxd",
  "asynq/batching.pxd",
  "asynq/scheduler.pxd",
  "asynq/async_task.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5qcore_6events_EventHook;
struct __pyx_obj_5qcore_6events_SinkingEventHook;
struct __pyx_obj_5qcore_6events_EventInterceptor;
struct __pyx_obj_5qcore_7helpers_MarkerObject;
struct __pyx_obj_5qcore_7helpers_EmptyContext;
struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper;
struct __pyx_obj_5qcore_7helpers_ScopedValue;
struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext;
struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext;
struct __pyx_obj_5asynq_6_debug_DebugOptions;
struct __pyx_obj_5asynq_7futures_FutureBase;
struct __pyx_obj_5asynq_7futures_Future;
struct __pyx_obj_5asynq_7futures_ConstFuture;
struct __pyx_obj_5asynq_7futures_ErrorFuture;
struct __pyx_obj_5asynq_8batching_BatchBase;
struct __pyx_obj_5asynq_8batching_BatchItemBase;
struct __pyx_obj_5asynq_8batching_DebugBatchItem;
struct __pyx_obj_5asynq_8batching_DebugBatch;
struct __pyx_obj_5asynq_9scheduler_TaskScheduler;
struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator;
struct __pyx_obj_5asynq_10async_task_AsyncTask;
struct __pyx_opt_args_5qcore_7helpers_safe_str;
struct __pyx_opt_args_5qcore_7helpers_safe_repr;

/* "qcore/helpers.pxd":75
 * 
 * cpdef object ellipsis(object source, int max_length)
 * cpdef object safe_str(object source, int max_length=?)             # <<<<<<<<<<<<<<
 * cpdef object safe_repr(object source, int max_length=?)
 * 
 */
struct __pyx_opt_args_5qcore_7helpers_safe_str {
  int __pyx_n;
  int max_length;
};

/* "qcore/helpers.pxd":76
 * cpdef object ellipsis(object source, int max_length)
 * cpdef object safe_str(object source, int max_length=?)
 * cpdef object safe_repr(object source, int max_length=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5qcore_7helpers_safe_repr {
  int __pyx_n;
  int max_length;
};
struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL;

/* "_debug.pxd":44
 *     cdef public object ENABLE_COMPLEX_ASSERTIONS
 * 
 *     cpdef DUMP_ALL(self, value=?)             # <<<<<<<<<<<<<<
 * 
 * cdef public DebugOptions options
 */
struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL {
  int __pyx_n;
  PyObject *value;
};
struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump;

/* "futures.pxd":46
 *     cdef inline raise_if_error(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump {
  int __pyx_n;
  int indent;
};
struct __pyx_opt_args_5qcore_10inspection_wraps;

/* "qcore/inspection.pxd":38
 * cpdef bint is_classmethod(object fn) except -1
 * 
 * cpdef wraps(object wrapped, object assigned=?, object updated=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5qcore_10inspection_wraps {
  int __pyx_n;
  PyObject *assigned;
  PyObject *updated;
};
struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel;
struct __pyx_opt_args_5asynq_8batching_9BatchBase_dump;
struct __pyx_opt_args_5asynq_8batching_sync;

/* "batching.pxd":35
 * 
 *     cpdef flush(self)
 *     cpdef cancel(self, object error=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _compute(self)
 */
struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel {
  int __pyx_n;
  PyObject *error;
};

/* "batching.pxd":43
 *     cpdef _try_switch_active_batch(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class BatchItemBase(futures.FutureBase):
 */
struct __pyx_opt_args_5asynq_8batching_9BatchBase_dump {
  int __pyx_n;
  int indent;
};

/* "batching.pxd":66
 * cdef object _debug_batch_state
 * 
 * cpdef sync(tag=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5asynq_8batching_sync {
  int __pyx_n;
  PyObject *tag;
};
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump;

/* "scheduler.pxd":45
 *     cpdef reset(self)
 * 
 *     cpdef async_task.AsyncTask start(self, async_task.AsyncTask task,             # <<<<<<<<<<<<<<
 *             bint run_immediately=?)
 *     cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start {
  int __pyx_n;
  int run_immediately;
};

/* "scheduler.pxd":48
 *             bint run_immediately=?)
 *     cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
 *     cpdef int flush_batches(self, bint must_await=?) except -1             # <<<<<<<<<<<<<<
 * 
 *     cpdef add_dependencies(self, async_task.AsyncTask task, object yield_result)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches {
  int __pyx_n;
  int must_await;
};

/* "scheduler.pxd":76
 *     @cython.locals(dump_continue_task=bint, task=async_task.AsyncTask)
 *     cdef async_task.AsyncTask _continue_with_task(self)
 *     cdef batching.BatchBase _continue_with_batch(self, bint assert_no_batch=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef batching.BatchBase _select_batch_to_flush(self)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch {
  int __pyx_n;
  int assert_no_batch;
};

/* "scheduler.pxd":84
 *     cpdef _accept_changes(self, TaskScheduler clone)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 *     @cython.locals(current_time=float)
 *     cpdef try_time_based_dump(self, object last_task=?)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump {
  int __pyx_n;
  int indent;
};

/* "scheduler.pxd":86
 *     cpdef dump(self, int indent=?)
 *     @cython.locals(current_time=float)
 *     cpdef try_time_based_dump(self, object last_task=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class TaskSchedulerActivator(object):
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump {
  int __pyx_n;
  PyObject *last_task;
};
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start;
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump;

/* "async_task.pxd":53
 * 
 *     @cython.locals(s=scheduler.TaskScheduler)
 *     cpdef AsyncTask start(self, bint run_immediately=?)             # <<<<<<<<<<<<<<
 *     @cython.locals(s=scheduler.TaskScheduler)
 *     cpdef AsyncTask after(self, futures.FutureBase future)
 */
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start {
  int __pyx_n;
  int run_immediately;
};

/* "async_task.pxd":84
 *     cpdef list traceback(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 *     # Contexts support
 */
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump {
  int __pyx_n;
  int indent;
};
struct __pyx_opt_args_5asynq_5utils_execute_on_separate_scheduler;

/* "asynq/utils.pxd":28
 * 
 * @cython.locals(task=async_task.AsyncTask, current_scheduler=scheduler.TaskScheduler, new_scheduler=scheduler.TaskScheduler)
 * cpdef object execute_on_separate_scheduler(object async_pull_fn, list contexts, tuple args=?, dict kwargs=?)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object _await(tuple args)
 */
struct __pyx_opt_args_5asynq_5utils_execute_on_separate_scheduler {
  int __pyx_n;
  PyObject *args;
  PyObject *kwargs;
};

/* "qcore/events.pxd":15
 * # limitations under the License.
 * 
 * cdef class EventHook(object):             # <<<<<<<<<<<<<<
 *     cdef list handlers
 * 
 */
struct __pyx_obj_5qcore_6events_EventHook {
  PyObject_HEAD
  PyObject *handlers;
};


/* "qcore/events.pxd":19
 * 
 * 
 * cdef class SinkingEventHook(EventHook):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5qcore_6events_SinkingEventHook {
  struct __pyx_obj_5qcore_6events_EventHook __pyx_base;
};


/* "qcore/events.pxd":25
 * 
 * 
 * cdef class EventInterceptor(object):             # <<<<<<<<<<<<<<
 *     cdef object source
 *     cdef dict events
 */
struct __pyx_obj_5qcore_6events_EventInterceptor {
  PyObject_HEAD
  PyObject *source;
  PyObject *events;
};


/* "qcore/helpers.pxd":24
 * 
 * 
 * cdef class MarkerObject(object):             # <<<<<<<<<<<<<<
 *     cpdef unicode name
 * 
 */
struct __pyx_obj_5qcore_7helpers_MarkerObject {
  PyObject_HEAD
  PyObject *name;
};


/* "qcore/helpers.pxd":33
 * 
 * 
 * cdef class EmptyContext(object):             # <<<<<<<<<<<<<<
 *     cpdef __enter__(self)
 *     cpdef __exit__(self, exc_type, exc_val, exc_tb)
 */
struct __pyx_obj_5qcore_7helpers_EmptyContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext *__pyx_vtab;
};


/* "qcore/helpers.pxd":40
 * 
 * 
 * cdef class CythonCachedHashWrapper(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 *     cdef int _hash
 */
struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper *__pyx_vtab;
  PyObject *_value;
  int _hash;
};


/* "qcore/helpers.pxd":52
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration
 * 
 * cdef class ScopedValue(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 * 
 */
struct __pyx_obj_5qcore_7helpers_ScopedValue {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue *__pyx_vtab;
  PyObject *_value;
};


/* "qcore/helpers.pxd":50
 * 
 * 
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class ScopedValue(object):
 */
struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext {
  PyObject_HEAD
  struct __pyx_obj_5qcore_7helpers_ScopedValue *_target;
  PyObject *_value;
  PyObject *_old_value;
};


/* "qcore/helpers.pxd":65
 * 
 * 
 * cdef class _PropertyOverrideContext(object):             # <<<<<<<<<<<<<<
 *     cdef object _target
 *     cdef object _property_name
 */
struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext {
  PyObject_HEAD
  PyObject *_target;
  PyObject *_property_name;
  PyObject *_value;
  PyObject *_old_value;
};


/* "_debug.pxd":18
 * 
 * 
 * cdef class DebugOptions(object):             # <<<<<<<<<<<<<<
 *     cdef public bint DUMP_PRE_ERROR_STATE
 *     cdef public bint DUMP_EXCEPTIONS
 */
struct __pyx_obj_5asynq_6_debug_DebugOptions {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions *__pyx_vtab;
  int DUMP_PRE_ERROR_STATE;
  int DUMP_EXCEPTIONS;
  int DUMP_AWAIT_RECURSION;
  int DUMP_SCHEDULE_TASK;
  int DUMP_CONTINUE_TASK;
  int DUMP_SCHEDULE_BATCH;
  int DUMP_FLUSH_BATCH;
  int DUMP_DEPENDENCIES;
  int DUMP_COMPUTED;
  int DUMP_NEW_TASKS;
  int DUMP_YIELD_RESULTS;
  int DUMP_QUEUED_RESULTS;
  int DUMP_CONTEXTS;
  int DUMP_SCHEDULER_CHANGE;
  int DUMP_SYNC;
  int DUMP_PRIMER;
  int DUMP_STACK;
  int DUMP_SCHEDULER_STATE;
  float SCHEDULER_STATE_DUMP_INTERVAL;
  int DEBUG_STR_REPR_MAX_LENGTH;
  PyObject *STACK_DUMP_LIMIT;
  PyObject *ENABLE_COMPLEX_ASSERTIONS;
};


/* "futures.pxd":26
 * 
 * 
 * cdef class FutureBase(object):             # <<<<<<<<<<<<<<
 *     cdef public object _value
 *     cdef public object _error
 */
struct __pyx_obj_5asynq_7futures_FutureBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase *__pyx_vtab;
  PyObject *_value;
  PyObject *_error;
  int _in_repr;
  struct __pyx_obj_5qcore_6events_EventHook *on_computed;
};


/* "futures.pxd":49
 * 
 * 
 * cdef class Future(FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public object _value_provider
 *     cpdef _compute(self)
 */
struct __pyx_obj_5asynq_7futures_Future {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  PyObject *_value_provider;
};


/* "futures.pxd":54
 * 
 * 
 * cdef class ConstFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5asynq_7futures_ConstFuture {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
};


/* "futures.pxd":57
 *     pass
 * 
 * cdef class ErrorFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_5asynq_7futures_ErrorFuture {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
};


/* "batching.pxd":26
 * 
 * 
 * cdef class BatchBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public list items
 * 
 */
struct __pyx_obj_5asynq_8batching_BatchBase {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  PyObject *items;
};


/* "batching.pxd":45
 *     cpdef dump(self, int indent=?)
 * 
 * cdef class BatchItemBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public BatchBase batch
 *     cdef public long index
 */
struct __pyx_obj_5asynq_8batching_BatchItemBase {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  struct __pyx_obj_5asynq_8batching_BatchBase *batch;
  long index;
};


/* "batching.pxd":53
 * 
 * 
 * cdef class DebugBatchItem(BatchItemBase):             # <<<<<<<<<<<<<<
 *     cdef public object _result
 * 
 */
struct __pyx_obj_5asynq_8batching_DebugBatchItem {
  struct __pyx_obj_5asynq_8batching_BatchItemBase __pyx_base;
  PyObject *_result;
};


/* "batching.pxd":56
 *     cdef public object _result
 * 
 * cdef class DebugBatch(BatchBase):             # <<<<<<<<<<<<<<
 *     cdef public str name
 *     cdef public long index
 */
struct __pyx_obj_5asynq_8batching_DebugBatch {
  struct __pyx_obj_5asynq_8batching_BatchBase __pyx_base;
  PyObject *name;
  long index;
};


/* "scheduler.pxd":29
 * 
 * 
 * cdef class TaskScheduler(object):             # <<<<<<<<<<<<<<
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time
 */
struct __pyx_obj_5asynq_9scheduler_TaskScheduler {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *__pyx_vtab;
  PyObject *_empty_set;
  float _last_dump_time;
  PyObject *_new_tasks;
  PyObject *_tasks;
  PyObject *_batches;
  int await_recursion_depth;
  int max_await_recursion_depth;
  struct __pyx_obj_5qcore_6events_EventHook *on_before_batch_flush;
  struct __pyx_obj_5qcore_6events_EventHook *on_after_batch_flush;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *active_task;
  PyObject *name;
};


/* "scheduler.pxd":88
 *     cpdef try_time_based_dump(self, object last_task=?)
 * 
 * cdef class TaskSchedulerActivator(object):             # <<<<<<<<<<<<<<
 *     cdef TaskScheduler _scheduler
 *     cdef TaskScheduler _old_scheduler
 */
struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator {
  PyObject_HEAD
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *_scheduler;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *_old_scheduler;
  int _must_await_on_exit;
  int _must_flush_batches_on_exit;
};


/* "async_task.pxd":31
 * 
 * 
 * cdef class AsyncTask(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index
 */
struct __pyx_obj_5asynq_10async_task_AsyncTask {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  int is_scheduled;
  long iteration_index;
  int group_cancel;
  PyObject *fn;
  PyObject *args;
  PyObject *kwargs;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *caller;
  int depth;
  int max_depth;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *scheduler;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *creator;
  PyObject *_generator;
  PyObject *_last_value;
  PyObject *_frame_info;
  PyObject *_dependencies;
  PyObject *_contexts;
};



/* "qcore/helpers.pxd":33
 * 
 * 
 * cdef class EmptyContext(object):             # <<<<<<<<<<<<<<
 *     cpdef __enter__(self)
 *     cpdef __exit__(self, exc_type, exc_val, exc_tb)
 */

struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext {
  PyObject *(*__pyx___enter__)(struct __pyx_obj_5qcore_7helpers_EmptyContext *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___exit__)(struct __pyx_obj_5qcore_7helpers_EmptyContext *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext *__pyx_vtabptr_5qcore_7helpers_EmptyContext;


/* "qcore/helpers.pxd":40
 * 
 * 
 * cdef class CythonCachedHashWrapper(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 *     cdef int _hash
 */

struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper {
  PyObject *(*value)(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper *, int __pyx_skip_dispatch);
  PyObject *(*hash)(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper *__pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper;


/* "qcore/helpers.pxd":52
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration
 * 
 * cdef class ScopedValue(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 * 
 */

struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue {
  PyObject *(*get)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, int __pyx_skip_dispatch);
  PyObject *(*set)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*override)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue *__pyx_vtabptr_5qcore_7helpers_ScopedValue;


/* "_debug.pxd":18
 * 
 * 
 * cdef class DebugOptions(object):             # <<<<<<<<<<<<<<
 *     cdef public bint DUMP_PRE_ERROR_STATE
 *     cdef public bint DUMP_EXCEPTIONS
 */

struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions {
  PyObject *(*DUMP_ALL)(struct __pyx_obj_5asynq_6_debug_DebugOptions *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions *__pyx_vtabptr_5asynq_6_debug_DebugOptions;


/* "futures.pxd":26
 * 
 * 
 * cdef class FutureBase(object):             # <<<<<<<<<<<<<<
 *     cdef public object _value
 *     cdef public object _error
 */

struct __pyx_vtabstruct_5asynq_7futures_FutureBase {
  PyObject *(*value)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*set_value)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*reset_unsafe)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*error)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*set_error)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, int __pyx_skip_dispatch);
  int (*is_computed)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*_compute)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*_computed)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*make_dependency)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*raise_if_error)(struct __pyx_obj_5asynq_7futures_FutureBase *);
  PyObject *(*dump)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_7futures_FutureBase *__pyx_vtabptr_5asynq_7futures_FutureBase;


/* "futures.pxd":49
 * 
 * 
 * cdef class Future(FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public object _value_provider
 *     cpdef _compute(self)
 */

struct __pyx_vtabstruct_5asynq_7futures_Future {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_Future *__pyx_vtabptr_5asynq_7futures_Future;


/* "futures.pxd":54
 * 
 * 
 * cdef class ConstFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5asynq_7futures_ConstFuture {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_ConstFuture *__pyx_vtabptr_5asynq_7futures_ConstFuture;


/* "futures.pxd":57
 *     pass
 * 
 * cdef class ErrorFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture *__pyx_vtabptr_5asynq_7futures_ErrorFuture;


/* "batching.pxd":26
 * 
 * 
 * cdef class BatchBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public list items
 * 
 */

struct __pyx_vtabstruct_5asynq_8batching_BatchBase {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
  int (*is_flushed)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*is_cancelled)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*is_empty)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*get_priority)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*flush)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*cancel)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel *__pyx_optional_args);
  PyObject *(*_flush)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*_cancel)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*_try_switch_active_batch)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5asynq_8batching_BatchBase *__pyx_vtabptr_5asynq_8batching_BatchBase;


/* "batching.pxd":45
 *     cpdef dump(self, int indent=?)
 * 
 * cdef class BatchItemBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public BatchBase batch
 *     cdef public long index
 */

struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase *__pyx_vtabptr_5asynq_8batching_BatchItemBase;


/* "batching.pxd":53
 * 
 * 
 * cdef class DebugBatchItem(BatchItemBase):             # <<<<<<<<<<<<<<
 *     cdef public object _result
 * 
 */

struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem {
  struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem *__pyx_vtabptr_5asynq_8batching_DebugBatchItem;


/* "batching.pxd":56
 *     cdef public object _result
 * 
 * cdef class DebugBatch(BatchBase):             # <<<<<<<<<<<<<<
 *     cdef public str name
 *     cdef public long index
 */

struct __pyx_vtabstruct_5asynq_8batching_DebugBatch {
  struct __pyx_vtabstruct_5asynq_8batching_BatchBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_DebugBatch *__pyx_vtabptr_5asynq_8batching_DebugBatch;


/* "scheduler.pxd":29
 * 
 * 
 * cdef class TaskScheduler(object):             # <<<<<<<<<<<<<<
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time
 */

struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler {
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *(*activate)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, int, int __pyx_skip_dispatch);
  PyObject *(*reset)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*start)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start *__pyx_optional_args);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*continue_with)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_7futures_FutureBase *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  int (*flush_batches)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches *__pyx_optional_args);
  PyObject *(*add_dependencies)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_dependency)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*schedule_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*_await)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch);
  int (*_execute)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_try_execute_on_new_scheduler)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch);
  int (*_schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  int (*_schedule_without_checks)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  int (*_try_schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int);
  int (*_schedule_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *);
  int (*_flush_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *);
  PyObject *(*_enqueue_new_tasks)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*_continue_with_task)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);
  struct __pyx_obj_5asynq_8batching_BatchBase *(*_continue_with_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch *__pyx_optional_args);
  struct __pyx_obj_5asynq_8batching_BatchBase *(*_select_batch_to_flush)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*_clone)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*_accept_changes)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*dump)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump *__pyx_optional_args);
  PyObject *(*try_time_based_dump)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *__pyx_vtabptr_5asynq_9scheduler_TaskScheduler;


/* "async_task.pxd":31
 * 
 * 
 * cdef class AsyncTask(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index
 */

struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
  int (*is_blocked)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  int (*can_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*start)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start *__pyx_optional_args);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*after)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*cancel_dependencies)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*_continue_on_generator)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, PyObject *);
  PyObject *(*_accept_yield_result)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_accept_error)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_queue_exit)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_queue_throw_error)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_remove_dependency_cython)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_before_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*_after_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*traceback)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*_enter_context)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_leave_context)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_pause_contexts)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*_resume_contexts)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
};
static struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *__pyx_vtabptr_5asynq_10async_task_AsyncTask;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectLookupSpecial.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ListCompAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* SaveResetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* SwapException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'qcore.events' */
static PyTypeObject *__pyx_ptype_5qcore_6events_EventHook = 0;
static PyTypeObject *__pyx_ptype_5qcore_6events_SinkingEventHook = 0;
static PyTypeObject *__pyx_ptype_5qcore_6events_EventInterceptor = 0;
static struct __pyx_obj_5qcore_6events_SinkingEventHook **__pyx_vp_5qcore_6events_sinking_event_hook = 0;
#define __pyx_v_5qcore_6events_sinking_event_hook (*__pyx_vp_5qcore_6events_sinking_event_hook)

/* Module declarations from 'qcore.helpers' */
static PyTypeObject *__pyx_ptype_5qcore_7helpers_MarkerObject = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_EmptyContext = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_ScopedValue = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers__PropertyOverrideContext = 0;
static PyObject **__pyx_vp_5qcore_7helpers_empty_list = 0;
#define __pyx_v_5qcore_7helpers_empty_list (*__pyx_vp_5qcore_7helpers_empty_list)
static PyObject **__pyx_vp_5qcore_7helpers_empty_tuple = 0;
#define __pyx_v_5qcore_7helpers_empty_tuple (*__pyx_vp_5qcore_7helpers_empty_tuple)
static PyObject **__pyx_vp_5qcore_7helpers_empty_dict = 0;
#define __pyx_v_5qcore_7helpers_empty_dict (*__pyx_vp_5qcore_7helpers_empty_dict)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_none = 0;
#define __pyx_v_5qcore_7helpers_none (*__pyx_vp_5qcore_7helpers_none)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_miss = 0;
#define __pyx_v_5qcore_7helpers_miss (*__pyx_vp_5qcore_7helpers_miss)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_same = 0;
#define __pyx_v_5qcore_7helpers_same (*__pyx_vp_5qcore_7helpers_same)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_unspecified = 0;
#define __pyx_v_5qcore_7helpers_unspecified (*__pyx_vp_5qcore_7helpers_unspecified)
static struct __pyx_obj_5qcore_7helpers_EmptyContext **__pyx_vp_5qcore_7helpers_empty_context = 0;
#define __pyx_v_5qcore_7helpers_empty_context (*__pyx_vp_5qcore_7helpers_empty_context)
static PyObject **__pyx_vp_5qcore_7helpers_CachedHashWrapper = 0;
#define __pyx_v_5qcore_7helpers_CachedHashWrapper (*__pyx_vp_5qcore_7helpers_CachedHashWrapper)
static PyObject **__pyx_vp_5qcore_7helpers_override = 0;
#define __pyx_v_5qcore_7helpers_override (*__pyx_vp_5qcore_7helpers_override)

/* Module declarations from 'asynq._debug' */
static PyTypeObject *__pyx_ptype_5asynq_6_debug_DebugOptions = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_6_debug_options = 0;
#define __pyx_v_5asynq_6_debug_options (*__pyx_vp_5asynq_6_debug_options)

/* Module declarations from 'asynq.futures' */
static PyTypeObject *__pyx_ptype_5asynq_7futures_FutureBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_Future = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_ConstFuture = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_ErrorFuture = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_7futures__debug_options = 0;
#define __pyx_v_5asynq_7futures__debug_options (*__pyx_vp_5asynq_7futures__debug_options)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5asynq_7futures__none = 0;
#define __pyx_v_5asynq_7futures__none (*__pyx_vp_5asynq_7futures__none)

/* Module declarations from 'qcore.inspection' */

/* Module declarations from 'asynq.batching' */
static PyTypeObject *__pyx_ptype_5asynq_8batching_BatchBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_BatchItemBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_DebugBatchItem = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_DebugBatch = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_8batching__debug_options = 0;
#define __pyx_v_5asynq_8batching__debug_options (*__pyx_vp_5asynq_8batching__debug_options)
static PyObject **__pyx_vp_5asynq_8batching__debug_batch_state = 0;
#define __pyx_v_5asynq_8batching__debug_batch_state (*__pyx_vp_5asynq_8batching__debug_batch_state)

/* Module declarations from 'asynq.scheduler' */
static PyTypeObject *__pyx_ptype_5asynq_9scheduler_TaskScheduler = 0;
static PyTypeObject *__pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_9scheduler__debug_options = 0;
#define __pyx_v_5asynq_9scheduler__debug_options (*__pyx_vp_5asynq_9scheduler__debug_options)
static PyObject **__pyx_vp_5asynq_9scheduler__futures_none = 0;
#define __pyx_v_5asynq_9scheduler__futures_none (*__pyx_vp_5asynq_9scheduler__futures_none)
static PyObject **__pyx_vp_5asynq_9scheduler__state = 0;
#define __pyx_v_5asynq_9scheduler__state (*__pyx_vp_5asynq_9scheduler__state)
static struct __pyx_obj_5asynq_9scheduler_TaskScheduler *(*__pyx_f_5asynq_9scheduler_get_scheduler)(int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'asynq.async_task' */
static PyTypeObject *__pyx_ptype_5asynq_10async_task_AsyncTask = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_10async_task__debug_options = 0;
#define __pyx_v_5asynq_10async_task__debug_options (*__pyx_vp_5asynq_10async_task__debug_options)
static PyObject **__pyx_vp_5asynq_10async_task__futures_none = 0;
#define __pyx_v_5asynq_10async_task__futures_none (*__pyx_vp_5asynq_10async_task__futures_none)
static PyObject **__pyx_vp_5asynq_10async_task__none_future = 0;
#define __pyx_v_5asynq_10async_task__none_future (*__pyx_vp_5asynq_10async_task__none_future)
static PyObject **__pyx_vp_5asynq_10async_task__empty_tuple = 0;
#define __pyx_v_5asynq_10async_task__empty_tuple (*__pyx_vp_5asynq_10async_task__empty_tuple)
static PyObject **__pyx_vp_5asynq_10async_task__empty_dictionary = 0;
#define __pyx_v_5asynq_10async_task__empty_dictionary (*__pyx_vp_5asynq_10async_task__empty_dictionary)
static PyObject *(*__pyx_f_5asynq_10async_task_unwrap)(PyObject *); /*proto*/

/* Module declarations from 'asynq.utils' */
static struct __pyx_obj_5asynq_6_debug_DebugOptions *__pyx_v_5asynq_5utils__debug_options = 0;
static PyObject *__pyx_f_5asynq_5utils_execute_on_separate_scheduler(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_5utils_execute_on_separate_scheduler *__pyx_optional_args); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_5asynq_5utils__await(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_5asynq_5utils__unwrap(PyObject *); /*proto*/
static PyObject *__pyx_f_5asynq_5utils_result(PyObject *, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_5asynq_5utils__extract_tasks(PyObject *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "asynq.utils"
int __pyx_module_is_main_asynq__utils = 0;

/* Implementation of 'asynq.utils' */
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k__3[] = "";
static const char __pyx_k_six[] = "six";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_await[] = "await";
static const char __pyx_k_debug[] = "_debug";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_futures[] = "futures";
static const char __pyx_k_contexts[] = "contexts";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_is_active[] = "_is_active";
static const char __pyx_k_scheduler[] = "scheduler";
static const char __pyx_k_FutureBase[] = "FutureBase";
static const char __pyx_k_async_task[] = "async_task";
static const char __pyx_k_itervalues[] = "itervalues";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_asynq_utils[] = "asynq.utils";
static const char __pyx_k_async_pull_fn[] = "async_pull_fn";
static const char __pyx_k_debug_options[] = "_debug_options";
static const char __pyx_k_AsyncTaskResult[] = "AsyncTaskResult";
static const char __pyx_k_No_arguments_to_unwrap[] = "No arguments to unwrap.";
static const char __pyx_k_home_jelle_github_asynq_asynq_u[] = "/home/jelle/github/asynq/asynq/utils.py";
static const char __pyx_k_Cannot_process_an_object_of_type[] = "Cannot process an object of type '%s': only futures and None are allowed.";
static const char __pyx_k_You_probably_forgot_to_yield_thi[] = "You probably forgot to yield this value before returning";
static const char __pyx_k_the_task_should_have_been_comput[] = "the task should have been computed at this point";
static PyObject *__pyx_n_s_AsyncTaskResult;
static PyObject *__pyx_kp_s_Cannot_process_an_object_of_type;
static PyObject *__pyx_n_s_FutureBase;
static PyObject *__pyx_kp_s_No_arguments_to_unwrap;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_You_probably_forgot_to_yield_thi;
static PyObject *__pyx_n_s__3;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_async_pull_fn;
static PyObject *__pyx_n_s_async_task;
static PyObject *__pyx_n_s_asynq_utils;
static PyObject *__pyx_n_s_await;
static PyObject *__pyx_n_s_contexts;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_debug_options;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_futures;
static PyObject *__pyx_kp_s_home_jelle_github_asynq_asynq_u;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_is_active;
static PyObject *__pyx_n_s_itervalues;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_scheduler;
static PyObject *__pyx_n_s_six;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_the_task_should_have_been_comput;
static PyObject *__pyx_pf_5asynq_5utils_execute_on_separate_scheduler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_async_pull_fn, PyObject *__pyx_v_contexts, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5asynq_5utils_2await(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_5asynq_5utils_4result(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_k_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_codeobj__5;

/* "asynq/utils.py":26
 * 
 * 
 * def execute_on_separate_scheduler(async_pull_fn, contexts, args=(), kwargs={}):             # <<<<<<<<<<<<<<
 *     current_scheduler = scheduler.get_scheduler()
 *     new_scheduler = current_scheduler._clone()
 */

static PyObject *__pyx_pw_5asynq_5utils_1execute_on_separate_scheduler(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_5utils_execute_on_separate_scheduler(PyObject *__pyx_v_async_pull_fn, PyObject *__pyx_v_contexts, CYTHON_UNUSED int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_5utils_execute_on_separate_scheduler *__pyx_optional_args) {
  PyObject *__pyx_v_args = ((PyObject*)__pyx_empty_tuple);
  PyObject *__pyx_v_kwargs = __pyx_k_;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_current_scheduler = 0;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_new_scheduler = 0;
  PyObject *__pyx_v_ctx = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("execute_on_separate_scheduler", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_args = __pyx_optional_args->args;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_kwargs = __pyx_optional_args->kwargs;
      }
    }
  }

  /* "asynq/utils.py":27
 * 
 * def execute_on_separate_scheduler(async_pull_fn, contexts, args=(), kwargs={}):
 *     current_scheduler = scheduler.get_scheduler()             # <<<<<<<<<<<<<<
 *     new_scheduler = current_scheduler._clone()
 *     try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_current_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/utils.py":28
 * def execute_on_separate_scheduler(async_pull_fn, contexts, args=(), kwargs={}):
 *     current_scheduler = scheduler.get_scheduler()
 *     new_scheduler = current_scheduler._clone()             # <<<<<<<<<<<<<<
 *     try:
 *         with new_scheduler.activate(True, False):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_current_scheduler->__pyx_vtab)->_clone(__pyx_v_current_scheduler, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5asynq_9scheduler_TaskScheduler))))) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_v_new_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/utils.py":29
 *     current_scheduler = scheduler.get_scheduler()
 *     new_scheduler = current_scheduler._clone()
 *     try:             # <<<<<<<<<<<<<<
 *         with new_scheduler.activate(True, False):
 *             task = async_task.AsyncTask(async_pull_fn(*args, **kwargs), async_pull_fn, args, kwargs)
 */
  /*try:*/ {

    /* "asynq/utils.py":30
 *     new_scheduler = current_scheduler._clone()
 *     try:
 *         with new_scheduler.activate(True, False):             # <<<<<<<<<<<<<<
 *             task = async_task.AsyncTask(async_pull_fn(*args, **kwargs), async_pull_fn, args, kwargs)
 *             # The above tasks should have no contexts right now (because it is created in a vacuum).
 */
    /*with:*/ {
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_new_scheduler->__pyx_vtab)->activate(__pyx_v_new_scheduler, 1, 0, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 30, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L6_error)
      }
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_7);
          __Pyx_XGOTREF(__pyx_t_8);
          /*try:*/ {

            /* "asynq/utils.py":31
 *     try:
 *         with new_scheduler.activate(True, False):
 *             task = async_task.AsyncTask(async_pull_fn(*args, **kwargs), async_pull_fn, args, kwargs)             # <<<<<<<<<<<<<<
 *             # The above tasks should have no contexts right now (because it is created in a vacuum).
 *             # We filter the contexts that are already active because they will remain active
 */
            if (unlikely(__pyx_v_args == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
              __PYX_ERR(0, 31, __pyx_L10_error)
            }
            if (unlikely(__pyx_v_kwargs == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "argument after ** must be a mapping, not NoneType");
              __PYX_ERR(0, 31, __pyx_L10_error)
            }
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_async_pull_fn, __pyx_v_args, __pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 31, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
            __Pyx_INCREF(__pyx_v_async_pull_fn);
            __Pyx_GIVEREF(__pyx_v_async_pull_fn);
            PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_async_pull_fn);
            __Pyx_INCREF(__pyx_v_args);
            __Pyx_GIVEREF(__pyx_v_args);
            PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_args);
            __Pyx_INCREF(__pyx_v_kwargs);
            __Pyx_GIVEREF(__pyx_v_kwargs);
            PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_kwargs);
            __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5asynq_10async_task_AsyncTask), __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_v_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
            __pyx_t_1 = 0;

            /* "asynq/utils.py":36
 *             # throughout the execution of this task. Not filtering them would result in calling
 *             # resume on them twice in a row (which most async_contexts do not support).
 *             task._contexts = [ctx for ctx in contexts if not ctx._is_active]             # <<<<<<<<<<<<<<
 *             new_scheduler.await([task])
 *     finally:
 */
            __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (unlikely(__pyx_v_contexts == Py_None)) {
              PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
              __PYX_ERR(0, 36, __pyx_L10_error)
            }
            __pyx_t_3 = __pyx_v_contexts; __Pyx_INCREF(__pyx_t_3); __pyx_t_9 = 0;
            for (;;) {
              if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_3)) break;
              #if CYTHON_COMPILING_IN_CPYTHON
              __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 36, __pyx_L10_error)
              #else
              __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 36, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_4);
              #endif
              __Pyx_XDECREF_SET(__pyx_v_ctx, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ctx, __pyx_n_s_is_active); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 36, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 36, __pyx_L10_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __pyx_t_11 = ((!__pyx_t_10) != 0);
              if (__pyx_t_11) {
                if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_v_ctx))) __PYX_ERR(0, 36, __pyx_L10_error)
              }
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_v_task->_contexts);
            __Pyx_DECREF(__pyx_v_task->_contexts);
            __pyx_v_task->_contexts = ((PyObject*)__pyx_t_1);
            __pyx_t_1 = 0;

            /* "asynq/utils.py":37
 *             # resume on them twice in a row (which most async_contexts do not support).
 *             task._contexts = [ctx for ctx in contexts if not ctx._is_active]
 *             new_scheduler.await([task])             # <<<<<<<<<<<<<<
 *     finally:
 *         current_scheduler._accept_changes(new_scheduler)
 */
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_new_scheduler), __pyx_n_s_await); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L10_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_INCREF(((PyObject *)__pyx_v_task));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
            PyList_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_v_task));
            __pyx_t_5 = NULL;
            if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L10_error)
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              __pyx_t_12 = PyTuple_New(1+1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 37, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_12);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_GIVEREF(__pyx_t_4);
              PyTuple_SET_ITEM(__pyx_t_12, 0+1, __pyx_t_4);
              __pyx_t_4 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L10_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "asynq/utils.py":30
 *     new_scheduler = current_scheduler._clone()
 *     try:
 *         with new_scheduler.activate(True, False):             # <<<<<<<<<<<<<<
 *             task = async_task.AsyncTask(async_pull_fn(*args, **kwargs), async_pull_fn, args, kwargs)
 *             # The above tasks should have no contexts right now (because it is created in a vacuum).
 */
          }
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          goto __pyx_L17_try_end;
          __pyx_L10_error:;
          __Pyx_PyThreadState_assign
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("asynq.utils.execute_on_separate_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_12) < 0) __PYX_ERR(0, 30, __pyx_L12_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_4 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_3, __pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 30, __pyx_L12_except_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 30, __pyx_L12_except_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (__pyx_t_11 < 0) __PYX_ERR(0, 30, __pyx_L12_except_error)
            __pyx_t_10 = ((!(__pyx_t_11 != 0)) != 0);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_12);
              __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_12);
              __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_12 = 0; 
              __PYX_ERR(0, 30, __pyx_L12_except_error)
            }
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            goto __pyx_L11_exception_handled;
          }
          __pyx_L12_except_error:;
          __Pyx_PyThreadState_assign
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
          goto __pyx_L4_error;
          __pyx_L11_exception_handled:;
          __Pyx_PyThreadState_assign
          __Pyx_XGIVEREF(__pyx_t_6);
          __Pyx_XGIVEREF(__pyx_t_7);
          __Pyx_XGIVEREF(__pyx_t_8);
          __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
          __pyx_L17_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_2) {
            __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__2, NULL);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 30, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          goto __pyx_L9;
        }
        __pyx_L9:;
      }
      goto __pyx_L24;
      __pyx_L6_error:;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L4_error;
      __pyx_L24:;
    }
  }

  /* "asynq/utils.py":39
 *             new_scheduler.await([task])
 *     finally:
 *         current_scheduler._accept_changes(new_scheduler)             # <<<<<<<<<<<<<<
 *     assert task.is_computed(), 'the task should have been computed at this point'
 *     return task.value()
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_12 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_current_scheduler->__pyx_vtab)->_accept_changes(__pyx_v_current_scheduler, __pyx_v_new_scheduler, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 39, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_2 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_13 = 0; __pyx_t_17 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_6, &__pyx_t_13, &__pyx_t_17);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_2, &__pyx_t_8, &__pyx_t_7) < 0)) __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_8, &__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_2);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_17);
      __pyx_t_14 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
      {
        __pyx_t_12 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_current_scheduler->__pyx_vtab)->_accept_changes(__pyx_v_current_scheduler, __pyx_v_new_scheduler, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 39, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_13, __pyx_t_17);
      }
      __Pyx_XGIVEREF(__pyx_t_2);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_ErrRestore(__pyx_t_2, __pyx_t_8, __pyx_t_7);
      __pyx_t_2 = 0; __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_13 = 0; __pyx_t_17 = 0;
      __pyx_lineno = __pyx_t_14; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
      goto __pyx_L1_error;
      __pyx_L26_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_13, __pyx_t_17);
      }
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_6 = 0; __pyx_t_13 = 0; __pyx_t_17 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "asynq/utils.py":40
 *     finally:
 *         current_scheduler._accept_changes(new_scheduler)
 *     assert task.is_computed(), 'the task should have been computed at this point'             # <<<<<<<<<<<<<<
 *     return task.value()
 * 
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!__pyx_v_task)) { __Pyx_RaiseUnboundLocalError("task"); __PYX_ERR(0, 40, __pyx_L1_error) }
    __pyx_t_10 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 40, __pyx_L1_error)
    if (unlikely(!(__pyx_t_10 != 0))) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_s_the_task_should_have_been_comput);
      __PYX_ERR(0, 40, __pyx_L1_error)
    }
  }
  #endif

  /* "asynq/utils.py":41
 *         current_scheduler._accept_changes(new_scheduler)
 *     assert task.is_computed(), 'the task should have been computed at this point'
 *     return task.value()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_v_task)) { __Pyx_RaiseUnboundLocalError("task"); __PYX_ERR(0, 41, __pyx_L1_error) }
  __pyx_t_12 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.value(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_r = __pyx_t_12;
  __pyx_t_12 = 0;
  goto __pyx_L0;

  /* "asynq/utils.py":26
 * 
 * 
 * def execute_on_separate_scheduler(async_pull_fn, contexts, args=(), kwargs={}):             # <<<<<<<<<<<<<<
 *     current_scheduler = scheduler.get_scheduler()
 *     new_scheduler = current_scheduler._clone()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("asynq.utils.execute_on_separate_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_current_scheduler);
  __Pyx_XDECREF((PyObject *)__pyx_v_task);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_scheduler);
  __Pyx_XDECREF(__pyx_v_ctx);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_5utils_1execute_on_separate_scheduler(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_5utils_1execute_on_separate_scheduler(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_async_pull_fn = 0;
  PyObject *__pyx_v_contexts = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_on_separate_scheduler (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_async_pull_fn,&__pyx_n_s_contexts,&__pyx_n_s_args,&__pyx_n_s_kwargs,0};
    PyObject* values[4] = {0,0,0,0};
    values[2] = ((PyObject*)__pyx_empty_tuple);
    values[3] = __pyx_k_;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_async_pull_fn)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_contexts)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("execute_on_separate_scheduler", 0, 2, 4, 1); __PYX_ERR(0, 26, __pyx_L3_error)
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_args);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_kwargs);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "execute_on_separate_scheduler") < 0)) __PYX_ERR(0, 26, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_async_pull_fn = values[0];
    __pyx_v_contexts = ((PyObject*)values[1]);
    __pyx_v_args = ((PyObject*)values[2]);
    __pyx_v_kwargs = ((PyObject*)values[3]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("execute_on_separate_scheduler", 0, 2, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 26, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.utils.execute_on_separate_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_contexts), (&PyList_Type), 1, "contexts", 1))) __PYX_ERR(0, 26, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_args), (&PyTuple_Type), 1, "args", 1))) __PYX_ERR(0, 26, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_kwargs), (&PyDict_Type), 1, "kwargs", 1))) __PYX_ERR(0, 26, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_5utils_execute_on_separate_scheduler(__pyx_self, __pyx_v_async_pull_fn, __pyx_v_contexts, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_5utils_execute_on_separate_scheduler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_async_pull_fn, PyObject *__pyx_v_contexts, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5asynq_5utils_execute_on_separate_scheduler __pyx_t_2;
  __Pyx_RefNannySetupContext("execute_on_separate_scheduler", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 2;
  __pyx_t_2.args = __pyx_v_args;
  __pyx_t_2.kwargs = __pyx_v_kwargs;
  __pyx_t_1 = __pyx_f_5asynq_5utils_execute_on_separate_scheduler(__pyx_v_async_pull_fn, __pyx_v_contexts, 0, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.utils.execute_on_separate_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/utils.py":44
 * 
 * 
 * def await(*args):             # <<<<<<<<<<<<<<
 *     """Runs all the tasks specified in args,
 *     and finally returns args unwrapped.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_5utils_3await(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_5utils_2await[] = "Runs all the tasks specified in args,\n    and finally returns args unwrapped.\n\n    ";
static PyMethodDef __pyx_mdef_5asynq_5utils_3await = {"await", (PyCFunction)__pyx_pw_5asynq_5utils_3await, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_5utils_2await};
static PyObject *__pyx_pw_5asynq_5utils_3await(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("await (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "await", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_5asynq_5utils_2await(__pyx_self, __pyx_v_args);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_5utils_2await(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("await", 0);

  /* "asynq/utils.py":49
 * 
 *     """
 *     return _await(args)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_5utils__await(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/utils.py":44
 * 
 * 
 * def await(*args):             # <<<<<<<<<<<<<<
 *     """Runs all the tasks specified in args,
 *     and finally returns args unwrapped.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.utils.await", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/utils.py":52
 * 
 * 
 * def _await(args):             # <<<<<<<<<<<<<<
 *     tasks = set()
 *     _extract_tasks(args, tasks)
 */

static CYTHON_INLINE PyObject *__pyx_f_5asynq_5utils__await(PyObject *__pyx_v_args) {
  PyObject *__pyx_v_tasks = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_await", 0);

  /* "asynq/utils.py":53
 * 
 * def _await(args):
 *     tasks = set()             # <<<<<<<<<<<<<<
 *     _extract_tasks(args, tasks)
 *     scheduler.get_scheduler()._execute(True, tasks)
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tasks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/utils.py":54
 * def _await(args):
 *     tasks = set()
 *     _extract_tasks(args, tasks)             # <<<<<<<<<<<<<<
 *     scheduler.get_scheduler()._execute(True, tasks)
 *     return _unwrap(args)
 */
  __pyx_t_1 = __pyx_f_5asynq_5utils__extract_tasks(__pyx_v_args, __pyx_v_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/utils.py":55
 *     tasks = set()
 *     _extract_tasks(args, tasks)
 *     scheduler.get_scheduler()._execute(True, tasks)             # <<<<<<<<<<<<<<
 *     return _unwrap(args)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1)->__pyx_vtab)->_execute(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1), 1, __pyx_v_tasks, 0); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/utils.py":56
 *     _extract_tasks(args, tasks)
 *     scheduler.get_scheduler()._execute(True, tasks)
 *     return _unwrap(args)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_5utils__unwrap(__pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/utils.py":52
 * 
 * 
 * def _await(args):             # <<<<<<<<<<<<<<
 *     tasks = set()
 *     _extract_tasks(args, tasks)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.utils._await", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tasks);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/utils.py":59
 * 
 * 
 * def _unwrap(args):             # <<<<<<<<<<<<<<
 *     l = len(args)
 *     assert l != 0, "No arguments to unwrap."
 */

static CYTHON_INLINE PyObject *__pyx_f_5asynq_5utils__unwrap(PyObject *__pyx_v_args) {
  int __pyx_v_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_unwrap", 0);

  /* "asynq/utils.py":60
 * 
 * def _unwrap(args):
 *     l = len(args)             # <<<<<<<<<<<<<<
 *     assert l != 0, "No arguments to unwrap."
 *     if l == 1:
 */
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 60, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_args); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 60, __pyx_L1_error)
  __pyx_v_l = __pyx_t_1;

  /* "asynq/utils.py":61
 * def _unwrap(args):
 *     l = len(args)
 *     assert l != 0, "No arguments to unwrap."             # <<<<<<<<<<<<<<
 *     if l == 1:
 *         return async_task.unwrap(args[0])  # Special case: no need to return tuple here
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_l != 0) != 0))) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_s_No_arguments_to_unwrap);
      __PYX_ERR(0, 61, __pyx_L1_error)
    }
  }
  #endif

  /* "asynq/utils.py":62
 *     l = len(args)
 *     assert l != 0, "No arguments to unwrap."
 *     if l == 1:             # <<<<<<<<<<<<<<
 *         return async_task.unwrap(args[0])  # Special case: no need to return tuple here
 *     return async_task.unwrap(args)
 */
  __pyx_t_2 = ((__pyx_v_l == 1) != 0);
  if (__pyx_t_2) {

    /* "asynq/utils.py":63
 *     assert l != 0, "No arguments to unwrap."
 *     if l == 1:
 *         return async_task.unwrap(args[0])  # Special case: no need to return tuple here             # <<<<<<<<<<<<<<
 *     return async_task.unwrap(args)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 63, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_args, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_f_5asynq_10async_task_unwrap(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 63, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "asynq/utils.py":62
 *     l = len(args)
 *     assert l != 0, "No arguments to unwrap."
 *     if l == 1:             # <<<<<<<<<<<<<<
 *         return async_task.unwrap(args[0])  # Special case: no need to return tuple here
 *     return async_task.unwrap(args)
 */
  }

  /* "asynq/utils.py":64
 *     if l == 1:
 *         return async_task.unwrap(args[0])  # Special case: no need to return tuple here
 *     return async_task.unwrap(args)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __pyx_f_5asynq_10async_task_unwrap(__pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "asynq/utils.py":59
 * 
 * 
 * def _unwrap(args):             # <<<<<<<<<<<<<<
 *     l = len(args)
 *     assert l != 0, "No arguments to unwrap."
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.utils._unwrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/utils.py":67
 * 
 * 
 * def result(value):             # <<<<<<<<<<<<<<
 *     """An async equivalent for return for async methods: result(x); return."""
 *     assert not isinstance(value, FutureBase), \
 */

static PyObject *__pyx_pw_5asynq_5utils_5result(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static PyObject *__pyx_f_5asynq_5utils_result(PyObject *__pyx_v_value, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("result", 0);

  /* "asynq/utils.py":69
 * def result(value):
 *     """An async equivalent for return for async methods: result(x); return."""
 *     assert not isinstance(value, FutureBase), \             # <<<<<<<<<<<<<<
 *         "You probably forgot to yield this value before returning"
 *     raise async_task.AsyncTaskResult(value)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_7futures_FutureBase); 
    if (unlikely(!((!(__pyx_t_1 != 0)) != 0))) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_s_You_probably_forgot_to_yield_thi);
      __PYX_ERR(0, 69, __pyx_L1_error)
    }
  }
  #endif

  /* "asynq/utils.py":71
 *     assert not isinstance(value, FutureBase), \
 *         "You probably forgot to yield this value before returning"
 *     raise async_task.AsyncTaskResult(value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_async_task); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_AsyncTaskResult); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_value);
    __Pyx_GIVEREF(__pyx_v_value);
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_value);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_2, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __PYX_ERR(0, 71, __pyx_L1_error)

  /* "asynq/utils.py":67
 * 
 * 
 * def result(value):             # <<<<<<<<<<<<<<
 *     """An async equivalent for return for async methods: result(x); return."""
 *     assert not isinstance(value, FutureBase), \
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.utils.result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_5utils_5result(PyObject *__pyx_self, PyObject *__pyx_v_value); /*proto*/
static char __pyx_doc_5asynq_5utils_4result[] = "An async equivalent for return for async methods: result(x); return.";
static PyObject *__pyx_pw_5asynq_5utils_5result(PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("result (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_5utils_4result(__pyx_self, ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_5utils_4result(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("result", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_5utils_result(__pyx_v_value, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.utils.result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/utils.py":76
 * # Private part
 * 
 * def _extract_tasks(value, result):             # <<<<<<<<<<<<<<
 *     if value is None:
 *         pass
 */

static PyObject *__pyx_f_5asynq_5utils__extract_tasks(PyObject *__pyx_v_value, PyObject *__pyx_v_result) {
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("_extract_tasks", 0);

  /* "asynq/utils.py":77
 * 
 * def _extract_tasks(value, result):
 *     if value is None:             # <<<<<<<<<<<<<<
 *         pass
 *     elif isinstance(value, async_task.AsyncTask):
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {
    goto __pyx_L3;
  }

  /* "asynq/utils.py":79
 *     if value is None:
 *         pass
 *     elif isinstance(value, async_task.AsyncTask):             # <<<<<<<<<<<<<<
 *         result.add(value)
 *     elif type(value) is tuple or type(value) is list:
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_10async_task_AsyncTask); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/utils.py":80
 *         pass
 *     elif isinstance(value, async_task.AsyncTask):
 *         result.add(value)             # <<<<<<<<<<<<<<
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:
 */
    if (unlikely(__pyx_v_result == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "add");
      __PYX_ERR(0, 80, __pyx_L1_error)
    }
    __pyx_t_3 = PySet_Add(__pyx_v_result, __pyx_v_value); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 80, __pyx_L1_error)

    /* "asynq/utils.py":79
 *     if value is None:
 *         pass
 *     elif isinstance(value, async_task.AsyncTask):             # <<<<<<<<<<<<<<
 *         result.add(value)
 *     elif type(value) is tuple or type(value) is list:
 */
    goto __pyx_L3;
  }

  /* "asynq/utils.py":81
 *     elif isinstance(value, async_task.AsyncTask):
 *         result.add(value)
 *     elif type(value) is tuple or type(value) is list:             # <<<<<<<<<<<<<<
 *         for item in value:
 *             _extract_tasks(item, result)
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyTuple_Type)));
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyList_Type)));
  __pyx_t_2 = (__pyx_t_4 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynq/utils.py":82
 *         result.add(value)
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:             # <<<<<<<<<<<<<<
 *             _extract_tasks(item, result)
 *     elif type(value) is dict:
 */
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_5 = __pyx_v_value; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 82, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 82, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 82, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 82, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "asynq/utils.py":83
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:
 *             _extract_tasks(item, result)             # <<<<<<<<<<<<<<
 *     elif type(value) is dict:
 *         for item in six.itervalues(value):
 */
      __pyx_t_8 = __pyx_f_5asynq_5utils__extract_tasks(__pyx_v_item, __pyx_v_result); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "asynq/utils.py":82
 *         result.add(value)
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:             # <<<<<<<<<<<<<<
 *             _extract_tasks(item, result)
 *     elif type(value) is dict:
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "asynq/utils.py":81
 *     elif isinstance(value, async_task.AsyncTask):
 *         result.add(value)
 *     elif type(value) is tuple or type(value) is list:             # <<<<<<<<<<<<<<
 *         for item in value:
 *             _extract_tasks(item, result)
 */
    goto __pyx_L3;
  }

  /* "asynq/utils.py":84
 *         for item in value:
 *             _extract_tasks(item, result)
 *     elif type(value) is dict:             # <<<<<<<<<<<<<<
 *         for item in six.itervalues(value):
 *             _extract_tasks(item, result)
 */
  __pyx_t_1 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyDict_Type)));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/utils.py":85
 *             _extract_tasks(item, result)
 *     elif type(value) is dict:
 *         for item in six.itervalues(value):             # <<<<<<<<<<<<<<
 *             _extract_tasks(item, result)
 *     elif isinstance(value, FutureBase):
 */
    __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_six); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_itervalues); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_v_value);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
      __pyx_t_9 = __pyx_t_5; __Pyx_INCREF(__pyx_t_9); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 85, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_7 = Py_TYPE(__pyx_t_9)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 85, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_9))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_5 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 85, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_9, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_9)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_6); __Pyx_INCREF(__pyx_t_5); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 85, __pyx_L1_error)
          #else
          __pyx_t_5 = PySequence_ITEM(__pyx_t_9, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 85, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          #endif
        }
      } else {
        __pyx_t_5 = __pyx_t_7(__pyx_t_9);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 85, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "asynq/utils.py":86
 *     elif type(value) is dict:
 *         for item in six.itervalues(value):
 *             _extract_tasks(item, result)             # <<<<<<<<<<<<<<
 *     elif isinstance(value, FutureBase):
 *         pass
 */
      __pyx_t_5 = __pyx_f_5asynq_5utils__extract_tasks(__pyx_v_item, __pyx_v_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "asynq/utils.py":85
 *             _extract_tasks(item, result)
 *     elif type(value) is dict:
 *         for item in six.itervalues(value):             # <<<<<<<<<<<<<<
 *             _extract_tasks(item, result)
 *     elif isinstance(value, FutureBase):
 */
    }
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "asynq/utils.py":84
 *         for item in value:
 *             _extract_tasks(item, result)
 *     elif type(value) is dict:             # <<<<<<<<<<<<<<
 *         for item in six.itervalues(value):
 *             _extract_tasks(item, result)
 */
    goto __pyx_L3;
  }

  /* "asynq/utils.py":87
 *         for item in six.itervalues(value):
 *             _extract_tasks(item, result)
 *     elif isinstance(value, FutureBase):             # <<<<<<<<<<<<<<
 *         pass
 *     else:
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_7futures_FutureBase); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {
    goto __pyx_L3;
  }

  /* "asynq/utils.py":90
 *         pass
 *     else:
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "Cannot process an object of type '%s': only futures and None are allowed."
 *             % type(value)
 */
  /*else*/ {

    /* "asynq/utils.py":92
 *         raise TypeError(
 *             "Cannot process an object of type '%s': only futures and None are allowed."
 *             % type(value)             # <<<<<<<<<<<<<<
 *         )
 */
    __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_Cannot_process_an_object_of_type, ((PyObject *)Py_TYPE(__pyx_v_value))); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 92, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "asynq/utils.py":90
 *         pass
 *     else:
 *         raise TypeError(             # <<<<<<<<<<<<<<
 *             "Cannot process an object of type '%s': only futures and None are allowed."
 *             % type(value)
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_5, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_9, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __PYX_ERR(0, 90, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "asynq/utils.py":76
 * # Private part
 * 
 * def _extract_tasks(value, result):             # <<<<<<<<<<<<<<
 *     if value is None:
 *         pass
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynq.utils._extract_tasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"execute_on_separate_scheduler", (PyCFunction)__pyx_pw_5asynq_5utils_1execute_on_separate_scheduler, METH_VARARGS|METH_KEYWORDS, 0},
  {"result", (PyCFunction)__pyx_pw_5asynq_5utils_5result, METH_O, __pyx_doc_5asynq_5utils_4result},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "utils",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AsyncTaskResult, __pyx_k_AsyncTaskResult, sizeof(__pyx_k_AsyncTaskResult), 0, 0, 1, 1},
  {&__pyx_kp_s_Cannot_process_an_object_of_type, __pyx_k_Cannot_process_an_object_of_type, sizeof(__pyx_k_Cannot_process_an_object_of_type), 0, 0, 1, 0},
  {&__pyx_n_s_FutureBase, __pyx_k_FutureBase, sizeof(__pyx_k_FutureBase), 0, 0, 1, 1},
  {&__pyx_kp_s_No_arguments_to_unwrap, __pyx_k_No_arguments_to_unwrap, sizeof(__pyx_k_No_arguments_to_unwrap), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_You_probably_forgot_to_yield_thi, __pyx_k_You_probably_forgot_to_yield_thi, sizeof(__pyx_k_You_probably_forgot_to_yield_thi), 0, 0, 1, 0},
  {&__pyx_n_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_async_pull_fn, __pyx_k_async_pull_fn, sizeof(__pyx_k_async_pull_fn), 0, 0, 1, 1},
  {&__pyx_n_s_async_task, __pyx_k_async_task, sizeof(__pyx_k_async_task), 0, 0, 1, 1},
  {&__pyx_n_s_asynq_utils, __pyx_k_asynq_utils, sizeof(__pyx_k_asynq_utils), 0, 0, 1, 1},
  {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
  {&__pyx_n_s_contexts, __pyx_k_contexts, sizeof(__pyx_k_contexts), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_debug_options, __pyx_k_debug_options, sizeof(__pyx_k_debug_options), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_futures, __pyx_k_futures, sizeof(__pyx_k_futures), 0, 0, 1, 1},
  {&__pyx_kp_s_home_jelle_github_asynq_asynq_u, __pyx_k_home_jelle_github_asynq_asynq_u, sizeof(__pyx_k_home_jelle_github_asynq_asynq_u), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_is_active, __pyx_k_is_active, sizeof(__pyx_k_is_active), 0, 0, 1, 1},
  {&__pyx_n_s_itervalues, __pyx_k_itervalues, sizeof(__pyx_k_itervalues), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_scheduler, __pyx_k_scheduler, sizeof(__pyx_k_scheduler), 0, 0, 1, 1},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_the_task_should_have_been_comput, __pyx_k_the_task_should_have_been_comput, sizeof(__pyx_k_the_task_should_have_been_comput), 0, 0, 1, 0},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 90, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "asynq/utils.py":30
 *     new_scheduler = current_scheduler._clone()
 *     try:
 *         with new_scheduler.activate(True, False):             # <<<<<<<<<<<<<<
 *             task = async_task.AsyncTask(async_pull_fn(*args, **kwargs), async_pull_fn, args, kwargs)
 *             # The above tasks should have no contexts right now (because it is created in a vacuum).
 */
  __pyx_tuple__2 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "asynq/utils.py":44
 * 
 * 
 * def await(*args):             # <<<<<<<<<<<<<<
 *     """Runs all the tasks specified in args,
 *     and finally returns args unwrapped.
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_n_s_args); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(0, 0, 1, 0, CO_VARARGS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jelle_github_asynq_asynq_u, __pyx_n_s_await, 44, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initutils(void); /*proto*/
PyMODINIT_FUNC initutils(void)
#else
PyMODINIT_FUNC PyInit_utils(void); /*proto*/
PyMODINIT_FUNC PyInit_utils(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_utils(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("utils", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_asynq__utils) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "asynq.utils")) {
      if (unlikely(PyDict_SetItemString(modules, "asynq.utils", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_5asynq_5utils__debug_options = ((struct __pyx_obj_5asynq_6_debug_DebugOptions *)Py_None); Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_debug_options, (void *)&__pyx_v_5asynq_5utils__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("execute_on_separate_scheduler", (void (*)(void))__pyx_f_5asynq_5utils_execute_on_separate_scheduler, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_5utils_execute_on_separate_scheduler *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_await", (void (*)(void))__pyx_f_5asynq_5utils__await, "PyObject *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_unwrap", (void (*)(void))__pyx_f_5asynq_5utils__unwrap, "PyObject *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("result", (void (*)(void))__pyx_f_5asynq_5utils_result, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_extract_tasks", (void (*)(void))__pyx_f_5asynq_5utils__extract_tasks, "PyObject *(PyObject *, PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(1, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(2, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(3, 15, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_EventHook = __Pyx_ImportType("qcore.events", "EventHook", sizeof(struct __pyx_obj_5qcore_6events_EventHook), 1); if (unlikely(!__pyx_ptype_5qcore_6events_EventHook)) __PYX_ERR(4, 15, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_SinkingEventHook = __Pyx_ImportType("qcore.events", "SinkingEventHook", sizeof(struct __pyx_obj_5qcore_6events_SinkingEventHook), 1); if (unlikely(!__pyx_ptype_5qcore_6events_SinkingEventHook)) __PYX_ERR(4, 19, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_EventInterceptor = __Pyx_ImportType("qcore.events", "EventInterceptor", sizeof(struct __pyx_obj_5qcore_6events_EventInterceptor), 1); if (unlikely(!__pyx_ptype_5qcore_6events_EventInterceptor)) __PYX_ERR(4, 25, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_MarkerObject = __Pyx_ImportType("qcore.helpers", "MarkerObject", sizeof(struct __pyx_obj_5qcore_7helpers_MarkerObject), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_MarkerObject)) __PYX_ERR(5, 24, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_EmptyContext = __Pyx_ImportType("qcore.helpers", "EmptyContext", sizeof(struct __pyx_obj_5qcore_7helpers_EmptyContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_EmptyContext)) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_EmptyContext = (struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_EmptyContext->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_EmptyContext)) __PYX_ERR(5, 33, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper = __Pyx_ImportType("qcore.helpers", "CythonCachedHashWrapper", sizeof(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper)) __PYX_ERR(5, 40, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper = (struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper)) __PYX_ERR(5, 40, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext = __Pyx_ImportType("qcore.helpers", "_ScopedValueOverrideContext", sizeof(struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext)) __PYX_ERR(5, 50, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_ScopedValue = __Pyx_ImportType("qcore.helpers", "ScopedValue", sizeof(struct __pyx_obj_5qcore_7helpers_ScopedValue), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_ScopedValue)) __PYX_ERR(5, 52, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_ScopedValue = (struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_ScopedValue->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_ScopedValue)) __PYX_ERR(5, 52, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers__PropertyOverrideContext = __Pyx_ImportType("qcore.helpers", "_PropertyOverrideContext", sizeof(struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers__PropertyOverrideContext)) __PYX_ERR(5, 65, __pyx_L1_error)
  __pyx_ptype_5asynq_6_debug_DebugOptions = __Pyx_ImportType("asynq._debug", "DebugOptions", sizeof(struct __pyx_obj_5asynq_6_debug_DebugOptions), 1); if (unlikely(!__pyx_ptype_5asynq_6_debug_DebugOptions)) __PYX_ERR(6, 18, __pyx_L1_error)
  __pyx_vtabptr_5asynq_6_debug_DebugOptions = (struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions*)__Pyx_GetVtable(__pyx_ptype_5asynq_6_debug_DebugOptions->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_6_debug_DebugOptions)) __PYX_ERR(6, 18, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_FutureBase = __Pyx_ImportType("asynq.futures", "FutureBase", sizeof(struct __pyx_obj_5asynq_7futures_FutureBase), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_FutureBase)) __PYX_ERR(7, 26, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_FutureBase = (struct __pyx_vtabstruct_5asynq_7futures_FutureBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_FutureBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_FutureBase)) __PYX_ERR(7, 26, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_Future = __Pyx_ImportType("asynq.futures", "Future", sizeof(struct __pyx_obj_5asynq_7futures_Future), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_Future)) __PYX_ERR(7, 49, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_Future = (struct __pyx_vtabstruct_5asynq_7futures_Future*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_Future->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_Future)) __PYX_ERR(7, 49, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_ConstFuture = __Pyx_ImportType("asynq.futures", "ConstFuture", sizeof(struct __pyx_obj_5asynq_7futures_ConstFuture), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_ConstFuture)) __PYX_ERR(7, 54, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_ConstFuture = (struct __pyx_vtabstruct_5asynq_7futures_ConstFuture*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_ConstFuture->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_ConstFuture)) __PYX_ERR(7, 54, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_ErrorFuture = __Pyx_ImportType("asynq.futures", "ErrorFuture", sizeof(struct __pyx_obj_5asynq_7futures_ErrorFuture), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_ErrorFuture)) __PYX_ERR(7, 57, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_ErrorFuture = (struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_ErrorFuture->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_ErrorFuture)) __PYX_ERR(7, 57, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_BatchBase = __Pyx_ImportType("asynq.batching", "BatchBase", sizeof(struct __pyx_obj_5asynq_8batching_BatchBase), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_BatchBase)) __PYX_ERR(8, 26, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_BatchBase = (struct __pyx_vtabstruct_5asynq_8batching_BatchBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_BatchBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_BatchBase)) __PYX_ERR(8, 26, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_BatchItemBase = __Pyx_ImportType("asynq.batching", "BatchItemBase", sizeof(struct __pyx_obj_5asynq_8batching_BatchItemBase), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_BatchItemBase)) __PYX_ERR(8, 45, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_BatchItemBase = (struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_BatchItemBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_BatchItemBase)) __PYX_ERR(8, 45, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_DebugBatchItem = __Pyx_ImportType("asynq.batching", "DebugBatchItem", sizeof(struct __pyx_obj_5asynq_8batching_DebugBatchItem), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_DebugBatchItem)) __PYX_ERR(8, 53, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_DebugBatchItem = (struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_DebugBatchItem->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_DebugBatchItem)) __PYX_ERR(8, 53, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_DebugBatch = __Pyx_ImportType("asynq.batching", "DebugBatch", sizeof(struct __pyx_obj_5asynq_8batching_DebugBatch), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_DebugBatch)) __PYX_ERR(8, 56, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_DebugBatch = (struct __pyx_vtabstruct_5asynq_8batching_DebugBatch*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_DebugBatch->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_DebugBatch)) __PYX_ERR(8, 56, __pyx_L1_error)
  __pyx_ptype_5asynq_9scheduler_TaskScheduler = __Pyx_ImportType("asynq.scheduler", "TaskScheduler", sizeof(struct __pyx_obj_5asynq_9scheduler_TaskScheduler), 1); if (unlikely(!__pyx_ptype_5asynq_9scheduler_TaskScheduler)) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_vtabptr_5asynq_9scheduler_TaskScheduler = (struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler*)__Pyx_GetVtable(__pyx_ptype_5asynq_9scheduler_TaskScheduler->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_9scheduler_TaskScheduler)) __PYX_ERR(9, 29, __pyx_L1_error)
  __pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator = __Pyx_ImportType("asynq.scheduler", "TaskSchedulerActivator", sizeof(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator), 1); if (unlikely(!__pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator)) __PYX_ERR(9, 88, __pyx_L1_error)
  __pyx_ptype_5asynq_10async_task_AsyncTask = __Pyx_ImportType("asynq.async_task", "AsyncTask", sizeof(struct __pyx_obj_5asynq_10async_task_AsyncTask), 1); if (unlikely(!__pyx_ptype_5asynq_10async_task_AsyncTask)) __PYX_ERR(10, 31, __pyx_L1_error)
  __pyx_vtabptr_5asynq_10async_task_AsyncTask = (struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask*)__Pyx_GetVtable(__pyx_ptype_5asynq_10async_task_AsyncTask->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_10async_task_AsyncTask)) __PYX_ERR(10, 31, __pyx_L1_error)
  /*--- Variable import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("qcore.events"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "sinking_event_hook", (void **)&__pyx_vp_5qcore_6events_sinking_event_hook, "struct __pyx_obj_5qcore_6events_SinkingEventHook *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("qcore.helpers"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_list", (void **)&__pyx_vp_5qcore_7helpers_empty_list, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_tuple", (void **)&__pyx_vp_5qcore_7helpers_empty_tuple, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_dict", (void **)&__pyx_vp_5qcore_7helpers_empty_dict, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "none", (void **)&__pyx_vp_5qcore_7helpers_none, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "miss", (void **)&__pyx_vp_5qcore_7helpers_miss, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "same", (void **)&__pyx_vp_5qcore_7helpers_same, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "unspecified", (void **)&__pyx_vp_5qcore_7helpers_unspecified, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_context", (void **)&__pyx_vp_5qcore_7helpers_empty_context, "struct __pyx_obj_5qcore_7helpers_EmptyContext *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "CachedHashWrapper", (void **)&__pyx_vp_5qcore_7helpers_CachedHashWrapper, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "override", (void **)&__pyx_vp_5qcore_7helpers_override, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_ImportModule("asynq._debug"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_3, "options", (void **)&__pyx_vp_5asynq_6_debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_ImportModule("asynq.futures"); if (!__pyx_t_4) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "_debug_options", (void **)&__pyx_vp_5asynq_7futures__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "_none", (void **)&__pyx_vp_5asynq_7futures__none, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_ImportModule("asynq.batching"); if (!__pyx_t_5) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_debug_options", (void **)&__pyx_vp_5asynq_8batching__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_debug_batch_state", (void **)&__pyx_vp_5asynq_8batching__debug_batch_state, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_ImportModule("asynq.scheduler"); if (!__pyx_t_6) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_debug_options", (void **)&__pyx_vp_5asynq_9scheduler__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_futures_none", (void **)&__pyx_vp_5asynq_9scheduler__futures_none, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_state", (void **)&__pyx_vp_5asynq_9scheduler__state, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_7 = __Pyx_ImportModule("asynq.async_task"); if (!__pyx_t_7) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_7, "_debug_options", (void **)&__pyx_vp_5asynq_10async_task__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_7, "_futures_none", (void **)&__pyx_vp_5asynq_10async_task__futures_none, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_7, "_none_future", (void **)&__pyx_vp_5asynq_10async_task__none_future, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_7, "_empty_tuple", (void **)&__pyx_vp_5asynq_10async_task__empty_tuple, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_7, "_empty_dictionary", (void **)&__pyx_vp_5asynq_10async_task__empty_dictionary, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  /*--- Function import code ---*/
  __pyx_t_8 = __Pyx_ImportModule("asynq.scheduler"); if (!__pyx_t_8) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_8, "get_scheduler", (void (**)(void))&__pyx_f_5asynq_9scheduler_get_scheduler, "struct __pyx_obj_5asynq_9scheduler_TaskScheduler *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_9 = __Pyx_ImportModule("asynq.async_task"); if (!__pyx_t_9) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_9, "unwrap", (void (**)(void))&__pyx_f_5asynq_10async_task_unwrap, "PyObject *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "asynq/utils.py":15
 * # limitations under the License.
 * 
 * import six             # <<<<<<<<<<<<<<
 * 
 * from .futures import FutureBase
 */
  __pyx_t_10 = __Pyx_Import(__pyx_n_s_six, 0, -1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_six, __pyx_t_10) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/utils.py":17
 * import six
 * 
 * from .futures import FutureBase             # <<<<<<<<<<<<<<
 * from . import async_task
 * from . import scheduler
 */
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_FutureBase);
  __Pyx_GIVEREF(__pyx_n_s_FutureBase);
  PyList_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_FutureBase);
  __pyx_t_11 = __Pyx_Import(__pyx_n_s_futures, __pyx_t_10, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "asynq/utils.py":18
 * 
 * from .futures import FutureBase
 * from . import async_task             # <<<<<<<<<<<<<<
 * from . import scheduler
 * from . import _debug
 */
  __pyx_t_11 = PyList_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_async_task);
  __Pyx_GIVEREF(__pyx_n_s_async_task);
  PyList_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_async_task);
  __pyx_t_10 = __Pyx_Import(__pyx_n_s__3, __pyx_t_11, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_ImportFrom(__pyx_t_10, __pyx_n_s_async_task); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_async_task, __pyx_t_11) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/utils.py":19
 * from .futures import FutureBase
 * from . import async_task
 * from . import scheduler             # <<<<<<<<<<<<<<
 * from . import _debug
 * 
 */
  __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_INCREF(__pyx_n_s_scheduler);
  __Pyx_GIVEREF(__pyx_n_s_scheduler);
  PyList_SET_ITEM(__pyx_t_10, 0, __pyx_n_s_scheduler);
  __pyx_t_11 = __Pyx_Import(__pyx_n_s__3, __pyx_t_10, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_ImportFrom(__pyx_t_11, __pyx_n_s_scheduler); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_scheduler, __pyx_t_10) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "asynq/utils.py":20
 * from . import async_task
 * from . import scheduler
 * from . import _debug             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_11 = PyList_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_INCREF(__pyx_n_s_debug);
  __Pyx_GIVEREF(__pyx_n_s_debug);
  PyList_SET_ITEM(__pyx_t_11, 0, __pyx_n_s_debug);
  __pyx_t_10 = __Pyx_Import(__pyx_n_s__3, __pyx_t_11, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_ImportFrom(__pyx_t_10, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_debug, __pyx_t_11) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/utils.py":23
 * 
 * 
 * _debug_options = _debug.options             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_5asynq_6_debug_options));
  __Pyx_XGOTREF(((PyObject *)__pyx_v_5asynq_5utils__debug_options));
  __Pyx_DECREF_SET(__pyx_v_5asynq_5utils__debug_options, __pyx_v_5asynq_6_debug_options);
  __Pyx_GIVEREF(((PyObject *)__pyx_v_5asynq_6_debug_options));

  /* "asynq/utils.py":26
 * 
 * 
 * def execute_on_separate_scheduler(async_pull_fn, contexts, args=(), kwargs={}):             # <<<<<<<<<<<<<<
 *     current_scheduler = scheduler.get_scheduler()
 *     new_scheduler = current_scheduler._clone()
 */
  __pyx_t_10 = PyDict_New(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_k_ = ((PyObject*)__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  __pyx_t_10 = 0;
  __pyx_t_10 = PyDict_New(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_k_ = ((PyObject*)__pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_10);
  __pyx_t_10 = 0;

  /* "asynq/utils.py":44
 * 
 * 
 * def await(*args):             # <<<<<<<<<<<<<<
 *     """Runs all the tasks specified in args,
 *     and finally returns args unwrapped.
 */
  __pyx_t_10 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5asynq_5utils_3await, 0, __pyx_n_s_await, NULL, __pyx_n_s_asynq_utils, __pyx_d, ((PyObject *)__pyx_codeobj__5)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_await, __pyx_t_10) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "asynq/utils.py":1
 * # Copyright 2016 Quora, Inc.             # <<<<<<<<<<<<<<
 * #
 * # Licensed under the Apache License, Version 2.0 (the "License");
 */
  __pyx_t_10 = PyDict_New(); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_10) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init asynq.utils", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init asynq.utils");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* SaveResetException */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* SwapException */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* None */
      static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* RaiseArgTupleInvalid */
      static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
      static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
      static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* ArgTypeTest */
      static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* KeywordStringCheck */
      static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetItemInt */
      static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* GetModuleGlobalName */
      static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* GetVTable */
          static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* FetchCommonType */
          static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
          static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CodeObjectCache */
              static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
              #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
              #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
              static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
              static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CheckBinaryVersion */
              static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrExport */
              static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
              static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
              #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
              #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* VoidPtrImport */
              #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
              #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
              static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
