/* Generated by Cython 0.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {}
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__asynq__scheduler
#define __PYX_HAVE_API__asynq__scheduler
#include "string.h"
#include "stdio.h"
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "asynq/scheduler.py",
  "asynq/scheduler.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "events.pxd",
  "helpers.pxd",
  "asynq/_debug.pxd",
  "asynq/futures.pxd",
  "asynq/async_task.pxd",
  "asynq/batching.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5qcore_6events_EventHook;
struct __pyx_obj_5qcore_6events_SinkingEventHook;
struct __pyx_obj_5qcore_6events_EventInterceptor;
struct __pyx_obj_5qcore_7helpers_MarkerObject;
struct __pyx_obj_5qcore_7helpers_EmptyContext;
struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper;
struct __pyx_obj_5qcore_7helpers_ScopedValue;
struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext;
struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext;
struct __pyx_obj_5asynq_6_debug_DebugOptions;
struct __pyx_obj_5asynq_7futures_FutureBase;
struct __pyx_obj_5asynq_7futures_Future;
struct __pyx_obj_5asynq_7futures_ConstFuture;
struct __pyx_obj_5asynq_7futures_ErrorFuture;
struct __pyx_obj_5asynq_10async_task_AsyncTask;
struct __pyx_obj_5asynq_8batching_BatchBase;
struct __pyx_obj_5asynq_8batching_BatchItemBase;
struct __pyx_obj_5asynq_8batching_DebugBatchItem;
struct __pyx_obj_5asynq_8batching_DebugBatch;
struct __pyx_obj_5asynq_9scheduler_TaskScheduler;
struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator;
struct __pyx_opt_args_5qcore_7helpers_safe_str;
struct __pyx_opt_args_5qcore_7helpers_safe_repr;

/* "qcore/helpers.pxd":75
 * 
 * cpdef object ellipsis(object source, int max_length)
 * cpdef object safe_str(object source, int max_length=?)             # <<<<<<<<<<<<<<
 * cpdef object safe_repr(object source, int max_length=?)
 * 
 */
struct __pyx_opt_args_5qcore_7helpers_safe_str {
  int __pyx_n;
  int max_length;
};

/* "qcore/helpers.pxd":76
 * cpdef object ellipsis(object source, int max_length)
 * cpdef object safe_str(object source, int max_length=?)
 * cpdef object safe_repr(object source, int max_length=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5qcore_7helpers_safe_repr {
  int __pyx_n;
  int max_length;
};
struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL;

/* "_debug.pxd":44
 *     cdef public object ENABLE_COMPLEX_ASSERTIONS
 * 
 *     cpdef DUMP_ALL(self, value=?)             # <<<<<<<<<<<<<<
 * 
 * cdef public DebugOptions options
 */
struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL {
  int __pyx_n;
  PyObject *value;
};
struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump;

/* "futures.pxd":46
 *     cdef inline raise_if_error(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump {
  int __pyx_n;
  int indent;
};
struct __pyx_opt_args_5qcore_10inspection_wraps;

/* "qcore/inspection.pxd":38
 * cpdef bint is_classmethod(object fn) except -1
 * 
 * cpdef wraps(object wrapped, object assigned=?, object updated=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5qcore_10inspection_wraps {
  int __pyx_n;
  PyObject *assigned;
  PyObject *updated;
};
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start;
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump;

/* "async_task.pxd":53
 * 
 *     @cython.locals(s=scheduler.TaskScheduler)
 *     cpdef AsyncTask start(self, bint run_immediately=?)             # <<<<<<<<<<<<<<
 *     @cython.locals(s=scheduler.TaskScheduler)
 *     cpdef AsyncTask after(self, futures.FutureBase future)
 */
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start {
  int __pyx_n;
  int run_immediately;
};

/* "async_task.pxd":84
 *     cpdef list traceback(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 *     # Contexts support
 */
struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_dump {
  int __pyx_n;
  int indent;
};
struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel;
struct __pyx_opt_args_5asynq_8batching_9BatchBase_dump;
struct __pyx_opt_args_5asynq_8batching_sync;

/* "batching.pxd":35
 * 
 *     cpdef flush(self)
 *     cpdef cancel(self, object error=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _compute(self)
 */
struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel {
  int __pyx_n;
  PyObject *error;
};

/* "batching.pxd":43
 *     cpdef _try_switch_active_batch(self)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class BatchItemBase(futures.FutureBase):
 */
struct __pyx_opt_args_5asynq_8batching_9BatchBase_dump {
  int __pyx_n;
  int indent;
};

/* "batching.pxd":66
 * cdef object _debug_batch_state
 * 
 * cpdef sync(tag=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5asynq_8batching_sync {
  int __pyx_n;
  PyObject *tag;
};
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump;
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump;

/* "asynq/scheduler.pxd":45
 *     cpdef reset(self)
 * 
 *     cpdef async_task.AsyncTask start(self, async_task.AsyncTask task,             # <<<<<<<<<<<<<<
 *             bint run_immediately=?)
 *     cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start {
  int __pyx_n;
  int run_immediately;
};

/* "asynq/scheduler.pxd":48
 *             bint run_immediately=?)
 *     cpdef async_task.AsyncTask continue_with(self, futures.FutureBase future, async_task.AsyncTask task)
 *     cpdef int flush_batches(self, bint must_await=?) except -1             # <<<<<<<<<<<<<<
 * 
 *     cpdef add_dependencies(self, async_task.AsyncTask task, object yield_result)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches {
  int __pyx_n;
  int must_await;
};

/* "asynq/scheduler.pxd":76
 *     @cython.locals(dump_continue_task=bint, task=async_task.AsyncTask)
 *     cdef async_task.AsyncTask _continue_with_task(self)
 *     cdef batching.BatchBase _continue_with_batch(self, bint assert_no_batch=?)             # <<<<<<<<<<<<<<
 * 
 *     cpdef batching.BatchBase _select_batch_to_flush(self)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch {
  int __pyx_n;
  int assert_no_batch;
};

/* "asynq/scheduler.pxd":84
 *     cpdef _accept_changes(self, TaskScheduler clone)
 * 
 *     cpdef dump(self, int indent=?)             # <<<<<<<<<<<<<<
 *     @cython.locals(current_time=float)
 *     cpdef try_time_based_dump(self, object last_task=?)
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump {
  int __pyx_n;
  int indent;
};

/* "asynq/scheduler.pxd":86
 *     cpdef dump(self, int indent=?)
 *     @cython.locals(current_time=float)
 *     cpdef try_time_based_dump(self, object last_task=?)             # <<<<<<<<<<<<<<
 * 
 * cdef class TaskSchedulerActivator(object):
 */
struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump {
  int __pyx_n;
  PyObject *last_task;
};

/* "qcore/events.pxd":15
 * # limitations under the License.
 * 
 * cdef class EventHook(object):             # <<<<<<<<<<<<<<
 *     cdef list handlers
 * 
 */
struct __pyx_obj_5qcore_6events_EventHook {
  PyObject_HEAD
  PyObject *handlers;
};


/* "qcore/events.pxd":19
 * 
 * 
 * cdef class SinkingEventHook(EventHook):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5qcore_6events_SinkingEventHook {
  struct __pyx_obj_5qcore_6events_EventHook __pyx_base;
};


/* "qcore/events.pxd":25
 * 
 * 
 * cdef class EventInterceptor(object):             # <<<<<<<<<<<<<<
 *     cdef object source
 *     cdef dict events
 */
struct __pyx_obj_5qcore_6events_EventInterceptor {
  PyObject_HEAD
  PyObject *source;
  PyObject *events;
};


/* "qcore/helpers.pxd":24
 * 
 * 
 * cdef class MarkerObject(object):             # <<<<<<<<<<<<<<
 *     cpdef unicode name
 * 
 */
struct __pyx_obj_5qcore_7helpers_MarkerObject {
  PyObject_HEAD
  PyObject *name;
};


/* "qcore/helpers.pxd":33
 * 
 * 
 * cdef class EmptyContext(object):             # <<<<<<<<<<<<<<
 *     cpdef __enter__(self)
 *     cpdef __exit__(self, exc_type, exc_val, exc_tb)
 */
struct __pyx_obj_5qcore_7helpers_EmptyContext {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext *__pyx_vtab;
};


/* "qcore/helpers.pxd":40
 * 
 * 
 * cdef class CythonCachedHashWrapper(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 *     cdef int _hash
 */
struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper *__pyx_vtab;
  PyObject *_value;
  int _hash;
};


/* "qcore/helpers.pxd":52
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration
 * 
 * cdef class ScopedValue(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 * 
 */
struct __pyx_obj_5qcore_7helpers_ScopedValue {
  PyObject_HEAD
  struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue *__pyx_vtab;
  PyObject *_value;
};


/* "qcore/helpers.pxd":50
 * 
 * 
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration             # <<<<<<<<<<<<<<
 * 
 * cdef class ScopedValue(object):
 */
struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext {
  PyObject_HEAD
  struct __pyx_obj_5qcore_7helpers_ScopedValue *_target;
  PyObject *_value;
  PyObject *_old_value;
};


/* "qcore/helpers.pxd":65
 * 
 * 
 * cdef class _PropertyOverrideContext(object):             # <<<<<<<<<<<<<<
 *     cdef object _target
 *     cdef object _property_name
 */
struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext {
  PyObject_HEAD
  PyObject *_target;
  PyObject *_property_name;
  PyObject *_value;
  PyObject *_old_value;
};


/* "_debug.pxd":18
 * 
 * 
 * cdef class DebugOptions(object):             # <<<<<<<<<<<<<<
 *     cdef public bint DUMP_PRE_ERROR_STATE
 *     cdef public bint DUMP_EXCEPTIONS
 */
struct __pyx_obj_5asynq_6_debug_DebugOptions {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions *__pyx_vtab;
  int DUMP_PRE_ERROR_STATE;
  int DUMP_EXCEPTIONS;
  int DUMP_AWAIT_RECURSION;
  int DUMP_SCHEDULE_TASK;
  int DUMP_CONTINUE_TASK;
  int DUMP_SCHEDULE_BATCH;
  int DUMP_FLUSH_BATCH;
  int DUMP_DEPENDENCIES;
  int DUMP_COMPUTED;
  int DUMP_NEW_TASKS;
  int DUMP_YIELD_RESULTS;
  int DUMP_QUEUED_RESULTS;
  int DUMP_CONTEXTS;
  int DUMP_SCHEDULER_CHANGE;
  int DUMP_SYNC;
  int DUMP_PRIMER;
  int DUMP_STACK;
  int DUMP_SCHEDULER_STATE;
  float SCHEDULER_STATE_DUMP_INTERVAL;
  int DEBUG_STR_REPR_MAX_LENGTH;
  PyObject *STACK_DUMP_LIMIT;
  PyObject *ENABLE_COMPLEX_ASSERTIONS;
};


/* "futures.pxd":26
 * 
 * 
 * cdef class FutureBase(object):             # <<<<<<<<<<<<<<
 *     cdef public object _value
 *     cdef public object _error
 */
struct __pyx_obj_5asynq_7futures_FutureBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase *__pyx_vtab;
  PyObject *_value;
  PyObject *_error;
  int _in_repr;
  struct __pyx_obj_5qcore_6events_EventHook *on_computed;
};


/* "futures.pxd":49
 * 
 * 
 * cdef class Future(FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public object _value_provider
 *     cpdef _compute(self)
 */
struct __pyx_obj_5asynq_7futures_Future {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  PyObject *_value_provider;
};


/* "futures.pxd":54
 * 
 * 
 * cdef class ConstFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_5asynq_7futures_ConstFuture {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
};


/* "futures.pxd":57
 *     pass
 * 
 * cdef class ErrorFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 */
struct __pyx_obj_5asynq_7futures_ErrorFuture {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
};


/* "async_task.pxd":31
 * 
 * 
 * cdef class AsyncTask(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index
 */
struct __pyx_obj_5asynq_10async_task_AsyncTask {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  int is_scheduled;
  long iteration_index;
  int group_cancel;
  PyObject *fn;
  PyObject *args;
  PyObject *kwargs;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *caller;
  int depth;
  int max_depth;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *scheduler;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *creator;
  PyObject *_generator;
  PyObject *_last_value;
  PyObject *_frame_info;
  PyObject *_dependencies;
  PyObject *_contexts;
};


/* "batching.pxd":26
 * 
 * 
 * cdef class BatchBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public list items
 * 
 */
struct __pyx_obj_5asynq_8batching_BatchBase {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  PyObject *items;
};


/* "batching.pxd":45
 *     cpdef dump(self, int indent=?)
 * 
 * cdef class BatchItemBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public BatchBase batch
 *     cdef public long index
 */
struct __pyx_obj_5asynq_8batching_BatchItemBase {
  struct __pyx_obj_5asynq_7futures_FutureBase __pyx_base;
  struct __pyx_obj_5asynq_8batching_BatchBase *batch;
  long index;
};


/* "batching.pxd":53
 * 
 * 
 * cdef class DebugBatchItem(BatchItemBase):             # <<<<<<<<<<<<<<
 *     cdef public object _result
 * 
 */
struct __pyx_obj_5asynq_8batching_DebugBatchItem {
  struct __pyx_obj_5asynq_8batching_BatchItemBase __pyx_base;
  PyObject *_result;
};


/* "batching.pxd":56
 *     cdef public object _result
 * 
 * cdef class DebugBatch(BatchBase):             # <<<<<<<<<<<<<<
 *     cdef public str name
 *     cdef public long index
 */
struct __pyx_obj_5asynq_8batching_DebugBatch {
  struct __pyx_obj_5asynq_8batching_BatchBase __pyx_base;
  PyObject *name;
  long index;
};


/* "asynq/scheduler.pxd":29
 * 
 * 
 * cdef class TaskScheduler(object):             # <<<<<<<<<<<<<<
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time
 */
struct __pyx_obj_5asynq_9scheduler_TaskScheduler {
  PyObject_HEAD
  struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *__pyx_vtab;
  PyObject *_empty_set;
  float _last_dump_time;
  PyObject *_new_tasks;
  PyObject *_tasks;
  PyObject *_batches;
  int await_recursion_depth;
  int max_await_recursion_depth;
  struct __pyx_obj_5qcore_6events_EventHook *on_before_batch_flush;
  struct __pyx_obj_5qcore_6events_EventHook *on_after_batch_flush;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *active_task;
  PyObject *name;
};


/* "asynq/scheduler.pxd":88
 *     cpdef try_time_based_dump(self, object last_task=?)
 * 
 * cdef class TaskSchedulerActivator(object):             # <<<<<<<<<<<<<<
 *     cdef TaskScheduler _scheduler
 *     cdef TaskScheduler _old_scheduler
 */
struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator {
  PyObject_HEAD
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *_scheduler;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *_old_scheduler;
  int _must_await_on_exit;
  int _must_flush_batches_on_exit;
};



/* "qcore/helpers.pxd":33
 * 
 * 
 * cdef class EmptyContext(object):             # <<<<<<<<<<<<<<
 *     cpdef __enter__(self)
 *     cpdef __exit__(self, exc_type, exc_val, exc_tb)
 */

struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext {
  PyObject *(*__pyx___enter__)(struct __pyx_obj_5qcore_7helpers_EmptyContext *, int __pyx_skip_dispatch);
  PyObject *(*__pyx___exit__)(struct __pyx_obj_5qcore_7helpers_EmptyContext *, PyObject *, PyObject *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext *__pyx_vtabptr_5qcore_7helpers_EmptyContext;


/* "qcore/helpers.pxd":40
 * 
 * 
 * cdef class CythonCachedHashWrapper(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 *     cdef int _hash
 */

struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper {
  PyObject *(*value)(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper *, int __pyx_skip_dispatch);
  PyObject *(*hash)(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper *__pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper;


/* "qcore/helpers.pxd":52
 * cdef class _ScopedValueOverrideContext(object)  # Forward declaration
 * 
 * cdef class ScopedValue(object):             # <<<<<<<<<<<<<<
 *     cdef object _value
 * 
 */

struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue {
  PyObject *(*get)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, int __pyx_skip_dispatch);
  PyObject *(*set)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*override)(struct __pyx_obj_5qcore_7helpers_ScopedValue *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue *__pyx_vtabptr_5qcore_7helpers_ScopedValue;


/* "_debug.pxd":18
 * 
 * 
 * cdef class DebugOptions(object):             # <<<<<<<<<<<<<<
 *     cdef public bint DUMP_PRE_ERROR_STATE
 *     cdef public bint DUMP_EXCEPTIONS
 */

struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions {
  PyObject *(*DUMP_ALL)(struct __pyx_obj_5asynq_6_debug_DebugOptions *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_6_debug_12DebugOptions_DUMP_ALL *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions *__pyx_vtabptr_5asynq_6_debug_DebugOptions;


/* "futures.pxd":26
 * 
 * 
 * cdef class FutureBase(object):             # <<<<<<<<<<<<<<
 *     cdef public object _value
 *     cdef public object _error
 */

struct __pyx_vtabstruct_5asynq_7futures_FutureBase {
  PyObject *(*value)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*set_value)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*reset_unsafe)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*error)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*set_error)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, int __pyx_skip_dispatch);
  int (*is_computed)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*_compute)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*_computed)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*make_dependency)(struct __pyx_obj_5asynq_7futures_FutureBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*raise_if_error)(struct __pyx_obj_5asynq_7futures_FutureBase *);
  PyObject *(*dump)(struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_7futures_FutureBase *__pyx_vtabptr_5asynq_7futures_FutureBase;


/* "futures.pxd":49
 * 
 * 
 * cdef class Future(FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public object _value_provider
 *     cpdef _compute(self)
 */

struct __pyx_vtabstruct_5asynq_7futures_Future {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_Future *__pyx_vtabptr_5asynq_7futures_Future;


/* "futures.pxd":54
 * 
 * 
 * cdef class ConstFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_5asynq_7futures_ConstFuture {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_ConstFuture *__pyx_vtabptr_5asynq_7futures_ConstFuture;


/* "futures.pxd":57
 *     pass
 * 
 * cdef class ErrorFuture(FutureBase):             # <<<<<<<<<<<<<<
 *     pass
 */

struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture *__pyx_vtabptr_5asynq_7futures_ErrorFuture;


/* "async_task.pxd":31
 * 
 * 
 * cdef class AsyncTask(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public bint is_scheduled
 *     cdef public long iteration_index
 */

struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
  int (*is_blocked)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  int (*can_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*start)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_10async_task_9AsyncTask_start *__pyx_optional_args);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*after)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*cancel_dependencies)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*_continue_on_generator)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, PyObject *);
  PyObject *(*_accept_yield_result)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_accept_error)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_queue_exit)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_queue_throw_error)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_remove_dependency_cython)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_before_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*_after_continue)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*traceback)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*_enter_context)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_leave_context)(struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *);
  PyObject *(*_pause_contexts)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  PyObject *(*_resume_contexts)(struct __pyx_obj_5asynq_10async_task_AsyncTask *);
};
static struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *__pyx_vtabptr_5asynq_10async_task_AsyncTask;


/* "batching.pxd":26
 * 
 * 
 * cdef class BatchBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public list items
 * 
 */

struct __pyx_vtabstruct_5asynq_8batching_BatchBase {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
  int (*is_flushed)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*is_cancelled)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*is_empty)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*get_priority)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*flush)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*cancel)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_8batching_9BatchBase_cancel *__pyx_optional_args);
  PyObject *(*_flush)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*_cancel)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  PyObject *(*_try_switch_active_batch)(struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5asynq_8batching_BatchBase *__pyx_vtabptr_5asynq_8batching_BatchBase;


/* "batching.pxd":45
 *     cpdef dump(self, int indent=?)
 * 
 * cdef class BatchItemBase(futures.FutureBase):             # <<<<<<<<<<<<<<
 *     cdef public BatchBase batch
 *     cdef public long index
 */

struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase {
  struct __pyx_vtabstruct_5asynq_7futures_FutureBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase *__pyx_vtabptr_5asynq_8batching_BatchItemBase;


/* "batching.pxd":53
 * 
 * 
 * cdef class DebugBatchItem(BatchItemBase):             # <<<<<<<<<<<<<<
 *     cdef public object _result
 * 
 */

struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem {
  struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem *__pyx_vtabptr_5asynq_8batching_DebugBatchItem;


/* "batching.pxd":56
 *     cdef public object _result
 * 
 * cdef class DebugBatch(BatchBase):             # <<<<<<<<<<<<<<
 *     cdef public str name
 *     cdef public long index
 */

struct __pyx_vtabstruct_5asynq_8batching_DebugBatch {
  struct __pyx_vtabstruct_5asynq_8batching_BatchBase __pyx_base;
};
static struct __pyx_vtabstruct_5asynq_8batching_DebugBatch *__pyx_vtabptr_5asynq_8batching_DebugBatch;


/* "asynq/scheduler.py":44
 * 
 * 
 * class TaskScheduler(object):             # <<<<<<<<<<<<<<
 *     """Schedules and runs AsyncTask objects flushing batches when needed."""
 * 
 */

struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler {
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *(*activate)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, int, int __pyx_skip_dispatch);
  PyObject *(*reset)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*start)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start *__pyx_optional_args);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*continue_with)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_7futures_FutureBase *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  int (*flush_batches)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches *__pyx_optional_args);
  PyObject *(*add_dependencies)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_dependency)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch);
  PyObject *(*schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch);
  PyObject *(*schedule_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch);
  int (*_await)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch);
  int (*_execute)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_try_execute_on_new_scheduler)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch);
  int (*_schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  int (*_schedule_without_checks)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *);
  int (*_try_schedule)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int);
  int (*_schedule_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *);
  int (*_flush_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *);
  PyObject *(*_enqueue_new_tasks)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *(*_continue_with_task)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);
  struct __pyx_obj_5asynq_8batching_BatchBase *(*_continue_with_batch)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch *__pyx_optional_args);
  struct __pyx_obj_5asynq_8batching_BatchBase *(*_select_batch_to_flush)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*_clone)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*_accept_changes)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch);
  PyObject *(*dump)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump *__pyx_optional_args);
  PyObject *(*try_time_based_dump)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *__pyx_vtabptr_5asynq_9scheduler_TaskScheduler;
static CYTHON_INLINE PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyIntBinop.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyThreadStateGet.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* SaveResetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* ListAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SliceObject.proto */
#define __Pyx_PyObject_DelSlice(obj, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)\
    __Pyx_PyObject_SetSlice(obj, (PyObject*)NULL, cstart, cstop, py_start, py_stop, py_slice, has_cstart, has_cstop, wraparound)
static CYTHON_INLINE int __Pyx_PyObject_SetSlice(
        PyObject* obj, PyObject* value, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* ListCompAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_f_5asynq_9scheduler_13TaskScheduler_activate(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_must_await_on_exit, int __pyx_v_must_flush_batches_on_exit, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_reset(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_13TaskScheduler_start(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_13TaskScheduler_continue_with(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler_flush_batches(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependencies(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, PyObject *__pyx_v_yield_result, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependency(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_dependency, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__await(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__execute(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_is_partial_execution, PyObject *__pyx_v_tasks, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__try_execute_on_new_scheduler(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__schedule_without_checks(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__try_schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_v_enqueue_new_tasks); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_schedule_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__schedule_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch); /* proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__flush_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch); /* proto*/
static CYTHON_INLINE PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_13TaskScheduler__continue_with_task(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto*/
static struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_f_5asynq_9scheduler_13TaskScheduler__continue_with_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch *__pyx_optional_args); /* proto*/
static struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_f_5asynq_9scheduler_13TaskScheduler__select_batch_to_flush(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__clone(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__accept_changes(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_clone, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_try_time_based_dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump *__pyx_optional_args); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'qcore.events' */
static PyTypeObject *__pyx_ptype_5qcore_6events_EventHook = 0;
static PyTypeObject *__pyx_ptype_5qcore_6events_SinkingEventHook = 0;
static PyTypeObject *__pyx_ptype_5qcore_6events_EventInterceptor = 0;
static struct __pyx_obj_5qcore_6events_SinkingEventHook **__pyx_vp_5qcore_6events_sinking_event_hook = 0;
#define __pyx_v_5qcore_6events_sinking_event_hook (*__pyx_vp_5qcore_6events_sinking_event_hook)

/* Module declarations from 'qcore.helpers' */
static PyTypeObject *__pyx_ptype_5qcore_7helpers_MarkerObject = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_EmptyContext = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers_ScopedValue = 0;
static PyTypeObject *__pyx_ptype_5qcore_7helpers__PropertyOverrideContext = 0;
static PyObject **__pyx_vp_5qcore_7helpers_empty_list = 0;
#define __pyx_v_5qcore_7helpers_empty_list (*__pyx_vp_5qcore_7helpers_empty_list)
static PyObject **__pyx_vp_5qcore_7helpers_empty_tuple = 0;
#define __pyx_v_5qcore_7helpers_empty_tuple (*__pyx_vp_5qcore_7helpers_empty_tuple)
static PyObject **__pyx_vp_5qcore_7helpers_empty_dict = 0;
#define __pyx_v_5qcore_7helpers_empty_dict (*__pyx_vp_5qcore_7helpers_empty_dict)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_none = 0;
#define __pyx_v_5qcore_7helpers_none (*__pyx_vp_5qcore_7helpers_none)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_miss = 0;
#define __pyx_v_5qcore_7helpers_miss (*__pyx_vp_5qcore_7helpers_miss)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_same = 0;
#define __pyx_v_5qcore_7helpers_same (*__pyx_vp_5qcore_7helpers_same)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5qcore_7helpers_unspecified = 0;
#define __pyx_v_5qcore_7helpers_unspecified (*__pyx_vp_5qcore_7helpers_unspecified)
static struct __pyx_obj_5qcore_7helpers_EmptyContext **__pyx_vp_5qcore_7helpers_empty_context = 0;
#define __pyx_v_5qcore_7helpers_empty_context (*__pyx_vp_5qcore_7helpers_empty_context)
static PyObject **__pyx_vp_5qcore_7helpers_CachedHashWrapper = 0;
#define __pyx_v_5qcore_7helpers_CachedHashWrapper (*__pyx_vp_5qcore_7helpers_CachedHashWrapper)
static PyObject **__pyx_vp_5qcore_7helpers_override = 0;
#define __pyx_v_5qcore_7helpers_override (*__pyx_vp_5qcore_7helpers_override)

/* Module declarations from 'asynq._debug' */
static PyTypeObject *__pyx_ptype_5asynq_6_debug_DebugOptions = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_6_debug_options = 0;
#define __pyx_v_5asynq_6_debug_options (*__pyx_vp_5asynq_6_debug_options)

/* Module declarations from 'asynq.futures' */
static PyTypeObject *__pyx_ptype_5asynq_7futures_FutureBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_Future = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_ConstFuture = 0;
static PyTypeObject *__pyx_ptype_5asynq_7futures_ErrorFuture = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_7futures__debug_options = 0;
#define __pyx_v_5asynq_7futures__debug_options (*__pyx_vp_5asynq_7futures__debug_options)
static struct __pyx_obj_5qcore_7helpers_MarkerObject **__pyx_vp_5asynq_7futures__none = 0;
#define __pyx_v_5asynq_7futures__none (*__pyx_vp_5asynq_7futures__none)

/* Module declarations from 'qcore.inspection' */

/* Module declarations from 'asynq.async_task' */
static PyTypeObject *__pyx_ptype_5asynq_10async_task_AsyncTask = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_10async_task__debug_options = 0;
#define __pyx_v_5asynq_10async_task__debug_options (*__pyx_vp_5asynq_10async_task__debug_options)
static PyObject **__pyx_vp_5asynq_10async_task__futures_none = 0;
#define __pyx_v_5asynq_10async_task__futures_none (*__pyx_vp_5asynq_10async_task__futures_none)
static PyObject **__pyx_vp_5asynq_10async_task__none_future = 0;
#define __pyx_v_5asynq_10async_task__none_future (*__pyx_vp_5asynq_10async_task__none_future)
static PyObject **__pyx_vp_5asynq_10async_task__empty_tuple = 0;
#define __pyx_v_5asynq_10async_task__empty_tuple (*__pyx_vp_5asynq_10async_task__empty_tuple)
static PyObject **__pyx_vp_5asynq_10async_task__empty_dictionary = 0;
#define __pyx_v_5asynq_10async_task__empty_dictionary (*__pyx_vp_5asynq_10async_task__empty_dictionary)

/* Module declarations from 'asynq.batching' */
static PyTypeObject *__pyx_ptype_5asynq_8batching_BatchBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_BatchItemBase = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_DebugBatchItem = 0;
static PyTypeObject *__pyx_ptype_5asynq_8batching_DebugBatch = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions **__pyx_vp_5asynq_8batching__debug_options = 0;
#define __pyx_v_5asynq_8batching__debug_options (*__pyx_vp_5asynq_8batching__debug_options)
static PyObject **__pyx_vp_5asynq_8batching__debug_batch_state = 0;
#define __pyx_v_5asynq_8batching__debug_batch_state (*__pyx_vp_5asynq_8batching__debug_batch_state)

/* Module declarations from 'asynq.scheduler' */
static PyTypeObject *__pyx_ptype_5asynq_9scheduler_TaskScheduler = 0;
static PyTypeObject *__pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator = 0;
static struct __pyx_obj_5asynq_6_debug_DebugOptions *__pyx_v_5asynq_9scheduler__debug_options = 0;
static PyObject *__pyx_v_5asynq_9scheduler__futures_none = 0;
static PyObject *__pyx_v_5asynq_9scheduler__state = 0;
static PyObject *__pyx_f_5asynq_9scheduler_extract_futures(PyObject *, PyObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_f_5asynq_9scheduler_get_scheduler(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_set_scheduler(PyObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_get_active_task(int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_5asynq_9scheduler__no_blocked_tasks(PyObject *, int __pyx_skip_dispatch); /*proto*/
static struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_f_5asynq_9scheduler__create_blocking_dependency(int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "asynq.scheduler"
int __pyx_module_is_main_asynq__scheduler = 0;

/* Implementation of 'asynq.scheduler' */
static PyObject *__pyx_builtin_Exception;
static PyObject *__pyx_builtin_BaseException;
static PyObject *__pyx_builtin_AssertionError;
static const char __pyx_k__8[] = "----------------------------------------------------------------------\n";
static const char __pyx_k__20[] = "*";
static const char __pyx_k__21[] = "";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_s_d[] = "%s / %d";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_dump[] = "dump";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_task[] = "task";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_await[] = "await";
static const char __pyx_k_clone[] = "_clone";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_flush[] = "flush";
static const char __pyx_k_ident[] = "ident";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_local[] = "local";
static const char __pyx_k_reset[] = "reset";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "_state";
static const char __pyx_k_str_2[] = "__str__";
static const char __pyx_k_tasks[] = "tasks";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_write[] = "write";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_future[] = "future";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_indent[] = "indent";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_stderr[] = "stderr";
static const char __pyx_k_stdout[] = "stdout";
static const char __pyx_k_Batches[] = "Batches:";
static const char __pyx_k_await_2[] = "_await";
static const char __pyx_k_current[] = "current";
static const char __pyx_k_debug_2[] = "_debug";
static const char __pyx_k_execute[] = "_execute";
static const char __pyx_k_futures[] = "futures";
static const char __pyx_k_last_id[] = "last_id";
static const char __pyx_k_popleft[] = "popleft";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_activate[] = "activate";
static const char __pyx_k_batching[] = "batching";
static const char __pyx_k_endswith[] = "endswith";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_schedule[] = "schedule";
static const char __pyx_k_Exception[] = "Exception";
static const char __pyx_k_New_tasks[] = "New tasks:";
static const char __pyx_k_last_task[] = "last_task";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_scheduler[] = "scheduler";
static const char __pyx_k_threading[] = "threading";
static const char __pyx_k_traceback[] = "traceback";
static const char __pyx_k_Task_queue[] = "Task queue:";
static const char __pyx_k_dependency[] = "dependency";
static const char __pyx_k_is_flushed[] = "is_flushed";
static const char __pyx_k_must_await[] = "must_await";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_Last_task_s[] = "Last task: %s";
static const char __pyx_k_active_task[] = "active_task";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_core_events[] = "core_events";
static const char __pyx_k_is_computed[] = "is_computed";
static const char __pyx_k_futures_none[] = "_futures_none";
static const char __pyx_k_get_priority[] = "get_priority";
static const char __pyx_k_qcore_events[] = "qcore.events";
static const char __pyx_k_yield_result[] = "yield_result";
static const char __pyx_k_BaseException[] = "BaseException";
static const char __pyx_k_continue_with[] = "continue_with";
static const char __pyx_k_debug_options[] = "_debug_options";
static const char __pyx_k_flush_batches[] = "flush_batches";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_AsyncTaskError[] = "AsyncTaskError";
static const char __pyx_k_accept_changes[] = "_accept_changes";
static const char __pyx_k_add_dependency[] = "add_dependency";
static const char __pyx_k_current_thread[] = "current_thread";
static const char __pyx_k_schedule_batch[] = "schedule_batch";
static const char __pyx_k_asynq_scheduler[] = "asynq.scheduler";
static const char __pyx_k_run_immediately[] = "run_immediately";
static const char __pyx_k_add_dependencies[] = "add_dependencies";
static const char __pyx_k_async_skipping_s[] = "@async: skipping %s";
static const char __pyx_k_async_continued_s[] = "@async: <- continued %s";
static const char __pyx_k_async_scheduler_s[] = "@async: scheduler = %s";
static const char __pyx_k_remove_dependency[] = "_remove_dependency";
static const char __pyx_k_async_continuing_s[] = "@async: -> continuing %s";
static const char __pyx_k_must_await_on_exit[] = "must_await_on_exit";
static const char __pyx_k_try_time_based_dump[] = "try_time_based_dump";
static const char __pyx_k_Scheduler_state_dump[] = "\n--- Scheduler state dump: --------------------------------------------";
static const char __pyx_k_is_partial_execution[] = "is_partial_execution";
static const char __pyx_k_asynq_lib64_scheduler[] = "asynq.lib64.scheduler";
static const char __pyx_k_select_batch_to_flush[] = "_select_batch_to_flush";
static const char __pyx_k_No_tasks_in_task_queue[] = "No tasks in task queue.";
static const char __pyx_k_LocalTaskSchedulerState[] = "LocalTaskSchedulerState";
static const char __pyx_k_async_no_batch_to_flush[] = "@async: no batch to flush";
static const char __pyx_k_async_scheduling_task_s[] = "@async: scheduling task %s";
static const char __pyx_k_Task_is_already_computed[] = "Task is already computed.";
static const char __pyx_k_async_scheduling_batch_s[] = "@async: scheduling batch %s";
static const char __pyx_k_MAX_AWAIT_RECURSION_DEPTH[] = "MAX_AWAIT_RECURSION_DEPTH";
static const char __pyx_k_Scheduler_is_awaiting_for[] = "Scheduler is awaiting for:";
static const char __pyx_k_Task_is_already_scheduled[] = "Task is already scheduled.";
static const char __pyx_k_async_blocking_dependency[] = "async_blocking_dependency";
static const char __pyx_k_async_no_task_to_continue[] = "@async: no task to continue";
static const char __pyx_k_async_dependency_s_needs_s[] = "@async: +dependency: %s needs %s";
static const char __pyx_k_async_scheduler_activation[] = "@async: -> scheduler activation";
static const char __pyx_k_must_flush_batches_on_exit[] = "must_flush_batches_on_exit";
static const char __pyx_k_async_scheduler_activation_2[] = "@async: <- scheduler activation";
static const char __pyx_k_try_execute_on_new_scheduler[] = "_try_execute_on_new_scheduler";
static const char __pyx_k_LocalTaskSchedulerState_reset[] = "LocalTaskSchedulerState.reset";
static const char __pyx_k_LocalTaskSchedulerState___init[] = "LocalTaskSchedulerState.__init__";
static const char __pyx_k_async_await_recursion_depth_is[] = "@async: await recursion: depth is %d, but the limit is %d";
static const char __pyx_k_async_await_recursion_going_to[] = "@async: await recursion: going to run %d tasks on a new scheduler";
static const char __pyx_k_async_await_recursion_ignoring[] = "@async: await recursion: ignoring await()";
static const char __pyx_k_s_s_d_d_tasks_d_batches_active[] = "%s %s (%d + %d tasks, %d batches; active task: %s)";
static const char __pyx_k_Task_is_blocked_so_it_shouldn_t[] = "Task is blocked, so it shouldn't be scheduled.";
static const char __pyx_k_async_can_t_schedule_blocked_ta[] = "@async: can't schedule blocked task %s";
static const char __pyx_k_async_can_t_schedule_computed_t[] = "@async: can't schedule computed task %s";
static const char __pyx_k_async_can_t_schedule_flushed_ba[] = "@async: can't schedule flushed batch %s";
static const char __pyx_k_async_dependency_s_already_depe[] = "@async: +dependency: %s already depends on %s";
static const char __pyx_k_home_jelle_github_asynq_asynq_s[] = "/home/jelle/github/asynq/asynq/scheduler.py";
static const char __pyx_k_Are_you_importing_asynq_from_the[] = "Are you importing asynq from the wrong directory?";
static const char __pyx_k_Can_t_proceed_further_no_task_to[] = "Can't proceed further: no task to continue or batch to flush, active tasks: %s";
static const char __pyx_k_Task_is_already_scheduled_on_ano[] = "Task is already scheduled on another scheduler.";
static const char __pyx_k_Task_s_is_already_scheduled_on_a[] = "Task %s is already scheduled on another scheduler.";
static PyObject *__pyx_kp_s_Are_you_importing_asynq_from_the;
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_AsyncTaskError;
static PyObject *__pyx_n_s_BaseException;
static PyObject *__pyx_kp_s_Batches;
static PyObject *__pyx_kp_s_Can_t_proceed_further_no_task_to;
static PyObject *__pyx_n_s_Exception;
static PyObject *__pyx_kp_s_Last_task_s;
static PyObject *__pyx_n_s_LocalTaskSchedulerState;
static PyObject *__pyx_n_s_LocalTaskSchedulerState___init;
static PyObject *__pyx_n_s_LocalTaskSchedulerState_reset;
static PyObject *__pyx_n_s_MAX_AWAIT_RECURSION_DEPTH;
static PyObject *__pyx_kp_s_New_tasks;
static PyObject *__pyx_kp_s_No_tasks_in_task_queue;
static PyObject *__pyx_kp_s_Scheduler_is_awaiting_for;
static PyObject *__pyx_kp_s_Scheduler_state_dump;
static PyObject *__pyx_kp_s_Task_is_already_computed;
static PyObject *__pyx_kp_s_Task_is_already_scheduled;
static PyObject *__pyx_kp_s_Task_is_already_scheduled_on_ano;
static PyObject *__pyx_kp_s_Task_is_blocked_so_it_shouldn_t;
static PyObject *__pyx_kp_s_Task_queue;
static PyObject *__pyx_kp_s_Task_s_is_already_scheduled_on_a;
static PyObject *__pyx_n_s__20;
static PyObject *__pyx_n_s__21;
static PyObject *__pyx_kp_s__8;
static PyObject *__pyx_n_s_accept_changes;
static PyObject *__pyx_n_s_activate;
static PyObject *__pyx_n_s_active_task;
static PyObject *__pyx_n_s_add_dependencies;
static PyObject *__pyx_n_s_add_dependency;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_kp_s_async_await_recursion_depth_is;
static PyObject *__pyx_kp_s_async_await_recursion_going_to;
static PyObject *__pyx_kp_s_async_await_recursion_ignoring;
static PyObject *__pyx_n_s_async_blocking_dependency;
static PyObject *__pyx_kp_s_async_can_t_schedule_blocked_ta;
static PyObject *__pyx_kp_s_async_can_t_schedule_computed_t;
static PyObject *__pyx_kp_s_async_can_t_schedule_flushed_ba;
static PyObject *__pyx_kp_s_async_continued_s;
static PyObject *__pyx_kp_s_async_continuing_s;
static PyObject *__pyx_kp_s_async_dependency_s_already_depe;
static PyObject *__pyx_kp_s_async_dependency_s_needs_s;
static PyObject *__pyx_kp_s_async_no_batch_to_flush;
static PyObject *__pyx_kp_s_async_no_task_to_continue;
static PyObject *__pyx_kp_s_async_scheduler_activation;
static PyObject *__pyx_kp_s_async_scheduler_activation_2;
static PyObject *__pyx_kp_s_async_scheduler_s;
static PyObject *__pyx_kp_s_async_scheduling_batch_s;
static PyObject *__pyx_kp_s_async_scheduling_task_s;
static PyObject *__pyx_kp_s_async_skipping_s;
static PyObject *__pyx_kp_s_asynq_lib64_scheduler;
static PyObject *__pyx_kp_s_asynq_scheduler;
static PyObject *__pyx_n_s_await;
static PyObject *__pyx_n_s_await_2;
static PyObject *__pyx_n_s_batching;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_collections;
static PyObject *__pyx_n_s_continue_with;
static PyObject *__pyx_n_s_core_events;
static PyObject *__pyx_n_s_current;
static PyObject *__pyx_n_s_current_thread;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_debug_2;
static PyObject *__pyx_n_s_debug_options;
static PyObject *__pyx_n_s_dependency;
static PyObject *__pyx_n_s_deque;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_dump;
static PyObject *__pyx_n_s_endswith;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_execute;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_flush;
static PyObject *__pyx_n_s_flush_batches;
static PyObject *__pyx_n_s_future;
static PyObject *__pyx_n_s_futures;
static PyObject *__pyx_n_s_futures_none;
static PyObject *__pyx_n_s_get_priority;
static PyObject *__pyx_kp_s_home_jelle_github_asynq_asynq_s;
static PyObject *__pyx_n_s_ident;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_indent;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_is_computed;
static PyObject *__pyx_n_s_is_flushed;
static PyObject *__pyx_n_s_is_partial_execution;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_last_id;
static PyObject *__pyx_n_s_last_task;
static PyObject *__pyx_n_s_local;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_must_await;
static PyObject *__pyx_n_s_must_await_on_exit;
static PyObject *__pyx_n_s_must_flush_batches_on_exit;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_popleft;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_qcore_events;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_n_s_remove_dependency;
static PyObject *__pyx_n_s_reset;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_run_immediately;
static PyObject *__pyx_kp_s_s_d;
static PyObject *__pyx_kp_s_s_s_d_d_tasks_d_batches_active;
static PyObject *__pyx_n_s_schedule;
static PyObject *__pyx_n_s_schedule_batch;
static PyObject *__pyx_n_s_scheduler;
static PyObject *__pyx_n_s_select_batch_to_flush;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_stderr;
static PyObject *__pyx_n_s_stdout;
static PyObject *__pyx_n_s_str;
static PyObject *__pyx_n_s_str_2;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_task;
static PyObject *__pyx_n_s_tasks;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_threading;
static PyObject *__pyx_n_s_time;
static PyObject *__pyx_n_s_traceback;
static PyObject *__pyx_n_s_try_execute_on_new_scheduler;
static PyObject *__pyx_n_s_try_time_based_dump;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_write;
static PyObject *__pyx_n_s_yield_result;
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler___init__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_2activate(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_must_await_on_exit, int __pyx_v_must_flush_batches_on_exit); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_4reset(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_6start(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_v_run_immediately); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_8continue_with(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_10flush_batches(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_must_await); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_12await(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_14add_dependencies(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, PyObject *__pyx_v_yield_result); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_16add_dependency(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_dependency); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_18_await(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_20_execute(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_is_partial_execution, PyObject *__pyx_v_tasks); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_22_try_execute_on_new_scheduler(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_24schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_26schedule_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_28_select_batch_to_flush(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_30_clone(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_32_accept_changes(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_clone); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_34__str__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_36__repr__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_38dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_indent); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_40try_time_based_dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_last_task); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_15_last_dump_time___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_4name___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self); /* proto */
static int __pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator___init__(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_v_self, PyObject *__pyx_v_scheduler, PyObject *__pyx_v_must_await_on_exit, PyObject *__pyx_v_must_flush_batches_on_exit); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator_2__enter__(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator_4__exit__(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_type, CYTHON_UNUSED PyObject *__pyx_v_value, CYTHON_UNUSED PyObject *__pyx_v_traceback); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_23LocalTaskSchedulerState___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_23LocalTaskSchedulerState_2reset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_get_scheduler(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_2reset(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_4set_scheduler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_scheduler); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_6get_active_task(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_8extract_futures(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value, PyObject *__pyx_v_result); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_10_no_blocked_tasks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tasks); /* proto */
static PyObject *__pyx_pf_5asynq_9scheduler_12_create_blocking_dependency(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_tp_new_5asynq_9scheduler_TaskScheduler(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5asynq_9scheduler_TaskSchedulerActivator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_10;
static PyObject *__pyx_k__5;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__28;

/* "asynq/scheduler.py":47
 *     """Schedules and runs AsyncTask objects flushing batches when needed."""
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self._last_dump_time = time.time()
 *         self._empty_set = set()  # Just a perf. optimization
 */

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler___init__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler___init__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_v_thread = NULL;
  PyObject *__pyx_v_thread_name = NULL;
  PyObject *__pyx_v__id = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  float __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "asynq/scheduler.py":48
 * 
 *     def __init__(self):
 *         self._last_dump_time = time.time()             # <<<<<<<<<<<<<<
 *         self._empty_set = set()  # Just a perf. optimization
 *         self.await_recursion_depth = 0
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 48, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->_last_dump_time = __pyx_t_4;

  /* "asynq/scheduler.py":49
 *     def __init__(self):
 *         self._last_dump_time = time.time()
 *         self._empty_set = set()  # Just a perf. optimization             # <<<<<<<<<<<<<<
 *         self.await_recursion_depth = 0
 *         self.max_await_recursion_depth = MAX_AWAIT_RECURSION_DEPTH
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_empty_set);
  __Pyx_DECREF(__pyx_v_self->_empty_set);
  __pyx_v_self->_empty_set = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":50
 *         self._last_dump_time = time.time()
 *         self._empty_set = set()  # Just a perf. optimization
 *         self.await_recursion_depth = 0             # <<<<<<<<<<<<<<
 *         self.max_await_recursion_depth = MAX_AWAIT_RECURSION_DEPTH
 *         self.on_before_batch_flush = core_events.EventHook()
 */
  __pyx_v_self->await_recursion_depth = 0;

  /* "asynq/scheduler.py":51
 *         self._empty_set = set()  # Just a perf. optimization
 *         self.await_recursion_depth = 0
 *         self.max_await_recursion_depth = MAX_AWAIT_RECURSION_DEPTH             # <<<<<<<<<<<<<<
 *         self.on_before_batch_flush = core_events.EventHook()
 *         self.on_after_batch_flush = core_events.EventHook()
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_AWAIT_RECURSION_DEPTH); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->max_await_recursion_depth = __pyx_t_5;

  /* "asynq/scheduler.py":52
 *         self.await_recursion_depth = 0
 *         self.max_await_recursion_depth = MAX_AWAIT_RECURSION_DEPTH
 *         self.on_before_batch_flush = core_events.EventHook()             # <<<<<<<<<<<<<<
 *         self.on_after_batch_flush = core_events.EventHook()
 *         self.reset()
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5qcore_6events_EventHook), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->on_before_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_before_batch_flush));
  __pyx_v_self->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":53
 *         self.max_await_recursion_depth = MAX_AWAIT_RECURSION_DEPTH
 *         self.on_before_batch_flush = core_events.EventHook()
 *         self.on_after_batch_flush = core_events.EventHook()             # <<<<<<<<<<<<<<
 *         self.reset()
 *         thread = threading.current_thread()
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5qcore_6events_EventHook), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->on_after_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
  __pyx_v_self->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":54
 *         self.on_before_batch_flush = core_events.EventHook()
 *         self.on_after_batch_flush = core_events.EventHook()
 *         self.reset()             # <<<<<<<<<<<<<<
 *         thread = threading.current_thread()
 *         thread_name = thread.name if thread.name else str(thread.ident)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->reset(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":55
 *         self.on_after_batch_flush = core_events.EventHook()
 *         self.reset()
 *         thread = threading.current_thread()             # <<<<<<<<<<<<<<
 *         thread_name = thread.name if thread.name else str(thread.ident)
 *         if '_state' in globals():
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_threading); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_current_thread); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_thread = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":56
 *         self.reset()
 *         thread = threading.current_thread()
 *         thread_name = thread.name if thread.name else str(thread.ident)             # <<<<<<<<<<<<<<
 *         if '_state' in globals():
 *             _state.last_id += 1
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_thread, __pyx_n_s_ident); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_v_thread_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":57
 *         thread = threading.current_thread()
 *         thread_name = thread.name if thread.name else str(thread.ident)
 *         if '_state' in globals():             # <<<<<<<<<<<<<<
 *             _state.last_id += 1
 *             _id = _state.last_id
 */
  __pyx_t_1 = __Pyx_Globals(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_state, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {

    /* "asynq/scheduler.py":58
 *         thread_name = thread.name if thread.name else str(thread.ident)
 *         if '_state' in globals():
 *             _state.last_id += 1             # <<<<<<<<<<<<<<
 *             _id = _state.last_id
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_last_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_last_id, __pyx_t_2) < 0) __PYX_ERR(0, 58, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/scheduler.py":59
 *         if '_state' in globals():
 *             _state.last_id += 1
 *             _id = _state.last_id             # <<<<<<<<<<<<<<
 *         else:
 *             _id = 0
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_last_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 59, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v__id = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":57
 *         thread = threading.current_thread()
 *         thread_name = thread.name if thread.name else str(thread.ident)
 *         if '_state' in globals():             # <<<<<<<<<<<<<<
 *             _state.last_id += 1
 *             _id = _state.last_id
 */
    goto __pyx_L3;
  }

  /* "asynq/scheduler.py":61
 *             _id = _state.last_id
 *         else:
 *             _id = 0             # <<<<<<<<<<<<<<
 *         self.name = '%s / %d' % (thread_name, _id)
 * 
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_int_0);
    __pyx_v__id = __pyx_int_0;
  }
  __pyx_L3:;

  /* "asynq/scheduler.py":62
 *         else:
 *             _id = 0
 *         self.name = '%s / %d' % (thread_name, _id)             # <<<<<<<<<<<<<<
 * 
 *     def activate(self, must_await_on_exit, must_flush_batches_on_exit):
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_thread_name);
  __Pyx_GIVEREF(__pyx_v_thread_name);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_thread_name);
  __Pyx_INCREF(__pyx_v__id);
  __Pyx_GIVEREF(__pyx_v__id);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v__id);
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s_d, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyString_CheckExact(__pyx_t_1))||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 62, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":47
 *     """Schedules and runs AsyncTask objects flushing batches when needed."""
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self._last_dump_time = time.time()
 *         self._empty_set = set()  # Just a perf. optimization
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_thread);
  __Pyx_XDECREF(__pyx_v_thread_name);
  __Pyx_XDECREF(__pyx_v__id);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":64
 *         self.name = '%s / %d' % (thread_name, _id)
 * 
 *     def activate(self, must_await_on_exit, must_flush_batches_on_exit):             # <<<<<<<<<<<<<<
 *         return TaskSchedulerActivator(self, must_await_on_exit, must_flush_batches_on_exit)
 * 
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_3activate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_f_5asynq_9scheduler_13TaskScheduler_activate(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_must_await_on_exit, int __pyx_v_must_flush_batches_on_exit, int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("activate", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_activate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_3activate)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_must_await_on_exit); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_must_flush_batches_on_exit); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator))))) __PYX_ERR(0, 64, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":65
 * 
 *     def activate(self, must_await_on_exit, must_flush_batches_on_exit):
 *         return TaskSchedulerActivator(self, must_await_on_exit, must_flush_batches_on_exit)             # <<<<<<<<<<<<<<
 * 
 *     def reset(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_must_await_on_exit); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_must_flush_batches_on_exit); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator), __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = ((struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":64
 *         self.name = '%s / %d' % (thread_name, _id)
 * 
 *     def activate(self, must_await_on_exit, must_flush_batches_on_exit):             # <<<<<<<<<<<<<<
 *         return TaskSchedulerActivator(self, must_await_on_exit, must_flush_batches_on_exit)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.activate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_3activate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_3activate(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_must_await_on_exit;
  int __pyx_v_must_flush_batches_on_exit;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("activate (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_must_await_on_exit,&__pyx_n_s_must_flush_batches_on_exit,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_must_await_on_exit)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_must_flush_batches_on_exit)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("activate", 1, 2, 2, 1); __PYX_ERR(0, 64, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "activate") < 0)) __PYX_ERR(0, 64, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_must_await_on_exit = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_must_await_on_exit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L3_error)
    __pyx_v_must_flush_batches_on_exit = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_must_flush_batches_on_exit == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 64, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("activate", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 64, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.activate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_2activate(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_must_await_on_exit, __pyx_v_must_flush_batches_on_exit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_2activate(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_must_await_on_exit, int __pyx_v_must_flush_batches_on_exit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("activate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_13TaskScheduler_activate(__pyx_v_self, __pyx_v_must_await_on_exit, __pyx_v_must_flush_batches_on_exit, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.activate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":67
 *         return TaskSchedulerActivator(self, must_await_on_exit, must_flush_batches_on_exit)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """Resets the scheduler to its initial state."""
 *         # Used to reverse the sequence of newly scheduled tasks
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_5reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_reset(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("reset", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_5reset)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":71
 *         # Used to reverse the sequence of newly scheduled tasks
 *         # to execute them in correct order
 *         self._new_tasks = []             # <<<<<<<<<<<<<<
 *         self._tasks = collections.deque()
 *         self._batches = set()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_new_tasks);
  __Pyx_DECREF(__pyx_v_self->_new_tasks);
  __pyx_v_self->_new_tasks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":72
 *         # to execute them in correct order
 *         self._new_tasks = []
 *         self._tasks = collections.deque()             # <<<<<<<<<<<<<<
 *         self._batches = set()
 *         self.active_task = None
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_collections); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_deque); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_tasks);
  __Pyx_DECREF(__pyx_v_self->_tasks);
  __pyx_v_self->_tasks = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":73
 *         self._new_tasks = []
 *         self._tasks = collections.deque()
 *         self._batches = set()             # <<<<<<<<<<<<<<
 *         self.active_task = None
 * 
 */
  __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_batches);
  __Pyx_DECREF(__pyx_v_self->_batches);
  __pyx_v_self->_batches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":74
 *         self._tasks = collections.deque()
 *         self._batches = set()
 *         self.active_task = None             # <<<<<<<<<<<<<<
 * 
 *     def start(self, task, run_immediately=False):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->active_task);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
  __pyx_v_self->active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);

  /* "asynq/scheduler.py":67
 *         return TaskSchedulerActivator(self, must_await_on_exit, must_flush_batches_on_exit)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """Resets the scheduler to its initial state."""
 *         # Used to reverse the sequence of newly scheduled tasks
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_5reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_4reset[] = "Resets the scheduler to its initial state.";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_5reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_4reset(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_4reset(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("reset", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler_reset(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":76
 *         self.active_task = None
 * 
 *     def start(self, task, run_immediately=False):             # <<<<<<<<<<<<<<
 *         """Starts the specified task by simply scheduling it.
 *         Returns the task passed to this method.
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_7start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_13TaskScheduler_start(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start *__pyx_optional_args) {
  int __pyx_v_run_immediately = ((int)0);
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("start", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_run_immediately = __pyx_optional_args->run_immediately;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_start); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_7start)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_run_immediately); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 76, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":81
 * 
 *         """
 *         if task.is_computed():             # <<<<<<<<<<<<<<
 *             return task
 *         if task.scheduler is not None:
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "asynq/scheduler.py":82
 *         """
 *         if task.is_computed():
 *             return task             # <<<<<<<<<<<<<<
 *         if task.scheduler is not None:
 *             raise AsyncTaskError('Task is already scheduled.')
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_task));
    __pyx_r = __pyx_v_task;
    goto __pyx_L0;

    /* "asynq/scheduler.py":81
 * 
 *         """
 *         if task.is_computed():             # <<<<<<<<<<<<<<
 *             return task
 *         if task.scheduler is not None:
 */
  }

  /* "asynq/scheduler.py":83
 *         if task.is_computed():
 *             return task
 *         if task.scheduler is not None:             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError('Task is already scheduled.')
 *         if task.is_blocked():
 */
  __pyx_t_9 = (((PyObject *)__pyx_v_task->scheduler) != Py_None);
  __pyx_t_8 = (__pyx_t_9 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":84
 *             return task
 *         if task.scheduler is not None:
 *             raise AsyncTaskError('Task is already scheduled.')             # <<<<<<<<<<<<<<
 *         if task.is_blocked():
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 84, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 84, __pyx_L1_error)

    /* "asynq/scheduler.py":83
 *         if task.is_computed():
 *             return task
 *         if task.scheduler is not None:             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError('Task is already scheduled.')
 *         if task.is_blocked():
 */
  }

  /* "asynq/scheduler.py":85
 *         if task.scheduler is not None:
 *             raise AsyncTaskError('Task is already scheduled.')
 *         if task.is_blocked():             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->is_blocked(__pyx_v_task, 0); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  if (__pyx_t_9) {

    /* "asynq/scheduler.py":86
 *             raise AsyncTaskError('Task is already scheduled.')
 *         if task.is_blocked():
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")             # <<<<<<<<<<<<<<
 * 
 *         if run_immediately:
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 86, __pyx_L1_error)

    /* "asynq/scheduler.py":85
 *         if task.scheduler is not None:
 *             raise AsyncTaskError('Task is already scheduled.')
 *         if task.is_blocked():             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")
 * 
 */
  }

  /* "asynq/scheduler.py":88
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")
 * 
 *         if run_immediately:             # <<<<<<<<<<<<<<
 *             self._enqueue_new_tasks()
 *             self._schedule(task)
 */
  __pyx_t_9 = (__pyx_v_run_immediately != 0);
  if (__pyx_t_9) {

    /* "asynq/scheduler.py":89
 * 
 *         if run_immediately:
 *             self._enqueue_new_tasks()             # <<<<<<<<<<<<<<
 *             self._schedule(task)
 *             self._continue_with_task()
 */
    __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":90
 *         if run_immediately:
 *             self._enqueue_new_tasks()
 *             self._schedule(task)             # <<<<<<<<<<<<<<
 *             self._continue_with_task()
 *         else:
 */
    __pyx_t_9 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule(__pyx_v_self, __pyx_v_task); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 90, __pyx_L1_error)

    /* "asynq/scheduler.py":91
 *             self._enqueue_new_tasks()
 *             self._schedule(task)
 *             self._continue_with_task()             # <<<<<<<<<<<<<<
 *         else:
 *             self._schedule(task)
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_continue_with_task(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":88
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")
 * 
 *         if run_immediately:             # <<<<<<<<<<<<<<
 *             self._enqueue_new_tasks()
 *             self._schedule(task)
 */
    goto __pyx_L6;
  }

  /* "asynq/scheduler.py":93
 *             self._continue_with_task()
 *         else:
 *             self._schedule(task)             # <<<<<<<<<<<<<<
 *         return task
 * 
 */
  /*else*/ {
    __pyx_t_9 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule(__pyx_v_self, __pyx_v_task); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 93, __pyx_L1_error)
  }
  __pyx_L6:;

  /* "asynq/scheduler.py":94
 *         else:
 *             self._schedule(task)
 *         return task             # <<<<<<<<<<<<<<
 * 
 *     def continue_with(self, future, task):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_task));
  __pyx_r = __pyx_v_task;
  goto __pyx_L0;

  /* "asynq/scheduler.py":76
 *         self.active_task = None
 * 
 *     def start(self, task, run_immediately=False):             # <<<<<<<<<<<<<<
 *         """Starts the specified task by simply scheduling it.
 *         Returns the task passed to this method.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_7start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_6start[] = "Starts the specified task by simply scheduling it.\n        Returns the task passed to this method.\n\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_7start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  int __pyx_v_run_immediately;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_task,&__pyx_n_s_run_immediately,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_run_immediately);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "start") < 0)) __PYX_ERR(0, 76, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)values[0]);
    if (values[1]) {
      __pyx_v_run_immediately = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_run_immediately == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 76, __pyx_L3_error)
    } else {
      __pyx_v_run_immediately = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 76, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_task), __pyx_ptype_5asynq_10async_task_AsyncTask, 1, "task", 0))) __PYX_ERR(0, 76, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_6start(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_task, __pyx_v_run_immediately);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_6start(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_v_run_immediately) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start __pyx_t_2;
  __Pyx_RefNannySetupContext("start", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.run_immediately = __pyx_v_run_immediately;
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_5asynq_9scheduler_TaskScheduler->start(__pyx_v_self, __pyx_v_task, 1, &__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":96
 *         return task
 * 
 *     def continue_with(self, future, task):             # <<<<<<<<<<<<<<
 *         """Starts (or continues) the task after specified future gets computed.
 *         Returns the task passed to this method.
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_9continue_with(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_13TaskScheduler_continue_with(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("continue_with", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_continue_with); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_9continue_with)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_future));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_future));
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_future));
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_task));
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 96, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 96, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":101
 * 
 *         """
 *         if task.is_computed():             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError('Task is already computed.')
 *         if future.is_computed():
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":102
 *         """
 *         if task.is_computed():
 *             raise AsyncTaskError('Task is already computed.')             # <<<<<<<<<<<<<<
 *         if future.is_computed():
 *             if task.scheduler is None:
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 102, __pyx_L1_error)

    /* "asynq/scheduler.py":101
 * 
 *         """
 *         if task.is_computed():             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError('Task is already computed.')
 *         if future.is_computed():
 */
  }

  /* "asynq/scheduler.py":103
 *         if task.is_computed():
 *             raise AsyncTaskError('Task is already computed.')
 *         if future.is_computed():             # <<<<<<<<<<<<<<
 *             if task.scheduler is None:
 *                 return self.start(task)
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_5asynq_7futures_FutureBase *)__pyx_v_future->__pyx_vtab)->is_computed(__pyx_v_future, 0); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 103, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_8 != 0);
  if (__pyx_t_7) {

    /* "asynq/scheduler.py":104
 *             raise AsyncTaskError('Task is already computed.')
 *         if future.is_computed():
 *             if task.scheduler is None:             # <<<<<<<<<<<<<<
 *                 return self.start(task)
 *             else:
 */
    __pyx_t_7 = (((PyObject *)__pyx_v_task->scheduler) == Py_None);
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "asynq/scheduler.py":105
 *         if future.is_computed():
 *             if task.scheduler is None:
 *                 return self.start(task)             # <<<<<<<<<<<<<<
 *             else:
 *                 # Nothing to do in this case
 */
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->start(__pyx_v_self, __pyx_v_task, 0, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 105, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2);
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "asynq/scheduler.py":104
 *             raise AsyncTaskError('Task is already computed.')
 *         if future.is_computed():
 *             if task.scheduler is None:             # <<<<<<<<<<<<<<
 *                 return self.start(task)
 *             else:
 */
    }

    /* "asynq/scheduler.py":108
 *             else:
 *                 # Nothing to do in this case
 *                 return task             # <<<<<<<<<<<<<<
 *         if task.scheduler is None:
 *             task.scheduler = self
 */
    /*else*/ {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __pyx_r = __pyx_v_task;
      goto __pyx_L0;
    }

    /* "asynq/scheduler.py":103
 *         if task.is_computed():
 *             raise AsyncTaskError('Task is already computed.')
 *         if future.is_computed():             # <<<<<<<<<<<<<<
 *             if task.scheduler is None:
 *                 return self.start(task)
 */
  }

  /* "asynq/scheduler.py":109
 *                 # Nothing to do in this case
 *                 return task
 *         if task.scheduler is None:             # <<<<<<<<<<<<<<
 *             task.scheduler = self
 *         elif task.scheduler is not self:
 */
  __pyx_t_8 = (((PyObject *)__pyx_v_task->scheduler) == Py_None);
  __pyx_t_7 = (__pyx_t_8 != 0);
  if (__pyx_t_7) {

    /* "asynq/scheduler.py":110
 *                 return task
 *         if task.scheduler is None:
 *             task.scheduler = self             # <<<<<<<<<<<<<<
 *         elif task.scheduler is not self:
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_task->scheduler);
    __Pyx_DECREF(((PyObject *)__pyx_v_task->scheduler));
    __pyx_v_task->scheduler = __pyx_v_self;

    /* "asynq/scheduler.py":109
 *                 # Nothing to do in this case
 *                 return task
 *         if task.scheduler is None:             # <<<<<<<<<<<<<<
 *             task.scheduler = self
 *         elif task.scheduler is not self:
 */
    goto __pyx_L6;
  }

  /* "asynq/scheduler.py":111
 *         if task.scheduler is None:
 *             task.scheduler = self
 *         elif task.scheduler is not self:             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')
 *         self.add_dependency(task, future)
 */
  __pyx_t_7 = (__pyx_v_task->scheduler != __pyx_v_self);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":112
 *             task.scheduler = self
 *         elif task.scheduler is not self:
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')             # <<<<<<<<<<<<<<
 *         self.add_dependency(task, future)
 *         self._try_schedule(task, True)
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_AsyncTaskError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 112, __pyx_L1_error)

    /* "asynq/scheduler.py":111
 *         if task.scheduler is None:
 *             task.scheduler = self
 *         elif task.scheduler is not self:             # <<<<<<<<<<<<<<
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')
 *         self.add_dependency(task, future)
 */
  }
  __pyx_L6:;

  /* "asynq/scheduler.py":113
 *         elif task.scheduler is not self:
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')
 *         self.add_dependency(task, future)             # <<<<<<<<<<<<<<
 *         self._try_schedule(task, True)
 *         return task
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->add_dependency(__pyx_v_self, __pyx_v_task, __pyx_v_future, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":114
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')
 *         self.add_dependency(task, future)
 *         self._try_schedule(task, True)             # <<<<<<<<<<<<<<
 *         return task
 * 
 */
  __pyx_t_8 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_try_schedule(__pyx_v_self, __pyx_v_task, 1); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 114, __pyx_L1_error)

  /* "asynq/scheduler.py":115
 *         self.add_dependency(task, future)
 *         self._try_schedule(task, True)
 *         return task             # <<<<<<<<<<<<<<
 * 
 *     def flush_batches(self, must_await=False):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_task));
  __pyx_r = __pyx_v_task;
  goto __pyx_L0;

  /* "asynq/scheduler.py":96
 *         return task
 * 
 *     def continue_with(self, future, task):             # <<<<<<<<<<<<<<
 *         """Starts (or continues) the task after specified future gets computed.
 *         Returns the task passed to this method.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.continue_with", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_9continue_with(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_8continue_with[] = "Starts (or continues) the task after specified future gets computed.\n        Returns the task passed to this method.\n\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_9continue_with(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future = 0;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("continue_with (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_future,&__pyx_n_s_task,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_future)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("continue_with", 1, 2, 2, 1); __PYX_ERR(0, 96, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "continue_with") < 0)) __PYX_ERR(0, 96, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_future = ((struct __pyx_obj_5asynq_7futures_FutureBase *)values[0]);
    __pyx_v_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("continue_with", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 96, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.continue_with", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_future), __pyx_ptype_5asynq_7futures_FutureBase, 1, "future", 0))) __PYX_ERR(0, 96, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_task), __pyx_ptype_5asynq_10async_task_AsyncTask, 1, "task", 0))) __PYX_ERR(0, 96, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_8continue_with(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_future, __pyx_v_task);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_8continue_with(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_future, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("continue_with", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_13TaskScheduler_continue_with(__pyx_v_self, __pyx_v_future, __pyx_v_task, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.continue_with", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":117
 *         return task
 * 
 *     def flush_batches(self, must_await=False):             # <<<<<<<<<<<<<<
 *         """Flushes all the batches that are currently 'known'
 *         to scheduler. Normally you shouldn't call this method.
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_11flush_batches(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler_flush_batches(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches *__pyx_optional_args) {
  int __pyx_v_must_await = ((int)0);
  PyObject *__pyx_v_error = NULL;
  PyObject *__pyx_v_batches = NULL;
  PyObject *__pyx_v_batch = NULL;
  PyObject *__pyx_v_e = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  __Pyx_RefNannySetupContext("flush_batches", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_must_await = __pyx_optional_args->must_await;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_flush_batches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_11flush_batches)) {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_must_await); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_7;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":122
 * 
 *         """
 *         if must_await:             # <<<<<<<<<<<<<<
 *             self.await()
 * 
 */
  __pyx_t_8 = (__pyx_v_must_await != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":123
 *         """
 *         if must_await:
 *             self.await()             # <<<<<<<<<<<<<<
 * 
 *         error = None
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_await); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":122
 * 
 *         """
 *         if must_await:             # <<<<<<<<<<<<<<
 *             self.await()
 * 
 */
  }

  /* "asynq/scheduler.py":125
 *             self.await()
 * 
 *         error = None             # <<<<<<<<<<<<<<
 *         while self._batches:
 *             batches = self._batches
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_error = Py_None;

  /* "asynq/scheduler.py":126
 * 
 *         error = None
 *         while self._batches:             # <<<<<<<<<<<<<<
 *             batches = self._batches
 *             self._batches = set()
 */
  while (1) {
    __pyx_t_8 = (__pyx_v_self->_batches != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_batches) != 0);
    if (!__pyx_t_8) break;

    /* "asynq/scheduler.py":127
 *         error = None
 *         while self._batches:
 *             batches = self._batches             # <<<<<<<<<<<<<<
 *             self._batches = set()
 *             for batch in batches:
 */
    __pyx_t_1 = __pyx_v_self->_batches;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_batches, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynq/scheduler.py":128
 *         while self._batches:
 *             batches = self._batches
 *             self._batches = set()             # <<<<<<<<<<<<<<
 *             for batch in batches:
 *                 try:
 */
    __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_batches);
    __Pyx_DECREF(__pyx_v_self->_batches);
    __pyx_v_self->_batches = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynq/scheduler.py":129
 *             batches = self._batches
 *             self._batches = set()
 *             for batch in batches:             # <<<<<<<<<<<<<<
 *                 try:
 *                     if batch.items and not batch.is_flushed():
 */
    __pyx_t_1 = PyObject_GetIter(__pyx_v_batches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 129, __pyx_L1_error)
    for (;;) {
      {
        __pyx_t_2 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 129, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_batch, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynq/scheduler.py":130
 *             self._batches = set()
 *             for batch in batches:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if batch.items and not batch.is_flushed():
 *                         self._flush_batch(batch)
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "asynq/scheduler.py":131
 *             for batch in batches:
 *                 try:
 *                     if batch.items and not batch.is_flushed():             # <<<<<<<<<<<<<<
 *                         self._flush_batch(batch)
 *                 except BaseException as e:
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_batch, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 131, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (__pyx_t_13) {
          } else {
            __pyx_t_8 = __pyx_t_13;
            goto __pyx_L17_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_batch, __pyx_n_s_is_flushed); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 131, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (__pyx_t_6) {
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          } else {
            __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L8_error)
          }
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 131, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_14 = ((!__pyx_t_13) != 0);
          __pyx_t_8 = __pyx_t_14;
          __pyx_L17_bool_binop_done:;
          if (__pyx_t_8) {

            /* "asynq/scheduler.py":132
 *                 try:
 *                     if batch.items and not batch.is_flushed():
 *                         self._flush_batch(batch)             # <<<<<<<<<<<<<<
 *                 except BaseException as e:
 *                     if error is None:
 */
            if (!(likely(((__pyx_v_batch) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_batch, __pyx_ptype_5asynq_8batching_BatchBase))))) __PYX_ERR(0, 132, __pyx_L8_error)
            __pyx_t_7 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_flush_batch(__pyx_v_self, ((struct __pyx_obj_5asynq_8batching_BatchBase *)__pyx_v_batch)); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 132, __pyx_L8_error)

            /* "asynq/scheduler.py":131
 *             for batch in batches:
 *                 try:
 *                     if batch.items and not batch.is_flushed():             # <<<<<<<<<<<<<<
 *                         self._flush_batch(batch)
 *                 except BaseException as e:
 */
          }

          /* "asynq/scheduler.py":130
 *             self._batches = set()
 *             for batch in batches:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if batch.items and not batch.is_flushed():
 *                         self._flush_batch(batch)
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L15_try_end;
        __pyx_L8_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynq/scheduler.py":133
 *                     if batch.items and not batch.is_flushed():
 *                         self._flush_batch(batch)
 *                 except BaseException as e:             # <<<<<<<<<<<<<<
 *                     if error is None:
 *                         error = e
 */
        __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_BaseException);
        if (__pyx_t_7) {
          __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.flush_batches", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_4, &__pyx_t_6) < 0) __PYX_ERR(0, 133, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_XDECREF_SET(__pyx_v_e, __pyx_t_4);

          /* "asynq/scheduler.py":134
 *                         self._flush_batch(batch)
 *                 except BaseException as e:
 *                     if error is None:             # <<<<<<<<<<<<<<
 *                         error = e
 *         if error is not None:
 */
          __pyx_t_8 = (__pyx_v_error == Py_None);
          __pyx_t_14 = (__pyx_t_8 != 0);
          if (__pyx_t_14) {

            /* "asynq/scheduler.py":135
 *                 except BaseException as e:
 *                     if error is None:
 *                         error = e             # <<<<<<<<<<<<<<
 *         if error is not None:
 *             raise error
 */
            __Pyx_INCREF(__pyx_v_e);
            __Pyx_DECREF_SET(__pyx_v_error, __pyx_v_e);

            /* "asynq/scheduler.py":134
 *                         self._flush_batch(batch)
 *                 except BaseException as e:
 *                     if error is None:             # <<<<<<<<<<<<<<
 *                         error = e
 *         if error is not None:
 */
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L9_exception_handled;
        }
        goto __pyx_L10_except_error;
        __pyx_L10_except_error:;

        /* "asynq/scheduler.py":130
 *             self._batches = set()
 *             for batch in batches:
 *                 try:             # <<<<<<<<<<<<<<
 *                     if batch.items and not batch.is_flushed():
 *                         self._flush_batch(batch)
 */
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L1_error;
        __pyx_L9_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_L15_try_end:;
      }

      /* "asynq/scheduler.py":129
 *             batches = self._batches
 *             self._batches = set()
 *             for batch in batches:             # <<<<<<<<<<<<<<
 *                 try:
 *                     if batch.items and not batch.is_flushed():
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":136
 *                     if error is None:
 *                         error = e
 *         if error is not None:             # <<<<<<<<<<<<<<
 *             raise error
 *         return 0
 */
  __pyx_t_14 = (__pyx_v_error != Py_None);
  __pyx_t_8 = (__pyx_t_14 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":137
 *                         error = e
 *         if error is not None:
 *             raise error             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(0, 137, __pyx_L1_error)

    /* "asynq/scheduler.py":136
 *                     if error is None:
 *                         error = e
 *         if error is not None:             # <<<<<<<<<<<<<<
 *             raise error
 *         return 0
 */
  }

  /* "asynq/scheduler.py":138
 *         if error is not None:
 *             raise error
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def await(self, tasks=[]):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":117
 *         return task
 * 
 *     def flush_batches(self, must_await=False):             # <<<<<<<<<<<<<<
 *         """Flushes all the batches that are currently 'known'
 *         to scheduler. Normally you shouldn't call this method.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.flush_batches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_XDECREF(__pyx_v_batches);
  __Pyx_XDECREF(__pyx_v_batch);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_11flush_batches(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_10flush_batches[] = "Flushes all the batches that are currently 'known'\n        to scheduler. Normally you shouldn't call this method.\n\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_11flush_batches(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_must_await;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("flush_batches (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_must_await,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_must_await);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "flush_batches") < 0)) __PYX_ERR(0, 117, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_must_await = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_must_await == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L3_error)
    } else {
      __pyx_v_must_await = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("flush_batches", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 117, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.flush_batches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10flush_batches(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_must_await);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_10flush_batches(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_must_await) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("flush_batches", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.must_await = __pyx_v_must_await;
  __pyx_t_1 = __pyx_vtabptr_5asynq_9scheduler_TaskScheduler->flush_batches(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.flush_batches", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":140
 *         return 0
 * 
 *     def await(self, tasks=[]):             # <<<<<<<<<<<<<<
 *         """
 *         Executes a list of tasks ensuring all of them are completed when this method returns.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_13await(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_12await[] = "\n        Executes a list of tasks ensuring all of them are completed when this method returns.\n\n        :param tasks: tasks to await for\n        :return: ``None``\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_13await(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_tasks = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("await (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tasks,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__5;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tasks);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "await") < 0)) __PYX_ERR(0, 140, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_tasks = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("await", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 140, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.await", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_12await(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_tasks);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_12await(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("await", 0);

  /* "asynq/scheduler.py":147
 *         :return: ``None``
 *         """
 *         self._await(tasks)             # <<<<<<<<<<<<<<
 * 
 *     def add_dependencies(self, task, yield_result):
 */
  if (!(likely(PyList_CheckExact(__pyx_v_tasks))||((__pyx_v_tasks) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_tasks)->tp_name), 0))) __PYX_ERR(0, 147, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_await(__pyx_v_self, ((PyObject*)__pyx_v_tasks), 0); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 147, __pyx_L1_error)

  /* "asynq/scheduler.py":140
 *         return 0
 * 
 *     def await(self, tasks=[]):             # <<<<<<<<<<<<<<
 *         """
 *         Executes a list of tasks ensuring all of them are completed when this method returns.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.await", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":149
 *         self._await(tasks)
 * 
 *     def add_dependencies(self, task, yield_result):             # <<<<<<<<<<<<<<
 *         """Mark all futures in yield_result as dependencies for task.
 * 
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_15add_dependencies(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependencies(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, PyObject *__pyx_v_yield_result, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_future = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  __Pyx_RefNannySetupContext("add_dependencies", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add_dependencies); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_15add_dependencies)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_task));
      __Pyx_INCREF(__pyx_v_yield_result);
      __Pyx_GIVEREF(__pyx_v_yield_result);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_yield_result);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":155
 * 
 *         """
 *         if yield_result is None:  # Frequent case             # <<<<<<<<<<<<<<
 *             return
 *         for future in extract_futures(yield_result, []):
 */
  __pyx_t_7 = (__pyx_v_yield_result == Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":156
 *         """
 *         if yield_result is None:  # Frequent case
 *             return             # <<<<<<<<<<<<<<
 *         for future in extract_futures(yield_result, []):
 *             self.add_dependency(task, future)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/scheduler.py":155
 * 
 *         """
 *         if yield_result is None:  # Frequent case             # <<<<<<<<<<<<<<
 *             return
 *         for future in extract_futures(yield_result, []):
 */
  }

  /* "asynq/scheduler.py":157
 *         if yield_result is None:  # Frequent case
 *             return
 *         for future in extract_futures(yield_result, []):             # <<<<<<<<<<<<<<
 *             self.add_dependency(task, future)
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_f_5asynq_9scheduler_extract_futures(__pyx_v_yield_result, ((PyObject*)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_5 = 0;
    __pyx_t_9 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 157, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_9)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 157, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 157, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_9(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 157, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_future, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":158
 *             return
 *         for future in extract_futures(yield_result, []):
 *             self.add_dependency(task, future)             # <<<<<<<<<<<<<<
 * 
 *     def add_dependency(self, task, dependency):
 */
    if (!(likely(((__pyx_v_future) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_future, __pyx_ptype_5asynq_7futures_FutureBase))))) __PYX_ERR(0, 158, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->add_dependency(__pyx_v_self, __pyx_v_task, ((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_future), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 158, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/scheduler.py":157
 *         if yield_result is None:  # Frequent case
 *             return
 *         for future in extract_futures(yield_result, []):             # <<<<<<<<<<<<<<
 *             self.add_dependency(task, future)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":149
 *         self._await(tasks)
 * 
 *     def add_dependencies(self, task, yield_result):             # <<<<<<<<<<<<<<
 *         """Mark all futures in yield_result as dependencies for task.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.add_dependencies", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_future);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_15add_dependencies(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_14add_dependencies[] = "Mark all futures in yield_result as dependencies for task.\n\n        i.e. everything in yield_result needs to be computed before task can be continued.\n\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_15add_dependencies(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  PyObject *__pyx_v_yield_result = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_dependencies (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_task,&__pyx_n_s_yield_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_yield_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_dependencies", 1, 2, 2, 1); __PYX_ERR(0, 149, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_dependencies") < 0)) __PYX_ERR(0, 149, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)values[0]);
    __pyx_v_yield_result = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_dependencies", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 149, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.add_dependencies", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_task), __pyx_ptype_5asynq_10async_task_AsyncTask, 1, "task", 0))) __PYX_ERR(0, 149, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_14add_dependencies(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_task, __pyx_v_yield_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_14add_dependencies(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, PyObject *__pyx_v_yield_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("add_dependencies", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependencies(__pyx_v_self, __pyx_v_task, __pyx_v_yield_result, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.add_dependencies", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":160
 *             self.add_dependency(task, future)
 * 
 *     def add_dependency(self, task, dependency):             # <<<<<<<<<<<<<<
 *         if dependency.is_computed():
 *             return
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_17add_dependency(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependency(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_dependency, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  __Pyx_RefNannySetupContext("add_dependency", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_add_dependency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_17add_dependency)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      __pyx_t_5 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_5 = 1;
        }
      }
      __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, ((PyObject *)__pyx_v_task));
      __Pyx_INCREF(((PyObject *)__pyx_v_dependency));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_dependency));
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, ((PyObject *)__pyx_v_dependency));
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":161
 * 
 *     def add_dependency(self, task, dependency):
 *         if dependency.is_computed():             # <<<<<<<<<<<<<<
 *             return
 *         if dependency in task._dependencies:
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_5asynq_7futures_FutureBase *)__pyx_v_dependency->__pyx_vtab)->is_computed(__pyx_v_dependency, 0); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":162
 *     def add_dependency(self, task, dependency):
 *         if dependency.is_computed():
 *             return             # <<<<<<<<<<<<<<
 *         if dependency in task._dependencies:
 *             if _debug_options.DUMP_DEPENDENCIES:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/scheduler.py":161
 * 
 *     def add_dependency(self, task, dependency):
 *         if dependency.is_computed():             # <<<<<<<<<<<<<<
 *             return
 *         if dependency in task._dependencies:
 */
  }

  /* "asynq/scheduler.py":163
 *         if dependency.is_computed():
 *             return
 *         if dependency in task._dependencies:             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_DEPENDENCIES:
 *                 debug.write('@async: +dependency: %s already depends on %s' %
 */
  __pyx_t_8 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_dependency), __pyx_v_task->_dependencies, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 163, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_8 != 0);
  if (__pyx_t_7) {

    /* "asynq/scheduler.py":164
 *             return
 *         if dependency in task._dependencies:
 *             if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: +dependency: %s already depends on %s' %
 *                     (debug.str(task), debug.str(dependency)))
 */
    __pyx_t_7 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_DEPENDENCIES != 0);
    if (__pyx_t_7) {

      /* "asynq/scheduler.py":165
 *         if dependency in task._dependencies:
 *             if _debug_options.DUMP_DEPENDENCIES:
 *                 debug.write('@async: +dependency: %s already depends on %s' %             # <<<<<<<<<<<<<<
 *                     (debug.str(task), debug.str(dependency)))
 *             return
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynq/scheduler.py":166
 *             if _debug_options.DUMP_DEPENDENCIES:
 *                 debug.write('@async: +dependency: %s already depends on %s' %
 *                     (debug.str(task), debug.str(dependency)))             # <<<<<<<<<<<<<<
 *             return
 *         if _debug_options.DUMP_DEPENDENCIES:
 */
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_task));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, ((PyObject *)__pyx_v_task));
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
        }
      }
      if (!__pyx_t_9) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_dependency)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_dependency));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_dependency));
        PyTuple_SET_ITEM(__pyx_t_10, 0+1, ((PyObject *)__pyx_v_dependency));
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_4);
      __pyx_t_2 = 0;
      __pyx_t_4 = 0;

      /* "asynq/scheduler.py":165
 *         if dependency in task._dependencies:
 *             if _debug_options.DUMP_DEPENDENCIES:
 *                 debug.write('@async: +dependency: %s already depends on %s' %             # <<<<<<<<<<<<<<
 *                     (debug.str(task), debug.str(dependency)))
 *             return
 */
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_dependency_s_already_depe, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":164
 *             return
 *         if dependency in task._dependencies:
 *             if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: +dependency: %s already depends on %s' %
 *                     (debug.str(task), debug.str(dependency)))
 */
    }

    /* "asynq/scheduler.py":167
 *                 debug.write('@async: +dependency: %s already depends on %s' %
 *                     (debug.str(task), debug.str(dependency)))
 *             return             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: +dependency: %s needs %s' %
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/scheduler.py":163
 *         if dependency.is_computed():
 *             return
 *         if dependency in task._dependencies:             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_DEPENDENCIES:
 *                 debug.write('@async: +dependency: %s already depends on %s' %
 */
  }

  /* "asynq/scheduler.py":168
 *                     (debug.str(task), debug.str(dependency)))
 *             return
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: +dependency: %s needs %s' %
 *                 (debug.str(task), debug.str(dependency)))
 */
  __pyx_t_7 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_DEPENDENCIES != 0);
  if (__pyx_t_7) {

    /* "asynq/scheduler.py":169
 *             return
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: +dependency: %s needs %s' %             # <<<<<<<<<<<<<<
 *                 (debug.str(task), debug.str(dependency)))
 *         dependency.make_dependency(task, self)
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynq/scheduler.py":170
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: +dependency: %s needs %s' %
 *                 (debug.str(task), debug.str(dependency)))             # <<<<<<<<<<<<<<
 *         dependency.make_dependency(task, self)
 * 
 */
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_10, 0+1, ((PyObject *)__pyx_v_task));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_10) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_dependency)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_dependency));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_dependency));
      PyTuple_SET_ITEM(__pyx_t_9, 0+1, ((PyObject *)__pyx_v_dependency));
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
    __pyx_t_3 = 0;
    __pyx_t_6 = 0;

    /* "asynq/scheduler.py":169
 *             return
 *         if _debug_options.DUMP_DEPENDENCIES:
 *             debug.write('@async: +dependency: %s needs %s' %             # <<<<<<<<<<<<<<
 *                 (debug.str(task), debug.str(dependency)))
 *         dependency.make_dependency(task, self)
 */
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_dependency_s_needs_s, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":168
 *                     (debug.str(task), debug.str(dependency)))
 *             return
 *         if _debug_options.DUMP_DEPENDENCIES:             # <<<<<<<<<<<<<<
 *             debug.write('@async: +dependency: %s needs %s' %
 *                 (debug.str(task), debug.str(dependency)))
 */
  }

  /* "asynq/scheduler.py":171
 *             debug.write('@async: +dependency: %s needs %s' %
 *                 (debug.str(task), debug.str(dependency)))
 *         dependency.make_dependency(task, self)             # <<<<<<<<<<<<<<
 * 
 *     def _await(self, tasks):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_7futures_FutureBase *)__pyx_v_dependency->__pyx_vtab)->make_dependency(__pyx_v_dependency, ((PyObject *)__pyx_v_task), ((PyObject *)__pyx_v_self), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":160
 *             self.add_dependency(task, future)
 * 
 *     def add_dependency(self, task, dependency):             # <<<<<<<<<<<<<<
 *         if dependency.is_computed():
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.add_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_17add_dependency(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_17add_dependency(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_dependency = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_dependency (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_task,&__pyx_n_s_dependency,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_task)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_dependency)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("add_dependency", 1, 2, 2, 1); __PYX_ERR(0, 160, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "add_dependency") < 0)) __PYX_ERR(0, 160, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)values[0]);
    __pyx_v_dependency = ((struct __pyx_obj_5asynq_7futures_FutureBase *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_dependency", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.add_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_task), __pyx_ptype_5asynq_10async_task_AsyncTask, 1, "task", 0))) __PYX_ERR(0, 160, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_dependency), __pyx_ptype_5asynq_7futures_FutureBase, 1, "dependency", 0))) __PYX_ERR(0, 160, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_16add_dependency(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_task, __pyx_v_dependency);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_16add_dependency(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_dependency) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("add_dependency", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependency(__pyx_v_self, __pyx_v_task, __pyx_v_dependency, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.add_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":173
 *         dependency.make_dependency(task, self)
 * 
 *     def _await(self, tasks):             # <<<<<<<<<<<<<<
 *         active_task = self.active_task
 *         if active_task is not None:
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_19_await(PyObject *__pyx_v_self, PyObject *__pyx_v_tasks); /*proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__await(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks, int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_active_task = 0;
  struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_v_blocking_dependency = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  __Pyx_RefNannySetupContext("_await", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_await_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_19_await)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_tasks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 173, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_tasks);
        __Pyx_GIVEREF(__pyx_v_tasks);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_tasks);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":174
 * 
 *     def _await(self, tasks):
 *         active_task = self.active_task             # <<<<<<<<<<<<<<
 *         if active_task is not None:
 *             # Since task we await for can be (and most likely is)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->active_task);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":175
 *     def _await(self, tasks):
 *         active_task = self.active_task
 *         if active_task is not None:             # <<<<<<<<<<<<<<
 *             # Since task we await for can be (and most likely is)
 *             # a child of the current one, it already copied the
 */
  __pyx_t_7 = (((PyObject *)__pyx_v_active_task) != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":182
 *             # to resume already active contexts. That's why we
 *             # need to pause them.
 *             active_task._pause_contexts()             # <<<<<<<<<<<<<<
 *             # We must temporarily block active task in case
 *             # it's somehow (c) added to scheduler's queue
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_active_task->__pyx_base.__pyx_vtab)->_pause_contexts(__pyx_v_active_task); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":185
 *             # We must temporarily block active task in case
 *             # it's somehow (c) added to scheduler's queue
 *             blocking_dependency = _create_blocking_dependency()             # <<<<<<<<<<<<<<
 *             active_task._dependencies.add(blocking_dependency)
 *         try:
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler__create_blocking_dependency(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_blocking_dependency = ((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynq/scheduler.py":186
 *             # it's somehow (c) added to scheduler's queue
 *             blocking_dependency = _create_blocking_dependency()
 *             active_task._dependencies.add(blocking_dependency)             # <<<<<<<<<<<<<<
 *         try:
 *             # Must never throw an exception
 */
    if (unlikely(__pyx_v_active_task->_dependencies == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "add");
      __PYX_ERR(0, 186, __pyx_L1_error)
    }
    __pyx_t_9 = PySet_Add(__pyx_v_active_task->_dependencies, ((PyObject *)__pyx_v_blocking_dependency)); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 186, __pyx_L1_error)

    /* "asynq/scheduler.py":175
 *     def _await(self, tasks):
 *         active_task = self.active_task
 *         if active_task is not None:             # <<<<<<<<<<<<<<
 *             # Since task we await for can be (and most likely is)
 *             # a child of the current one, it already copied the
 */
  }

  /* "asynq/scheduler.py":187
 *             blocking_dependency = _create_blocking_dependency()
 *             active_task._dependencies.add(blocking_dependency)
 *         try:             # <<<<<<<<<<<<<<
 *             # Must never throw an exception
 *             if len(tasks) == 0:
 */
  /*try:*/ {

    /* "asynq/scheduler.py":189
 *         try:
 *             # Must never throw an exception
 *             if len(tasks) == 0:             # <<<<<<<<<<<<<<
 *                 # No arguments = run() must empty task queue
 *                 return self._execute(False, self._empty_set)
 */
    if (unlikely(__pyx_v_tasks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 189, __pyx_L5_error)
    }
    __pyx_t_10 = PyList_GET_SIZE(__pyx_v_tasks); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 189, __pyx_L5_error)
    __pyx_t_8 = ((__pyx_t_10 == 0) != 0);
    if (__pyx_t_8) {

      /* "asynq/scheduler.py":191
 *             if len(tasks) == 0:
 *                 # No arguments = run() must empty task queue
 *                 return self._execute(False, self._empty_set)             # <<<<<<<<<<<<<<
 *             else:
 *                 return self._execute(True, set(tasks))
 */
      __pyx_t_1 = __pyx_v_self->_empty_set;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_execute(__pyx_v_self, 0, ((PyObject*)__pyx_t_1), 0); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 191, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_6;
      goto __pyx_L4_return;

      /* "asynq/scheduler.py":189
 *         try:
 *             # Must never throw an exception
 *             if len(tasks) == 0:             # <<<<<<<<<<<<<<
 *                 # No arguments = run() must empty task queue
 *                 return self._execute(False, self._empty_set)
 */
    }

    /* "asynq/scheduler.py":193
 *                 return self._execute(False, self._empty_set)
 *             else:
 *                 return self._execute(True, set(tasks))             # <<<<<<<<<<<<<<
 *         finally:
 *             if active_task is not None:
 */
    /*else*/ {
      __pyx_t_1 = PySet_New(__pyx_v_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_execute(__pyx_v_self, 1, ((PyObject*)__pyx_t_1), 0); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 193, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_r = __pyx_t_6;
      goto __pyx_L4_return;
    }
  }

  /* "asynq/scheduler.py":195
 *                 return self._execute(True, set(tasks))
 *         finally:
 *             if active_task is not None:             # <<<<<<<<<<<<<<
 *                 active_task._dependencies.remove(blocking_dependency)
 *                 active_task._resume_contexts()
 */
  /*finally:*/ {
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L5_error:;
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_14, &__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __pyx_t_6 = __pyx_lineno; __pyx_t_11 = __pyx_clineno; __pyx_t_12 = __pyx_filename;
      {
        __pyx_t_8 = (((PyObject *)__pyx_v_active_task) != Py_None);
        __pyx_t_7 = (__pyx_t_8 != 0);
        if (__pyx_t_7) {

          /* "asynq/scheduler.py":196
 *         finally:
 *             if active_task is not None:
 *                 active_task._dependencies.remove(blocking_dependency)             # <<<<<<<<<<<<<<
 *                 active_task._resume_contexts()
 *         return 0
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_active_task->_dependencies, __pyx_n_s_remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 196, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(!__pyx_v_blocking_dependency)) { __Pyx_RaiseUnboundLocalError("blocking_dependency"); __PYX_ERR(0, 196, __pyx_L9_error) }
          __pyx_t_3 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_blocking_dependency)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 196, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_INCREF(((PyObject *)__pyx_v_blocking_dependency));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_blocking_dependency));
            PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_blocking_dependency));
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L9_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/scheduler.py":197
 *             if active_task is not None:
 *                 active_task._dependencies.remove(blocking_dependency)
 *                 active_task._resume_contexts()             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
          __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_active_task->__pyx_base.__pyx_vtab)->_resume_contexts(__pyx_v_active_task); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L9_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/scheduler.py":195
 *                 return self._execute(True, set(tasks))
 *         finally:
 *             if active_task is not None:             # <<<<<<<<<<<<<<
 *                 active_task._dependencies.remove(blocking_dependency)
 *                 active_task._resume_contexts()
 */
        }
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
      }
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_ErrRestore(__pyx_t_13, __pyx_t_14, __pyx_t_15);
      __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_11; __pyx_filename = __pyx_t_12;
      goto __pyx_L1_error;
      __pyx_L9_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_17, __pyx_t_18);
      }
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L4_return: {
      __pyx_t_11 = __pyx_r;
      __pyx_t_7 = (((PyObject *)__pyx_v_active_task) != Py_None);
      __pyx_t_8 = (__pyx_t_7 != 0);
      if (__pyx_t_8) {

        /* "asynq/scheduler.py":196
 *         finally:
 *             if active_task is not None:
 *                 active_task._dependencies.remove(blocking_dependency)             # <<<<<<<<<<<<<<
 *                 active_task._resume_contexts()
 *         return 0
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_active_task->_dependencies, __pyx_n_s_remove); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 196, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (unlikely(!__pyx_v_blocking_dependency)) { __Pyx_RaiseUnboundLocalError("blocking_dependency"); __PYX_ERR(0, 196, __pyx_L1_error) }
        __pyx_t_5 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_5) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_blocking_dependency)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 196, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_blocking_dependency));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_blocking_dependency));
          PyTuple_SET_ITEM(__pyx_t_3, 0+1, ((PyObject *)__pyx_v_blocking_dependency));
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 196, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":197
 *             if active_task is not None:
 *                 active_task._dependencies.remove(blocking_dependency)
 *                 active_task._resume_contexts()             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_active_task->__pyx_base.__pyx_vtab)->_resume_contexts(__pyx_v_active_task); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":195
 *                 return self._execute(True, set(tasks))
 *         finally:
 *             if active_task is not None:             # <<<<<<<<<<<<<<
 *                 active_task._dependencies.remove(blocking_dependency)
 *                 active_task._resume_contexts()
 */
      }
      __pyx_r = __pyx_t_11;
      goto __pyx_L0;
    }
  }

  /* "asynq/scheduler.py":173
 *         dependency.make_dependency(task, self)
 * 
 *     def _await(self, tasks):             # <<<<<<<<<<<<<<
 *         active_task = self.active_task
 *         if active_task is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._await", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_active_task);
  __Pyx_XDECREF((PyObject *)__pyx_v_blocking_dependency);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_19_await(PyObject *__pyx_v_self, PyObject *__pyx_v_tasks); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_19_await(PyObject *__pyx_v_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_await (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tasks), (&PyList_Type), 1, "tasks", 1))) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_18_await(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject*)__pyx_v_tasks));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_18_await(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_await", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__await(__pyx_v_self, __pyx_v_tasks, 1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._await", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":200
 *         return 0
 * 
 *     def _execute(self, is_partial_execution, tasks):             # <<<<<<<<<<<<<<
 *         """Implements task execution loop."""
 *         # Must never throw an exception
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21_execute(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_5asynq_9scheduler_13TaskScheduler__execute(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_is_partial_execution, PyObject *__pyx_v_tasks, int __pyx_skip_dispatch) {
  int __pyx_v_dump_await_recursion;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  int __pyx_v_dump_scheduler_state;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_creator = 0;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_current_active_task = NULL;
  struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch __pyx_t_17;
  struct __pyx_opt_args_5asynq_7futures_10FutureBase_dump __pyx_t_18;
  int __pyx_t_19;
  char const *__pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  __Pyx_RefNannySetupContext("_execute", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_execute); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_21_execute)) {
      __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_is_partial_execution); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      __pyx_t_6 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
          __pyx_t_6 = 1;
        }
      }
      __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_3);
      __Pyx_INCREF(__pyx_v_tasks);
      __Pyx_GIVEREF(__pyx_v_tasks);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_tasks);
      __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 200, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_8;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":203
 *         """Implements task execution loop."""
 *         # Must never throw an exception
 *         self.await_recursion_depth += 1             # <<<<<<<<<<<<<<
 *         current_active_task = self.active_task
 *         self.active_task = None
 */
  __pyx_v_self->await_recursion_depth = (__pyx_v_self->await_recursion_depth + 1);

  /* "asynq/scheduler.py":204
 *         # Must never throw an exception
 *         self.await_recursion_depth += 1
 *         current_active_task = self.active_task             # <<<<<<<<<<<<<<
 *         self.active_task = None
 *         try:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->active_task);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_current_active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":205
 *         self.await_recursion_depth += 1
 *         current_active_task = self.active_task
 *         self.active_task = None             # <<<<<<<<<<<<<<
 *         try:
 *             if self.await_recursion_depth > self.max_await_recursion_depth:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->active_task);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
  __pyx_v_self->active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);

  /* "asynq/scheduler.py":206
 *         current_active_task = self.active_task
 *         self.active_task = None
 *         try:             # <<<<<<<<<<<<<<
 *             if self.await_recursion_depth > self.max_await_recursion_depth:
 *                 # We don't want to execute every next task on
 */
  /*try:*/ {

    /* "asynq/scheduler.py":207
 *         self.active_task = None
 *         try:
 *             if self.await_recursion_depth > self.max_await_recursion_depth:             # <<<<<<<<<<<<<<
 *                 # We don't want to execute every next task on
 *                 # stack in case all of these tasks invoke
 */
    __pyx_t_9 = ((__pyx_v_self->await_recursion_depth > __pyx_v_self->max_await_recursion_depth) != 0);
    if (__pyx_t_9) {

      /* "asynq/scheduler.py":211
 *                 # stack in case all of these tasks invoke
 *                 # await at some point
 *                 dump_await_recursion = _debug_options.DUMP_AWAIT_RECURSION             # <<<<<<<<<<<<<<
 *                 if dump_await_recursion:
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %
 */
      __pyx_t_9 = __pyx_v_5asynq_9scheduler__debug_options->DUMP_AWAIT_RECURSION;
      __pyx_v_dump_await_recursion = __pyx_t_9;

      /* "asynq/scheduler.py":212
 *                 # await at some point
 *                 dump_await_recursion = _debug_options.DUMP_AWAIT_RECURSION
 *                 if dump_await_recursion:             # <<<<<<<<<<<<<<
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 */
      __pyx_t_9 = (__pyx_v_dump_await_recursion != 0);
      if (__pyx_t_9) {

        /* "asynq/scheduler.py":213
 *                 dump_await_recursion = _debug_options.DUMP_AWAIT_RECURSION
 *                 if dump_await_recursion:
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %             # <<<<<<<<<<<<<<
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 *                 if not is_partial_execution:
 */
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynq/scheduler.py":214
 *                 if dump_await_recursion:
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))             # <<<<<<<<<<<<<<
 *                 if not is_partial_execution:
 *                     if dump_await_recursion:
 */
        __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->await_recursion_depth); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->max_await_recursion_depth); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_7);
        __pyx_t_2 = 0;
        __pyx_t_7 = 0;

        /* "asynq/scheduler.py":213
 *                 dump_await_recursion = _debug_options.DUMP_AWAIT_RECURSION
 *                 if dump_await_recursion:
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %             # <<<<<<<<<<<<<<
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 *                 if not is_partial_execution:
 */
        __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_async_await_recursion_depth_is, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 213, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L4_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          __pyx_t_2 = PyTuple_New(1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_2, 0+1, __pyx_t_7);
          __pyx_t_7 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":212
 *                 # await at some point
 *                 dump_await_recursion = _debug_options.DUMP_AWAIT_RECURSION
 *                 if dump_await_recursion:             # <<<<<<<<<<<<<<
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 */
      }

      /* "asynq/scheduler.py":215
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 *                 if not is_partial_execution:             # <<<<<<<<<<<<<<
 *                     if dump_await_recursion:
 *                         debug.write("@async: await recursion: ignoring await()")
 */
      __pyx_t_9 = ((!(__pyx_v_is_partial_execution != 0)) != 0);
      if (__pyx_t_9) {

        /* "asynq/scheduler.py":216
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 *                 if not is_partial_execution:
 *                     if dump_await_recursion:             # <<<<<<<<<<<<<<
 *                         debug.write("@async: await recursion: ignoring await()")
 *                     return 0
 */
        __pyx_t_9 = (__pyx_v_dump_await_recursion != 0);
        if (__pyx_t_9) {

          /* "asynq/scheduler.py":217
 *                 if not is_partial_execution:
 *                     if dump_await_recursion:
 *                         debug.write("@async: await recursion: ignoring await()")             # <<<<<<<<<<<<<<
 *                     return 0
 *                 self._try_execute_on_new_scheduler(tasks)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 217, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/scheduler.py":216
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 *                 if not is_partial_execution:
 *                     if dump_await_recursion:             # <<<<<<<<<<<<<<
 *                         debug.write("@async: await recursion: ignoring await()")
 *                     return 0
 */
        }

        /* "asynq/scheduler.py":218
 *                     if dump_await_recursion:
 *                         debug.write("@async: await recursion: ignoring await()")
 *                     return 0             # <<<<<<<<<<<<<<
 *                 self._try_execute_on_new_scheduler(tasks)
 *                 # The rest part of tasks will be processed by this scheduler
 */
        __pyx_r = 0;
        goto __pyx_L3_return;

        /* "asynq/scheduler.py":215
 *                     debug.write("@async: await recursion: depth is %d, but the limit is %d" %
 *                                 (self.await_recursion_depth, self.max_await_recursion_depth))
 *                 if not is_partial_execution:             # <<<<<<<<<<<<<<
 *                     if dump_await_recursion:
 *                         debug.write("@async: await recursion: ignoring await()")
 */
      }

      /* "asynq/scheduler.py":219
 *                         debug.write("@async: await recursion: ignoring await()")
 *                     return 0
 *                 self._try_execute_on_new_scheduler(tasks)             # <<<<<<<<<<<<<<
 *                 # The rest part of tasks will be processed by this scheduler
 *             dump_scheduler_state = _debug_options.DUMP_SCHEDULER_STATE
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_try_execute_on_new_scheduler(__pyx_v_self, __pyx_v_tasks, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":207
 *         self.active_task = None
 *         try:
 *             if self.await_recursion_depth > self.max_await_recursion_depth:             # <<<<<<<<<<<<<<
 *                 # We don't want to execute every next task on
 *                 # stack in case all of these tasks invoke
 */
    }

    /* "asynq/scheduler.py":221
 *                 self._try_execute_on_new_scheduler(tasks)
 *                 # The rest part of tasks will be processed by this scheduler
 *             dump_scheduler_state = _debug_options.DUMP_SCHEDULER_STATE             # <<<<<<<<<<<<<<
 *             if is_partial_execution and len(tasks) == 0:
 *                 return 0
 */
    __pyx_t_9 = __pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULER_STATE;
    __pyx_v_dump_scheduler_state = __pyx_t_9;

    /* "asynq/scheduler.py":222
 *                 # The rest part of tasks will be processed by this scheduler
 *             dump_scheduler_state = _debug_options.DUMP_SCHEDULER_STATE
 *             if is_partial_execution and len(tasks) == 0:             # <<<<<<<<<<<<<<
 *                 return 0
 *             for task in tasks:
 */
    __pyx_t_10 = (__pyx_v_is_partial_execution != 0);
    if (__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L11_bool_binop_done;
    }
    if (unlikely(__pyx_v_tasks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 222, __pyx_L4_error)
    }
    __pyx_t_6 = PySet_GET_SIZE(__pyx_v_tasks); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 222, __pyx_L4_error)
    __pyx_t_10 = ((__pyx_t_6 == 0) != 0);
    __pyx_t_9 = __pyx_t_10;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_9) {

      /* "asynq/scheduler.py":223
 *             dump_scheduler_state = _debug_options.DUMP_SCHEDULER_STATE
 *             if is_partial_execution and len(tasks) == 0:
 *                 return 0             # <<<<<<<<<<<<<<
 *             for task in tasks:
 *                 creator = task.creator
 */
      __pyx_r = 0;
      goto __pyx_L3_return;

      /* "asynq/scheduler.py":222
 *                 # The rest part of tasks will be processed by this scheduler
 *             dump_scheduler_state = _debug_options.DUMP_SCHEDULER_STATE
 *             if is_partial_execution and len(tasks) == 0:             # <<<<<<<<<<<<<<
 *                 return 0
 *             for task in tasks:
 */
    }

    /* "asynq/scheduler.py":224
 *             if is_partial_execution and len(tasks) == 0:
 *                 return 0
 *             for task in tasks:             # <<<<<<<<<<<<<<
 *                 creator = task.creator
 *                 if creator is not None:
 */
    __pyx_t_1 = PyObject_GetIter(__pyx_v_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 224, __pyx_L4_error)
    for (;;) {
      {
        __pyx_t_4 = __pyx_t_11(__pyx_t_1);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 224, __pyx_L4_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 224, __pyx_L4_error)
      __Pyx_XDECREF_SET(__pyx_v_task, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "asynq/scheduler.py":225
 *                 return 0
 *             for task in tasks:
 *                 creator = task.creator             # <<<<<<<<<<<<<<
 *                 if creator is not None:
 *                     if task in creator._dependencies:
 */
      __pyx_t_4 = ((PyObject *)__pyx_v_task->creator);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_creator, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "asynq/scheduler.py":226
 *             for task in tasks:
 *                 creator = task.creator
 *                 if creator is not None:             # <<<<<<<<<<<<<<
 *                     if task in creator._dependencies:
 *                         creator._remove_dependency(task)
 */
      __pyx_t_9 = (((PyObject *)__pyx_v_creator) != Py_None);
      __pyx_t_10 = (__pyx_t_9 != 0);
      if (__pyx_t_10) {

        /* "asynq/scheduler.py":227
 *                 creator = task.creator
 *                 if creator is not None:
 *                     if task in creator._dependencies:             # <<<<<<<<<<<<<<
 *                         creator._remove_dependency(task)
 *                 self._schedule(task)
 */
        __pyx_t_10 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_task), __pyx_v_creator->_dependencies, Py_EQ)); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 227, __pyx_L4_error)
        __pyx_t_9 = (__pyx_t_10 != 0);
        if (__pyx_t_9) {

          /* "asynq/scheduler.py":228
 *                 if creator is not None:
 *                     if task in creator._dependencies:
 *                         creator._remove_dependency(task)             # <<<<<<<<<<<<<<
 *                 self._schedule(task)
 *             while True:  # Batch loop
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_creator), __pyx_n_s_remove_dependency); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 228, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_7 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          if (!__pyx_t_7) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 228, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_7); __pyx_t_7 = NULL;
            __Pyx_INCREF(((PyObject *)__pyx_v_task));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, ((PyObject *)__pyx_v_task));
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "asynq/scheduler.py":227
 *                 creator = task.creator
 *                 if creator is not None:
 *                     if task in creator._dependencies:             # <<<<<<<<<<<<<<
 *                         creator._remove_dependency(task)
 *                 self._schedule(task)
 */
        }

        /* "asynq/scheduler.py":226
 *             for task in tasks:
 *                 creator = task.creator
 *                 if creator is not None:             # <<<<<<<<<<<<<<
 *                     if task in creator._dependencies:
 *                         creator._remove_dependency(task)
 */
      }

      /* "asynq/scheduler.py":229
 *                     if task in creator._dependencies:
 *                         creator._remove_dependency(task)
 *                 self._schedule(task)             # <<<<<<<<<<<<<<
 *             while True:  # Batch loop
 *                 while True:  # Task loop
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule(__pyx_v_self, __pyx_v_task); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 229, __pyx_L4_error)

      /* "asynq/scheduler.py":224
 *             if is_partial_execution and len(tasks) == 0:
 *                 return 0
 *             for task in tasks:             # <<<<<<<<<<<<<<
 *                 creator = task.creator
 *                 if creator is not None:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":230
 *                         creator._remove_dependency(task)
 *                 self._schedule(task)
 *             while True:  # Batch loop             # <<<<<<<<<<<<<<
 *                 while True:  # Task loop
 *                     task = self._continue_with_task()
 */
    while (1) {

      /* "asynq/scheduler.py":231
 *                 self._schedule(task)
 *             while True:  # Batch loop
 *                 while True:  # Task loop             # <<<<<<<<<<<<<<
 *                     task = self._continue_with_task()
 *                     if dump_scheduler_state:
 */
      while (1) {

        /* "asynq/scheduler.py":232
 *             while True:  # Batch loop
 *                 while True:  # Task loop
 *                     task = self._continue_with_task()             # <<<<<<<<<<<<<<
 *                     if dump_scheduler_state:
 *                         self.try_time_based_dump(task)
 */
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_continue_with_task(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_task, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "asynq/scheduler.py":233
 *                 while True:  # Task loop
 *                     task = self._continue_with_task()
 *                     if dump_scheduler_state:             # <<<<<<<<<<<<<<
 *                         self.try_time_based_dump(task)
 *                     if task is None:
 */
        __pyx_t_9 = (__pyx_v_dump_scheduler_state != 0);
        if (__pyx_t_9) {

          /* "asynq/scheduler.py":234
 *                     task = self._continue_with_task()
 *                     if dump_scheduler_state:
 *                         self.try_time_based_dump(task)             # <<<<<<<<<<<<<<
 *                     if task is None:
 *                         if _no_blocked_tasks(tasks):
 */
          __pyx_t_12.__pyx_n = 1;
          __pyx_t_12.last_task = ((PyObject *)__pyx_v_task);
          __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->try_time_based_dump(__pyx_v_self, 0, &__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/scheduler.py":233
 *                 while True:  # Task loop
 *                     task = self._continue_with_task()
 *                     if dump_scheduler_state:             # <<<<<<<<<<<<<<
 *                         self.try_time_based_dump(task)
 *                     if task is None:
 */
        }

        /* "asynq/scheduler.py":235
 *                     if dump_scheduler_state:
 *                         self.try_time_based_dump(task)
 *                     if task is None:             # <<<<<<<<<<<<<<
 *                         if _no_blocked_tasks(tasks):
 *                             return 0  # All specified tasks are computed
 */
        __pyx_t_9 = (((PyObject *)__pyx_v_task) == Py_None);
        __pyx_t_10 = (__pyx_t_9 != 0);
        if (__pyx_t_10) {

          /* "asynq/scheduler.py":236
 *                         self.try_time_based_dump(task)
 *                     if task is None:
 *                         if _no_blocked_tasks(tasks):             # <<<<<<<<<<<<<<
 *                             return 0  # All specified tasks are computed
 *                         break  # No more tasks - let's try to flush the batch
 */
          __pyx_t_10 = __pyx_f_5asynq_9scheduler__no_blocked_tasks(__pyx_v_tasks, 0); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 236, __pyx_L4_error)
          __pyx_t_9 = (__pyx_t_10 != 0);
          if (__pyx_t_9) {

            /* "asynq/scheduler.py":237
 *                     if task is None:
 *                         if _no_blocked_tasks(tasks):
 *                             return 0  # All specified tasks are computed             # <<<<<<<<<<<<<<
 *                         break  # No more tasks - let's try to flush the batch
 *                     if is_partial_execution and task.is_computed() and task in tasks:
 */
            __pyx_r = 0;
            goto __pyx_L3_return;

            /* "asynq/scheduler.py":236
 *                         self.try_time_based_dump(task)
 *                     if task is None:
 *                         if _no_blocked_tasks(tasks):             # <<<<<<<<<<<<<<
 *                             return 0  # All specified tasks are computed
 *                         break  # No more tasks - let's try to flush the batch
 */
          }

          /* "asynq/scheduler.py":238
 *                         if _no_blocked_tasks(tasks):
 *                             return 0  # All specified tasks are computed
 *                         break  # No more tasks - let's try to flush the batch             # <<<<<<<<<<<<<<
 *                     if is_partial_execution and task.is_computed() and task in tasks:
 *                         tasks.remove(task)
 */
          goto __pyx_L20_break;

          /* "asynq/scheduler.py":235
 *                     if dump_scheduler_state:
 *                         self.try_time_based_dump(task)
 *                     if task is None:             # <<<<<<<<<<<<<<
 *                         if _no_blocked_tasks(tasks):
 *                             return 0  # All specified tasks are computed
 */
        }

        /* "asynq/scheduler.py":239
 *                             return 0  # All specified tasks are computed
 *                         break  # No more tasks - let's try to flush the batch
 *                     if is_partial_execution and task.is_computed() and task in tasks:             # <<<<<<<<<<<<<<
 *                         tasks.remove(task)
 *                         if len(tasks) == 0:
 */
        __pyx_t_10 = (__pyx_v_is_partial_execution != 0);
        if (__pyx_t_10) {
        } else {
          __pyx_t_9 = __pyx_t_10;
          goto __pyx_L25_bool_binop_done;
        }
        __pyx_t_10 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 239, __pyx_L4_error)
        __pyx_t_13 = (__pyx_t_10 != 0);
        if (__pyx_t_13) {
        } else {
          __pyx_t_9 = __pyx_t_13;
          goto __pyx_L25_bool_binop_done;
        }
        __pyx_t_13 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_task), __pyx_v_tasks, Py_EQ)); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 239, __pyx_L4_error)
        __pyx_t_10 = (__pyx_t_13 != 0);
        __pyx_t_9 = __pyx_t_10;
        __pyx_L25_bool_binop_done:;
        if (__pyx_t_9) {

          /* "asynq/scheduler.py":240
 *                         break  # No more tasks - let's try to flush the batch
 *                     if is_partial_execution and task.is_computed() and task in tasks:
 *                         tasks.remove(task)             # <<<<<<<<<<<<<<
 *                         if len(tasks) == 0:
 *                             return 0  # All specified tasks are computed
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_tasks, __pyx_n_s_remove); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 240, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_2 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_2) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 240, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
            __Pyx_INCREF(((PyObject *)__pyx_v_task));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, ((PyObject *)__pyx_v_task));
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 240, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/scheduler.py":241
 *                     if is_partial_execution and task.is_computed() and task in tasks:
 *                         tasks.remove(task)
 *                         if len(tasks) == 0:             # <<<<<<<<<<<<<<
 *                             return 0  # All specified tasks are computed
 *                 try:
 */
          if (unlikely(__pyx_v_tasks == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
            __PYX_ERR(0, 241, __pyx_L4_error)
          }
          __pyx_t_6 = PySet_GET_SIZE(__pyx_v_tasks); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 241, __pyx_L4_error)
          __pyx_t_9 = ((__pyx_t_6 == 0) != 0);
          if (__pyx_t_9) {

            /* "asynq/scheduler.py":242
 *                         tasks.remove(task)
 *                         if len(tasks) == 0:
 *                             return 0  # All specified tasks are computed             # <<<<<<<<<<<<<<
 *                 try:
 *                     batch = self._continue_with_batch(is_partial_execution)
 */
            __pyx_r = 0;
            goto __pyx_L3_return;

            /* "asynq/scheduler.py":241
 *                     if is_partial_execution and task.is_computed() and task in tasks:
 *                         tasks.remove(task)
 *                         if len(tasks) == 0:             # <<<<<<<<<<<<<<
 *                             return 0  # All specified tasks are computed
 *                 try:
 */
          }

          /* "asynq/scheduler.py":239
 *                             return 0  # All specified tasks are computed
 *                         break  # No more tasks - let's try to flush the batch
 *                     if is_partial_execution and task.is_computed() and task in tasks:             # <<<<<<<<<<<<<<
 *                         tasks.remove(task)
 *                         if len(tasks) == 0:
 */
        }
      }
      __pyx_L20_break:;

      /* "asynq/scheduler.py":243
 *                         if len(tasks) == 0:
 *                             return 0  # All specified tasks are computed
 *                 try:             # <<<<<<<<<<<<<<
 *                     batch = self._continue_with_batch(is_partial_execution)
 *                     if batch is None:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        /*try:*/ {

          /* "asynq/scheduler.py":244
 *                             return 0  # All specified tasks are computed
 *                 try:
 *                     batch = self._continue_with_batch(is_partial_execution)             # <<<<<<<<<<<<<<
 *                     if batch is None:
 *                         return 0  # Nothing more to execute
 */
          __pyx_t_17.__pyx_n = 1;
          __pyx_t_17.assert_no_batch = __pyx_v_is_partial_execution;
          __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_continue_with_batch(__pyx_v_self, &__pyx_t_17)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L29_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_batch, ((struct __pyx_obj_5asynq_8batching_BatchBase *)__pyx_t_1));
          __pyx_t_1 = 0;

          /* "asynq/scheduler.py":245
 *                 try:
 *                     batch = self._continue_with_batch(is_partial_execution)
 *                     if batch is None:             # <<<<<<<<<<<<<<
 *                         return 0  # Nothing more to execute
 *                 except AssertionError:
 */
          __pyx_t_9 = (((PyObject *)__pyx_v_batch) == Py_None);
          __pyx_t_10 = (__pyx_t_9 != 0);
          if (__pyx_t_10) {

            /* "asynq/scheduler.py":246
 *                     batch = self._continue_with_batch(is_partial_execution)
 *                     if batch is None:
 *                         return 0  # Nothing more to execute             # <<<<<<<<<<<<<<
 *                 except AssertionError:
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:
 */
            __pyx_r = 0;
            goto __pyx_L33_try_return;

            /* "asynq/scheduler.py":245
 *                 try:
 *                     batch = self._continue_with_batch(is_partial_execution)
 *                     if batch is None:             # <<<<<<<<<<<<<<
 *                         return 0  # Nothing more to execute
 *                 except AssertionError:
 */
          }

          /* "asynq/scheduler.py":243
 *                         if len(tasks) == 0:
 *                             return 0  # All specified tasks are computed
 *                 try:             # <<<<<<<<<<<<<<
 *                     batch = self._continue_with_batch(is_partial_execution)
 *                     if batch is None:
 */
        }
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        goto __pyx_L36_try_end;
        __pyx_L29_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":247
 *                     if batch is None:
 *                         return 0  # Nothing more to execute
 *                 except AssertionError:             # <<<<<<<<<<<<<<
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:
 *                         debug.write('Scheduler is awaiting for:')
 */
        __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
        if (__pyx_t_8) {
          __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_4, &__pyx_t_3) < 0) __PYX_ERR(0, 247, __pyx_L31_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_3);

          /* "asynq/scheduler.py":248
 *                         return 0  # Nothing more to execute
 *                 except AssertionError:
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:             # <<<<<<<<<<<<<<
 *                         debug.write('Scheduler is awaiting for:')
 *                         for task in tasks:
 */
          __pyx_t_10 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_PRE_ERROR_STATE != 0);
          if (__pyx_t_10) {

            /* "asynq/scheduler.py":249
 *                 except AssertionError:
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:
 *                         debug.write('Scheduler is awaiting for:')             # <<<<<<<<<<<<<<
 *                         for task in tasks:
 *                             task.dump(2)
 */
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 249, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 249, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "asynq/scheduler.py":250
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:
 *                         debug.write('Scheduler is awaiting for:')
 *                         for task in tasks:             # <<<<<<<<<<<<<<
 *                             task.dump(2)
 *                         debug.write('----------------------------------------------------------------------\n')
 */
            __pyx_t_2 = PyObject_GetIter(__pyx_v_tasks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_11 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 250, __pyx_L31_except_error)
            for (;;) {
              {
                __pyx_t_7 = __pyx_t_11(__pyx_t_2);
                if (unlikely(!__pyx_t_7)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 250, __pyx_L31_except_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_7);
              }
              if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 250, __pyx_L31_except_error)
              __Pyx_XDECREF_SET(__pyx_v_task, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_7));
              __pyx_t_7 = 0;

              /* "asynq/scheduler.py":251
 *                         debug.write('Scheduler is awaiting for:')
 *                         for task in tasks:
 *                             task.dump(2)             # <<<<<<<<<<<<<<
 *                         debug.write('----------------------------------------------------------------------\n')
 *                         raise
 */
              __pyx_t_18.__pyx_n = 1;
              __pyx_t_18.indent = 2;
              __pyx_t_7 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.dump(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0, &__pyx_t_18); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 251, __pyx_L31_except_error)
              __Pyx_GOTREF(__pyx_t_7);
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

              /* "asynq/scheduler.py":250
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:
 *                         debug.write('Scheduler is awaiting for:')
 *                         for task in tasks:             # <<<<<<<<<<<<<<
 *                             task.dump(2)
 *                         debug.write('----------------------------------------------------------------------\n')
 */
            }
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "asynq/scheduler.py":252
 *                         for task in tasks:
 *                             task.dump(2)
 *                         debug.write('----------------------------------------------------------------------\n')             # <<<<<<<<<<<<<<
 *                         raise
 *         finally:
 */
            __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 252, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 252, __pyx_L31_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "asynq/scheduler.py":253
 *                             task.dump(2)
 *                         debug.write('----------------------------------------------------------------------\n')
 *                         raise             # <<<<<<<<<<<<<<
 *         finally:
 *             self.active_task = current_active_task
 */
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_XGIVEREF(__pyx_t_3);
            __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_4, __pyx_t_3);
            __pyx_t_1 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; 
            __PYX_ERR(0, 253, __pyx_L31_except_error)

            /* "asynq/scheduler.py":248
 *                         return 0  # Nothing more to execute
 *                 except AssertionError:
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:             # <<<<<<<<<<<<<<
 *                         debug.write('Scheduler is awaiting for:')
 *                         for task in tasks:
 */
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L30_exception_handled;
        }
        goto __pyx_L31_except_error;
        __pyx_L31_except_error:;

        /* "asynq/scheduler.py":243
 *                         if len(tasks) == 0:
 *                             return 0  # All specified tasks are computed
 *                 try:             # <<<<<<<<<<<<<<
 *                     batch = self._continue_with_batch(is_partial_execution)
 *                     if batch is None:
 */
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        goto __pyx_L4_error;
        __pyx_L33_try_return:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        goto __pyx_L3_return;
        __pyx_L30_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        __pyx_L36_try_end:;
      }
    }
  }

  /* "asynq/scheduler.py":255
 *                         raise
 *         finally:
 *             self.active_task = current_active_task             # <<<<<<<<<<<<<<
 *             self.await_recursion_depth -= 1
 *         return 0
 */
  /*finally:*/ {
    /*normal exit:*/{
      __Pyx_INCREF(((PyObject *)__pyx_v_current_active_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_current_active_task));
      __Pyx_GOTREF(__pyx_v_self->active_task);
      __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
      __pyx_v_self->active_task = __pyx_v_current_active_task;

      /* "asynq/scheduler.py":256
 *         finally:
 *             self.active_task = current_active_task
 *             self.await_recursion_depth -= 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_v_self->await_recursion_depth = (__pyx_v_self->await_recursion_depth - 1);
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_21);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_23);
      __pyx_t_8 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
      {

        /* "asynq/scheduler.py":255
 *                         raise
 *         finally:
 *             self.active_task = current_active_task             # <<<<<<<<<<<<<<
 *             self.await_recursion_depth -= 1
 *         return 0
 */
        __Pyx_INCREF(((PyObject *)__pyx_v_current_active_task));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_current_active_task));
        __Pyx_GOTREF(__pyx_v_self->active_task);
        __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
        __pyx_v_self->active_task = __pyx_v_current_active_task;

        /* "asynq/scheduler.py":256
 *         finally:
 *             self.active_task = current_active_task
 *             self.await_recursion_depth -= 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
        __pyx_v_self->await_recursion_depth = (__pyx_v_self->await_recursion_depth - 1);
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_XGIVEREF(__pyx_t_23);
        __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
      }
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_ErrRestore(__pyx_t_16, __pyx_t_15, __pyx_t_14);
      __pyx_t_16 = 0; __pyx_t_15 = 0; __pyx_t_14 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_20;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_19 = __pyx_r;

      /* "asynq/scheduler.py":255
 *                         raise
 *         finally:
 *             self.active_task = current_active_task             # <<<<<<<<<<<<<<
 *             self.await_recursion_depth -= 1
 *         return 0
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_current_active_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_current_active_task));
      __Pyx_GOTREF(__pyx_v_self->active_task);
      __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
      __pyx_v_self->active_task = __pyx_v_current_active_task;

      /* "asynq/scheduler.py":256
 *         finally:
 *             self.active_task = current_active_task
 *             self.await_recursion_depth -= 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_v_self->await_recursion_depth = (__pyx_v_self->await_recursion_depth - 1);
      __pyx_r = __pyx_t_19;
      goto __pyx_L0;
    }
    __pyx_L5:;
  }

  /* "asynq/scheduler.py":257
 *             self.active_task = current_active_task
 *             self.await_recursion_depth -= 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _try_execute_on_new_scheduler(self, tasks):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":200
 *         return 0
 * 
 *     def _execute(self, is_partial_execution, tasks):             # <<<<<<<<<<<<<<
 *         """Implements task execution loop."""
 *         # Must never throw an exception
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_task);
  __Pyx_XDECREF((PyObject *)__pyx_v_creator);
  __Pyx_XDECREF((PyObject *)__pyx_v_current_active_task);
  __Pyx_XDECREF((PyObject *)__pyx_v_batch);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21_execute(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_20_execute[] = "Implements task execution loop.";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21_execute(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_is_partial_execution;
  PyObject *__pyx_v_tasks = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_execute (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_is_partial_execution,&__pyx_n_s_tasks,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_is_partial_execution)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_tasks)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_execute", 1, 2, 2, 1); __PYX_ERR(0, 200, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_execute") < 0)) __PYX_ERR(0, 200, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_is_partial_execution = __Pyx_PyObject_IsTrue(values[0]); if (unlikely((__pyx_v_is_partial_execution == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 200, __pyx_L3_error)
    __pyx_v_tasks = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_execute", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 200, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tasks), (&PySet_Type), 1, "tasks", 1))) __PYX_ERR(0, 200, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_20_execute(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_is_partial_execution, __pyx_v_tasks);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_20_execute(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_is_partial_execution, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_execute", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__execute(__pyx_v_self, __pyx_v_is_partial_execution, __pyx_v_tasks, 1); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 200, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":259
 *         return 0
 * 
 *     def _try_execute_on_new_scheduler(self, tasks):             # <<<<<<<<<<<<<<
 *         new_scheduler_tasks = []
 *         for task in tasks:
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_23_try_execute_on_new_scheduler(PyObject *__pyx_v_self, PyObject *__pyx_v_tasks); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__try_execute_on_new_scheduler(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks, int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  PyObject *__pyx_v_new_scheduler_tasks = 0;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_new_scheduler = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("_try_execute_on_new_scheduler", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_try_execute_on_new_scheduler); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_23_try_execute_on_new_scheduler)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_tasks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_tasks);
        __Pyx_GIVEREF(__pyx_v_tasks);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_tasks);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":260
 * 
 *     def _try_execute_on_new_scheduler(self, tasks):
 *         new_scheduler_tasks = []             # <<<<<<<<<<<<<<
 *         for task in tasks:
 *             if task.scheduler is None:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_new_scheduler_tasks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":261
 *     def _try_execute_on_new_scheduler(self, tasks):
 *         new_scheduler_tasks = []
 *         for task in tasks:             # <<<<<<<<<<<<<<
 *             if task.scheduler is None:
 *                 new_scheduler_tasks.append(task)
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 261, __pyx_L1_error)
  for (;;) {
    {
      __pyx_t_2 = __pyx_t_6(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 261, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 261, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_task, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":262
 *         new_scheduler_tasks = []
 *         for task in tasks:
 *             if task.scheduler is None:             # <<<<<<<<<<<<<<
 *                 new_scheduler_tasks.append(task)
 *         for task in new_scheduler_tasks:
 */
    __pyx_t_7 = (((PyObject *)__pyx_v_task->scheduler) == Py_None);
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "asynq/scheduler.py":263
 *         for task in tasks:
 *             if task.scheduler is None:
 *                 new_scheduler_tasks.append(task)             # <<<<<<<<<<<<<<
 *         for task in new_scheduler_tasks:
 *             tasks.remove(task)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_new_scheduler_tasks, ((PyObject *)__pyx_v_task)); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 263, __pyx_L1_error)

      /* "asynq/scheduler.py":262
 *         new_scheduler_tasks = []
 *         for task in tasks:
 *             if task.scheduler is None:             # <<<<<<<<<<<<<<
 *                 new_scheduler_tasks.append(task)
 *         for task in new_scheduler_tasks:
 */
    }

    /* "asynq/scheduler.py":261
 *     def _try_execute_on_new_scheduler(self, tasks):
 *         new_scheduler_tasks = []
 *         for task in tasks:             # <<<<<<<<<<<<<<
 *             if task.scheduler is None:
 *                 new_scheduler_tasks.append(task)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":264
 *             if task.scheduler is None:
 *                 new_scheduler_tasks.append(task)
 *         for task in new_scheduler_tasks:             # <<<<<<<<<<<<<<
 *             tasks.remove(task)
 * 
 */
  __pyx_t_1 = __pyx_v_new_scheduler_tasks; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
  for (;;) {
    if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 264, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_task, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":265
 *                 new_scheduler_tasks.append(task)
 *         for task in new_scheduler_tasks:
 *             tasks.remove(task)             # <<<<<<<<<<<<<<
 * 
 *         if _debug_options.DUMP_AWAIT_RECURSION:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_tasks, __pyx_n_s_remove); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_task));
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 265, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/scheduler.py":264
 *             if task.scheduler is None:
 *                 new_scheduler_tasks.append(task)
 *         for task in new_scheduler_tasks:             # <<<<<<<<<<<<<<
 *             tasks.remove(task)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":267
 *             tasks.remove(task)
 * 
 *         if _debug_options.DUMP_AWAIT_RECURSION:             # <<<<<<<<<<<<<<
 *             debug.write("@async: await recursion: going to run %d tasks on a new scheduler" % len(new_scheduler_tasks))
 * 
 */
  __pyx_t_8 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_AWAIT_RECURSION != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":268
 * 
 *         if _debug_options.DUMP_AWAIT_RECURSION:
 *             debug.write("@async: await recursion: going to run %d tasks on a new scheduler" % len(new_scheduler_tasks))             # <<<<<<<<<<<<<<
 * 
 *         if len(new_scheduler_tasks) == 0:
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_10 = PyList_GET_SIZE(__pyx_v_new_scheduler_tasks); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_await_recursion_going_to, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_2) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":267
 *             tasks.remove(task)
 * 
 *         if _debug_options.DUMP_AWAIT_RECURSION:             # <<<<<<<<<<<<<<
 *             debug.write("@async: await recursion: going to run %d tasks on a new scheduler" % len(new_scheduler_tasks))
 * 
 */
  }

  /* "asynq/scheduler.py":270
 *             debug.write("@async: await recursion: going to run %d tasks on a new scheduler" % len(new_scheduler_tasks))
 * 
 *         if len(new_scheduler_tasks) == 0:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
  __pyx_t_10 = PyList_GET_SIZE(__pyx_v_new_scheduler_tasks); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_10 == 0) != 0);
  if (__pyx_t_8) {
    goto __pyx_L9;
  }

  /* "asynq/scheduler.py":273
 *             pass
 *         else:
 *             new_scheduler = self._clone()             # <<<<<<<<<<<<<<
 *             try:
 *                 with new_scheduler.activate(True, False):
 */
  /*else*/ {
    __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_clone(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5asynq_9scheduler_TaskScheduler))))) __PYX_ERR(0, 273, __pyx_L1_error)
    __pyx_v_new_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynq/scheduler.py":274
 *         else:
 *             new_scheduler = self._clone()
 *             try:             # <<<<<<<<<<<<<<
 *                 with new_scheduler.activate(True, False):
 *                     new_scheduler.await(new_scheduler_tasks)
 */
    /*try:*/ {

      /* "asynq/scheduler.py":275
 *             new_scheduler = self._clone()
 *             try:
 *                 with new_scheduler.activate(True, False):             # <<<<<<<<<<<<<<
 *                     new_scheduler.await(new_scheduler_tasks)
 *             finally:
 */
      /*with:*/ {
        __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_new_scheduler->__pyx_vtab)->activate(__pyx_v_new_scheduler, 1, 0, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 275, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 275, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L13_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 275, __pyx_L13_error)
        }
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        /*try:*/ {
          {
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
            __Pyx_XGOTREF(__pyx_t_12);
            __Pyx_XGOTREF(__pyx_t_13);
            __Pyx_XGOTREF(__pyx_t_14);
            /*try:*/ {

              /* "asynq/scheduler.py":276
 *             try:
 *                 with new_scheduler.activate(True, False):
 *                     new_scheduler.await(new_scheduler_tasks)             # <<<<<<<<<<<<<<
 *             finally:
 *                 self._accept_changes(new_scheduler)
 */
              __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_new_scheduler), __pyx_n_s_await); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 276, __pyx_L17_error)
              __Pyx_GOTREF(__pyx_t_3);
              __pyx_t_5 = NULL;
              if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
                __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
                if (likely(__pyx_t_5)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                  __Pyx_INCREF(__pyx_t_5);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_3, function);
                }
              }
              if (!__pyx_t_5) {
                __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_new_scheduler_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_1);
              } else {
                __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_4);
                __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
                __Pyx_INCREF(__pyx_v_new_scheduler_tasks);
                __Pyx_GIVEREF(__pyx_v_new_scheduler_tasks);
                PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_new_scheduler_tasks);
                __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 276, __pyx_L17_error)
                __Pyx_GOTREF(__pyx_t_1);
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              }
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

              /* "asynq/scheduler.py":275
 *             new_scheduler = self._clone()
 *             try:
 *                 with new_scheduler.activate(True, False):             # <<<<<<<<<<<<<<
 *                     new_scheduler.await(new_scheduler_tasks)
 *             finally:
 */
            }
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            goto __pyx_L24_try_end;
            __pyx_L17_error:;
            __Pyx_PyThreadState_assign
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            /*except:*/ {
              __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._try_execute_on_new_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
              if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 275, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_GOTREF(__pyx_t_4);
              __pyx_t_5 = PyTuple_Pack(3, __pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 275, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_5);
              __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_5, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
              if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 275, __pyx_L19_except_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_15);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (__pyx_t_8 < 0) __PYX_ERR(0, 275, __pyx_L19_except_error)
              __pyx_t_7 = ((!(__pyx_t_8 != 0)) != 0);
              if (__pyx_t_7) {
                __Pyx_GIVEREF(__pyx_t_1);
                __Pyx_GIVEREF(__pyx_t_3);
                __Pyx_XGIVEREF(__pyx_t_4);
                __Pyx_ErrRestoreWithState(__pyx_t_1, __pyx_t_3, __pyx_t_4);
                __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0; 
                __PYX_ERR(0, 275, __pyx_L19_except_error)
              }
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
              goto __pyx_L18_exception_handled;
            }
            __pyx_L19_except_error:;
            __Pyx_PyThreadState_assign
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            goto __pyx_L11_error;
            __pyx_L18_exception_handled:;
            __Pyx_PyThreadState_assign
            __Pyx_XGIVEREF(__pyx_t_12);
            __Pyx_XGIVEREF(__pyx_t_13);
            __Pyx_XGIVEREF(__pyx_t_14);
            __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
            __pyx_L24_try_end:;
          }
        }
        /*finally:*/ {
          /*normal exit:*/{
            if (__pyx_t_11) {
              __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_tuple__10, NULL);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 275, __pyx_L11_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            goto __pyx_L16;
          }
          __pyx_L16:;
        }
        goto __pyx_L28;
        __pyx_L13_error:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L11_error;
        __pyx_L28:;
      }
    }

    /* "asynq/scheduler.py":278
 *                     new_scheduler.await(new_scheduler_tasks)
 *             finally:
 *                 self._accept_changes(new_scheduler)             # <<<<<<<<<<<<<<
 * 
 *     def schedule(self, task):
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_t_4 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_accept_changes(__pyx_v_self, __pyx_v_new_scheduler, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L12;
      }
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __pyx_L11_error:;
        __pyx_t_11 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_15 = 0; __pyx_t_19 = 0;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_12, &__pyx_t_15, &__pyx_t_19);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_14, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_14, &__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_19);
        __pyx_t_16 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
        {
          __pyx_t_4 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_accept_changes(__pyx_v_self, __pyx_v_new_scheduler, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 278, __pyx_L30_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_15, __pyx_t_19);
        }
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ErrRestore(__pyx_t_11, __pyx_t_14, __pyx_t_13);
        __pyx_t_11 = 0; __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_15 = 0; __pyx_t_19 = 0;
        __pyx_lineno = __pyx_t_16; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
        goto __pyx_L1_error;
        __pyx_L30_error:;
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_15, __pyx_t_19);
        }
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_12 = 0; __pyx_t_15 = 0; __pyx_t_19 = 0;
        goto __pyx_L1_error;
      }
      __pyx_L12:;
    }
  }
  __pyx_L9:;

  /* "asynq/scheduler.py":259
 *         return 0
 * 
 *     def _try_execute_on_new_scheduler(self, tasks):             # <<<<<<<<<<<<<<
 *         new_scheduler_tasks = []
 *         for task in tasks:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._try_execute_on_new_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_task);
  __Pyx_XDECREF(__pyx_v_new_scheduler_tasks);
  __Pyx_XDECREF((PyObject *)__pyx_v_new_scheduler);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_23_try_execute_on_new_scheduler(PyObject *__pyx_v_self, PyObject *__pyx_v_tasks); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_23_try_execute_on_new_scheduler(PyObject *__pyx_v_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_try_execute_on_new_scheduler (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tasks), (&PySet_Type), 1, "tasks", 1))) __PYX_ERR(0, 259, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_22_try_execute_on_new_scheduler(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject*)__pyx_v_tasks));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_22_try_execute_on_new_scheduler(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_try_execute_on_new_scheduler", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__try_execute_on_new_scheduler(__pyx_v_self, __pyx_v_tasks, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._try_execute_on_new_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":280
 *                 self._accept_changes(new_scheduler)
 * 
 *     def schedule(self, task):             # <<<<<<<<<<<<<<
 *         self._schedule(task)
 * 
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_25schedule(PyObject *__pyx_v_self, PyObject *__pyx_v_task); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("schedule", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_schedule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_25schedule)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_task));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_task));
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":281
 * 
 *     def schedule(self, task):
 *         self._schedule(task)             # <<<<<<<<<<<<<<
 * 
 *     def _schedule(self, task):
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule(__pyx_v_self, __pyx_v_task); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 281, __pyx_L1_error)

  /* "asynq/scheduler.py":280
 *                 self._accept_changes(new_scheduler)
 * 
 *     def schedule(self, task):             # <<<<<<<<<<<<<<
 *         self._schedule(task)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.schedule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_25schedule(PyObject *__pyx_v_self, PyObject *__pyx_v_task); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_25schedule(PyObject *__pyx_v_self, PyObject *__pyx_v_task) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schedule (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_task), __pyx_ptype_5asynq_10async_task_AsyncTask, 1, "task", 0))) __PYX_ERR(0, 280, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_24schedule(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_v_task));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_24schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("schedule", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler_schedule(__pyx_v_self, __pyx_v_task, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.schedule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":283
 *         self._schedule(task)
 * 
 *     def _schedule(self, task):             # <<<<<<<<<<<<<<
 *         if task.is_blocked():
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 */

static int __pyx_f_5asynq_9scheduler_13TaskScheduler__schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("_schedule", 0);

  /* "asynq/scheduler.py":284
 * 
 *     def _schedule(self, task):
 *         if task.is_blocked():             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->is_blocked(__pyx_v_task, 0); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/scheduler.py":285
 *     def _schedule(self, task):
 *         if task.is_blocked():
 *             if _debug_options.DUMP_SCHEDULE_TASK:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))
 *             assert task.scheduler is self, \
 */
    __pyx_t_2 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULE_TASK != 0);
    if (__pyx_t_2) {

      /* "asynq/scheduler.py":286
 *         if task.is_blocked():
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))             # <<<<<<<<<<<<<<
 *             assert task.scheduler is self, \
 *                 'Task %s is already scheduled on another scheduler.' % debug.str(task)
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_str); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_task));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_task));
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_async_can_t_schedule_blocked_ta, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/scheduler.py":285
 *     def _schedule(self, task):
 *         if task.is_blocked():
 *             if _debug_options.DUMP_SCHEDULE_TASK:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))
 *             assert task.scheduler is self, \
 */
    }

    /* "asynq/scheduler.py":287
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))
 *             assert task.scheduler is self, \             # <<<<<<<<<<<<<<
 *                 'Task %s is already scheduled on another scheduler.' % debug.str(task)
 *             return False
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_2 = (__pyx_v_task->scheduler == __pyx_v_self);
      if (unlikely(!(__pyx_t_2 != 0))) {

        /* "asynq/scheduler.py":288
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))
 *             assert task.scheduler is self, \
 *                 'Task %s is already scheduled on another scheduler.' % debug.str(task)             # <<<<<<<<<<<<<<
 *             return False
 *         if task.is_computed():
 */
        __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 288, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 288, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_8, function);
          }
        }
        if (!__pyx_t_5) {
          __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_8, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
        } else {
          __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_task));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
          PyTuple_SET_ITEM(__pyx_t_7, 0+1, ((PyObject *)__pyx_v_task));
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_Task_s_is_already_scheduled_on_a, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 288, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        PyErr_SetObject(PyExc_AssertionError, __pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 287, __pyx_L1_error)
      }
    }
    #endif

    /* "asynq/scheduler.py":289
 *             assert task.scheduler is self, \
 *                 'Task %s is already scheduled on another scheduler.' % debug.str(task)
 *             return False             # <<<<<<<<<<<<<<
 *         if task.is_computed():
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "asynq/scheduler.py":284
 * 
 *     def _schedule(self, task):
 *         if task.is_blocked():             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule blocked task %s" % debug.str(task))
 */
  }

  /* "asynq/scheduler.py":290
 *                 'Task %s is already scheduled on another scheduler.' % debug.str(task)
 *             return False
 *         if task.is_computed():             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 290, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":291
 *             return False
 *         if task.is_computed():
 *             if _debug_options.DUMP_SCHEDULE_TASK:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))
 *             return False
 */
    __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULE_TASK != 0);
    if (__pyx_t_1) {

      /* "asynq/scheduler.py":292
 *         if task.is_computed():
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))             # <<<<<<<<<<<<<<
 *             return False
 *         return self._schedule_without_checks(task)
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_task));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, ((PyObject *)__pyx_v_task));
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_can_t_schedule_computed_t, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 292, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (!__pyx_t_3) {
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_8);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_6, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 292, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "asynq/scheduler.py":291
 *             return False
 *         if task.is_computed():
 *             if _debug_options.DUMP_SCHEDULE_TASK:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))
 *             return False
 */
    }

    /* "asynq/scheduler.py":293
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))
 *             return False             # <<<<<<<<<<<<<<
 *         return self._schedule_without_checks(task)
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "asynq/scheduler.py":290
 *                 'Task %s is already scheduled on another scheduler.' % debug.str(task)
 *             return False
 *         if task.is_computed():             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))
 */
  }

  /* "asynq/scheduler.py":294
 *                 debug.write("@async: can't schedule computed task %s" % debug.str(task))
 *             return False
 *         return self._schedule_without_checks(task)             # <<<<<<<<<<<<<<
 * 
 *     def _schedule_without_checks(self, task):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule_without_checks(__pyx_v_self, __pyx_v_task); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 294, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "asynq/scheduler.py":283
 *         self._schedule(task)
 * 
 *     def _schedule(self, task):             # <<<<<<<<<<<<<<
 *         if task.is_blocked():
 *             if _debug_options.DUMP_SCHEDULE_TASK:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._schedule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":296
 *         return self._schedule_without_checks(task)
 * 
 *     def _schedule_without_checks(self, task):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_SCHEDULE_TASK:
 *             debug.write('@async: scheduling task %s' % debug.str(task))
 */

static int __pyx_f_5asynq_9scheduler_13TaskScheduler__schedule_without_checks(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("_schedule_without_checks", 0);

  /* "asynq/scheduler.py":297
 * 
 *     def _schedule_without_checks(self, task):
 *         if _debug_options.DUMP_SCHEDULE_TASK:             # <<<<<<<<<<<<<<
 *             debug.write('@async: scheduling task %s' % debug.str(task))
 * 
 */
  __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULE_TASK != 0);
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":298
 *     def _schedule_without_checks(self, task):
 *         if _debug_options.DUMP_SCHEDULE_TASK:
 *             debug.write('@async: scheduling task %s' % debug.str(task))             # <<<<<<<<<<<<<<
 * 
 *         if task.scheduler is None:
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_task));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, ((PyObject *)__pyx_v_task));
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_scheduling_task_s, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 298, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/scheduler.py":297
 * 
 *     def _schedule_without_checks(self, task):
 *         if _debug_options.DUMP_SCHEDULE_TASK:             # <<<<<<<<<<<<<<
 *             debug.write('@async: scheduling task %s' % debug.str(task))
 * 
 */
  }

  /* "asynq/scheduler.py":300
 *             debug.write('@async: scheduling task %s' % debug.str(task))
 * 
 *         if task.scheduler is None:             # <<<<<<<<<<<<<<
 *             task.scheduler = self
 *         else:
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_task->scheduler) == Py_None);
  __pyx_t_8 = (__pyx_t_1 != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":301
 * 
 *         if task.scheduler is None:
 *             task.scheduler = self             # <<<<<<<<<<<<<<
 *         else:
 *             assert task.scheduler is self, 'Task is already scheduled on another scheduler.'
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_task->scheduler);
    __Pyx_DECREF(((PyObject *)__pyx_v_task->scheduler));
    __pyx_v_task->scheduler = __pyx_v_self;

    /* "asynq/scheduler.py":300
 *             debug.write('@async: scheduling task %s' % debug.str(task))
 * 
 *         if task.scheduler is None:             # <<<<<<<<<<<<<<
 *             task.scheduler = self
 *         else:
 */
    goto __pyx_L4;
  }

  /* "asynq/scheduler.py":303
 *             task.scheduler = self
 *         else:
 *             assert task.scheduler is self, 'Task is already scheduled on another scheduler.'             # <<<<<<<<<<<<<<
 * 
 *         if not task.is_scheduled:
 */
  /*else*/ {
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      __pyx_t_8 = (__pyx_v_task->scheduler == __pyx_v_self);
      if (unlikely(!(__pyx_t_8 != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_s_Task_is_already_scheduled_on_ano);
        __PYX_ERR(0, 303, __pyx_L1_error)
      }
    }
    #endif
  }
  __pyx_L4:;

  /* "asynq/scheduler.py":305
 *             assert task.scheduler is self, 'Task is already scheduled on another scheduler.'
 * 
 *         if not task.is_scheduled:             # <<<<<<<<<<<<<<
 *             task.is_scheduled = True
 *             self._new_tasks.append(task)
 */
  __pyx_t_8 = ((!(__pyx_v_task->is_scheduled != 0)) != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":306
 * 
 *         if not task.is_scheduled:
 *             task.is_scheduled = True             # <<<<<<<<<<<<<<
 *             self._new_tasks.append(task)
 *         return True
 */
    __pyx_v_task->is_scheduled = 1;

    /* "asynq/scheduler.py":307
 *         if not task.is_scheduled:
 *             task.is_scheduled = True
 *             self._new_tasks.append(task)             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
    if (unlikely(__pyx_v_self->_new_tasks == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(0, 307, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_self->_new_tasks, ((PyObject *)__pyx_v_task)); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 307, __pyx_L1_error)

    /* "asynq/scheduler.py":305
 *             assert task.scheduler is self, 'Task is already scheduled on another scheduler.'
 * 
 *         if not task.is_scheduled:             # <<<<<<<<<<<<<<
 *             task.is_scheduled = True
 *             self._new_tasks.append(task)
 */
  }

  /* "asynq/scheduler.py":308
 *             task.is_scheduled = True
 *             self._new_tasks.append(task)
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def _try_schedule(self, task, enqueue_new_tasks):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "asynq/scheduler.py":296
 *         return self._schedule_without_checks(task)
 * 
 *     def _schedule_without_checks(self, task):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_SCHEDULE_TASK:
 *             debug.write('@async: scheduling task %s' % debug.str(task))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._schedule_without_checks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":310
 *         return True
 * 
 *     def _try_schedule(self, task, enqueue_new_tasks):             # <<<<<<<<<<<<<<
 *         if task.is_computed() or task.is_blocked():
 *             return False
 */

static int __pyx_f_5asynq_9scheduler_13TaskScheduler__try_schedule(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task, int __pyx_v_enqueue_new_tasks) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_try_schedule", 0);

  /* "asynq/scheduler.py":311
 * 
 *     def _try_schedule(self, task, enqueue_new_tasks):
 *         if task.is_computed() or task.is_blocked():             # <<<<<<<<<<<<<<
 *             return False
 *         if enqueue_new_tasks:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->__pyx_base.is_computed(((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_v_task), 0); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->is_blocked(__pyx_v_task, 0); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":312
 *     def _try_schedule(self, task, enqueue_new_tasks):
 *         if task.is_computed() or task.is_blocked():
 *             return False             # <<<<<<<<<<<<<<
 *         if enqueue_new_tasks:
 *             self._enqueue_new_tasks()
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "asynq/scheduler.py":311
 * 
 *     def _try_schedule(self, task, enqueue_new_tasks):
 *         if task.is_computed() or task.is_blocked():             # <<<<<<<<<<<<<<
 *             return False
 *         if enqueue_new_tasks:
 */
  }

  /* "asynq/scheduler.py":313
 *         if task.is_computed() or task.is_blocked():
 *             return False
 *         if enqueue_new_tasks:             # <<<<<<<<<<<<<<
 *             self._enqueue_new_tasks()
 *         return self._schedule_without_checks(task)
 */
  __pyx_t_1 = (__pyx_v_enqueue_new_tasks != 0);
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":314
 *             return False
 *         if enqueue_new_tasks:
 *             self._enqueue_new_tasks()             # <<<<<<<<<<<<<<
 *         return self._schedule_without_checks(task)
 * 
 */
    __pyx_t_4 = __pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "asynq/scheduler.py":313
 *         if task.is_computed() or task.is_blocked():
 *             return False
 *         if enqueue_new_tasks:             # <<<<<<<<<<<<<<
 *             self._enqueue_new_tasks()
 *         return self._schedule_without_checks(task)
 */
  }

  /* "asynq/scheduler.py":315
 *         if enqueue_new_tasks:
 *             self._enqueue_new_tasks()
 *         return self._schedule_without_checks(task)             # <<<<<<<<<<<<<<
 * 
 *     def schedule_batch(self, batch):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule_without_checks(__pyx_v_self, __pyx_v_task); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "asynq/scheduler.py":310
 *         return True
 * 
 *     def _try_schedule(self, task, enqueue_new_tasks):             # <<<<<<<<<<<<<<
 *         if task.is_computed() or task.is_blocked():
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._try_schedule", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":317
 *         return self._schedule_without_checks(task)
 * 
 *     def schedule_batch(self, batch):             # <<<<<<<<<<<<<<
 *         self._schedule_batch(batch)
 * 
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_27schedule_batch(PyObject *__pyx_v_self, PyObject *__pyx_v_batch); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_schedule_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("schedule_batch", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_schedule_batch); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_27schedule_batch)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_batch));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_batch));
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":318
 * 
 *     def schedule_batch(self, batch):
 *         self._schedule_batch(batch)             # <<<<<<<<<<<<<<
 * 
 *     def _schedule_batch(self, batch):
 */
  __pyx_t_6 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_schedule_batch(__pyx_v_self, __pyx_v_batch); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 318, __pyx_L1_error)

  /* "asynq/scheduler.py":317
 *         return self._schedule_without_checks(task)
 * 
 *     def schedule_batch(self, batch):             # <<<<<<<<<<<<<<
 *         self._schedule_batch(batch)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.schedule_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_27schedule_batch(PyObject *__pyx_v_self, PyObject *__pyx_v_batch); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_27schedule_batch(PyObject *__pyx_v_self, PyObject *__pyx_v_batch) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("schedule_batch (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_batch), __pyx_ptype_5asynq_8batching_BatchBase, 1, "batch", 0))) __PYX_ERR(0, 317, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_26schedule_batch(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((struct __pyx_obj_5asynq_8batching_BatchBase *)__pyx_v_batch));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_26schedule_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("schedule_batch", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler_schedule_batch(__pyx_v_self, __pyx_v_batch, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.schedule_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":320
 *         self._schedule_batch(batch)
 * 
 *     def _schedule_batch(self, batch):             # <<<<<<<<<<<<<<
 *         if batch.is_flushed():
 *             if _debug_options.DUMP_SCHEDULE_BATCH:
 */

static int __pyx_f_5asynq_9scheduler_13TaskScheduler__schedule_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("_schedule_batch", 0);

  /* "asynq/scheduler.py":321
 * 
 *     def _schedule_batch(self, batch):
 *         if batch.is_flushed():             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULE_BATCH:
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_8batching_BatchBase *)__pyx_v_batch->__pyx_base.__pyx_vtab)->is_flushed(__pyx_v_batch, 0); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "asynq/scheduler.py":322
 *     def _schedule_batch(self, batch):
 *         if batch.is_flushed():
 *             if _debug_options.DUMP_SCHEDULE_BATCH:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 *             return False
 */
    __pyx_t_2 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULE_BATCH != 0);
    if (__pyx_t_2) {

      /* "asynq/scheduler.py":323
 *         if batch.is_flushed():
 *             if _debug_options.DUMP_SCHEDULE_BATCH:
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))             # <<<<<<<<<<<<<<
 *             return False
 *         if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_str); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_batch));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_batch));
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_async_can_t_schedule_flushed_ba, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 323, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_7);
        __pyx_t_7 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/scheduler.py":322
 *     def _schedule_batch(self, batch):
 *         if batch.is_flushed():
 *             if _debug_options.DUMP_SCHEDULE_BATCH:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 *             return False
 */
    }

    /* "asynq/scheduler.py":324
 *             if _debug_options.DUMP_SCHEDULE_BATCH:
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 *             return False             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:
 *             debug.write('@async: scheduling batch %s' % debug.str(batch))
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "asynq/scheduler.py":321
 * 
 *     def _schedule_batch(self, batch):
 *         if batch.is_flushed():             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULE_BATCH:
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 */
  }

  /* "asynq/scheduler.py":325
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 *             return False
 *         if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:             # <<<<<<<<<<<<<<
 *             debug.write('@async: scheduling batch %s' % debug.str(batch))
 *         self._batches.add(batch)
 */
  __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULE_BATCH != 0);
  if (__pyx_t_1) {
  } else {
    __pyx_t_2 = __pyx_t_1;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(((PyObject *)__pyx_v_batch), __pyx_v_self->_batches, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 325, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_1 != 0);
  __pyx_t_2 = __pyx_t_9;
  __pyx_L6_bool_binop_done:;
  if (__pyx_t_2) {

    /* "asynq/scheduler.py":326
 *             return False
 *         if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:
 *             debug.write('@async: scheduling batch %s' % debug.str(batch))             # <<<<<<<<<<<<<<
 *         self._batches.add(batch)
 *         return True
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_write); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(((PyObject *)__pyx_v_batch));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, ((PyObject *)__pyx_v_batch));
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_scheduling_batch_s, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "asynq/scheduler.py":325
 *                 debug.write("@async: can't schedule flushed batch %s" % debug.str(batch))
 *             return False
 *         if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:             # <<<<<<<<<<<<<<
 *             debug.write('@async: scheduling batch %s' % debug.str(batch))
 *         self._batches.add(batch)
 */
  }

  /* "asynq/scheduler.py":327
 *         if _debug_options.DUMP_SCHEDULE_BATCH and batch not in self._batches:
 *             debug.write('@async: scheduling batch %s' % debug.str(batch))
 *         self._batches.add(batch)             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
  if (unlikely(__pyx_v_self->_batches == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "add");
    __PYX_ERR(0, 327, __pyx_L1_error)
  }
  __pyx_t_10 = PySet_Add(__pyx_v_self->_batches, ((PyObject *)__pyx_v_batch)); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 327, __pyx_L1_error)

  /* "asynq/scheduler.py":328
 *             debug.write('@async: scheduling batch %s' % debug.str(batch))
 *         self._batches.add(batch)
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def _flush_batch(self, batch):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "asynq/scheduler.py":320
 *         self._schedule_batch(batch)
 * 
 *     def _schedule_batch(self, batch):             # <<<<<<<<<<<<<<
 *         if batch.is_flushed():
 *             if _debug_options.DUMP_SCHEDULE_BATCH:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._schedule_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":330
 *         return True
 * 
 *     def _flush_batch(self, batch):             # <<<<<<<<<<<<<<
 *         self.on_before_batch_flush(batch)
 *         try:
 */

static int __pyx_f_5asynq_9scheduler_13TaskScheduler__flush_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  __Pyx_RefNannySetupContext("_flush_batch", 0);

  /* "asynq/scheduler.py":331
 * 
 *     def _flush_batch(self, batch):
 *         self.on_before_batch_flush(batch)             # <<<<<<<<<<<<<<
 *         try:
 *             batch.flush()
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self->on_before_batch_flush));
  __pyx_t_2 = ((PyObject *)__pyx_v_self->on_before_batch_flush); __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_batch));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_batch));
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 331, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":332
 *     def _flush_batch(self, batch):
 *         self.on_before_batch_flush(batch)
 *         try:             # <<<<<<<<<<<<<<
 *             batch.flush()
 *         finally:
 */
  /*try:*/ {

    /* "asynq/scheduler.py":333
 *         self.on_before_batch_flush(batch)
 *         try:
 *             batch.flush()             # <<<<<<<<<<<<<<
 *         finally:
 *             self.on_after_batch_flush(batch)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_8batching_BatchBase *)__pyx_v_batch->__pyx_base.__pyx_vtab)->flush(__pyx_v_batch, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":335
 *             batch.flush()
 *         finally:
 *             self.on_after_batch_flush(batch)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __Pyx_INCREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
      __pyx_t_2 = ((PyObject *)__pyx_v_self->on_after_batch_flush); __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_batch));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, ((PyObject *)__pyx_v_batch));
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
      {
        __Pyx_INCREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
        __pyx_t_2 = ((PyObject *)__pyx_v_self->on_after_batch_flush); __pyx_t_3 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (!__pyx_t_3) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_batch));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
          PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_batch));
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
      __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
      goto __pyx_L1_error;
      __pyx_L7_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "asynq/scheduler.py":336
 *         finally:
 *             self.on_after_batch_flush(batch)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _enqueue_new_tasks(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":330
 *         return True
 * 
 *     def _flush_batch(self, batch):             # <<<<<<<<<<<<<<
 *         self.on_before_batch_flush(batch)
 *         try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._flush_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":338
 *         return 0
 * 
 *     def _enqueue_new_tasks(self):             # <<<<<<<<<<<<<<
 *         l = len(self._new_tasks)
 *         if l == 0:
 */

static CYTHON_INLINE PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_l;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("_enqueue_new_tasks", 0);

  /* "asynq/scheduler.py":339
 * 
 *     def _enqueue_new_tasks(self):
 *         l = len(self._new_tasks)             # <<<<<<<<<<<<<<
 *         if l == 0:
 *             pass
 */
  __pyx_t_1 = __pyx_v_self->_new_tasks;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 339, __pyx_L1_error)
  }
  __pyx_t_2 = PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_l = __pyx_t_2;

  /* "asynq/scheduler.py":340
 *     def _enqueue_new_tasks(self):
 *         l = len(self._new_tasks)
 *         if l == 0:             # <<<<<<<<<<<<<<
 *             pass
 *         else:
 */
  __pyx_t_3 = ((__pyx_v_l == 0) != 0);
  if (__pyx_t_3) {
    goto __pyx_L3;
  }

  /* "asynq/scheduler.py":343
 *             pass
 *         else:
 *             i = 0             # <<<<<<<<<<<<<<
 *             while i < l:
 *                 self._tasks.append(self._new_tasks[i])
 */
  /*else*/ {
    __pyx_v_i = 0;

    /* "asynq/scheduler.py":344
 *         else:
 *             i = 0
 *             while i < l:             # <<<<<<<<<<<<<<
 *                 self._tasks.append(self._new_tasks[i])
 *                 i += 1
 */
    while (1) {
      __pyx_t_3 = ((__pyx_v_i < __pyx_v_l) != 0);
      if (!__pyx_t_3) break;

      /* "asynq/scheduler.py":345
 *             i = 0
 *             while i < l:
 *                 self._tasks.append(self._new_tasks[i])             # <<<<<<<<<<<<<<
 *                 i += 1
 *             del self._new_tasks[:]
 */
      if (unlikely(__pyx_v_self->_new_tasks == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 345, __pyx_L1_error)
      }
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_self->_new_tasks, __pyx_v_i, int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_Append(__pyx_v_self->_tasks, __pyx_t_1); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":346
 *             while i < l:
 *                 self._tasks.append(self._new_tasks[i])
 *                 i += 1             # <<<<<<<<<<<<<<
 *             del self._new_tasks[:]
 * 
 */
      __pyx_v_i = (__pyx_v_i + 1);
    }

    /* "asynq/scheduler.py":347
 *                 self._tasks.append(self._new_tasks[i])
 *                 i += 1
 *             del self._new_tasks[:]             # <<<<<<<<<<<<<<
 * 
 *     def _continue_with_task(self):
 */
    if (unlikely(__pyx_v_self->_new_tasks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 347, __pyx_L1_error)
    }
    if (__Pyx_PyObject_DelSlice(__pyx_v_self->_new_tasks, 0, 0, NULL, NULL, NULL, 0, 0, 1) < 0) __PYX_ERR(0, 347, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "asynq/scheduler.py":338
 *         return 0
 * 
 *     def _enqueue_new_tasks(self):             # <<<<<<<<<<<<<<
 *         l = len(self._new_tasks)
 *         if l == 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._enqueue_new_tasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":349
 *             del self._new_tasks[:]
 * 
 *     def _continue_with_task(self):             # <<<<<<<<<<<<<<
 *         """Continues the next task.
 *         Returns the task that made some progress or None.
 */

static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_13TaskScheduler__continue_with_task(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_v_dump_continue_task;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_task = 0;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_v_old_active_task = NULL;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("_continue_with_task", 0);

  /* "asynq/scheduler.py":355
 *         """
 *         # Queuing new tasks
 *         self._enqueue_new_tasks()             # <<<<<<<<<<<<<<
 *         dump_continue_task = _debug_options.DUMP_CONTINUE_TASK
 *         while len(self._tasks) != 0:
 */
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":356
 *         # Queuing new tasks
 *         self._enqueue_new_tasks()
 *         dump_continue_task = _debug_options.DUMP_CONTINUE_TASK             # <<<<<<<<<<<<<<
 *         while len(self._tasks) != 0:
 *             task = self._tasks.popleft()
 */
  __pyx_t_2 = __pyx_v_5asynq_9scheduler__debug_options->DUMP_CONTINUE_TASK;
  __pyx_v_dump_continue_task = __pyx_t_2;

  /* "asynq/scheduler.py":357
 *         self._enqueue_new_tasks()
 *         dump_continue_task = _debug_options.DUMP_CONTINUE_TASK
 *         while len(self._tasks) != 0:             # <<<<<<<<<<<<<<
 *             task = self._tasks.popleft()
 *             task.is_scheduled = False
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->_tasks;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = ((__pyx_t_3 != 0) != 0);
    if (!__pyx_t_2) break;

    /* "asynq/scheduler.py":358
 *         dump_continue_task = _debug_options.DUMP_CONTINUE_TASK
 *         while len(self._tasks) != 0:
 *             task = self._tasks.popleft()             # <<<<<<<<<<<<<<
 *             task.is_scheduled = False
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_tasks, __pyx_n_s_popleft); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_task, ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "asynq/scheduler.py":359
 *         while len(self._tasks) != 0:
 *             task = self._tasks.popleft()
 *             task.is_scheduled = False             # <<<<<<<<<<<<<<
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster
 *                 if dump_continue_task:
 */
    __pyx_v_task->is_scheduled = 0;

    /* "asynq/scheduler.py":360
 *             task = self._tasks.popleft()
 *             task.is_scheduled = False
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster             # <<<<<<<<<<<<<<
 *                 if dump_continue_task:
 *                     debug.write('@async: skipping %s' % debug.str(task))
 */
    __pyx_t_6 = (__pyx_v_task->_dependencies != Py_None) && (PySet_GET_SIZE(__pyx_v_task->_dependencies) != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_task->__pyx_base._value != __pyx_v_5asynq_9scheduler__futures_none);
    __pyx_t_7 = (__pyx_t_6 != 0);
    __pyx_t_2 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_2) {

      /* "asynq/scheduler.py":361
 *             task.is_scheduled = False
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: skipping %s' % debug.str(task))
 *                 continue
 */
      __pyx_t_2 = (__pyx_v_dump_continue_task != 0);
      if (__pyx_t_2) {

        /* "asynq/scheduler.py":362
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster
 *                 if dump_continue_task:
 *                     debug.write('@async: skipping %s' % debug.str(task))             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
        __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_str); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        if (!__pyx_t_8) {
          __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_9, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
        } else {
          __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_task));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
          PyTuple_SET_ITEM(__pyx_t_10, 0+1, ((PyObject *)__pyx_v_task));
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyString_Format(__pyx_kp_s_async_skipping_s, __pyx_t_4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 362, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
          }
        }
        if (!__pyx_t_4) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_GIVEREF(__pyx_t_9);
          PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_t_9);
          __pyx_t_9 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":361
 *             task.is_scheduled = False
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: skipping %s' % debug.str(task))
 *                 continue
 */
      }

      /* "asynq/scheduler.py":363
 *                 if dump_continue_task:
 *                     debug.write('@async: skipping %s' % debug.str(task))
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # Executing the task
 */
      goto __pyx_L3_continue;

      /* "asynq/scheduler.py":360
 *             task = self._tasks.popleft()
 *             task.is_scheduled = False
 *             if task._dependencies or task._value is not _futures_none:  # is_blocked() or is_computed(), but faster             # <<<<<<<<<<<<<<
 *                 if dump_continue_task:
 *                     debug.write('@async: skipping %s' % debug.str(task))
 */
    }

    /* "asynq/scheduler.py":366
 * 
 *             # Executing the task
 *             old_active_task = self.active_task             # <<<<<<<<<<<<<<
 *             self.active_task = task
 *             try:
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_self->active_task);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_old_active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "asynq/scheduler.py":367
 *             # Executing the task
 *             old_active_task = self.active_task
 *             self.active_task = task             # <<<<<<<<<<<<<<
 *             try:
 *                 if dump_continue_task:
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_task));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
    __Pyx_GOTREF(__pyx_v_self->active_task);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
    __pyx_v_self->active_task = __pyx_v_task;

    /* "asynq/scheduler.py":368
 *             old_active_task = self.active_task
 *             self.active_task = task
 *             try:             # <<<<<<<<<<<<<<
 *                 if dump_continue_task:
 *                     debug.write('@async: -> continuing %s' % debug.str(task))
 */
    /*try:*/ {

      /* "asynq/scheduler.py":369
 *             self.active_task = task
 *             try:
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: -> continuing %s' % debug.str(task))
 *                 task._continue()
 */
      __pyx_t_2 = (__pyx_v_dump_continue_task != 0);
      if (__pyx_t_2) {

        /* "asynq/scheduler.py":370
 *             try:
 *                 if dump_continue_task:
 *                     debug.write('@async: -> continuing %s' % debug.str(task))             # <<<<<<<<<<<<<<
 *                 task._continue()
 *             finally:
 */
        __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 370, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_write); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 370, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 370, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 370, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        if (!__pyx_t_9) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_4, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 370, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_5);
        } else {
          __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 370, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_9); __pyx_t_9 = NULL;
          __Pyx_INCREF(((PyObject *)__pyx_v_task));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
          PyTuple_SET_ITEM(__pyx_t_8, 0+1, ((PyObject *)__pyx_v_task));
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_8, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 370, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_async_continuing_s, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 370, __pyx_L12_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_5 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
          }
        }
        if (!__pyx_t_5) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L12_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else {
          __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 370, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":369
 *             self.active_task = task
 *             try:
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: -> continuing %s' % debug.str(task))
 *                 task._continue()
 */
      }

      /* "asynq/scheduler.py":371
 *                 if dump_continue_task:
 *                     debug.write('@async: -> continuing %s' % debug.str(task))
 *                 task._continue()             # <<<<<<<<<<<<<<
 *             finally:
 *                 # task._continue should never throw an error, but
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask *)__pyx_v_task->__pyx_base.__pyx_vtab)->_continue(__pyx_v_task, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "asynq/scheduler.py":375
 *                 # task._continue should never throw an error, but
 *                 # it's better to implement active task switch in safe way
 *                 self.active_task = old_active_task             # <<<<<<<<<<<<<<
 *                 # No need to reschedule: task._continue() ensures this
 *                 # anyway
 */
    /*finally:*/ {
      /*normal exit:*/{
        __Pyx_INCREF(((PyObject *)__pyx_v_old_active_task));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_old_active_task));
        __Pyx_GOTREF(__pyx_v_self->active_task);
        __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
        __pyx_v_self->active_task = __pyx_v_old_active_task;

        /* "asynq/scheduler.py":378
 *                 # No need to reschedule: task._continue() ensures this
 *                 # anyway
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task
 */
        __pyx_t_2 = (__pyx_v_dump_continue_task != 0);
        if (__pyx_t_2) {

          /* "asynq/scheduler.py":379
 *                 # anyway
 *                 if dump_continue_task:
 *                     debug.write('@async: <- continued %s' % debug.str(task))             # <<<<<<<<<<<<<<
 *             return task
 *         if dump_continue_task:
 */
          __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_write); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_5, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
          } else {
            __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 379, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_INCREF(((PyObject *)__pyx_v_task));
            __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
            PyTuple_SET_ITEM(__pyx_t_9, 0+1, ((PyObject *)__pyx_v_task));
            __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_async_continued_s, __pyx_t_10); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
            }
          }
          if (!__pyx_t_10) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 379, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
            __Pyx_GIVEREF(__pyx_t_5);
            PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_t_5);
            __pyx_t_5 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          }
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "asynq/scheduler.py":378
 *                 # No need to reschedule: task._continue() ensures this
 *                 # anyway
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task
 */
        }
        goto __pyx_L13;
      }
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __pyx_L12_error:;
        __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_17);
        __Pyx_XGOTREF(__pyx_t_18);
        __Pyx_XGOTREF(__pyx_t_19);
        __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
        {

          /* "asynq/scheduler.py":375
 *                 # task._continue should never throw an error, but
 *                 # it's better to implement active task switch in safe way
 *                 self.active_task = old_active_task             # <<<<<<<<<<<<<<
 *                 # No need to reschedule: task._continue() ensures this
 *                 # anyway
 */
          __Pyx_INCREF(((PyObject *)__pyx_v_old_active_task));
          __Pyx_GIVEREF(((PyObject *)__pyx_v_old_active_task));
          __Pyx_GOTREF(__pyx_v_self->active_task);
          __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
          __pyx_v_self->active_task = __pyx_v_old_active_task;

          /* "asynq/scheduler.py":378
 *                 # No need to reschedule: task._continue() ensures this
 *                 # anyway
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task
 */
          __pyx_t_2 = (__pyx_v_dump_continue_task != 0);
          if (__pyx_t_2) {

            /* "asynq/scheduler.py":379
 *                 # anyway
 *                 if dump_continue_task:
 *                     debug.write('@async: <- continued %s' % debug.str(task))             # <<<<<<<<<<<<<<
 *             return task
 *         if dump_continue_task:
 */
            __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 379, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_8);
            __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_write); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 379, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __pyx_t_5 = NULL;
            if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_5)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_5);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
              }
            }
            if (!__pyx_t_5) {
              __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_t_10, ((PyObject *)__pyx_v_task)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 379, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_8);
            } else {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
              __Pyx_INCREF(((PyObject *)__pyx_v_task));
              __Pyx_GIVEREF(((PyObject *)__pyx_v_task));
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_task));
              __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 379, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_8);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyString_Format(__pyx_kp_s_async_continued_s, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 379, __pyx_L19_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            __pyx_t_8 = NULL;
            if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_9))) {
              __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_9);
              if (likely(__pyx_t_8)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
                __Pyx_INCREF(__pyx_t_8);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_9, function);
              }
            }
            if (!__pyx_t_8) {
              __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L19_error)
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_1);
            } else {
              __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_8); __pyx_t_8 = NULL;
              __Pyx_GIVEREF(__pyx_t_10);
              PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_10);
              __pyx_t_10 = 0;
              __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L19_error)
              __Pyx_GOTREF(__pyx_t_1);
              __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            }
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

            /* "asynq/scheduler.py":378
 *                 # No need to reschedule: task._continue() ensures this
 *                 # anyway
 *                 if dump_continue_task:             # <<<<<<<<<<<<<<
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task
 */
          }
        }
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
        }
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
        __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
        __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
        goto __pyx_L1_error;
        __pyx_L19_error:;
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
        }
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
        goto __pyx_L1_error;
      }
      __pyx_L13:;
    }

    /* "asynq/scheduler.py":380
 *                 if dump_continue_task:
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task             # <<<<<<<<<<<<<<
 *         if dump_continue_task:
 *             debug.write('@async: no task to continue')
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_task));
    __pyx_r = __pyx_v_task;
    goto __pyx_L0;
    __pyx_L3_continue:;
  }

  /* "asynq/scheduler.py":381
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task
 *         if dump_continue_task:             # <<<<<<<<<<<<<<
 *             debug.write('@async: no task to continue')
 *         return None
 */
  __pyx_t_2 = (__pyx_v_dump_continue_task != 0);
  if (__pyx_t_2) {

    /* "asynq/scheduler.py":382
 *             return task
 *         if dump_continue_task:
 *             debug.write('@async: no task to continue')             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":381
 *                     debug.write('@async: <- continued %s' % debug.str(task))
 *             return task
 *         if dump_continue_task:             # <<<<<<<<<<<<<<
 *             debug.write('@async: no task to continue')
 *         return None
 */
  }

  /* "asynq/scheduler.py":383
 *         if dump_continue_task:
 *             debug.write('@async: no task to continue')
 *         return None             # <<<<<<<<<<<<<<
 * 
 *     def _continue_with_batch(self, assert_no_batch=True):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(Py_None);
  __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);
  goto __pyx_L0;

  /* "asynq/scheduler.py":349
 *             del self._new_tasks[:]
 * 
 *     def _continue_with_task(self):             # <<<<<<<<<<<<<<
 *         """Continues the next task.
 *         Returns the task that made some progress or None.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._continue_with_task", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_task);
  __Pyx_XDECREF((PyObject *)__pyx_v_old_active_task);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":385
 *         return None
 * 
 *     def _continue_with_batch(self, assert_no_batch=True):             # <<<<<<<<<<<<<<
 *         """
 *         Flushes one of batches (the longest one by default).
 */

static struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_f_5asynq_9scheduler_13TaskScheduler__continue_with_batch(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch *__pyx_optional_args) {
  int __pyx_v_assert_no_batch = ((int)1);
  struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_v_batch = NULL;
  struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_continue_with_batch", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_assert_no_batch = __pyx_optional_args->assert_no_batch;
    }
  }

  /* "asynq/scheduler.py":395
 * 
 *         """
 *         batch = self._select_batch_to_flush()             # <<<<<<<<<<<<<<
 *         if batch is None:
 *             if _debug_options.DUMP_FLUSH_BATCH:
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_select_batch_to_flush(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_batch = ((struct __pyx_obj_5asynq_8batching_BatchBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":396
 *         """
 *         batch = self._select_batch_to_flush()
 *         if batch is None:             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_FLUSH_BATCH:
 *                 debug.write('@async: no batch to flush')
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_batch) == Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "asynq/scheduler.py":397
 *         batch = self._select_batch_to_flush()
 *         if batch is None:
 *             if _debug_options.DUMP_FLUSH_BATCH:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: no batch to flush')
 *             if assert_no_batch:
 */
    __pyx_t_3 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_FLUSH_BATCH != 0);
    if (__pyx_t_3) {

      /* "asynq/scheduler.py":398
 *         if batch is None:
 *             if _debug_options.DUMP_FLUSH_BATCH:
 *                 debug.write('@async: no batch to flush')             # <<<<<<<<<<<<<<
 *             if assert_no_batch:
 *                 debug.dump(self)
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":397
 *         batch = self._select_batch_to_flush()
 *         if batch is None:
 *             if _debug_options.DUMP_FLUSH_BATCH:             # <<<<<<<<<<<<<<
 *                 debug.write('@async: no batch to flush')
 *             if assert_no_batch:
 */
    }

    /* "asynq/scheduler.py":399
 *             if _debug_options.DUMP_FLUSH_BATCH:
 *                 debug.write('@async: no batch to flush')
 *             if assert_no_batch:             # <<<<<<<<<<<<<<
 *                 debug.dump(self)
 *                 assert False, \
 */
    __pyx_t_3 = (__pyx_v_assert_no_batch != 0);
    if (__pyx_t_3) {

      /* "asynq/scheduler.py":400
 *                 debug.write('@async: no batch to flush')
 *             if assert_no_batch:
 *                 debug.dump(self)             # <<<<<<<<<<<<<<
 *                 assert False, \
 *                     "Can't proceed further: no task to continue or batch to flush, active tasks: " \
 */
      __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_dump); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 400, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_self));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, ((PyObject *)__pyx_v_self));
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":401
 *             if assert_no_batch:
 *                 debug.dump(self)
 *                 assert False, \             # <<<<<<<<<<<<<<
 *                     "Can't proceed further: no task to continue or batch to flush, active tasks: " \
 *                     "%s" % self._tasks
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        if (unlikely(!0)) {

          /* "asynq/scheduler.py":403
 *                 assert False, \
 *                     "Can't proceed further: no task to continue or batch to flush, active tasks: " \
 *                     "%s" % self._tasks             # <<<<<<<<<<<<<<
 *             else:
 *                 return None
 */
          __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Can_t_proceed_further_no_task_to, __pyx_v_self->_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          PyErr_SetObject(PyExc_AssertionError, __pyx_t_1);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __PYX_ERR(0, 401, __pyx_L1_error)
        }
      }
      #endif

      /* "asynq/scheduler.py":399
 *             if _debug_options.DUMP_FLUSH_BATCH:
 *                 debug.write('@async: no batch to flush')
 *             if assert_no_batch:             # <<<<<<<<<<<<<<
 *                 debug.dump(self)
 *                 assert False, \
 */
      goto __pyx_L5;
    }

    /* "asynq/scheduler.py":405
 *                     "%s" % self._tasks
 *             else:
 *                 return None             # <<<<<<<<<<<<<<
 *         self._batches.remove(batch)
 *         self._flush_batch(batch)
 */
    /*else*/ {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(Py_None);
      __pyx_r = ((struct __pyx_obj_5asynq_8batching_BatchBase *)Py_None);
      goto __pyx_L0;
    }
    __pyx_L5:;

    /* "asynq/scheduler.py":396
 *         """
 *         batch = self._select_batch_to_flush()
 *         if batch is None:             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_FLUSH_BATCH:
 *                 debug.write('@async: no batch to flush')
 */
  }

  /* "asynq/scheduler.py":406
 *             else:
 *                 return None
 *         self._batches.remove(batch)             # <<<<<<<<<<<<<<
 *         self._flush_batch(batch)
 *         return batch
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_batches, __pyx_n_s_remove); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, ((PyObject *)__pyx_v_batch)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_batch));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_batch));
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, ((PyObject *)__pyx_v_batch));
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":407
 *                 return None
 *         self._batches.remove(batch)
 *         self._flush_batch(batch)             # <<<<<<<<<<<<<<
 *         return batch
 * 
 */
  __pyx_t_7 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->_flush_batch(__pyx_v_self, __pyx_v_batch); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 407, __pyx_L1_error)

  /* "asynq/scheduler.py":408
 *         self._batches.remove(batch)
 *         self._flush_batch(batch)
 *         return batch             # <<<<<<<<<<<<<<
 * 
 *     def _select_batch_to_flush(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_batch));
  __pyx_r = __pyx_v_batch;
  goto __pyx_L0;

  /* "asynq/scheduler.py":385
 *         return None
 * 
 *     def _continue_with_batch(self, assert_no_batch=True):             # <<<<<<<<<<<<<<
 *         """
 *         Flushes one of batches (the longest one by default).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._continue_with_batch", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_batch);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":410
 *         return batch
 * 
 *     def _select_batch_to_flush(self):             # <<<<<<<<<<<<<<
 *         """Returns the batch having highest priority,
 *         or ``None``, if there are no batches.
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_29_select_batch_to_flush(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_f_5asynq_9scheduler_13TaskScheduler__select_batch_to_flush(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_best_batch = NULL;
  PyObject *__pyx_v_best_priority = NULL;
  PyObject *__pyx_v_batches_to_remove = NULL;
  PyObject *__pyx_v_batch = NULL;
  PyObject *__pyx_v_priority = NULL;
  struct __pyx_obj_5asynq_8batching_BatchBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("_select_batch_to_flush", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_select_batch_to_flush); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_29_select_batch_to_flush)) {
      __Pyx_XDECREF(((PyObject *)__pyx_r));
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5asynq_8batching_BatchBase))))) __PYX_ERR(0, 410, __pyx_L1_error)
      __pyx_r = ((struct __pyx_obj_5asynq_8batching_BatchBase *)__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":422
 * 
 *         """
 *         best_batch = None             # <<<<<<<<<<<<<<
 *         best_priority = None
 *         batches_to_remove = None
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_best_batch = Py_None;

  /* "asynq/scheduler.py":423
 *         """
 *         best_batch = None
 *         best_priority = None             # <<<<<<<<<<<<<<
 *         batches_to_remove = None
 *         for batch in self._batches:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_best_priority = Py_None;

  /* "asynq/scheduler.py":424
 *         best_batch = None
 *         best_priority = None
 *         batches_to_remove = None             # <<<<<<<<<<<<<<
 *         for batch in self._batches:
 *             if not batch.items or batch.is_flushed():
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_batches_to_remove = Py_None;

  /* "asynq/scheduler.py":425
 *         best_priority = None
 *         batches_to_remove = None
 *         for batch in self._batches:             # <<<<<<<<<<<<<<
 *             if not batch.items or batch.is_flushed():
 *                 if batches_to_remove is None:
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_self->_batches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 425, __pyx_L1_error)
  for (;;) {
    {
      __pyx_t_2 = __pyx_t_5(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 425, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_batch, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":426
 *         batches_to_remove = None
 *         for batch in self._batches:
 *             if not batch.items or batch.is_flushed():             # <<<<<<<<<<<<<<
 *                 if batches_to_remove is None:
 *                     batches_to_remove = [batch]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_batch, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = ((!__pyx_t_7) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_6 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_batch, __pyx_n_s_is_flushed); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 426, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 426, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "asynq/scheduler.py":427
 *         for batch in self._batches:
 *             if not batch.items or batch.is_flushed():
 *                 if batches_to_remove is None:             # <<<<<<<<<<<<<<
 *                     batches_to_remove = [batch]
 *                 else:
 */
      __pyx_t_6 = (__pyx_v_batches_to_remove == Py_None);
      __pyx_t_8 = (__pyx_t_6 != 0);
      if (__pyx_t_8) {

        /* "asynq/scheduler.py":428
 *             if not batch.items or batch.is_flushed():
 *                 if batches_to_remove is None:
 *                     batches_to_remove = [batch]             # <<<<<<<<<<<<<<
 *                 else:
 *                     batches_to_remove.append(batch)
 */
        __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_INCREF(__pyx_v_batch);
        __Pyx_GIVEREF(__pyx_v_batch);
        PyList_SET_ITEM(__pyx_t_2, 0, __pyx_v_batch);
        __Pyx_DECREF_SET(__pyx_v_batches_to_remove, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "asynq/scheduler.py":427
 *         for batch in self._batches:
 *             if not batch.items or batch.is_flushed():
 *                 if batches_to_remove is None:             # <<<<<<<<<<<<<<
 *                     batches_to_remove = [batch]
 *                 else:
 */
        goto __pyx_L8;
      }

      /* "asynq/scheduler.py":430
 *                     batches_to_remove = [batch]
 *                 else:
 *                     batches_to_remove.append(batch)             # <<<<<<<<<<<<<<
 *                 continue
 *             priority = batch.get_priority()
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyObject_Append(__pyx_v_batches_to_remove, __pyx_v_batch); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 430, __pyx_L1_error)
      }
      __pyx_L8:;

      /* "asynq/scheduler.py":431
 *                 else:
 *                     batches_to_remove.append(batch)
 *                 continue             # <<<<<<<<<<<<<<
 *             priority = batch.get_priority()
 *             if best_batch is None or best_priority < priority:
 */
      goto __pyx_L3_continue;

      /* "asynq/scheduler.py":426
 *         batches_to_remove = None
 *         for batch in self._batches:
 *             if not batch.items or batch.is_flushed():             # <<<<<<<<<<<<<<
 *                 if batches_to_remove is None:
 *                     batches_to_remove = [batch]
 */
    }

    /* "asynq/scheduler.py":432
 *                     batches_to_remove.append(batch)
 *                 continue
 *             priority = batch.get_priority()             # <<<<<<<<<<<<<<
 *             if best_batch is None or best_priority < priority:
 *                 best_batch = batch
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_batch, __pyx_n_s_get_priority); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_priority, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":433
 *                 continue
 *             priority = batch.get_priority()
 *             if best_batch is None or best_priority < priority:             # <<<<<<<<<<<<<<
 *                 best_batch = batch
 *                 best_priority = priority
 */
    __pyx_t_6 = (__pyx_v_best_batch == Py_None);
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_8 = __pyx_t_7;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_best_priority, __pyx_v_priority, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = __pyx_t_7;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_8) {

      /* "asynq/scheduler.py":434
 *             priority = batch.get_priority()
 *             if best_batch is None or best_priority < priority:
 *                 best_batch = batch             # <<<<<<<<<<<<<<
 *                 best_priority = priority
 *         if batches_to_remove:
 */
      __Pyx_INCREF(__pyx_v_batch);
      __Pyx_DECREF_SET(__pyx_v_best_batch, __pyx_v_batch);

      /* "asynq/scheduler.py":435
 *             if best_batch is None or best_priority < priority:
 *                 best_batch = batch
 *                 best_priority = priority             # <<<<<<<<<<<<<<
 *         if batches_to_remove:
 *             for batch in batches_to_remove:
 */
      __Pyx_INCREF(__pyx_v_priority);
      __Pyx_DECREF_SET(__pyx_v_best_priority, __pyx_v_priority);

      /* "asynq/scheduler.py":433
 *                 continue
 *             priority = batch.get_priority()
 *             if best_batch is None or best_priority < priority:             # <<<<<<<<<<<<<<
 *                 best_batch = batch
 *                 best_priority = priority
 */
    }

    /* "asynq/scheduler.py":425
 *         best_priority = None
 *         batches_to_remove = None
 *         for batch in self._batches:             # <<<<<<<<<<<<<<
 *             if not batch.items or batch.is_flushed():
 *                 if batches_to_remove is None:
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":436
 *                 best_batch = batch
 *                 best_priority = priority
 *         if batches_to_remove:             # <<<<<<<<<<<<<<
 *             for batch in batches_to_remove:
 *                 self._batches.remove(batch)
 */
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_batches_to_remove); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 436, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":437
 *                 best_priority = priority
 *         if batches_to_remove:
 *             for batch in batches_to_remove:             # <<<<<<<<<<<<<<
 *                 self._batches.remove(batch)
 *         return best_batch
 */
    if (likely(PyList_CheckExact(__pyx_v_batches_to_remove)) || PyTuple_CheckExact(__pyx_v_batches_to_remove)) {
      __pyx_t_1 = __pyx_v_batches_to_remove; __Pyx_INCREF(__pyx_t_1); __pyx_t_10 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_batches_to_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 437, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 437, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 437, __pyx_L1_error)
          #else
          __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 437, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_batch, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "asynq/scheduler.py":438
 *         if batches_to_remove:
 *             for batch in batches_to_remove:
 *                 self._batches.remove(batch)             # <<<<<<<<<<<<<<
 *         return best_batch
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_batches, __pyx_n_s_remove); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_batch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_11 = PyTuple_New(1+1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 438, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_batch);
        __Pyx_GIVEREF(__pyx_v_batch);
        PyTuple_SET_ITEM(__pyx_t_11, 0+1, __pyx_v_batch);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynq/scheduler.py":437
 *                 best_priority = priority
 *         if batches_to_remove:
 *             for batch in batches_to_remove:             # <<<<<<<<<<<<<<
 *                 self._batches.remove(batch)
 *         return best_batch
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":436
 *                 best_batch = batch
 *                 best_priority = priority
 *         if batches_to_remove:             # <<<<<<<<<<<<<<
 *             for batch in batches_to_remove:
 *                 self._batches.remove(batch)
 */
  }

  /* "asynq/scheduler.py":439
 *             for batch in batches_to_remove:
 *                 self._batches.remove(batch)
 *         return best_batch             # <<<<<<<<<<<<<<
 * 
 *     def _clone(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  if (!(likely(((__pyx_v_best_batch) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_best_batch, __pyx_ptype_5asynq_8batching_BatchBase))))) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_best_batch);
  __pyx_r = ((struct __pyx_obj_5asynq_8batching_BatchBase *)__pyx_v_best_batch);
  goto __pyx_L0;

  /* "asynq/scheduler.py":410
 *         return batch
 * 
 *     def _select_batch_to_flush(self):             # <<<<<<<<<<<<<<
 *         """Returns the batch having highest priority,
 *         or ``None``, if there are no batches.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._select_batch_to_flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_best_batch);
  __Pyx_XDECREF(__pyx_v_best_priority);
  __Pyx_XDECREF(__pyx_v_batches_to_remove);
  __Pyx_XDECREF(__pyx_v_batch);
  __Pyx_XDECREF(__pyx_v_priority);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_29_select_batch_to_flush(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_28_select_batch_to_flush[] = "Returns the batch having highest priority,\n        or ``None``, if there are no batches.\n\n        This method uses ``BatchBase.get_priority()`` to\n        determine the priority.\n\n        Side effect: this method removed flushed batches.\n\n        :return: selected batch or None.\n\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_29_select_batch_to_flush(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_select_batch_to_flush (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_28_select_batch_to_flush(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_28_select_batch_to_flush(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_select_batch_to_flush", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_13TaskScheduler__select_batch_to_flush(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._select_batch_to_flush", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":441
 *         return best_batch
 * 
 *     def _clone(self):             # <<<<<<<<<<<<<<
 *         """Protected method cloning the scheduler.
 * 
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_31_clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__clone(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_clone = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  float __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_31_clone)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":448
 * 
 *         """
 *         clone = type(self)()             # <<<<<<<<<<<<<<
 *         clone._last_dump_time = self._last_dump_time
 *         clone.max_await_recursion_depth = self.max_await_recursion_depth
 */
  __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  __pyx_t_2 = ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))); __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5asynq_9scheduler_TaskScheduler))))) __PYX_ERR(0, 448, __pyx_L1_error)
  __pyx_v_clone = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":449
 *         """
 *         clone = type(self)()
 *         clone._last_dump_time = self._last_dump_time             # <<<<<<<<<<<<<<
 *         clone.max_await_recursion_depth = self.max_await_recursion_depth
 *         clone.on_before_batch_flush = self.on_before_batch_flush
 */
  __pyx_t_5 = __pyx_v_self->_last_dump_time;
  __pyx_v_clone->_last_dump_time = __pyx_t_5;

  /* "asynq/scheduler.py":450
 *         clone = type(self)()
 *         clone._last_dump_time = self._last_dump_time
 *         clone.max_await_recursion_depth = self.max_await_recursion_depth             # <<<<<<<<<<<<<<
 *         clone.on_before_batch_flush = self.on_before_batch_flush
 *         clone.on_after_batch_flush = self.on_after_batch_flush
 */
  __pyx_t_6 = __pyx_v_self->max_await_recursion_depth;
  __pyx_v_clone->max_await_recursion_depth = __pyx_t_6;

  /* "asynq/scheduler.py":451
 *         clone._last_dump_time = self._last_dump_time
 *         clone.max_await_recursion_depth = self.max_await_recursion_depth
 *         clone.on_before_batch_flush = self.on_before_batch_flush             # <<<<<<<<<<<<<<
 *         clone.on_after_batch_flush = self.on_after_batch_flush
 *         return clone
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->on_before_batch_flush);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_clone->on_before_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_clone->on_before_batch_flush));
  __pyx_v_clone->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":452
 *         clone.max_await_recursion_depth = self.max_await_recursion_depth
 *         clone.on_before_batch_flush = self.on_before_batch_flush
 *         clone.on_after_batch_flush = self.on_after_batch_flush             # <<<<<<<<<<<<<<
 *         return clone
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->on_after_batch_flush);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_clone->on_after_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_clone->on_after_batch_flush));
  __pyx_v_clone->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":453
 *         clone.on_before_batch_flush = self.on_before_batch_flush
 *         clone.on_after_batch_flush = self.on_after_batch_flush
 *         return clone             # <<<<<<<<<<<<<<
 * 
 *     def _accept_changes(self, clone):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_clone));
  __pyx_r = ((PyObject *)__pyx_v_clone);
  goto __pyx_L0;

  /* "asynq/scheduler.py":441
 *         return best_batch
 * 
 *     def _clone(self):             # <<<<<<<<<<<<<<
 *         """Protected method cloning the scheduler.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_clone);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_31_clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_30_clone[] = "Protected method cloning the scheduler.\n\n        This method is used to create a new scheduler temporarily\n        replacing the current one to handle await recursion\n\n        ";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_31_clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clone (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_30_clone(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_30_clone(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__clone(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":455
 *         return clone
 * 
 *     def _accept_changes(self, clone):             # <<<<<<<<<<<<<<
 *         """Protected method accepting the changes made to this scheduler's clone."""
 * 
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_33_accept_changes(PyObject *__pyx_v_self, PyObject *__pyx_v_clone); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler__accept_changes(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_clone, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_batch = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  float __pyx_t_6;
  int __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("_accept_changes", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_accept_changes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_33_accept_changes)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_clone)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 455, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(((PyObject *)__pyx_v_clone));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_clone));
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_clone));
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 455, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":458
 *         """Protected method accepting the changes made to this scheduler's clone."""
 * 
 *         self._last_dump_time = clone._last_dump_time             # <<<<<<<<<<<<<<
 *         self.max_await_recursion_depth = clone.max_await_recursion_depth
 *         self.on_before_batch_flush = clone.on_before_batch_flush
 */
  __pyx_t_6 = __pyx_v_clone->_last_dump_time;
  __pyx_v_self->_last_dump_time = __pyx_t_6;

  /* "asynq/scheduler.py":459
 * 
 *         self._last_dump_time = clone._last_dump_time
 *         self.max_await_recursion_depth = clone.max_await_recursion_depth             # <<<<<<<<<<<<<<
 *         self.on_before_batch_flush = clone.on_before_batch_flush
 *         self.on_after_batch_flush = clone.on_after_batch_flush
 */
  __pyx_t_7 = __pyx_v_clone->max_await_recursion_depth;
  __pyx_v_self->max_await_recursion_depth = __pyx_t_7;

  /* "asynq/scheduler.py":460
 *         self._last_dump_time = clone._last_dump_time
 *         self.max_await_recursion_depth = clone.max_await_recursion_depth
 *         self.on_before_batch_flush = clone.on_before_batch_flush             # <<<<<<<<<<<<<<
 *         self.on_after_batch_flush = clone.on_after_batch_flush
 *         for batch in clone._batches:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_clone->on_before_batch_flush);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->on_before_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_before_batch_flush));
  __pyx_v_self->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":461
 *         self.max_await_recursion_depth = clone.max_await_recursion_depth
 *         self.on_before_batch_flush = clone.on_before_batch_flush
 *         self.on_after_batch_flush = clone.on_after_batch_flush             # <<<<<<<<<<<<<<
 *         for batch in clone._batches:
 *             self._batches.add(batch)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_clone->on_after_batch_flush);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->on_after_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
  __pyx_v_self->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":462
 *         self.on_before_batch_flush = clone.on_before_batch_flush
 *         self.on_after_batch_flush = clone.on_after_batch_flush
 *         for batch in clone._batches:             # <<<<<<<<<<<<<<
 *             self._batches.add(batch)
 * 
 */
  __pyx_t_1 = PyObject_GetIter(__pyx_v_clone->_batches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 462, __pyx_L1_error)
  for (;;) {
    {
      __pyx_t_2 = __pyx_t_8(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 462, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_batch, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":463
 *         self.on_after_batch_flush = clone.on_after_batch_flush
 *         for batch in clone._batches:
 *             self._batches.add(batch)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    if (unlikely(__pyx_v_self->_batches == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "add");
      __PYX_ERR(0, 463, __pyx_L1_error)
    }
    __pyx_t_9 = PySet_Add(__pyx_v_self->_batches, __pyx_v_batch); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 463, __pyx_L1_error)

    /* "asynq/scheduler.py":462
 *         self.on_before_batch_flush = clone.on_before_batch_flush
 *         self.on_after_batch_flush = clone.on_after_batch_flush
 *         for batch in clone._batches:             # <<<<<<<<<<<<<<
 *             self._batches.add(batch)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":455
 *         return clone
 * 
 *     def _accept_changes(self, clone):             # <<<<<<<<<<<<<<
 *         """Protected method accepting the changes made to this scheduler's clone."""
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._accept_changes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_batch);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_33_accept_changes(PyObject *__pyx_v_self, PyObject *__pyx_v_clone); /*proto*/
static char __pyx_doc_5asynq_9scheduler_13TaskScheduler_32_accept_changes[] = "Protected method accepting the changes made to this scheduler's clone.";
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_33_accept_changes(PyObject *__pyx_v_self, PyObject *__pyx_v_clone) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_accept_changes (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_clone), __pyx_ptype_5asynq_9scheduler_TaskScheduler, 1, "clone", 0))) __PYX_ERR(0, 455, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_32_accept_changes(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_clone));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_32_accept_changes(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_clone) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_accept_changes", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_13TaskScheduler__accept_changes(__pyx_v_self, __pyx_v_clone, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._accept_changes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":465
 *             self._batches.add(batch)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \
 *             (type(self), repr(self.name),
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_35__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_35__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_34__str__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_34__str__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "asynq/scheduler.py":466
 * 
 *     def __str__(self):
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \             # <<<<<<<<<<<<<<
 *             (type(self), repr(self.name),
 *              len(self._tasks), len(self._new_tasks), len(self._batches),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "asynq/scheduler.py":467
 *     def __str__(self):
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \
 *             (type(self), repr(self.name),             # <<<<<<<<<<<<<<
 *              len(self._tasks), len(self._new_tasks), len(self._batches),
 *              str(self.active_task))
 */
  __pyx_t_1 = __pyx_v_self->name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":468
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \
 *             (type(self), repr(self.name),
 *              len(self._tasks), len(self._new_tasks), len(self._batches),             # <<<<<<<<<<<<<<
 *              str(self.active_task))
 * 
 */
  __pyx_t_1 = __pyx_v_self->_tasks;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __pyx_v_self->_new_tasks;
  __Pyx_INCREF(__pyx_t_4);
  if (unlikely(__pyx_t_4 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 468, __pyx_L1_error)
  }
  __pyx_t_3 = PyList_GET_SIZE(__pyx_t_4); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __pyx_v_self->_batches;
  __Pyx_INCREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 468, __pyx_L1_error)
  }
  __pyx_t_3 = PySet_GET_SIZE(__pyx_t_5); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "asynq/scheduler.py":469
 *             (type(self), repr(self.name),
 *              len(self._tasks), len(self._new_tasks), len(self._batches),
 *              str(self.active_task))             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->active_task));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->active_task));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)__pyx_v_self->active_task));
  __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "asynq/scheduler.py":467
 *     def __str__(self):
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \
 *             (type(self), repr(self.name),             # <<<<<<<<<<<<<<
 *              len(self._tasks), len(self._new_tasks), len(self._batches),
 *              str(self.active_task))
 */
  __pyx_t_6 = PyTuple_New(6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_7);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_7 = 0;

  /* "asynq/scheduler.py":466
 * 
 *     def __str__(self):
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \             # <<<<<<<<<<<<<<
 *             (type(self), repr(self.name),
 *              len(self._tasks), len(self._new_tasks), len(self._batches),
 */
  __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_s_s_d_d_tasks_d_batches_active, __pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":465
 *             self._batches.add(batch)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return '%s %s (%d + %d tasks, %d batches; active task: %s)' % \
 *             (type(self), repr(self.name),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":471
 *              str(self.active_task))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.__str__()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_37__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_37__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_36__repr__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_36__repr__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "asynq/scheduler.py":472
 * 
 *     def __repr__(self):
 *         return self.__str__()             # <<<<<<<<<<<<<<
 * 
 *     def dump(self, indent=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_str_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":471
 *              str(self.active_task))
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.__str__()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":474
 *         return self.__str__()
 * 
 *     def dump(self, indent=0):             # <<<<<<<<<<<<<<
 *         debug.write(debug.str(self), indent)
 *         if self._tasks:
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_39dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump *__pyx_optional_args) {
  int __pyx_v_indent = ((int)0);
  PyObject *__pyx_v_task = NULL;
  PyObject *__pyx_v_batch = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  __Pyx_RefNannySetupContext("dump", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_indent = __pyx_optional_args->indent;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_39dump)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_indent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":475
 * 
 *     def dump(self, indent=0):
 *         debug.write(debug.str(self), indent)             # <<<<<<<<<<<<<<
 *         if self._tasks:
 *             debug.write('Task queue:', indent + 1)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, ((PyObject *)__pyx_v_self));
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_indent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_7 = 1;
    }
  }
  __pyx_t_6 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_5) {
    __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_7, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_7, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":476
 *     def dump(self, indent=0):
 *         debug.write(debug.str(self), indent)
 *         if self._tasks:             # <<<<<<<<<<<<<<
 *             debug.write('Task queue:', indent + 1)
 *             for task in self._tasks:
 */
  __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_self->_tasks); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 476, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":477
 *         debug.write(debug.str(self), indent)
 *         if self._tasks:
 *             debug.write('Task queue:', indent + 1)             # <<<<<<<<<<<<<<
 *             for task in self._tasks:
 *                 task.dump(indent + 2)
 */
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    __pyx_t_2 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Task_queue);
    __Pyx_GIVEREF(__pyx_kp_s_Task_queue);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_7, __pyx_kp_s_Task_queue);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_7, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":478
 *         if self._tasks:
 *             debug.write('Task queue:', indent + 1)
 *             for task in self._tasks:             # <<<<<<<<<<<<<<
 *                 task.dump(indent + 2)
 *         else:
 */
    if (likely(PyList_CheckExact(__pyx_v_self->_tasks)) || PyTuple_CheckExact(__pyx_v_self->_tasks)) {
      __pyx_t_1 = __pyx_v_self->_tasks; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->_tasks); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 478, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 478, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 478, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_6); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 478, __pyx_L1_error)
          #else
          __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 478, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 478, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_task, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "asynq/scheduler.py":479
 *             debug.write('Task queue:', indent + 1)
 *             for task in self._tasks:
 *                 task.dump(indent + 2)             # <<<<<<<<<<<<<<
 *         else:
 *             debug.write('No tasks in task queue.', indent + 1)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_dump); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 479, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_indent + 2)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 479, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (!__pyx_t_3) {
        __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_6);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "asynq/scheduler.py":478
 *         if self._tasks:
 *             debug.write('Task queue:', indent + 1)
 *             for task in self._tasks:             # <<<<<<<<<<<<<<
 *                 task.dump(indent + 2)
 *         else:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":476
 *     def dump(self, indent=0):
 *         debug.write(debug.str(self), indent)
 *         if self._tasks:             # <<<<<<<<<<<<<<
 *             debug.write('Task queue:', indent + 1)
 *             for task in self._tasks:
 */
    goto __pyx_L3;
  }

  /* "asynq/scheduler.py":481
 *                 task.dump(indent + 2)
 *         else:
 *             debug.write('No tasks in task queue.', indent + 1)             # <<<<<<<<<<<<<<
 *         if self._new_tasks:
 *             debug.write('New tasks:', indent + 1)
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_7 = 1;
      }
    }
    __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_No_tasks_in_task_queue);
    __Pyx_GIVEREF(__pyx_kp_s_No_tasks_in_task_queue);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_kp_s_No_tasks_in_task_queue);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "asynq/scheduler.py":482
 *         else:
 *             debug.write('No tasks in task queue.', indent + 1)
 *         if self._new_tasks:             # <<<<<<<<<<<<<<
 *             debug.write('New tasks:', indent + 1)
 *             for task in self._new_tasks:
 */
  __pyx_t_8 = (__pyx_v_self->_new_tasks != Py_None) && (PyList_GET_SIZE(__pyx_v_self->_new_tasks) != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":483
 *             debug.write('No tasks in task queue.', indent + 1)
 *         if self._new_tasks:
 *             debug.write('New tasks:', indent + 1)             # <<<<<<<<<<<<<<
 *             for task in self._new_tasks:
 *                 task.dump(indent + 2)
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    __pyx_t_5 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_New_tasks);
    __Pyx_GIVEREF(__pyx_kp_s_New_tasks);
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_7, __pyx_kp_s_New_tasks);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_7, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":484
 *         if self._new_tasks:
 *             debug.write('New tasks:', indent + 1)
 *             for task in self._new_tasks:             # <<<<<<<<<<<<<<
 *                 task.dump(indent + 2)
 *         if self._batches:
 */
    if (unlikely(__pyx_v_self->_new_tasks == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 484, __pyx_L1_error)
    }
    __pyx_t_1 = __pyx_v_self->_new_tasks; __Pyx_INCREF(__pyx_t_1); __pyx_t_7 = 0;
    for (;;) {
      if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_1)) break;
      #if CYTHON_COMPILING_IN_CPYTHON
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_7); __Pyx_INCREF(__pyx_t_4); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_v_task, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "asynq/scheduler.py":485
 *             debug.write('New tasks:', indent + 1)
 *             for task in self._new_tasks:
 *                 task.dump(indent + 2)             # <<<<<<<<<<<<<<
 *         if self._batches:
 *             debug.write('Batches:', indent + 1)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_dump); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_2 = __Pyx_PyInt_From_long((__pyx_v_indent + 2)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 485, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_GIVEREF(__pyx_t_2);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_2);
        __pyx_t_2 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 485, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "asynq/scheduler.py":484
 *         if self._new_tasks:
 *             debug.write('New tasks:', indent + 1)
 *             for task in self._new_tasks:             # <<<<<<<<<<<<<<
 *                 task.dump(indent + 2)
 *         if self._batches:
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":482
 *         else:
 *             debug.write('No tasks in task queue.', indent + 1)
 *         if self._new_tasks:             # <<<<<<<<<<<<<<
 *             debug.write('New tasks:', indent + 1)
 *             for task in self._new_tasks:
 */
  }

  /* "asynq/scheduler.py":486
 *             for task in self._new_tasks:
 *                 task.dump(indent + 2)
 *         if self._batches:             # <<<<<<<<<<<<<<
 *             debug.write('Batches:', indent + 1)
 *             for batch in self._batches:
 */
  __pyx_t_8 = (__pyx_v_self->_batches != Py_None) && (PySet_GET_SIZE(__pyx_v_self->_batches) != 0);
  if (__pyx_t_8) {

    /* "asynq/scheduler.py":487
 *                 task.dump(indent + 2)
 *         if self._batches:
 *             debug.write('Batches:', indent + 1)             # <<<<<<<<<<<<<<
 *             for batch in self._batches:
 *                 batch.dump(indent + 2)
 */
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_write); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_indent + 1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    __pyx_t_2 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (__pyx_t_3) {
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3); __pyx_t_3 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Batches);
    __Pyx_GIVEREF(__pyx_kp_s_Batches);
    PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_7, __pyx_kp_s_Batches);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_7, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":488
 *         if self._batches:
 *             debug.write('Batches:', indent + 1)
 *             for batch in self._batches:             # <<<<<<<<<<<<<<
 *                 batch.dump(indent + 2)
 * 
 */
    __pyx_t_1 = PyObject_GetIter(__pyx_v_self->_batches); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 488, __pyx_L1_error)
    for (;;) {
      {
        __pyx_t_5 = __pyx_t_9(__pyx_t_1);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 488, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_5);
      }
      __Pyx_XDECREF_SET(__pyx_v_batch, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "asynq/scheduler.py":489
 *             debug.write('Batches:', indent + 1)
 *             for batch in self._batches:
 *                 batch.dump(indent + 2)             # <<<<<<<<<<<<<<
 * 
 *     def try_time_based_dump(self, last_task=None):
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_batch, __pyx_n_s_dump); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyInt_From_long((__pyx_v_indent + 2)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 489, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (!__pyx_t_3) {
        __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 489, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_5);
      } else {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 489, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 489, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "asynq/scheduler.py":488
 *         if self._batches:
 *             debug.write('Batches:', indent + 1)
 *             for batch in self._batches:             # <<<<<<<<<<<<<<
 *                 batch.dump(indent + 2)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":486
 *             for task in self._new_tasks:
 *                 task.dump(indent + 2)
 *         if self._batches:             # <<<<<<<<<<<<<<
 *             debug.write('Batches:', indent + 1)
 *             for batch in self._batches:
 */
  }

  /* "asynq/scheduler.py":474
 *         return self.__str__()
 * 
 *     def dump(self, indent=0):             # <<<<<<<<<<<<<<
 *         debug.write(debug.str(self), indent)
 *         if self._tasks:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_task);
  __Pyx_XDECREF(__pyx_v_batch);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_39dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_39dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_indent;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dump (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_indent,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_indent);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dump") < 0)) __PYX_ERR(0, 474, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_indent = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_indent == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L3_error)
    } else {
      __pyx_v_indent = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dump", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 474, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_38dump(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_indent);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_38dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_v_indent) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump __pyx_t_2;
  __Pyx_RefNannySetupContext("dump", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.indent = __pyx_v_indent;
  __pyx_t_1 = __pyx_vtabptr_5asynq_9scheduler_TaskScheduler->dump(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":491
 *                 batch.dump(indent + 2)
 * 
 *     def try_time_based_dump(self, last_task=None):             # <<<<<<<<<<<<<<
 *         current_time = time.time()
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_41try_time_based_dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_13TaskScheduler_try_time_based_dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump *__pyx_optional_args) {
  PyObject *__pyx_v_last_task = ((PyObject *)Py_None);
  float __pyx_v_current_time;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  float __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("try_time_based_dump", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_last_task = __pyx_optional_args->last_task;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_try_time_based_dump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_41try_time_based_dump)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_last_task); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_last_task);
        __Pyx_GIVEREF(__pyx_v_last_task);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_last_task);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 491, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "asynq/scheduler.py":492
 * 
 *     def try_time_based_dump(self, last_task=None):
 *         current_time = time.time()             # <<<<<<<<<<<<<<
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
 *             return
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = __pyx_PyFloat_AsFloat(__pyx_t_1); if (unlikely((__pyx_t_6 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_current_time = __pyx_t_6;

  /* "asynq/scheduler.py":493
 *     def try_time_based_dump(self, last_task=None):
 *         current_time = time.time()
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:             # <<<<<<<<<<<<<<
 *             return
 *         self._last_dump_time = current_time
 */
  __pyx_t_7 = (((__pyx_v_current_time - __pyx_v_self->_last_dump_time) < __pyx_v_5asynq_9scheduler__debug_options->SCHEDULER_STATE_DUMP_INTERVAL) != 0);
  if (__pyx_t_7) {

    /* "asynq/scheduler.py":494
 *         current_time = time.time()
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
 *             return             # <<<<<<<<<<<<<<
 *         self._last_dump_time = current_time
 *         debug.write('\n--- Scheduler state dump: --------------------------------------------')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "asynq/scheduler.py":493
 *     def try_time_based_dump(self, last_task=None):
 *         current_time = time.time()
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:             # <<<<<<<<<<<<<<
 *             return
 *         self._last_dump_time = current_time
 */
  }

  /* "asynq/scheduler.py":495
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
 *             return
 *         self._last_dump_time = current_time             # <<<<<<<<<<<<<<
 *         debug.write('\n--- Scheduler state dump: --------------------------------------------')
 *         try:
 */
  __pyx_v_self->_last_dump_time = __pyx_v_current_time;

  /* "asynq/scheduler.py":496
 *             return
 *         self._last_dump_time = current_time
 *         debug.write('\n--- Scheduler state dump: --------------------------------------------')             # <<<<<<<<<<<<<<
 *         try:
 *             self.dump()
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":497
 *         self._last_dump_time = current_time
 *         debug.write('\n--- Scheduler state dump: --------------------------------------------')
 *         try:             # <<<<<<<<<<<<<<
 *             self.dump()
 *             if last_task is not None:
 */
  /*try:*/ {

    /* "asynq/scheduler.py":498
 *         debug.write('\n--- Scheduler state dump: --------------------------------------------')
 *         try:
 *             self.dump()             # <<<<<<<<<<<<<<
 *             if last_task is not None:
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->__pyx_vtab)->dump(__pyx_v_self, 0, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "asynq/scheduler.py":499
 *         try:
 *             self.dump()
 *             if last_task is not None:             # <<<<<<<<<<<<<<
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)
 *         finally:
 */
    __pyx_t_7 = (__pyx_v_last_task != Py_None);
    __pyx_t_8 = (__pyx_t_7 != 0);
    if (__pyx_t_8) {

      /* "asynq/scheduler.py":500
 *             self.dump()
 *             if last_task is not None:
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)             # <<<<<<<<<<<<<<
 *         finally:
 *             debug.write('----------------------------------------------------------------------\n')
 */
      __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_last_task); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 500, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_last_task);
        __Pyx_GIVEREF(__pyx_v_last_task);
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_v_last_task);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 500, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Last_task_s, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_10 = 1;
        }
      }
      __pyx_t_9 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_t_4);
      __Pyx_INCREF(__pyx_int_1);
      __Pyx_GIVEREF(__pyx_int_1);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_int_1);
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":499
 *         try:
 *             self.dump()
 *             if last_task is not None:             # <<<<<<<<<<<<<<
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)
 *         finally:
 */
    }
  }

  /* "asynq/scheduler.py":502
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)
 *         finally:
 *             debug.write('----------------------------------------------------------------------\n')             # <<<<<<<<<<<<<<
 *             stdout.flush()
 *             stderr.flush()
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":503
 *         finally:
 *             debug.write('----------------------------------------------------------------------\n')
 *             stdout.flush()             # <<<<<<<<<<<<<<
 *             stderr.flush()
 * 
 */
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 503, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_9))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_9, function);
        }
      }
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "asynq/scheduler.py":504
 *             debug.write('----------------------------------------------------------------------\n')
 *             stdout.flush()
 *             stderr.flush()             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_stderr); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_flush); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (__pyx_t_9) {
        __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L6;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L5_error:;
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
      {

        /* "asynq/scheduler.py":502
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)
 *         finally:
 *             debug.write('----------------------------------------------------------------------\n')             # <<<<<<<<<<<<<<
 *             stdout.flush()
 *             stderr.flush()
 */
        __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":503
 *         finally:
 *             debug.write('----------------------------------------------------------------------\n')
 *             stdout.flush()             # <<<<<<<<<<<<<<
 *             stderr.flush()
 * 
 */
        __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_stdout); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_flush); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 503, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
          }
        }
        if (__pyx_t_2) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L9_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "asynq/scheduler.py":504
 *             debug.write('----------------------------------------------------------------------\n')
 *             stdout.flush()
 *             stderr.flush()             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_stderr); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 504, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_flush); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
          }
        }
        if (__pyx_t_9) {
          __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L9_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        } else {
          __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L9_error)
        }
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
      goto __pyx_L1_error;
      __pyx_L9_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "asynq/scheduler.py":491
 *                 batch.dump(indent + 2)
 * 
 *     def try_time_based_dump(self, last_task=None):             # <<<<<<<<<<<<<<
 *         current_time = time.time()
 *         if (current_time - self._last_dump_time) < _debug_options.SCHEDULER_STATE_DUMP_INTERVAL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.try_time_based_dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_41try_time_based_dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_41try_time_based_dump(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_last_task = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("try_time_based_dump (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_last_task,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_last_task);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "try_time_based_dump") < 0)) __PYX_ERR(0, 491, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_last_task = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("try_time_based_dump", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 491, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.try_time_based_dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_40try_time_based_dump(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), __pyx_v_last_task);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_40try_time_based_dump(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_last_task) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump __pyx_t_2;
  __Pyx_RefNannySetupContext("try_time_based_dump", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.last_task = __pyx_v_last_task;
  __pyx_t_1 = __pyx_vtabptr_5asynq_9scheduler_TaskScheduler->try_time_based_dump(__pyx_v_self, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.try_time_based_dump", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":30
 * 
 * cdef class TaskScheduler(object):
 *     cdef public set _empty_set             # <<<<<<<<<<<<<<
 *     cdef public float _last_dump_time
 *     cdef public list _new_tasks
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_empty_set);
  __pyx_r = __pyx_v_self->_empty_set;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 30, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_empty_set);
  __Pyx_DECREF(__pyx_v_self->_empty_set);
  __pyx_v_self->_empty_set = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._empty_set.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_empty_set_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_empty_set);
  __Pyx_DECREF(__pyx_v_self->_empty_set);
  __pyx_v_self->_empty_set = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":31
 * cdef class TaskScheduler(object):
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time             # <<<<<<<<<<<<<<
 *     cdef public list _new_tasks
 *     cdef public object _tasks
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_15_last_dump_time___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_15_last_dump_time___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->_last_dump_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._last_dump_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  float __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsFloat(__pyx_v_value); if (unlikely((__pyx_t_1 == (float)-1) && PyErr_Occurred())) __PYX_ERR(1, 31, __pyx_L1_error)
  __pyx_v_self->_last_dump_time = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._last_dump_time.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":32
 *     cdef public set _empty_set
 *     cdef public float _last_dump_time
 *     cdef public list _new_tasks             # <<<<<<<<<<<<<<
 *     cdef public object _tasks
 *     cdef public set _batches
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_new_tasks);
  __pyx_r = __pyx_v_self->_new_tasks;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 32, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_new_tasks);
  __Pyx_DECREF(__pyx_v_self->_new_tasks);
  __pyx_v_self->_new_tasks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._new_tasks.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_10_new_tasks_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_new_tasks);
  __Pyx_DECREF(__pyx_v_self->_new_tasks);
  __pyx_v_self->_new_tasks = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":33
 *     cdef public float _last_dump_time
 *     cdef public list _new_tasks
 *     cdef public object _tasks             # <<<<<<<<<<<<<<
 *     cdef public set _batches
 *     cdef public int await_recursion_depth
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_tasks);
  __pyx_r = __pyx_v_self->_tasks;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_tasks);
  __Pyx_DECREF(__pyx_v_self->_tasks);
  __pyx_v_self->_tasks = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_6_tasks_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_tasks);
  __Pyx_DECREF(__pyx_v_self->_tasks);
  __pyx_v_self->_tasks = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":34
 *     cdef public list _new_tasks
 *     cdef public object _tasks
 *     cdef public set _batches             # <<<<<<<<<<<<<<
 *     cdef public int await_recursion_depth
 *     cdef public int max_await_recursion_depth
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_batches);
  __pyx_r = __pyx_v_self->_batches;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PySet_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "set", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 34, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_batches);
  __Pyx_DECREF(__pyx_v_self->_batches);
  __pyx_v_self->_batches = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler._batches.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_8_batches_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_batches);
  __Pyx_DECREF(__pyx_v_self->_batches);
  __pyx_v_self->_batches = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":35
 *     cdef public object _tasks
 *     cdef public set _batches
 *     cdef public int await_recursion_depth             # <<<<<<<<<<<<<<
 *     cdef public int max_await_recursion_depth
 *     cdef public core_events.EventHook on_before_batch_flush
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->await_recursion_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.await_recursion_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 35, __pyx_L1_error)
  __pyx_v_self->await_recursion_depth = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.await_recursion_depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":36
 *     cdef public set _batches
 *     cdef public int await_recursion_depth
 *     cdef public int max_await_recursion_depth             # <<<<<<<<<<<<<<
 *     cdef public core_events.EventHook on_before_batch_flush
 *     cdef public core_events.EventHook on_after_batch_flush
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->max_await_recursion_depth); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.max_await_recursion_depth.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 36, __pyx_L1_error)
  __pyx_v_self->max_await_recursion_depth = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.max_await_recursion_depth.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":37
 *     cdef public int await_recursion_depth
 *     cdef public int max_await_recursion_depth
 *     cdef public core_events.EventHook on_before_batch_flush             # <<<<<<<<<<<<<<
 *     cdef public core_events.EventHook on_after_batch_flush
 *     cdef public async_task.AsyncTask active_task
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->on_before_batch_flush));
  __pyx_r = ((PyObject *)__pyx_v_self->on_before_batch_flush);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5qcore_6events_EventHook))))) __PYX_ERR(1, 37, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->on_before_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_before_batch_flush));
  __pyx_v_self->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.on_before_batch_flush.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->on_before_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_before_batch_flush));
  __pyx_v_self->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":38
 *     cdef public int max_await_recursion_depth
 *     cdef public core_events.EventHook on_before_batch_flush
 *     cdef public core_events.EventHook on_after_batch_flush             # <<<<<<<<<<<<<<
 *     cdef public async_task.AsyncTask active_task
 *     cdef public str name
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
  __pyx_r = ((PyObject *)__pyx_v_self->on_after_batch_flush);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5qcore_6events_EventHook))))) __PYX_ERR(1, 38, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->on_after_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
  __pyx_v_self->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.on_after_batch_flush.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->on_after_batch_flush);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->on_after_batch_flush));
  __pyx_v_self->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":39
 *     cdef public core_events.EventHook on_before_batch_flush
 *     cdef public core_events.EventHook on_after_batch_flush
 *     cdef public async_task.AsyncTask active_task             # <<<<<<<<<<<<<<
 *     cdef public str name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->active_task));
  __pyx_r = ((PyObject *)__pyx_v_self->active_task);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(1, 39, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->active_task);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
  __pyx_v_self->active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.active_task.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_11active_task_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->active_task);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->active_task));
  __pyx_v_self->active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.pxd":40
 *     cdef public core_events.EventHook on_after_batch_flush
 *     cdef public async_task.AsyncTask active_task
 *     cdef public str name             # <<<<<<<<<<<<<<
 * 
 *     cpdef TaskSchedulerActivator activate(self, bint must_await_on_exit, bint must_flush_batches_on_exit)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name___get__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_13TaskScheduler_4name___get__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name_2__set__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name_2__set__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskScheduler.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name_4__del__(((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_13TaskScheduler_4name_4__del__(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":508
 * 
 * class TaskSchedulerActivator(object):
 *     def __init__(self, scheduler, must_await_on_exit, must_flush_batches_on_exit):             # <<<<<<<<<<<<<<
 *         self._scheduler = scheduler
 *         self._old_scheduler = None
 */

/* Python wrapper */
static int __pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_scheduler = 0;
  PyObject *__pyx_v_must_await_on_exit = 0;
  PyObject *__pyx_v_must_flush_batches_on_exit = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_scheduler,&__pyx_n_s_must_await_on_exit,&__pyx_n_s_must_flush_batches_on_exit,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_scheduler)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_must_await_on_exit)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 1); __PYX_ERR(0, 508, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_must_flush_batches_on_exit)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, 2); __PYX_ERR(0, 508, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 508, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_scheduler = values[0];
    __pyx_v_must_await_on_exit = values[1];
    __pyx_v_must_flush_batches_on_exit = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 508, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskSchedulerActivator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator___init__(((struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)__pyx_v_self), __pyx_v_scheduler, __pyx_v_must_await_on_exit, __pyx_v_must_flush_batches_on_exit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator___init__(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_v_self, PyObject *__pyx_v_scheduler, PyObject *__pyx_v_must_await_on_exit, PyObject *__pyx_v_must_flush_batches_on_exit) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "asynq/scheduler.py":509
 * class TaskSchedulerActivator(object):
 *     def __init__(self, scheduler, must_await_on_exit, must_flush_batches_on_exit):
 *         self._scheduler = scheduler             # <<<<<<<<<<<<<<
 *         self._old_scheduler = None
 *         self._must_await_on_exit = must_await_on_exit
 */
  if (!(likely(((__pyx_v_scheduler) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_scheduler, __pyx_ptype_5asynq_9scheduler_TaskScheduler))))) __PYX_ERR(0, 509, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_scheduler;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_scheduler);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_scheduler));
  __pyx_v_self->_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":510
 *     def __init__(self, scheduler, must_await_on_exit, must_flush_batches_on_exit):
 *         self._scheduler = scheduler
 *         self._old_scheduler = None             # <<<<<<<<<<<<<<
 *         self._must_await_on_exit = must_await_on_exit
 *         self._must_flush_batches_on_exit = must_flush_batches_on_exit
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_old_scheduler);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_old_scheduler));
  __pyx_v_self->_old_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None);

  /* "asynq/scheduler.py":511
 *         self._scheduler = scheduler
 *         self._old_scheduler = None
 *         self._must_await_on_exit = must_await_on_exit             # <<<<<<<<<<<<<<
 *         self._must_flush_batches_on_exit = must_flush_batches_on_exit
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_must_await_on_exit); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 511, __pyx_L1_error)
  __pyx_v_self->_must_await_on_exit = __pyx_t_2;

  /* "asynq/scheduler.py":512
 *         self._old_scheduler = None
 *         self._must_await_on_exit = must_await_on_exit
 *         self._must_flush_batches_on_exit = must_flush_batches_on_exit             # <<<<<<<<<<<<<<
 * 
 *     def __enter__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_must_flush_batches_on_exit); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 512, __pyx_L1_error)
  __pyx_v_self->_must_flush_batches_on_exit = __pyx_t_2;

  /* "asynq/scheduler.py":508
 * 
 * class TaskSchedulerActivator(object):
 *     def __init__(self, scheduler, must_await_on_exit, must_flush_batches_on_exit):             # <<<<<<<<<<<<<<
 *         self._scheduler = scheduler
 *         self._old_scheduler = None
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.TaskSchedulerActivator.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":514
 *         self._must_flush_batches_on_exit = must_flush_batches_on_exit
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:
 *             debug.write("@async: -> scheduler activation")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_3__enter__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_3__enter__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__enter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator_2__enter__(((struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator_2__enter__(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__enter__", 0);

  /* "asynq/scheduler.py":515
 * 
 *     def __enter__(self):
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *             debug.write("@async: -> scheduler activation")
 *         self._old_scheduler = get_scheduler()
 */
  __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULER_CHANGE != 0);
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":516
 *     def __enter__(self):
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:
 *             debug.write("@async: -> scheduler activation")             # <<<<<<<<<<<<<<
 *         self._old_scheduler = get_scheduler()
 *         set_scheduler(self._scheduler)
 */
    __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/scheduler.py":515
 * 
 *     def __enter__(self):
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *             debug.write("@async: -> scheduler activation")
 *         self._old_scheduler = get_scheduler()
 */
  }

  /* "asynq/scheduler.py":517
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:
 *             debug.write("@async: -> scheduler activation")
 *         self._old_scheduler = get_scheduler()             # <<<<<<<<<<<<<<
 *         set_scheduler(self._scheduler)
 * 
 */
  __pyx_t_2 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_old_scheduler);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_old_scheduler));
  __pyx_v_self->_old_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "asynq/scheduler.py":518
 *             debug.write("@async: -> scheduler activation")
 *         self._old_scheduler = get_scheduler()
 *         set_scheduler(self._scheduler)             # <<<<<<<<<<<<<<
 * 
 *     def __exit__(self, type, value, traceback):
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_scheduler);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = __pyx_f_5asynq_9scheduler_set_scheduler(__pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "asynq/scheduler.py":514
 *         self._must_flush_batches_on_exit = must_flush_batches_on_exit
 * 
 *     def __enter__(self):             # <<<<<<<<<<<<<<
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:
 *             debug.write("@async: -> scheduler activation")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.TaskSchedulerActivator.__enter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":520
 *         set_scheduler(self._scheduler)
 * 
 *     def __exit__(self, type, value, traceback):             # <<<<<<<<<<<<<<
 *         try:
 *             try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_5__exit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_5__exit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_type = 0;
  CYTHON_UNUSED PyObject *__pyx_v_value = 0;
  CYTHON_UNUSED PyObject *__pyx_v_traceback = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__exit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_type,&__pyx_n_s_value,&__pyx_n_s_traceback,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, 1); __PYX_ERR(0, 520, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_traceback)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, 2); __PYX_ERR(0, 520, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__exit__") < 0)) __PYX_ERR(0, 520, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_type = values[0];
    __pyx_v_value = values[1];
    __pyx_v_traceback = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__exit__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 520, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.TaskSchedulerActivator.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator_4__exit__(((struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)__pyx_v_self), __pyx_v_type, __pyx_v_value, __pyx_v_traceback);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_22TaskSchedulerActivator_4__exit__(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_type, CYTHON_UNUSED PyObject *__pyx_v_value, CYTHON_UNUSED PyObject *__pyx_v_traceback) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  char const *__pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  __Pyx_RefNannySetupContext("__exit__", 0);

  /* "asynq/scheduler.py":521
 * 
 *     def __exit__(self, type, value, traceback):
 *         try:             # <<<<<<<<<<<<<<
 *             try:
 *                 if self._must_await_on_exit:
 */
  /*try:*/ {

    /* "asynq/scheduler.py":522
 *     def __exit__(self, type, value, traceback):
 *         try:
 *             try:             # <<<<<<<<<<<<<<
 *                 if self._must_await_on_exit:
 *                     self._scheduler.await()
 */
    /*try:*/ {

      /* "asynq/scheduler.py":523
 *         try:
 *             try:
 *                 if self._must_await_on_exit:             # <<<<<<<<<<<<<<
 *                     self._scheduler.await()
 *             finally:
 */
      __pyx_t_1 = (__pyx_v_self->_must_await_on_exit != 0);
      if (__pyx_t_1) {

        /* "asynq/scheduler.py":524
 *             try:
 *                 if self._must_await_on_exit:
 *                     self._scheduler.await()             # <<<<<<<<<<<<<<
 *             finally:
 *                 if self._must_flush_batches_on_exit:
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_scheduler), __pyx_n_s_await); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (__pyx_t_4) {
          __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        } else {
          __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L7_error)
        }
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynq/scheduler.py":523
 *         try:
 *             try:
 *                 if self._must_await_on_exit:             # <<<<<<<<<<<<<<
 *                     self._scheduler.await()
 *             finally:
 */
      }
    }

    /* "asynq/scheduler.py":526
 *                     self._scheduler.await()
 *             finally:
 *                 if self._must_flush_batches_on_exit:             # <<<<<<<<<<<<<<
 *                     self._scheduler.flush_batches()
 *         finally:
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_t_1 = (__pyx_v_self->_must_flush_batches_on_exit != 0);
        if (__pyx_t_1) {

          /* "asynq/scheduler.py":527
 *             finally:
 *                 if self._must_flush_batches_on_exit:
 *                     self._scheduler.flush_batches()             # <<<<<<<<<<<<<<
 *         finally:
 *             set_scheduler(self._old_scheduler)
 */
          __pyx_t_5 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->_scheduler->__pyx_vtab)->flush_batches(__pyx_v_self->_scheduler, 0, NULL); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 527, __pyx_L4_error)

          /* "asynq/scheduler.py":526
 *                     self._scheduler.await()
 *             finally:
 *                 if self._must_flush_batches_on_exit:             # <<<<<<<<<<<<<<
 *                     self._scheduler.flush_batches()
 *         finally:
 */
        }
        goto __pyx_L8;
      }
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __pyx_L7_error:;
        __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __pyx_t_5 = __pyx_lineno; __pyx_t_6 = __pyx_clineno; __pyx_t_7 = __pyx_filename;
        {
          __pyx_t_1 = (__pyx_v_self->_must_flush_batches_on_exit != 0);
          if (__pyx_t_1) {

            /* "asynq/scheduler.py":527
 *             finally:
 *                 if self._must_flush_batches_on_exit:
 *                     self._scheduler.flush_batches()             # <<<<<<<<<<<<<<
 *         finally:
 *             set_scheduler(self._old_scheduler)
 */
            __pyx_t_14 = ((struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler *)__pyx_v_self->_scheduler->__pyx_vtab)->flush_batches(__pyx_v_self->_scheduler, 0, NULL); if (unlikely(__pyx_t_14 == -1)) __PYX_ERR(0, 527, __pyx_L12_error)

            /* "asynq/scheduler.py":526
 *                     self._scheduler.await()
 *             finally:
 *                 if self._must_flush_batches_on_exit:             # <<<<<<<<<<<<<<
 *                     self._scheduler.flush_batches()
 *         finally:
 */
          }
        }
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        }
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ErrRestore(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
        __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_6; __pyx_filename = __pyx_t_7;
        goto __pyx_L4_error;
        __pyx_L12_error:;
        __Pyx_PyThreadState_assign
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_12, __pyx_t_13);
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0;
        goto __pyx_L4_error;
      }
      __pyx_L8:;
    }
  }

  /* "asynq/scheduler.py":529
 *                     self._scheduler.flush_batches()
 *         finally:
 *             set_scheduler(self._old_scheduler)             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:
 *                 debug.write("@async: <- scheduler activation")
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_2 = ((PyObject *)__pyx_v_self->_old_scheduler);
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_3 = __pyx_f_5asynq_9scheduler_set_scheduler(__pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "asynq/scheduler.py":530
 *         finally:
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: <- scheduler activation")
 * 
 */
      __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULER_CHANGE != 0);
      if (__pyx_t_1) {

        /* "asynq/scheduler.py":531
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:
 *                 debug.write("@async: <- scheduler activation")             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "asynq/scheduler.py":530
 *         finally:
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: <- scheduler activation")
 * 
 */
      }
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_9, &__pyx_t_8);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11) < 0)) __Pyx_ErrFetch(&__pyx_t_13, &__pyx_t_12, &__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_8);
      __pyx_t_6 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
      {

        /* "asynq/scheduler.py":529
 *                     self._scheduler.flush_batches()
 *         finally:
 *             set_scheduler(self._old_scheduler)             # <<<<<<<<<<<<<<
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:
 *                 debug.write("@async: <- scheduler activation")
 */
        __pyx_t_3 = ((PyObject *)__pyx_v_self->_old_scheduler);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_2 = __pyx_f_5asynq_9scheduler_set_scheduler(__pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "asynq/scheduler.py":530
 *         finally:
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: <- scheduler activation")
 * 
 */
        __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULER_CHANGE != 0);
        if (__pyx_t_1) {

          /* "asynq/scheduler.py":531
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:
 *                 debug.write("@async: <- scheduler activation")             # <<<<<<<<<<<<<<
 * 
 * 
 */
          __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_write); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "asynq/scheduler.py":530
 *         finally:
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *                 debug.write("@async: <- scheduler activation")
 * 
 */
        }
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_ErrRestore(__pyx_t_13, __pyx_t_12, __pyx_t_11);
      __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_11 = 0; __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_15;
      goto __pyx_L1_error;
      __pyx_L16_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_9, __pyx_t_8);
      }
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_10 = 0; __pyx_t_9 = 0; __pyx_t_8 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "asynq/scheduler.py":520
 *         set_scheduler(self._scheduler)
 * 
 *     def __exit__(self, type, value, traceback):             # <<<<<<<<<<<<<<
 *         try:
 *             try:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("asynq.scheduler.TaskSchedulerActivator.__exit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":535
 * 
 * class LocalTaskSchedulerState(threading.local):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.last_id = 0
 *         self.reset()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_23LocalTaskSchedulerState_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_5asynq_9scheduler_23LocalTaskSchedulerState_1__init__ = {"__init__", (PyCFunction)__pyx_pw_5asynq_9scheduler_23LocalTaskSchedulerState_1__init__, METH_O, 0};
static PyObject *__pyx_pw_5asynq_9scheduler_23LocalTaskSchedulerState_1__init__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_23LocalTaskSchedulerState___init__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_23LocalTaskSchedulerState___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "asynq/scheduler.py":536
 * class LocalTaskSchedulerState(threading.local):
 *     def __init__(self):
 *         self.last_id = 0             # <<<<<<<<<<<<<<
 *         self.reset()
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_last_id, __pyx_int_0) < 0) __PYX_ERR(0, 536, __pyx_L1_error)

  /* "asynq/scheduler.py":537
 *     def __init__(self):
 *         self.last_id = 0
 *         self.reset()             # <<<<<<<<<<<<<<
 * 
 *     def reset(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_reset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 537, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":535
 * 
 * class LocalTaskSchedulerState(threading.local):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.last_id = 0
 *         self.reset()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.LocalTaskSchedulerState.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":539
 *         self.reset()
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.current = TaskScheduler()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_23LocalTaskSchedulerState_3reset(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static PyMethodDef __pyx_mdef_5asynq_9scheduler_23LocalTaskSchedulerState_3reset = {"reset", (PyCFunction)__pyx_pw_5asynq_9scheduler_23LocalTaskSchedulerState_3reset, METH_O, 0};
static PyObject *__pyx_pw_5asynq_9scheduler_23LocalTaskSchedulerState_3reset(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_23LocalTaskSchedulerState_2reset(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_23LocalTaskSchedulerState_2reset(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("reset", 0);

  /* "asynq/scheduler.py":540
 * 
 *     def reset(self):
 *         self.current = TaskScheduler()             # <<<<<<<<<<<<<<
 * 
 * _state = LocalTaskSchedulerState()
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5asynq_9scheduler_TaskScheduler), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_current, __pyx_t_1) < 0) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":539
 *         self.reset()
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.current = TaskScheduler()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.LocalTaskSchedulerState.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":546
 * 
 * 
 * def get_scheduler():             # <<<<<<<<<<<<<<
 *     global _state
 *     return _state.current
 */

static PyObject *__pyx_pw_5asynq_9scheduler_1get_scheduler(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_f_5asynq_9scheduler_get_scheduler(CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_scheduler", 0);

  /* "asynq/scheduler.py":548
 * def get_scheduler():
 *     global _state
 *     return _state.current             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_current); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5asynq_9scheduler_TaskScheduler))))) __PYX_ERR(0, 548, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":546
 * 
 * 
 * def get_scheduler():             # <<<<<<<<<<<<<<
 *     global _state
 *     return _state.current
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.get_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_1get_scheduler(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_1get_scheduler(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_scheduler (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_get_scheduler(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_get_scheduler(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_scheduler", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_scheduler(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.get_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":551
 * 
 * 
 * def reset():             # <<<<<<<<<<<<<<
 *     _state.reset()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_3reset(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_5asynq_9scheduler_3reset = {"reset", (PyCFunction)__pyx_pw_5asynq_9scheduler_3reset, METH_NOARGS, 0};
static PyObject *__pyx_pw_5asynq_9scheduler_3reset(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_2reset(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_2reset(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("reset", 0);

  /* "asynq/scheduler.py":552
 * 
 * def reset():
 *     _state.reset()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_reset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":551
 * 
 * 
 * def reset():             # <<<<<<<<<<<<<<
 *     _state.reset()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.reset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":555
 * 
 * 
 * def set_scheduler(scheduler):             # <<<<<<<<<<<<<<
 *     global _state
 *     _state.current = scheduler
 */

static PyObject *__pyx_pw_5asynq_9scheduler_5set_scheduler(PyObject *__pyx_self, PyObject *__pyx_v_scheduler); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_set_scheduler(PyObject *__pyx_v_scheduler, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("set_scheduler", 0);

  /* "asynq/scheduler.py":557
 * def set_scheduler(scheduler):
 *     global _state
 *     _state.current = scheduler             # <<<<<<<<<<<<<<
 *     if _debug_options.DUMP_SCHEDULER_CHANGE:
 *         debug.write("@async: scheduler = %s" % debug.str(scheduler))
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_current, __pyx_v_scheduler) < 0) __PYX_ERR(0, 557, __pyx_L1_error)

  /* "asynq/scheduler.py":558
 *     global _state
 *     _state.current = scheduler
 *     if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *         debug.write("@async: scheduler = %s" % debug.str(scheduler))
 * 
 */
  __pyx_t_1 = (__pyx_v_5asynq_9scheduler__debug_options->DUMP_SCHEDULER_CHANGE != 0);
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":559
 *     _state.current = scheduler
 *     if _debug_options.DUMP_SCHEDULER_CHANGE:
 *         debug.write("@async: scheduler = %s" % debug.str(scheduler))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_write); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_str); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_scheduler); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_scheduler);
      __Pyx_GIVEREF(__pyx_v_scheduler);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_scheduler);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_async_scheduler_s, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 559, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "asynq/scheduler.py":558
 *     global _state
 *     _state.current = scheduler
 *     if _debug_options.DUMP_SCHEDULER_CHANGE:             # <<<<<<<<<<<<<<
 *         debug.write("@async: scheduler = %s" % debug.str(scheduler))
 * 
 */
  }

  /* "asynq/scheduler.py":555
 * 
 * 
 * def set_scheduler(scheduler):             # <<<<<<<<<<<<<<
 *     global _state
 *     _state.current = scheduler
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.scheduler.set_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_5set_scheduler(PyObject *__pyx_self, PyObject *__pyx_v_scheduler); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_5set_scheduler(PyObject *__pyx_self, PyObject *__pyx_v_scheduler) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_scheduler (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_4set_scheduler(__pyx_self, ((PyObject *)__pyx_v_scheduler));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_4set_scheduler(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_scheduler) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("set_scheduler", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_set_scheduler(__pyx_v_scheduler, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.set_scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":562
 * 
 * 
 * def get_active_task():             # <<<<<<<<<<<<<<
 *     global _state
 *     s = _state.current
 */

static PyObject *__pyx_pw_5asynq_9scheduler_7get_active_task(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_f_5asynq_9scheduler_get_active_task(CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_s = NULL;
  struct __pyx_obj_5asynq_10async_task_AsyncTask *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get_active_task", 0);

  /* "asynq/scheduler.py":564
 * def get_active_task():
 *     global _state
 *     s = _state.current             # <<<<<<<<<<<<<<
 *     return None if s is None else s.active_task
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_5asynq_9scheduler__state, __pyx_n_s_current); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":565
 *     global _state
 *     s = _state.current
 *     return None if s is None else s.active_task             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_2 = (__pyx_v_s == Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  } else {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_s, __pyx_n_s_active_task); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5asynq_10async_task_AsyncTask))))) __PYX_ERR(0, 565, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_r = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":562
 * 
 * 
 * def get_active_task():             # <<<<<<<<<<<<<<
 *     global _state
 *     s = _state.current
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("asynq.scheduler.get_active_task", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_7get_active_task(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_7get_active_task(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_active_task (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_6get_active_task(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_6get_active_task(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("get_active_task", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler_get_active_task(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.get_active_task", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":568
 * 
 * 
 * def extract_futures(value, result):             # <<<<<<<<<<<<<<
 *     """Enumerates all the futures inside a particular value."""
 *     if value is None:
 */

static PyObject *__pyx_pw_5asynq_9scheduler_9extract_futures(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_5asynq_9scheduler_extract_futures(PyObject *__pyx_v_value, PyObject *__pyx_v_result, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_item = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("extract_futures", 0);

  /* "asynq/scheduler.py":570
 * def extract_futures(value, result):
 *     """Enumerates all the futures inside a particular value."""
 *     if value is None:             # <<<<<<<<<<<<<<
 *         pass
 *     elif isinstance(value, futures.FutureBase):
 */
  __pyx_t_1 = (__pyx_v_value == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {
    goto __pyx_L3;
  }

  /* "asynq/scheduler.py":572
 *     if value is None:
 *         pass
 *     elif isinstance(value, futures.FutureBase):             # <<<<<<<<<<<<<<
 *         result.append(value)
 *     elif type(value) is tuple or type(value) is list:
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_value, __pyx_ptype_5asynq_7futures_FutureBase); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":573
 *         pass
 *     elif isinstance(value, futures.FutureBase):
 *         result.append(value)             # <<<<<<<<<<<<<<
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:
 */
    if (unlikely(__pyx_v_result == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
      __PYX_ERR(0, 573, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_value); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 573, __pyx_L1_error)

    /* "asynq/scheduler.py":572
 *     if value is None:
 *         pass
 *     elif isinstance(value, futures.FutureBase):             # <<<<<<<<<<<<<<
 *         result.append(value)
 *     elif type(value) is tuple or type(value) is list:
 */
    goto __pyx_L3;
  }

  /* "asynq/scheduler.py":574
 *     elif isinstance(value, futures.FutureBase):
 *         result.append(value)
 *     elif type(value) is tuple or type(value) is list:             # <<<<<<<<<<<<<<
 *         for item in value:
 *             extract_futures(item, result)
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyTuple_Type)));
  __pyx_t_4 = (__pyx_t_2 != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = (((PyObject *)Py_TYPE(__pyx_v_value)) == ((PyObject *)(&PyList_Type)));
  __pyx_t_2 = (__pyx_t_4 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "asynq/scheduler.py":575
 *         result.append(value)
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:             # <<<<<<<<<<<<<<
 *             extract_futures(item, result)
 *     return result
 */
    if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
      __pyx_t_5 = __pyx_v_value; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 575, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 575, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
          #if CYTHON_COMPILING_IN_CPYTHON
          __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_8); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 575, __pyx_L1_error)
          #else
          __pyx_t_8 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 575, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_8);
          #endif
        }
      } else {
        __pyx_t_8 = __pyx_t_7(__pyx_t_5);
        if (unlikely(!__pyx_t_8)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 575, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_8);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "asynq/scheduler.py":576
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:
 *             extract_futures(item, result)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
      __pyx_t_8 = __pyx_f_5asynq_9scheduler_extract_futures(__pyx_v_item, __pyx_v_result, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 576, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "asynq/scheduler.py":575
 *         result.append(value)
 *     elif type(value) is tuple or type(value) is list:
 *         for item in value:             # <<<<<<<<<<<<<<
 *             extract_futures(item, result)
 *     return result
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "asynq/scheduler.py":574
 *     elif isinstance(value, futures.FutureBase):
 *         result.append(value)
 *     elif type(value) is tuple or type(value) is list:             # <<<<<<<<<<<<<<
 *         for item in value:
 *             extract_futures(item, result)
 */
  }
  __pyx_L3:;

  /* "asynq/scheduler.py":577
 *         for item in value:
 *             extract_futures(item, result)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "asynq/scheduler.py":568
 * 
 * 
 * def extract_futures(value, result):             # <<<<<<<<<<<<<<
 *     """Enumerates all the futures inside a particular value."""
 *     if value is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("asynq.scheduler.extract_futures", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_9extract_futures(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5asynq_9scheduler_8extract_futures[] = "Enumerates all the futures inside a particular value.";
static PyObject *__pyx_pw_5asynq_9scheduler_9extract_futures(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extract_futures (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,&__pyx_n_s_result,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_value)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_result)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("extract_futures", 1, 2, 2, 1); __PYX_ERR(0, 568, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "extract_futures") < 0)) __PYX_ERR(0, 568, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_value = values[0];
    __pyx_v_result = ((PyObject*)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("extract_futures", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 568, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("asynq.scheduler.extract_futures", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_result), (&PyList_Type), 1, "result", 1))) __PYX_ERR(0, 568, __pyx_L1_error)
  __pyx_r = __pyx_pf_5asynq_9scheduler_8extract_futures(__pyx_self, __pyx_v_value, __pyx_v_result);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_8extract_futures(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value, PyObject *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extract_futures", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler_extract_futures(__pyx_v_value, __pyx_v_result, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler.extract_futures", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":580
 * 
 * 
 * def _no_blocked_tasks(tasks):             # <<<<<<<<<<<<<<
 *     computed = [task for task in tasks if task.is_computed()]
 *     for task in computed:
 */

static PyObject *__pyx_pw_5asynq_9scheduler_11_no_blocked_tasks(PyObject *__pyx_self, PyObject *__pyx_v_tasks); /*proto*/
static int __pyx_f_5asynq_9scheduler__no_blocked_tasks(PyObject *__pyx_v_tasks, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_computed = NULL;
  PyObject *__pyx_v_task = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("_no_blocked_tasks", 0);

  /* "asynq/scheduler.py":581
 * 
 * def _no_blocked_tasks(tasks):
 *     computed = [task for task in tasks if task.is_computed()]             # <<<<<<<<<<<<<<
 *     for task in computed:
 *         if task in tasks:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_v_tasks)) || PyTuple_CheckExact(__pyx_v_tasks)) {
    __pyx_t_2 = __pyx_v_tasks; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_tasks); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 581, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 581, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 581, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_task, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_task, __pyx_n_s_is_computed); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (__pyx_t_7) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 581, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 581, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_8) {
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_v_task))) __PYX_ERR(0, 581, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_computed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "asynq/scheduler.py":582
 * def _no_blocked_tasks(tasks):
 *     computed = [task for task in tasks if task.is_computed()]
 *     for task in computed:             # <<<<<<<<<<<<<<
 *         if task in tasks:
 *             tasks.remove(task)
 */
  __pyx_t_1 = __pyx_v_computed; __Pyx_INCREF(__pyx_t_1); __pyx_t_3 = 0;
  for (;;) {
    if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 582, __pyx_L1_error)
    #else
    __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_task, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "asynq/scheduler.py":583
 *     computed = [task for task in tasks if task.is_computed()]
 *     for task in computed:
 *         if task in tasks:             # <<<<<<<<<<<<<<
 *             tasks.remove(task)
 *     return len(tasks) == 0
 */
    __pyx_t_8 = (__Pyx_PySequence_ContainsTF(__pyx_v_task, __pyx_v_tasks, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 583, __pyx_L1_error)
    __pyx_t_9 = (__pyx_t_8 != 0);
    if (__pyx_t_9) {

      /* "asynq/scheduler.py":584
 *     for task in computed:
 *         if task in tasks:
 *             tasks.remove(task)             # <<<<<<<<<<<<<<
 *     return len(tasks) == 0
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_tasks, __pyx_n_s_remove); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 584, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_6) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_task); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_task);
        __Pyx_GIVEREF(__pyx_v_task);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_task);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "asynq/scheduler.py":583
 *     computed = [task for task in tasks if task.is_computed()]
 *     for task in computed:
 *         if task in tasks:             # <<<<<<<<<<<<<<
 *             tasks.remove(task)
 *     return len(tasks) == 0
 */
    }

    /* "asynq/scheduler.py":582
 * def _no_blocked_tasks(tasks):
 *     computed = [task for task in tasks if task.is_computed()]
 *     for task in computed:             # <<<<<<<<<<<<<<
 *         if task in tasks:
 *             tasks.remove(task)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "asynq/scheduler.py":585
 *         if task in tasks:
 *             tasks.remove(task)
 *     return len(tasks) == 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = PyObject_Length(__pyx_v_tasks); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __pyx_r = (__pyx_t_3 == 0);
  goto __pyx_L0;

  /* "asynq/scheduler.py":580
 * 
 * 
 * def _no_blocked_tasks(tasks):             # <<<<<<<<<<<<<<
 *     computed = [task for task in tasks if task.is_computed()]
 *     for task in computed:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("asynq.scheduler._no_blocked_tasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_computed);
  __Pyx_XDECREF(__pyx_v_task);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_11_no_blocked_tasks(PyObject *__pyx_self, PyObject *__pyx_v_tasks); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_11_no_blocked_tasks(PyObject *__pyx_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_no_blocked_tasks (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_10_no_blocked_tasks(__pyx_self, ((PyObject *)__pyx_v_tasks));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_10_no_blocked_tasks(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tasks) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_no_blocked_tasks", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5asynq_9scheduler__no_blocked_tasks(__pyx_v_tasks, 0); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 580, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("asynq.scheduler._no_blocked_tasks", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "asynq/scheduler.py":588
 * 
 * 
 * def _create_blocking_dependency():             # <<<<<<<<<<<<<<
 *     return futures.ConstFuture("async_blocking_dependency")
 */

static PyObject *__pyx_pw_5asynq_9scheduler_13_create_blocking_dependency(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_f_5asynq_9scheduler__create_blocking_dependency(CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_5asynq_7futures_FutureBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_create_blocking_dependency", 0);

  /* "asynq/scheduler.py":589
 * 
 * def _create_blocking_dependency():
 *     return futures.ConstFuture("async_blocking_dependency")             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5asynq_7futures_ConstFuture), __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_5asynq_7futures_FutureBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "asynq/scheduler.py":588
 * 
 * 
 * def _create_blocking_dependency():             # <<<<<<<<<<<<<<
 *     return futures.ConstFuture("async_blocking_dependency")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler._create_blocking_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5asynq_9scheduler_13_create_blocking_dependency(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5asynq_9scheduler_13_create_blocking_dependency(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_create_blocking_dependency (wrapper)", 0);
  __pyx_r = __pyx_pf_5asynq_9scheduler_12_create_blocking_dependency(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5asynq_9scheduler_12_create_blocking_dependency(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_create_blocking_dependency", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_5asynq_9scheduler__create_blocking_dependency(0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("asynq.scheduler._create_blocking_dependency", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5asynq_9scheduler_TaskScheduler __pyx_vtable_5asynq_9scheduler_TaskScheduler;

static PyObject *__pyx_tp_new_5asynq_9scheduler_TaskScheduler(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)o);
  p->__pyx_vtab = __pyx_vtabptr_5asynq_9scheduler_TaskScheduler;
  p->_empty_set = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_new_tasks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_tasks = Py_None; Py_INCREF(Py_None);
  p->_batches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)Py_None); Py_INCREF(Py_None);
  p->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)Py_None); Py_INCREF(Py_None);
  p->active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None); Py_INCREF(Py_None);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5asynq_9scheduler_TaskScheduler(PyObject *o) {
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *p = (struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_empty_set);
  Py_CLEAR(p->_new_tasks);
  Py_CLEAR(p->_tasks);
  Py_CLEAR(p->_batches);
  Py_CLEAR(p->on_before_batch_flush);
  Py_CLEAR(p->on_after_batch_flush);
  Py_CLEAR(p->active_task);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5asynq_9scheduler_TaskScheduler(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *p = (struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)o;
  if (p->_empty_set) {
    e = (*v)(p->_empty_set, a); if (e) return e;
  }
  if (p->_new_tasks) {
    e = (*v)(p->_new_tasks, a); if (e) return e;
  }
  if (p->_tasks) {
    e = (*v)(p->_tasks, a); if (e) return e;
  }
  if (p->_batches) {
    e = (*v)(p->_batches, a); if (e) return e;
  }
  if (p->on_before_batch_flush) {
    e = (*v)(((PyObject*)p->on_before_batch_flush), a); if (e) return e;
  }
  if (p->on_after_batch_flush) {
    e = (*v)(((PyObject*)p->on_after_batch_flush), a); if (e) return e;
  }
  if (p->active_task) {
    e = (*v)(((PyObject*)p->active_task), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5asynq_9scheduler_TaskScheduler(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5asynq_9scheduler_TaskScheduler *p = (struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)o;
  tmp = ((PyObject*)p->_empty_set);
  p->_empty_set = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_new_tasks);
  p->_new_tasks = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_tasks);
  p->_tasks = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_batches);
  p->_batches = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->on_before_batch_flush);
  p->on_before_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->on_after_batch_flush);
  p->on_after_batch_flush = ((struct __pyx_obj_5qcore_6events_EventHook *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->active_task);
  p->active_task = ((struct __pyx_obj_5asynq_10async_task_AsyncTask *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler__empty_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler__empty_set(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_empty_set_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler__last_dump_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler__last_dump_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_15_last_dump_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler__new_tasks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler__new_tasks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_10_new_tasks_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler__tasks(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler__tasks(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_6_tasks_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler__batches(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler__batches(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_8_batches_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler_await_recursion_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler_await_recursion_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_21await_recursion_depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler_max_await_recursion_depth(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler_max_await_recursion_depth(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_25max_await_recursion_depth_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler_on_before_batch_flush(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler_on_before_batch_flush(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_21on_before_batch_flush_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler_on_after_batch_flush(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler_on_after_batch_flush(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_20on_after_batch_flush_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler_active_task(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler_active_task(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_11active_task_5__del__(o);
  }
}

static PyObject *__pyx_getprop_5asynq_9scheduler_13TaskScheduler_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_1__get__(o);
}

static int __pyx_setprop_5asynq_9scheduler_13TaskScheduler_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_5asynq_9scheduler_13TaskScheduler_4name_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_5asynq_9scheduler_TaskScheduler[] = {
  {"activate", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_3activate, METH_VARARGS|METH_KEYWORDS, 0},
  {"reset", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_5reset, METH_NOARGS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_4reset},
  {"start", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_7start, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_6start},
  {"continue_with", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_9continue_with, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_8continue_with},
  {"flush_batches", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_11flush_batches, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_10flush_batches},
  {"await", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_13await, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_12await},
  {"add_dependencies", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_15add_dependencies, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_14add_dependencies},
  {"add_dependency", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_17add_dependency, METH_VARARGS|METH_KEYWORDS, 0},
  {"_await", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_19_await, METH_O, 0},
  {"_execute", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_21_execute, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_20_execute},
  {"_try_execute_on_new_scheduler", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_23_try_execute_on_new_scheduler, METH_O, 0},
  {"schedule", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_25schedule, METH_O, 0},
  {"schedule_batch", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_27schedule_batch, METH_O, 0},
  {"_select_batch_to_flush", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_29_select_batch_to_flush, METH_NOARGS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_28_select_batch_to_flush},
  {"_clone", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_31_clone, METH_NOARGS, __pyx_doc_5asynq_9scheduler_13TaskScheduler_30_clone},
  {"_accept_changes", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_33_accept_changes, METH_O, __pyx_doc_5asynq_9scheduler_13TaskScheduler_32_accept_changes},
  {"dump", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_39dump, METH_VARARGS|METH_KEYWORDS, 0},
  {"try_time_based_dump", (PyCFunction)__pyx_pw_5asynq_9scheduler_13TaskScheduler_41try_time_based_dump, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_5asynq_9scheduler_TaskScheduler[] = {
  {(char *)"_empty_set", __pyx_getprop_5asynq_9scheduler_13TaskScheduler__empty_set, __pyx_setprop_5asynq_9scheduler_13TaskScheduler__empty_set, (char *)0, 0},
  {(char *)"_last_dump_time", __pyx_getprop_5asynq_9scheduler_13TaskScheduler__last_dump_time, __pyx_setprop_5asynq_9scheduler_13TaskScheduler__last_dump_time, (char *)0, 0},
  {(char *)"_new_tasks", __pyx_getprop_5asynq_9scheduler_13TaskScheduler__new_tasks, __pyx_setprop_5asynq_9scheduler_13TaskScheduler__new_tasks, (char *)0, 0},
  {(char *)"_tasks", __pyx_getprop_5asynq_9scheduler_13TaskScheduler__tasks, __pyx_setprop_5asynq_9scheduler_13TaskScheduler__tasks, (char *)0, 0},
  {(char *)"_batches", __pyx_getprop_5asynq_9scheduler_13TaskScheduler__batches, __pyx_setprop_5asynq_9scheduler_13TaskScheduler__batches, (char *)0, 0},
  {(char *)"await_recursion_depth", __pyx_getprop_5asynq_9scheduler_13TaskScheduler_await_recursion_depth, __pyx_setprop_5asynq_9scheduler_13TaskScheduler_await_recursion_depth, (char *)0, 0},
  {(char *)"max_await_recursion_depth", __pyx_getprop_5asynq_9scheduler_13TaskScheduler_max_await_recursion_depth, __pyx_setprop_5asynq_9scheduler_13TaskScheduler_max_await_recursion_depth, (char *)0, 0},
  {(char *)"on_before_batch_flush", __pyx_getprop_5asynq_9scheduler_13TaskScheduler_on_before_batch_flush, __pyx_setprop_5asynq_9scheduler_13TaskScheduler_on_before_batch_flush, (char *)0, 0},
  {(char *)"on_after_batch_flush", __pyx_getprop_5asynq_9scheduler_13TaskScheduler_on_after_batch_flush, __pyx_setprop_5asynq_9scheduler_13TaskScheduler_on_after_batch_flush, (char *)0, 0},
  {(char *)"active_task", __pyx_getprop_5asynq_9scheduler_13TaskScheduler_active_task, __pyx_setprop_5asynq_9scheduler_13TaskScheduler_active_task, (char *)0, 0},
  {(char *)"name", __pyx_getprop_5asynq_9scheduler_13TaskScheduler_name, __pyx_setprop_5asynq_9scheduler_13TaskScheduler_name, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5asynq_9scheduler_TaskScheduler = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynq.scheduler.TaskScheduler", /*tp_name*/
  sizeof(struct __pyx_obj_5asynq_9scheduler_TaskScheduler), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5asynq_9scheduler_TaskScheduler, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_5asynq_9scheduler_13TaskScheduler_37__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_5asynq_9scheduler_13TaskScheduler_35__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Schedules and runs AsyncTask objects flushing batches when needed.", /*tp_doc*/
  __pyx_tp_traverse_5asynq_9scheduler_TaskScheduler, /*tp_traverse*/
  __pyx_tp_clear_5asynq_9scheduler_TaskScheduler, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5asynq_9scheduler_TaskScheduler, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_5asynq_9scheduler_TaskScheduler, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5asynq_9scheduler_13TaskScheduler_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5asynq_9scheduler_TaskScheduler, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_5asynq_9scheduler_TaskSchedulerActivator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)o);
  p->_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None); Py_INCREF(Py_None);
  p->_old_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5asynq_9scheduler_TaskSchedulerActivator(PyObject *o) {
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *p = (struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_scheduler);
  Py_CLEAR(p->_old_scheduler);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5asynq_9scheduler_TaskSchedulerActivator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *p = (struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)o;
  if (p->_scheduler) {
    e = (*v)(((PyObject*)p->_scheduler), a); if (e) return e;
  }
  if (p->_old_scheduler) {
    e = (*v)(((PyObject*)p->_old_scheduler), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5asynq_9scheduler_TaskSchedulerActivator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *p = (struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *)o;
  tmp = ((PyObject*)p->_scheduler);
  p->_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_old_scheduler);
  p->_old_scheduler = ((struct __pyx_obj_5asynq_9scheduler_TaskScheduler *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_5asynq_9scheduler_TaskSchedulerActivator[] = {
  {"__enter__", (PyCFunction)__pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_3__enter__, METH_NOARGS, 0},
  {"__exit__", (PyCFunction)__pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_5__exit__, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5asynq_9scheduler_TaskSchedulerActivator = {
  PyVarObject_HEAD_INIT(0, 0)
  "asynq.scheduler.TaskSchedulerActivator", /*tp_name*/
  sizeof(struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5asynq_9scheduler_TaskSchedulerActivator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5asynq_9scheduler_TaskSchedulerActivator, /*tp_traverse*/
  __pyx_tp_clear_5asynq_9scheduler_TaskSchedulerActivator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5asynq_9scheduler_TaskSchedulerActivator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5asynq_9scheduler_22TaskSchedulerActivator_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5asynq_9scheduler_TaskSchedulerActivator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {"get_scheduler", (PyCFunction)__pyx_pw_5asynq_9scheduler_1get_scheduler, METH_NOARGS, 0},
  {"set_scheduler", (PyCFunction)__pyx_pw_5asynq_9scheduler_5set_scheduler, METH_O, 0},
  {"get_active_task", (PyCFunction)__pyx_pw_5asynq_9scheduler_7get_active_task, METH_NOARGS, 0},
  {"extract_futures", (PyCFunction)__pyx_pw_5asynq_9scheduler_9extract_futures, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5asynq_9scheduler_8extract_futures},
  {"_no_blocked_tasks", (PyCFunction)__pyx_pw_5asynq_9scheduler_11_no_blocked_tasks, METH_O, 0},
  {"_create_blocking_dependency", (PyCFunction)__pyx_pw_5asynq_9scheduler_13_create_blocking_dependency, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "scheduler",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_Are_you_importing_asynq_from_the, __pyx_k_Are_you_importing_asynq_from_the, sizeof(__pyx_k_Are_you_importing_asynq_from_the), 0, 0, 1, 0},
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_AsyncTaskError, __pyx_k_AsyncTaskError, sizeof(__pyx_k_AsyncTaskError), 0, 0, 1, 1},
  {&__pyx_n_s_BaseException, __pyx_k_BaseException, sizeof(__pyx_k_BaseException), 0, 0, 1, 1},
  {&__pyx_kp_s_Batches, __pyx_k_Batches, sizeof(__pyx_k_Batches), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_t_proceed_further_no_task_to, __pyx_k_Can_t_proceed_further_no_task_to, sizeof(__pyx_k_Can_t_proceed_further_no_task_to), 0, 0, 1, 0},
  {&__pyx_n_s_Exception, __pyx_k_Exception, sizeof(__pyx_k_Exception), 0, 0, 1, 1},
  {&__pyx_kp_s_Last_task_s, __pyx_k_Last_task_s, sizeof(__pyx_k_Last_task_s), 0, 0, 1, 0},
  {&__pyx_n_s_LocalTaskSchedulerState, __pyx_k_LocalTaskSchedulerState, sizeof(__pyx_k_LocalTaskSchedulerState), 0, 0, 1, 1},
  {&__pyx_n_s_LocalTaskSchedulerState___init, __pyx_k_LocalTaskSchedulerState___init, sizeof(__pyx_k_LocalTaskSchedulerState___init), 0, 0, 1, 1},
  {&__pyx_n_s_LocalTaskSchedulerState_reset, __pyx_k_LocalTaskSchedulerState_reset, sizeof(__pyx_k_LocalTaskSchedulerState_reset), 0, 0, 1, 1},
  {&__pyx_n_s_MAX_AWAIT_RECURSION_DEPTH, __pyx_k_MAX_AWAIT_RECURSION_DEPTH, sizeof(__pyx_k_MAX_AWAIT_RECURSION_DEPTH), 0, 0, 1, 1},
  {&__pyx_kp_s_New_tasks, __pyx_k_New_tasks, sizeof(__pyx_k_New_tasks), 0, 0, 1, 0},
  {&__pyx_kp_s_No_tasks_in_task_queue, __pyx_k_No_tasks_in_task_queue, sizeof(__pyx_k_No_tasks_in_task_queue), 0, 0, 1, 0},
  {&__pyx_kp_s_Scheduler_is_awaiting_for, __pyx_k_Scheduler_is_awaiting_for, sizeof(__pyx_k_Scheduler_is_awaiting_for), 0, 0, 1, 0},
  {&__pyx_kp_s_Scheduler_state_dump, __pyx_k_Scheduler_state_dump, sizeof(__pyx_k_Scheduler_state_dump), 0, 0, 1, 0},
  {&__pyx_kp_s_Task_is_already_computed, __pyx_k_Task_is_already_computed, sizeof(__pyx_k_Task_is_already_computed), 0, 0, 1, 0},
  {&__pyx_kp_s_Task_is_already_scheduled, __pyx_k_Task_is_already_scheduled, sizeof(__pyx_k_Task_is_already_scheduled), 0, 0, 1, 0},
  {&__pyx_kp_s_Task_is_already_scheduled_on_ano, __pyx_k_Task_is_already_scheduled_on_ano, sizeof(__pyx_k_Task_is_already_scheduled_on_ano), 0, 0, 1, 0},
  {&__pyx_kp_s_Task_is_blocked_so_it_shouldn_t, __pyx_k_Task_is_blocked_so_it_shouldn_t, sizeof(__pyx_k_Task_is_blocked_so_it_shouldn_t), 0, 0, 1, 0},
  {&__pyx_kp_s_Task_queue, __pyx_k_Task_queue, sizeof(__pyx_k_Task_queue), 0, 0, 1, 0},
  {&__pyx_kp_s_Task_s_is_already_scheduled_on_a, __pyx_k_Task_s_is_already_scheduled_on_a, sizeof(__pyx_k_Task_s_is_already_scheduled_on_a), 0, 0, 1, 0},
  {&__pyx_n_s__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 0, 1, 1},
  {&__pyx_n_s__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 0, 1, 1},
  {&__pyx_kp_s__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 0, 1, 0},
  {&__pyx_n_s_accept_changes, __pyx_k_accept_changes, sizeof(__pyx_k_accept_changes), 0, 0, 1, 1},
  {&__pyx_n_s_activate, __pyx_k_activate, sizeof(__pyx_k_activate), 0, 0, 1, 1},
  {&__pyx_n_s_active_task, __pyx_k_active_task, sizeof(__pyx_k_active_task), 0, 0, 1, 1},
  {&__pyx_n_s_add_dependencies, __pyx_k_add_dependencies, sizeof(__pyx_k_add_dependencies), 0, 0, 1, 1},
  {&__pyx_n_s_add_dependency, __pyx_k_add_dependency, sizeof(__pyx_k_add_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_kp_s_async_await_recursion_depth_is, __pyx_k_async_await_recursion_depth_is, sizeof(__pyx_k_async_await_recursion_depth_is), 0, 0, 1, 0},
  {&__pyx_kp_s_async_await_recursion_going_to, __pyx_k_async_await_recursion_going_to, sizeof(__pyx_k_async_await_recursion_going_to), 0, 0, 1, 0},
  {&__pyx_kp_s_async_await_recursion_ignoring, __pyx_k_async_await_recursion_ignoring, sizeof(__pyx_k_async_await_recursion_ignoring), 0, 0, 1, 0},
  {&__pyx_n_s_async_blocking_dependency, __pyx_k_async_blocking_dependency, sizeof(__pyx_k_async_blocking_dependency), 0, 0, 1, 1},
  {&__pyx_kp_s_async_can_t_schedule_blocked_ta, __pyx_k_async_can_t_schedule_blocked_ta, sizeof(__pyx_k_async_can_t_schedule_blocked_ta), 0, 0, 1, 0},
  {&__pyx_kp_s_async_can_t_schedule_computed_t, __pyx_k_async_can_t_schedule_computed_t, sizeof(__pyx_k_async_can_t_schedule_computed_t), 0, 0, 1, 0},
  {&__pyx_kp_s_async_can_t_schedule_flushed_ba, __pyx_k_async_can_t_schedule_flushed_ba, sizeof(__pyx_k_async_can_t_schedule_flushed_ba), 0, 0, 1, 0},
  {&__pyx_kp_s_async_continued_s, __pyx_k_async_continued_s, sizeof(__pyx_k_async_continued_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_continuing_s, __pyx_k_async_continuing_s, sizeof(__pyx_k_async_continuing_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_dependency_s_already_depe, __pyx_k_async_dependency_s_already_depe, sizeof(__pyx_k_async_dependency_s_already_depe), 0, 0, 1, 0},
  {&__pyx_kp_s_async_dependency_s_needs_s, __pyx_k_async_dependency_s_needs_s, sizeof(__pyx_k_async_dependency_s_needs_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_no_batch_to_flush, __pyx_k_async_no_batch_to_flush, sizeof(__pyx_k_async_no_batch_to_flush), 0, 0, 1, 0},
  {&__pyx_kp_s_async_no_task_to_continue, __pyx_k_async_no_task_to_continue, sizeof(__pyx_k_async_no_task_to_continue), 0, 0, 1, 0},
  {&__pyx_kp_s_async_scheduler_activation, __pyx_k_async_scheduler_activation, sizeof(__pyx_k_async_scheduler_activation), 0, 0, 1, 0},
  {&__pyx_kp_s_async_scheduler_activation_2, __pyx_k_async_scheduler_activation_2, sizeof(__pyx_k_async_scheduler_activation_2), 0, 0, 1, 0},
  {&__pyx_kp_s_async_scheduler_s, __pyx_k_async_scheduler_s, sizeof(__pyx_k_async_scheduler_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_scheduling_batch_s, __pyx_k_async_scheduling_batch_s, sizeof(__pyx_k_async_scheduling_batch_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_scheduling_task_s, __pyx_k_async_scheduling_task_s, sizeof(__pyx_k_async_scheduling_task_s), 0, 0, 1, 0},
  {&__pyx_kp_s_async_skipping_s, __pyx_k_async_skipping_s, sizeof(__pyx_k_async_skipping_s), 0, 0, 1, 0},
  {&__pyx_kp_s_asynq_lib64_scheduler, __pyx_k_asynq_lib64_scheduler, sizeof(__pyx_k_asynq_lib64_scheduler), 0, 0, 1, 0},
  {&__pyx_kp_s_asynq_scheduler, __pyx_k_asynq_scheduler, sizeof(__pyx_k_asynq_scheduler), 0, 0, 1, 0},
  {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
  {&__pyx_n_s_await_2, __pyx_k_await_2, sizeof(__pyx_k_await_2), 0, 0, 1, 1},
  {&__pyx_n_s_batching, __pyx_k_batching, sizeof(__pyx_k_batching), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_continue_with, __pyx_k_continue_with, sizeof(__pyx_k_continue_with), 0, 0, 1, 1},
  {&__pyx_n_s_core_events, __pyx_k_core_events, sizeof(__pyx_k_core_events), 0, 0, 1, 1},
  {&__pyx_n_s_current, __pyx_k_current, sizeof(__pyx_k_current), 0, 0, 1, 1},
  {&__pyx_n_s_current_thread, __pyx_k_current_thread, sizeof(__pyx_k_current_thread), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_debug_2, __pyx_k_debug_2, sizeof(__pyx_k_debug_2), 0, 0, 1, 1},
  {&__pyx_n_s_debug_options, __pyx_k_debug_options, sizeof(__pyx_k_debug_options), 0, 0, 1, 1},
  {&__pyx_n_s_dependency, __pyx_k_dependency, sizeof(__pyx_k_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_dump, __pyx_k_dump, sizeof(__pyx_k_dump), 0, 0, 1, 1},
  {&__pyx_n_s_endswith, __pyx_k_endswith, sizeof(__pyx_k_endswith), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_execute, __pyx_k_execute, sizeof(__pyx_k_execute), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_flush, __pyx_k_flush, sizeof(__pyx_k_flush), 0, 0, 1, 1},
  {&__pyx_n_s_flush_batches, __pyx_k_flush_batches, sizeof(__pyx_k_flush_batches), 0, 0, 1, 1},
  {&__pyx_n_s_future, __pyx_k_future, sizeof(__pyx_k_future), 0, 0, 1, 1},
  {&__pyx_n_s_futures, __pyx_k_futures, sizeof(__pyx_k_futures), 0, 0, 1, 1},
  {&__pyx_n_s_futures_none, __pyx_k_futures_none, sizeof(__pyx_k_futures_none), 0, 0, 1, 1},
  {&__pyx_n_s_get_priority, __pyx_k_get_priority, sizeof(__pyx_k_get_priority), 0, 0, 1, 1},
  {&__pyx_kp_s_home_jelle_github_asynq_asynq_s, __pyx_k_home_jelle_github_asynq_asynq_s, sizeof(__pyx_k_home_jelle_github_asynq_asynq_s), 0, 0, 1, 0},
  {&__pyx_n_s_ident, __pyx_k_ident, sizeof(__pyx_k_ident), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_indent, __pyx_k_indent, sizeof(__pyx_k_indent), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_is_computed, __pyx_k_is_computed, sizeof(__pyx_k_is_computed), 0, 0, 1, 1},
  {&__pyx_n_s_is_flushed, __pyx_k_is_flushed, sizeof(__pyx_k_is_flushed), 0, 0, 1, 1},
  {&__pyx_n_s_is_partial_execution, __pyx_k_is_partial_execution, sizeof(__pyx_k_is_partial_execution), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_last_id, __pyx_k_last_id, sizeof(__pyx_k_last_id), 0, 0, 1, 1},
  {&__pyx_n_s_last_task, __pyx_k_last_task, sizeof(__pyx_k_last_task), 0, 0, 1, 1},
  {&__pyx_n_s_local, __pyx_k_local, sizeof(__pyx_k_local), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_must_await, __pyx_k_must_await, sizeof(__pyx_k_must_await), 0, 0, 1, 1},
  {&__pyx_n_s_must_await_on_exit, __pyx_k_must_await_on_exit, sizeof(__pyx_k_must_await_on_exit), 0, 0, 1, 1},
  {&__pyx_n_s_must_flush_batches_on_exit, __pyx_k_must_flush_batches_on_exit, sizeof(__pyx_k_must_flush_batches_on_exit), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_popleft, __pyx_k_popleft, sizeof(__pyx_k_popleft), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_qcore_events, __pyx_k_qcore_events, sizeof(__pyx_k_qcore_events), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_remove_dependency, __pyx_k_remove_dependency, sizeof(__pyx_k_remove_dependency), 0, 0, 1, 1},
  {&__pyx_n_s_reset, __pyx_k_reset, sizeof(__pyx_k_reset), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_run_immediately, __pyx_k_run_immediately, sizeof(__pyx_k_run_immediately), 0, 0, 1, 1},
  {&__pyx_kp_s_s_d, __pyx_k_s_d, sizeof(__pyx_k_s_d), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s_d_d_tasks_d_batches_active, __pyx_k_s_s_d_d_tasks_d_batches_active, sizeof(__pyx_k_s_s_d_d_tasks_d_batches_active), 0, 0, 1, 0},
  {&__pyx_n_s_schedule, __pyx_k_schedule, sizeof(__pyx_k_schedule), 0, 0, 1, 1},
  {&__pyx_n_s_schedule_batch, __pyx_k_schedule_batch, sizeof(__pyx_k_schedule_batch), 0, 0, 1, 1},
  {&__pyx_n_s_scheduler, __pyx_k_scheduler, sizeof(__pyx_k_scheduler), 0, 0, 1, 1},
  {&__pyx_n_s_select_batch_to_flush, __pyx_k_select_batch_to_flush, sizeof(__pyx_k_select_batch_to_flush), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_stderr, __pyx_k_stderr, sizeof(__pyx_k_stderr), 0, 0, 1, 1},
  {&__pyx_n_s_stdout, __pyx_k_stdout, sizeof(__pyx_k_stdout), 0, 0, 1, 1},
  {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
  {&__pyx_n_s_str_2, __pyx_k_str_2, sizeof(__pyx_k_str_2), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_task, __pyx_k_task, sizeof(__pyx_k_task), 0, 0, 1, 1},
  {&__pyx_n_s_tasks, __pyx_k_tasks, sizeof(__pyx_k_tasks), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_threading, __pyx_k_threading, sizeof(__pyx_k_threading), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_s_traceback, __pyx_k_traceback, sizeof(__pyx_k_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_try_execute_on_new_scheduler, __pyx_k_try_execute_on_new_scheduler, sizeof(__pyx_k_try_execute_on_new_scheduler), 0, 0, 1, 1},
  {&__pyx_n_s_try_time_based_dump, __pyx_k_try_time_based_dump, sizeof(__pyx_k_try_time_based_dump), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_write, __pyx_k_write, sizeof(__pyx_k_write), 0, 0, 1, 1},
  {&__pyx_n_s_yield_result, __pyx_k_yield_result, sizeof(__pyx_k_yield_result), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_Exception = __Pyx_GetBuiltinName(__pyx_n_s_Exception); if (!__pyx_builtin_Exception) __PYX_ERR(0, 40, __pyx_L1_error)
  __pyx_builtin_BaseException = __Pyx_GetBuiltinName(__pyx_n_s_BaseException); if (!__pyx_builtin_BaseException) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 247, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "asynq/scheduler.py":84
 *             return task
 *         if task.scheduler is not None:
 *             raise AsyncTaskError('Task is already scheduled.')             # <<<<<<<<<<<<<<
 *         if task.is_blocked():
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Task_is_already_scheduled); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "asynq/scheduler.py":86
 *             raise AsyncTaskError('Task is already scheduled.')
 *         if task.is_blocked():
 *             raise AsyncTaskError("Task is blocked, so it shouldn't be scheduled.")             # <<<<<<<<<<<<<<
 * 
 *         if run_immediately:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Task_is_blocked_so_it_shouldn_t); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "asynq/scheduler.py":102
 *         """
 *         if task.is_computed():
 *             raise AsyncTaskError('Task is already computed.')             # <<<<<<<<<<<<<<
 *         if future.is_computed():
 *             if task.scheduler is None:
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_Task_is_already_computed); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "asynq/scheduler.py":112
 *             task.scheduler = self
 *         elif task.scheduler is not self:
 *             raise AsyncTaskError('Task is already scheduled on another scheduler.')             # <<<<<<<<<<<<<<
 *         self.add_dependency(task, future)
 *         self._try_schedule(task, True)
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_Task_is_already_scheduled_on_ano); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "asynq/scheduler.py":217
 *                 if not is_partial_execution:
 *                     if dump_await_recursion:
 *                         debug.write("@async: await recursion: ignoring await()")             # <<<<<<<<<<<<<<
 *                     return 0
 *                 self._try_execute_on_new_scheduler(tasks)
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_async_await_recursion_ignoring); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "asynq/scheduler.py":249
 *                 except AssertionError:
 *                     if _debug_options.DUMP_PRE_ERROR_STATE:
 *                         debug.write('Scheduler is awaiting for:')             # <<<<<<<<<<<<<<
 *                         for task in tasks:
 *                             task.dump(2)
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_Scheduler_is_awaiting_for); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "asynq/scheduler.py":252
 *                         for task in tasks:
 *                             task.dump(2)
 *                         debug.write('----------------------------------------------------------------------\n')             # <<<<<<<<<<<<<<
 *                         raise
 *         finally:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s__8); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "asynq/scheduler.py":275
 *             new_scheduler = self._clone()
 *             try:
 *                 with new_scheduler.activate(True, False):             # <<<<<<<<<<<<<<
 *                     new_scheduler.await(new_scheduler_tasks)
 *             finally:
 */
  __pyx_tuple__10 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "asynq/scheduler.py":382
 *             return task
 *         if dump_continue_task:
 *             debug.write('@async: no task to continue')             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_async_no_task_to_continue); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "asynq/scheduler.py":398
 *         if batch is None:
 *             if _debug_options.DUMP_FLUSH_BATCH:
 *                 debug.write('@async: no batch to flush')             # <<<<<<<<<<<<<<
 *             if assert_no_batch:
 *                 debug.dump(self)
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_async_no_batch_to_flush); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "asynq/scheduler.py":496
 *             return
 *         self._last_dump_time = current_time
 *         debug.write('\n--- Scheduler state dump: --------------------------------------------')             # <<<<<<<<<<<<<<
 *         try:
 *             self.dump()
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_Scheduler_state_dump); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "asynq/scheduler.py":502
 *                 debug.write('Last task: %s' % debug.str(last_task), 1)
 *         finally:
 *             debug.write('----------------------------------------------------------------------\n')             # <<<<<<<<<<<<<<
 *             stdout.flush()
 *             stderr.flush()
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s__8); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s__8); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "asynq/scheduler.py":516
 *     def __enter__(self):
 *         if _debug_options.DUMP_SCHEDULER_CHANGE:
 *             debug.write("@async: -> scheduler activation")             # <<<<<<<<<<<<<<
 *         self._old_scheduler = get_scheduler()
 *         set_scheduler(self._scheduler)
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_async_scheduler_activation); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "asynq/scheduler.py":531
 *             set_scheduler(self._old_scheduler)
 *             if _debug_options.DUMP_SCHEDULER_CHANGE:
 *                 debug.write("@async: <- scheduler activation")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_async_scheduler_activation_2); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_async_scheduler_activation_2); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "asynq/scheduler.py":589
 * 
 * def _create_blocking_dependency():
 *     return futures.ConstFuture("async_blocking_dependency")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_n_s_async_blocking_dependency); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "asynq/scheduler.py":29
 * 
 * 
 * assert str(__name__).endswith('asynq.scheduler') \             # <<<<<<<<<<<<<<
 *     or str(__name__).endswith('asynq.lib64.scheduler'), \
 *     "Are you importing asynq from the wrong directory?"
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_asynq_scheduler); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "asynq/scheduler.py":30
 * 
 * assert str(__name__).endswith('asynq.scheduler') \
 *     or str(__name__).endswith('asynq.lib64.scheduler'), \             # <<<<<<<<<<<<<<
 *     "Are you importing asynq from the wrong directory?"
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_asynq_lib64_scheduler); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "asynq/scheduler.py":535
 * 
 * class LocalTaskSchedulerState(threading.local):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.last_id = 0
 *         self.reset()
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jelle_github_asynq_asynq_s, __pyx_n_s_init, 535, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 535, __pyx_L1_error)

  /* "asynq/scheduler.py":539
 *         self.reset()
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.current = TaskScheduler()
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jelle_github_asynq_asynq_s, __pyx_n_s_reset, 539, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 539, __pyx_L1_error)

  /* "asynq/scheduler.py":551
 * 
 * 
 * def reset():             # <<<<<<<<<<<<<<
 *     _state.reset()
 * 
 */
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jelle_github_asynq_asynq_s, __pyx_n_s_reset, 551, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initscheduler(void); /*proto*/
PyMODINIT_FUNC initscheduler(void)
#else
PyMODINIT_FUNC PyInit_scheduler(void); /*proto*/
PyMODINIT_FUNC PyInit_scheduler(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_scheduler(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("scheduler", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_asynq__scheduler) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "asynq.scheduler")) {
      if (unlikely(PyDict_SetItemString(modules, "asynq.scheduler", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_5asynq_9scheduler__debug_options = ((struct __pyx_obj_5asynq_6_debug_DebugOptions *)Py_None); Py_INCREF(Py_None);
  __pyx_v_5asynq_9scheduler__futures_none = Py_None; Py_INCREF(Py_None);
  __pyx_v_5asynq_9scheduler__state = Py_None; Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_debug_options, (void *)&__pyx_v_5asynq_9scheduler__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_futures_none, (void *)&__pyx_v_5asynq_9scheduler__futures_none, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportVoidPtr(__pyx_n_s_state, (void *)&__pyx_v_5asynq_9scheduler__state, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("extract_futures", (void (*)(void))__pyx_f_5asynq_9scheduler_extract_futures, "PyObject *(PyObject *, PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_scheduler", (void (*)(void))__pyx_f_5asynq_9scheduler_get_scheduler, "struct __pyx_obj_5asynq_9scheduler_TaskScheduler *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("set_scheduler", (void (*)(void))__pyx_f_5asynq_9scheduler_set_scheduler, "PyObject *(PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("get_active_task", (void (*)(void))__pyx_f_5asynq_9scheduler_get_active_task, "struct __pyx_obj_5asynq_10async_task_AsyncTask *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_no_blocked_tasks", (void (*)(void))__pyx_f_5asynq_9scheduler__no_blocked_tasks, "int (PyObject *, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_create_blocking_dependency", (void (*)(void))__pyx_f_5asynq_9scheduler__create_blocking_dependency, "struct __pyx_obj_5asynq_7futures_FutureBase *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Type init code ---*/
  __pyx_vtabptr_5asynq_9scheduler_TaskScheduler = &__pyx_vtable_5asynq_9scheduler_TaskScheduler;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.activate = (struct __pyx_obj_5asynq_9scheduler_TaskSchedulerActivator *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, int, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_activate;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.reset = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_reset;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.start = (struct __pyx_obj_5asynq_10async_task_AsyncTask *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_start *__pyx_optional_args))__pyx_f_5asynq_9scheduler_13TaskScheduler_start;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.continue_with = (struct __pyx_obj_5asynq_10async_task_AsyncTask *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_7futures_FutureBase *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_continue_with;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.flush_batches = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_flush_batches *__pyx_optional_args))__pyx_f_5asynq_9scheduler_13TaskScheduler_flush_batches;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.add_dependencies = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependencies;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.add_dependency = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, struct __pyx_obj_5asynq_7futures_FutureBase *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_add_dependency;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.schedule = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_schedule;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.schedule_batch = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler_schedule_batch;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._await = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler__await;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._execute = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler__execute;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._try_execute_on_new_scheduler = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, PyObject *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler__try_execute_on_new_scheduler;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._schedule = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *))__pyx_f_5asynq_9scheduler_13TaskScheduler__schedule;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._schedule_without_checks = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *))__pyx_f_5asynq_9scheduler_13TaskScheduler__schedule_without_checks;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._try_schedule = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_10async_task_AsyncTask *, int))__pyx_f_5asynq_9scheduler_13TaskScheduler__try_schedule;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._schedule_batch = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *))__pyx_f_5asynq_9scheduler_13TaskScheduler__schedule_batch;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._flush_batch = (int (*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_8batching_BatchBase *))__pyx_f_5asynq_9scheduler_13TaskScheduler__flush_batch;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._enqueue_new_tasks = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *))__pyx_f_5asynq_9scheduler_13TaskScheduler__enqueue_new_tasks;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._continue_with_task = (struct __pyx_obj_5asynq_10async_task_AsyncTask *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *))__pyx_f_5asynq_9scheduler_13TaskScheduler__continue_with_task;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._continue_with_batch = (struct __pyx_obj_5asynq_8batching_BatchBase *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler__continue_with_batch *__pyx_optional_args))__pyx_f_5asynq_9scheduler_13TaskScheduler__continue_with_batch;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._select_batch_to_flush = (struct __pyx_obj_5asynq_8batching_BatchBase *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler__select_batch_to_flush;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._clone = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler__clone;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler._accept_changes = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch))__pyx_f_5asynq_9scheduler_13TaskScheduler__accept_changes;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.dump = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_dump *__pyx_optional_args))__pyx_f_5asynq_9scheduler_13TaskScheduler_dump;
  __pyx_vtable_5asynq_9scheduler_TaskScheduler.try_time_based_dump = (PyObject *(*)(struct __pyx_obj_5asynq_9scheduler_TaskScheduler *, int __pyx_skip_dispatch, struct __pyx_opt_args_5asynq_9scheduler_13TaskScheduler_try_time_based_dump *__pyx_optional_args))__pyx_f_5asynq_9scheduler_13TaskScheduler_try_time_based_dump;
  if (PyType_Ready(&__pyx_type_5asynq_9scheduler_TaskScheduler) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __pyx_type_5asynq_9scheduler_TaskScheduler.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_5asynq_9scheduler_TaskScheduler.tp_dict, __pyx_vtabptr_5asynq_9scheduler_TaskScheduler) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "TaskScheduler", (PyObject *)&__pyx_type_5asynq_9scheduler_TaskScheduler) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __pyx_ptype_5asynq_9scheduler_TaskScheduler = &__pyx_type_5asynq_9scheduler_TaskScheduler;
  if (PyType_Ready(&__pyx_type_5asynq_9scheduler_TaskSchedulerActivator) < 0) __PYX_ERR(0, 507, __pyx_L1_error)
  __pyx_type_5asynq_9scheduler_TaskSchedulerActivator.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "TaskSchedulerActivator", (PyObject *)&__pyx_type_5asynq_9scheduler_TaskSchedulerActivator) < 0) __PYX_ERR(0, 507, __pyx_L1_error)
  __pyx_ptype_5asynq_9scheduler_TaskSchedulerActivator = &__pyx_type_5asynq_9scheduler_TaskSchedulerActivator;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(3, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(4, 15, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_EventHook = __Pyx_ImportType("qcore.events", "EventHook", sizeof(struct __pyx_obj_5qcore_6events_EventHook), 1); if (unlikely(!__pyx_ptype_5qcore_6events_EventHook)) __PYX_ERR(5, 15, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_SinkingEventHook = __Pyx_ImportType("qcore.events", "SinkingEventHook", sizeof(struct __pyx_obj_5qcore_6events_SinkingEventHook), 1); if (unlikely(!__pyx_ptype_5qcore_6events_SinkingEventHook)) __PYX_ERR(5, 19, __pyx_L1_error)
  __pyx_ptype_5qcore_6events_EventInterceptor = __Pyx_ImportType("qcore.events", "EventInterceptor", sizeof(struct __pyx_obj_5qcore_6events_EventInterceptor), 1); if (unlikely(!__pyx_ptype_5qcore_6events_EventInterceptor)) __PYX_ERR(5, 25, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_MarkerObject = __Pyx_ImportType("qcore.helpers", "MarkerObject", sizeof(struct __pyx_obj_5qcore_7helpers_MarkerObject), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_MarkerObject)) __PYX_ERR(6, 24, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_EmptyContext = __Pyx_ImportType("qcore.helpers", "EmptyContext", sizeof(struct __pyx_obj_5qcore_7helpers_EmptyContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_EmptyContext)) __PYX_ERR(6, 33, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_EmptyContext = (struct __pyx_vtabstruct_5qcore_7helpers_EmptyContext*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_EmptyContext->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_EmptyContext)) __PYX_ERR(6, 33, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper = __Pyx_ImportType("qcore.helpers", "CythonCachedHashWrapper", sizeof(struct __pyx_obj_5qcore_7helpers_CythonCachedHashWrapper), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper)) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper = (struct __pyx_vtabstruct_5qcore_7helpers_CythonCachedHashWrapper*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_CythonCachedHashWrapper->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_CythonCachedHashWrapper)) __PYX_ERR(6, 40, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext = __Pyx_ImportType("qcore.helpers", "_ScopedValueOverrideContext", sizeof(struct __pyx_obj_5qcore_7helpers__ScopedValueOverrideContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers__ScopedValueOverrideContext)) __PYX_ERR(6, 50, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers_ScopedValue = __Pyx_ImportType("qcore.helpers", "ScopedValue", sizeof(struct __pyx_obj_5qcore_7helpers_ScopedValue), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers_ScopedValue)) __PYX_ERR(6, 52, __pyx_L1_error)
  __pyx_vtabptr_5qcore_7helpers_ScopedValue = (struct __pyx_vtabstruct_5qcore_7helpers_ScopedValue*)__Pyx_GetVtable(__pyx_ptype_5qcore_7helpers_ScopedValue->tp_dict); if (unlikely(!__pyx_vtabptr_5qcore_7helpers_ScopedValue)) __PYX_ERR(6, 52, __pyx_L1_error)
  __pyx_ptype_5qcore_7helpers__PropertyOverrideContext = __Pyx_ImportType("qcore.helpers", "_PropertyOverrideContext", sizeof(struct __pyx_obj_5qcore_7helpers__PropertyOverrideContext), 1); if (unlikely(!__pyx_ptype_5qcore_7helpers__PropertyOverrideContext)) __PYX_ERR(6, 65, __pyx_L1_error)
  __pyx_ptype_5asynq_6_debug_DebugOptions = __Pyx_ImportType("asynq._debug", "DebugOptions", sizeof(struct __pyx_obj_5asynq_6_debug_DebugOptions), 1); if (unlikely(!__pyx_ptype_5asynq_6_debug_DebugOptions)) __PYX_ERR(7, 18, __pyx_L1_error)
  __pyx_vtabptr_5asynq_6_debug_DebugOptions = (struct __pyx_vtabstruct_5asynq_6_debug_DebugOptions*)__Pyx_GetVtable(__pyx_ptype_5asynq_6_debug_DebugOptions->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_6_debug_DebugOptions)) __PYX_ERR(7, 18, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_FutureBase = __Pyx_ImportType("asynq.futures", "FutureBase", sizeof(struct __pyx_obj_5asynq_7futures_FutureBase), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_FutureBase)) __PYX_ERR(8, 26, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_FutureBase = (struct __pyx_vtabstruct_5asynq_7futures_FutureBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_FutureBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_FutureBase)) __PYX_ERR(8, 26, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_Future = __Pyx_ImportType("asynq.futures", "Future", sizeof(struct __pyx_obj_5asynq_7futures_Future), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_Future)) __PYX_ERR(8, 49, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_Future = (struct __pyx_vtabstruct_5asynq_7futures_Future*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_Future->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_Future)) __PYX_ERR(8, 49, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_ConstFuture = __Pyx_ImportType("asynq.futures", "ConstFuture", sizeof(struct __pyx_obj_5asynq_7futures_ConstFuture), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_ConstFuture)) __PYX_ERR(8, 54, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_ConstFuture = (struct __pyx_vtabstruct_5asynq_7futures_ConstFuture*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_ConstFuture->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_ConstFuture)) __PYX_ERR(8, 54, __pyx_L1_error)
  __pyx_ptype_5asynq_7futures_ErrorFuture = __Pyx_ImportType("asynq.futures", "ErrorFuture", sizeof(struct __pyx_obj_5asynq_7futures_ErrorFuture), 1); if (unlikely(!__pyx_ptype_5asynq_7futures_ErrorFuture)) __PYX_ERR(8, 57, __pyx_L1_error)
  __pyx_vtabptr_5asynq_7futures_ErrorFuture = (struct __pyx_vtabstruct_5asynq_7futures_ErrorFuture*)__Pyx_GetVtable(__pyx_ptype_5asynq_7futures_ErrorFuture->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_7futures_ErrorFuture)) __PYX_ERR(8, 57, __pyx_L1_error)
  __pyx_ptype_5asynq_10async_task_AsyncTask = __Pyx_ImportType("asynq.async_task", "AsyncTask", sizeof(struct __pyx_obj_5asynq_10async_task_AsyncTask), 1); if (unlikely(!__pyx_ptype_5asynq_10async_task_AsyncTask)) __PYX_ERR(9, 31, __pyx_L1_error)
  __pyx_vtabptr_5asynq_10async_task_AsyncTask = (struct __pyx_vtabstruct_5asynq_10async_task_AsyncTask*)__Pyx_GetVtable(__pyx_ptype_5asynq_10async_task_AsyncTask->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_10async_task_AsyncTask)) __PYX_ERR(9, 31, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_BatchBase = __Pyx_ImportType("asynq.batching", "BatchBase", sizeof(struct __pyx_obj_5asynq_8batching_BatchBase), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_BatchBase)) __PYX_ERR(10, 26, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_BatchBase = (struct __pyx_vtabstruct_5asynq_8batching_BatchBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_BatchBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_BatchBase)) __PYX_ERR(10, 26, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_BatchItemBase = __Pyx_ImportType("asynq.batching", "BatchItemBase", sizeof(struct __pyx_obj_5asynq_8batching_BatchItemBase), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_BatchItemBase)) __PYX_ERR(10, 45, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_BatchItemBase = (struct __pyx_vtabstruct_5asynq_8batching_BatchItemBase*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_BatchItemBase->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_BatchItemBase)) __PYX_ERR(10, 45, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_DebugBatchItem = __Pyx_ImportType("asynq.batching", "DebugBatchItem", sizeof(struct __pyx_obj_5asynq_8batching_DebugBatchItem), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_DebugBatchItem)) __PYX_ERR(10, 53, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_DebugBatchItem = (struct __pyx_vtabstruct_5asynq_8batching_DebugBatchItem*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_DebugBatchItem->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_DebugBatchItem)) __PYX_ERR(10, 53, __pyx_L1_error)
  __pyx_ptype_5asynq_8batching_DebugBatch = __Pyx_ImportType("asynq.batching", "DebugBatch", sizeof(struct __pyx_obj_5asynq_8batching_DebugBatch), 1); if (unlikely(!__pyx_ptype_5asynq_8batching_DebugBatch)) __PYX_ERR(10, 56, __pyx_L1_error)
  __pyx_vtabptr_5asynq_8batching_DebugBatch = (struct __pyx_vtabstruct_5asynq_8batching_DebugBatch*)__Pyx_GetVtable(__pyx_ptype_5asynq_8batching_DebugBatch->tp_dict); if (unlikely(!__pyx_vtabptr_5asynq_8batching_DebugBatch)) __PYX_ERR(10, 56, __pyx_L1_error)
  /*--- Variable import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("qcore.events"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "sinking_event_hook", (void **)&__pyx_vp_5qcore_6events_sinking_event_hook, "struct __pyx_obj_5qcore_6events_SinkingEventHook *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("qcore.helpers"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_list", (void **)&__pyx_vp_5qcore_7helpers_empty_list, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_tuple", (void **)&__pyx_vp_5qcore_7helpers_empty_tuple, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_dict", (void **)&__pyx_vp_5qcore_7helpers_empty_dict, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "none", (void **)&__pyx_vp_5qcore_7helpers_none, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "miss", (void **)&__pyx_vp_5qcore_7helpers_miss, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "same", (void **)&__pyx_vp_5qcore_7helpers_same, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "unspecified", (void **)&__pyx_vp_5qcore_7helpers_unspecified, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "empty_context", (void **)&__pyx_vp_5qcore_7helpers_empty_context, "struct __pyx_obj_5qcore_7helpers_EmptyContext *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "CachedHashWrapper", (void **)&__pyx_vp_5qcore_7helpers_CachedHashWrapper, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_2, "override", (void **)&__pyx_vp_5qcore_7helpers_override, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_ImportModule("asynq._debug"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_3, "options", (void **)&__pyx_vp_5asynq_6_debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_ImportModule("asynq.futures"); if (!__pyx_t_4) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "_debug_options", (void **)&__pyx_vp_5asynq_7futures__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_4, "_none", (void **)&__pyx_vp_5asynq_7futures__none, "struct __pyx_obj_5qcore_7helpers_MarkerObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_ImportModule("asynq.async_task"); if (!__pyx_t_5) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_debug_options", (void **)&__pyx_vp_5asynq_10async_task__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_futures_none", (void **)&__pyx_vp_5asynq_10async_task__futures_none, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_none_future", (void **)&__pyx_vp_5asynq_10async_task__none_future, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_empty_tuple", (void **)&__pyx_vp_5asynq_10async_task__empty_tuple, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_5, "_empty_dictionary", (void **)&__pyx_vp_5asynq_10async_task__empty_dictionary, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_6 = __Pyx_ImportModule("asynq.batching"); if (!__pyx_t_6) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_debug_options", (void **)&__pyx_vp_5asynq_8batching__debug_options, "struct __pyx_obj_5asynq_6_debug_DebugOptions *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_6, "_debug_batch_state", (void **)&__pyx_vp_5asynq_8batching__debug_batch_state, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "asynq/scheduler.py":15
 * # limitations under the License.
 * 
 * from sys import stderr, stdout             # <<<<<<<<<<<<<<
 * import time
 * import collections
 */
  __pyx_t_7 = PyList_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_stderr);
  __Pyx_GIVEREF(__pyx_n_s_stderr);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_stderr);
  __Pyx_INCREF(__pyx_n_s_stdout);
  __Pyx_GIVEREF(__pyx_n_s_stdout);
  PyList_SET_ITEM(__pyx_t_7, 1, __pyx_n_s_stdout);
  __pyx_t_8 = __Pyx_Import(__pyx_n_s_sys, __pyx_t_7, -1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_8, __pyx_n_s_stderr); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stderr, __pyx_t_7) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_8, __pyx_n_s_stdout); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stdout, __pyx_t_7) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "asynq/scheduler.py":16
 * 
 * from sys import stderr, stdout
 * import time             # <<<<<<<<<<<<<<
 * import collections
 * import threading
 */
  __pyx_t_8 = __Pyx_Import(__pyx_n_s_time, 0, -1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_8) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "asynq/scheduler.py":17
 * from sys import stderr, stdout
 * import time
 * import collections             # <<<<<<<<<<<<<<
 * import threading
 * 
 */
  __pyx_t_8 = __Pyx_Import(__pyx_n_s_collections, 0, -1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_collections, __pyx_t_8) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "asynq/scheduler.py":18
 * import time
 * import collections
 * import threading             # <<<<<<<<<<<<<<
 * 
 * import qcore.events as core_events
 */
  __pyx_t_8 = __Pyx_Import(__pyx_n_s_threading, 0, -1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_threading, __pyx_t_8) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "asynq/scheduler.py":20
 * import threading
 * 
 * import qcore.events as core_events             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s__20);
  __Pyx_GIVEREF(__pyx_n_s__20);
  PyList_SET_ITEM(__pyx_t_8, 0, __pyx_n_s__20);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s_qcore_events, __pyx_t_8, -1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_core_events, __pyx_t_7) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":23
 * 
 * 
 * from . import debug             # <<<<<<<<<<<<<<
 * from . import futures
 * from . import batching
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_debug);
  __Pyx_GIVEREF(__pyx_n_s_debug);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_debug);
  __pyx_t_8 = __Pyx_Import(__pyx_n_s__21, __pyx_t_7, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_8, __pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_debug, __pyx_t_7) < 0) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "asynq/scheduler.py":24
 * 
 * from . import debug
 * from . import futures             # <<<<<<<<<<<<<<
 * from . import batching
 * from . import _debug
 */
  __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s_futures);
  __Pyx_GIVEREF(__pyx_n_s_futures);
  PyList_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_futures);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s__21, __pyx_t_8, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_futures); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_futures, __pyx_t_8) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":25
 * from . import debug
 * from . import futures
 * from . import batching             # <<<<<<<<<<<<<<
 * from . import _debug
 * 
 */
  __pyx_t_7 = PyList_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_n_s_batching);
  __Pyx_GIVEREF(__pyx_n_s_batching);
  PyList_SET_ITEM(__pyx_t_7, 0, __pyx_n_s_batching);
  __pyx_t_8 = __Pyx_Import(__pyx_n_s__21, __pyx_t_7, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_ImportFrom(__pyx_t_8, __pyx_n_s_batching); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_batching, __pyx_t_7) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "asynq/scheduler.py":26
 * from . import futures
 * from . import batching
 * from . import _debug             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_8 = PyList_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_n_s_debug_2);
  __Pyx_GIVEREF(__pyx_n_s_debug_2);
  PyList_SET_ITEM(__pyx_t_8, 0, __pyx_n_s_debug_2);
  __pyx_t_7 = __Pyx_Import(__pyx_n_s__21, __pyx_t_8, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_ImportFrom(__pyx_t_7, __pyx_n_s_debug_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_debug_2, __pyx_t_8) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":29
 * 
 * 
 * assert str(__name__).endswith('asynq.scheduler') \             # <<<<<<<<<<<<<<
 *     or str(__name__).endswith('asynq.lib64.scheduler'), \
 *     "Are you importing asynq from the wrong directory?"
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {

    /* "asynq/scheduler.py":30
 * 
 * assert str(__name__).endswith('asynq.scheduler') \
 *     or str(__name__).endswith('asynq.lib64.scheduler'), \             # <<<<<<<<<<<<<<
 *     "Are you importing asynq from the wrong directory?"
 * 
 */
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "asynq/scheduler.py":29
 * 
 * 
 * assert str(__name__).endswith('asynq.scheduler') \             # <<<<<<<<<<<<<<
 *     or str(__name__).endswith('asynq.lib64.scheduler'), \
 *     "Are you importing asynq from the wrong directory?"
 */
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_endswith); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (!__pyx_t_10) {
    } else {
      __pyx_t_9 = __pyx_t_10;
      goto __pyx_L2_bool_binop_done;
    }

    /* "asynq/scheduler.py":30
 * 
 * assert str(__name__).endswith('asynq.scheduler') \
 *     or str(__name__).endswith('asynq.lib64.scheduler'), \             # <<<<<<<<<<<<<<
 *     "Are you importing asynq from the wrong directory?"
 * 
 */
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_name_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7);
    __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_endswith); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 30, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_9 = __pyx_t_10;
    __pyx_L2_bool_binop_done:;
    if (unlikely(!__pyx_t_9)) {
      PyErr_SetObject(PyExc_AssertionError, __pyx_kp_s_Are_you_importing_asynq_from_the);
      __PYX_ERR(0, 29, __pyx_L1_error)
    }
  }
  #endif

  /* "asynq/scheduler.py":34
 * 
 * 
 * MAX_AWAIT_RECURSION_DEPTH = 10             # <<<<<<<<<<<<<<
 * 
 * _debug_options = _debug.options
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAX_AWAIT_RECURSION_DEPTH, __pyx_int_10) < 0) __PYX_ERR(0, 34, __pyx_L1_error)

  /* "asynq/scheduler.py":36
 * MAX_AWAIT_RECURSION_DEPTH = 10
 * 
 * _debug_options = _debug.options             # <<<<<<<<<<<<<<
 * _futures_none = futures._none
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_5asynq_6_debug_options));
  __Pyx_XGOTREF(((PyObject *)__pyx_v_5asynq_9scheduler__debug_options));
  __Pyx_DECREF_SET(__pyx_v_5asynq_9scheduler__debug_options, __pyx_v_5asynq_6_debug_options);
  __Pyx_GIVEREF(((PyObject *)__pyx_v_5asynq_6_debug_options));

  /* "asynq/scheduler.py":37
 * 
 * _debug_options = _debug.options
 * _futures_none = futures._none             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_5asynq_7futures__none));
  __Pyx_XGOTREF(__pyx_v_5asynq_9scheduler__futures_none);
  __Pyx_DECREF_SET(__pyx_v_5asynq_9scheduler__futures_none, ((PyObject *)__pyx_v_5asynq_7futures__none));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_5asynq_7futures__none));

  /* "asynq/scheduler.py":40
 * 
 * 
 * class AsyncTaskError(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_builtin_Exception);
  __Pyx_GIVEREF(__pyx_builtin_Exception);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_builtin_Exception);
  __pyx_t_8 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_11 = __Pyx_Py3MetaclassPrepare(__pyx_t_8, __pyx_t_7, __pyx_n_s_AsyncTaskError, __pyx_n_s_AsyncTaskError, (PyObject *) NULL, __pyx_kp_s_asynq_scheduler, (PyObject *) NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_Py3ClassCreate(__pyx_t_8, __pyx_n_s_AsyncTaskError, __pyx_t_7, __pyx_t_11, NULL, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AsyncTaskError, __pyx_t_12) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":140
 *         return 0
 * 
 *     def await(self, tasks=[]):             # <<<<<<<<<<<<<<
 *         """
 *         Executes a list of tasks ensuring all of them are completed when this method returns.
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_k__5 = __pyx_t_7;
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "asynq/scheduler.py":534
 * 
 * 
 * class LocalTaskSchedulerState(threading.local):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         self.last_id = 0
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_threading); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_local); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_8);
  __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CalculateMetaclass(NULL, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_11 = __Pyx_Py3MetaclassPrepare(__pyx_t_8, __pyx_t_7, __pyx_n_s_LocalTaskSchedulerState, __pyx_n_s_LocalTaskSchedulerState, (PyObject *) NULL, __pyx_kp_s_asynq_scheduler, (PyObject *) NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "asynq/scheduler.py":535
 * 
 * class LocalTaskSchedulerState(threading.local):
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.last_id = 0
 *         self.reset()
 */
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5asynq_9scheduler_23LocalTaskSchedulerState_1__init__, 0, __pyx_n_s_LocalTaskSchedulerState___init, NULL, __pyx_kp_s_asynq_scheduler, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyObject_SetItem(__pyx_t_11, __pyx_n_s_init, __pyx_t_12) < 0) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "asynq/scheduler.py":539
 *         self.reset()
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         self.current = TaskScheduler()
 * 
 */
  __pyx_t_12 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5asynq_9scheduler_23LocalTaskSchedulerState_3reset, 0, __pyx_n_s_LocalTaskSchedulerState_reset, NULL, __pyx_kp_s_asynq_scheduler, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyObject_SetItem(__pyx_t_11, __pyx_n_s_reset, __pyx_t_12) < 0) __PYX_ERR(0, 539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

  /* "asynq/scheduler.py":534
 * 
 * 
 * class LocalTaskSchedulerState(threading.local):             # <<<<<<<<<<<<<<
 *     def __init__(self):
 *         self.last_id = 0
 */
  __pyx_t_12 = __Pyx_Py3ClassCreate(__pyx_t_8, __pyx_n_s_LocalTaskSchedulerState, __pyx_t_7, __pyx_t_11, NULL, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LocalTaskSchedulerState, __pyx_t_12) < 0) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":542
 *         self.current = TaskScheduler()
 * 
 * _state = LocalTaskSchedulerState()             # <<<<<<<<<<<<<<
 * globals()['_state'] = _state
 * 
 */
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_LocalTaskSchedulerState); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_11 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  if (__pyx_t_11) {
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_11); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  } else {
    __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 542, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_XGOTREF(__pyx_v_5asynq_9scheduler__state);
  __Pyx_DECREF_SET(__pyx_v_5asynq_9scheduler__state, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_7);
  __pyx_t_7 = 0;

  /* "asynq/scheduler.py":543
 * 
 * _state = LocalTaskSchedulerState()
 * globals()['_state'] = _state             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_Globals(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (unlikely(PyObject_SetItem(__pyx_t_7, __pyx_n_s_state, __pyx_v_5asynq_9scheduler__state) < 0)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":551
 * 
 * 
 * def reset():             # <<<<<<<<<<<<<<
 *     _state.reset()
 * 
 */
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5asynq_9scheduler_3reset, 0, __pyx_n_s_reset, NULL, __pyx_kp_s_asynq_scheduler, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reset, __pyx_t_7) < 0) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "asynq/scheduler.py":1
 * # Copyright 2016 Quora, Inc.             # <<<<<<<<<<<<<<
 * #
 * # Licensed under the Apache License, Version 2.0 (the "License");
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_7) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init asynq.scheduler", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init asynq.scheduler");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* GetAttr */
      static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_COMPILING_IN_CPYTHON
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
      static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* PyIntBinop */
      #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* ExtTypeTest */
      static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* RaiseDoubleKeywords */
      static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
      static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyErrFetchRestore */
      #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
      #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* ArgTypeTest */
        static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* SaveResetException */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
        #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
          #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* None */
          static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyObjectCallMethod1 */
          static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            args = PyTuple_New(2);
            if (unlikely(!args)) goto bad;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
bad:
    Py_XDECREF(method);
    return result;
}

/* append */
          static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* GetItemInt */
          static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SliceObject */
          static CYTHON_INLINE int __Pyx_PyObject_SetSlice(PyObject* obj, PyObject* value,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_ass_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_ass_slice(obj, cstart, cstop, value);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_ass_subscript))
#endif
    {
        int result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_COMPILING_IN_CPYTHON
        result = mp->mp_ass_subscript(obj, py_slice, value);
#else
        result = value ? PyObject_SetItem(obj, py_slice, value) : PyObject_DelItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object does not support slice %.10s",
        Py_TYPE(obj)->tp_name, value ? "assignment" : "deletion");
bad:
    return -1;
}

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
          static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
          static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* Py3ClassCreate */
          static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* FetchCommonType */
          static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
          static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CodeObjectCache */
              static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
              #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
              #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
              static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
              static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
              static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
              static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrExport */
              static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* FunctionExport */
              static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
              #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
              #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* VoidPtrImport */
              #ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
              static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
